﻿# -*- coding: utf-8 -*-

from brownie import *
from keymap import *
from nodes import *
from ConfigParser import SafeConfigParser
import os
import sys
import error
import resource
import user

#=============================================================================

def create_listview(owner):
    def pair(self, name):
        a = getattr(type(self), i)
        if hasattr(a, '__call__'):
            return (name + ' ()', a.__doc__)
        else:
            return (name, a.__doc__)
    dlg = Dialog(owner, size=(640, 480))
    lv = ListView(dlg, dock=CENTER)
    lv.columns = [ '属性', '説明' ]
    lv.items = [pair(lv, i) for i in dir(lv) if not i.startswith('__')]
    dlg.run()

#=============================================================================

def format_version(version):
    return '%d.%d.%d.%d' % (
        (version >> 24) & 0xFF,
        (version >> 16) & 0xFF,
        (version >>  8) & 0xFF,
        (version      ) & 0xFF)

def about():
    version = format_version(config.VERSION)
    winver = format_version(config.WINDOWS_VERSION)
    alert('brownie : %s\nWindows : %s\nPython : %s' % (version, winver, sys.version), 'バージョン情報', INFO)

#=============================================================================

class AddressBar(HeaderBar):
    __slots__ = ('__path',)

    def __init__(self, *args, **kw):
        super(AddressBar, self).__init__(size = (24, 24), *args, **kw)
        self.icons = resource.load(Icons, config.PATH_ROOT + '/res/16x16.png')
        self.__path = None

    def get_path(self):
        return self.__path

    def set_path(self, value):
        if self.__path == value:
            return
        self.__path = value
        icon = value.icon
        n = Item(icon = icon, children = [
            Item('閉じる(*C)'          , self.parent.close  , icon=46),
            Item('最新の情報に更新(*R)', self.parent.refresh, icon=17),
            Item('パスをコピー(*P)'    , lambda: Clipboard().copy(self.path), icon=4),
            Item('幅を調整(*A)'        , self.parent.adjust),
        ])
        elements = [ n ]
        while value:
            parent = value.parent
            if parent:
                n = PathItem(value)
            else:
                n = PathItem(value, '/')
            n.icon = None
            elements.insert(1, n)
            value = parent
        self.items = elements

    path = property(get_path, set_path)

#=============================================================================

class ExplorerView(FolderView):
    def __init__(self, *args, **kwds):
        super(ExplorerView, self).__init__(*args, **kwds)
        self.address = AddressBar(self)
        self.previous_mode = (ICON, 256)
        self.keymap = user.explorer_keymap(self)
        self.gesture = user.explorer_gesture(self)

    def copy_name(self):
        items = [i.name for i in self.selection]
        Clipboard().text = '\r\n'.join(items)
        return ('クリップボードに %d 個のファイル名をコピーしました。' % len(items), INFO)

    def copy_path(self):
        items = [i.path for i in self.selection]
        Clipboard().text = '\r\n'.join(items)
        return ('クリップボードに %d 個のパスをコピーしました。' % len(items), INFO)

    def on_execute(self, items):
        mods = modifiers()
        verb = None
        if (mods & KEYS) == SHIFT:
            verb = '.txt'
        elif (mods & KEYS) == CONTROL:
            verb = 'edit'
        go = (mods & KEYS) == CONTROL or (mods & BUTTONS) == MBUTTON
        for i in items:
            item = resource.resolve(i)
            if item.browsable > 1:
                if not go:
                    try:
                        self.go(item)
                        go = True
                    except:
                        pass
                else:
                    sub = ExplorerView(self.parent)
                    try:
                        sub.go(item)
                    except:
                        sub.close()
            else:
                i.execute(verb)
        return True

    def on_navigate(self):
        self.address.path = self.path
        self.root.on_view_update(self)

    def on_activate(self):
        self.root.on_view_activate(self)

    def on_notify(self, message, mode):
        self.root.on_view_notify(self, message, mode)

    def on_close(self):
        self.root.on_view_close(self)

    def open(self, path):
        if (modifiers() & (CONTROL | MBUTTON)) == 0:
            self.go(path)
        else:
            dst = self.parse(path)
            if dst:
                self.parent.open(dst)

#=============================================================================

class Form(Window):
    def __init__(self, *args, **kwds):
        super(Form, self).__init__(*args, **kwds)
        # bars
        self.rebar = ReBar(self) # self
        self.toolbar = ToolBar(self.rebar)
        self.tabbar = TabBar(self.rebar)
        self.statusbar = StatusBar(self, size = (24, 24))
        self.notification_mode = LOG
        self.notification_time = 0
        icon16x16 = resource.load(Icons, config.PATH_ROOT + '/res/16x16.png')

        # mru
        def open_mru():
            children = self.mru.children
            if len(children) > 2:
                self.open(children[2])
                del self.mru.children[2]
        self.mru = Item('履歴', open_mru, [Item('最後に閉じたフォルダを開く(*L)', open_mru), None], icon=2)

        # toolbar
        self.toolbar.name = 'ツールバー'
        self.toolbar.icons = icon16x16
        self.toolbar.items = [
            GoUp('上へ', icon=9),
            GoBack('戻る', icon=7),
            GoForward('進む', icon=8),
            None,
            FavoriteItem('お気に入り', icon=19),
            self.mru,
            PathItem(FOLDER_DRIVE, ''),
            PathItem(FOLDER_DOCUMENT, ''),
            None,
            Item('', None, [
                Item('新しいウィンドウ(*N)', lambda: System.fork(lambda: Form(size=self.size))),
                Item('ListViewテスト', lambda: create_listview(self)),
                Trash(),
                None,
                Item('バージョン情報(*A)...', about),
                PathItem(os.path.join(config.PATH_ROOT, 'man/index.html'), 'マニュアル(*M)'),
                PathItem(os.path.join(config.PATH_ROOT, 'TODO.txt'), '既知のバグ／要望(*T)'),
                None,
                Item('オプション(*O)', None, [
                    LISTVIEW_LOOP       ('カーソルのループ(*L)'),
                    LISTVIEW_WHEELSELECT('マウスホイールで選択(*W)'),
                    LISTVIEW_AUTOADJUST ('自動的に幅を調整する(*D)'),
                    LISTVIEW_BACKGROUND ('行の塗り分け(*B)'),
                    LISTVIEW_AUTOUNIT   ('ファイルサイズ単位の自動選択(*A)'),
                    RENAME_EXTENSION    ('ファイル名編集で拡張子を選択(*E)'),
                    None,
                    PathItem(os.path.join(config.PATH_USER, 'settings.ini'), '設定ファイルを編集(*S)'),
                ]),
                Item('再起動(*R)', System.restart),
                Item('終了(*X)', System.quit, icon=46),
            ], icon=47),
        ]

        # tabbar
        self.tabbar.name = 'タブ'
        self.tabbar.size = (24, 24)

        # statusbar
        self.statusbar.name = 'ステータスバー'
        self.statusbar.icons = icon16x16
        self.statusbar.items = [
            Mode('', icon=34),
            Grouping('グループ'),
            Current('幅を調整', ExplorerView.adjust),
        ]

        # treeview
        #self.tree = TreeView(self, dock = WEST, size = (200, 200))
        #self.tree.icons = self.toolbar.icons
        #self.tree.items = self.toolbar.items

        # attributes
        self.name = 'brownie'
        self.keymap = user.form_keymap(self)
        self.gesture = user.form_gesture(self)

    def on_notify(self, message, mode):
        self.notify(message, mode)

    def on_view_notify(self, view, message, mode):
        if view == self.focus:
            self.notify(message, mode)

    def on_view_update(self, view):
        self.tabbar.items = [TabItem(i) for i in self.children if i.dock == CENTER]

    def on_view_activate(self, view):
        items = self.tabbar.items
        for i in range(len(items)):
            if items[i].target == view:
                self.tabbar.selection = i
                break

    def on_view_close(self, view):
        # update tabbar
        self.tabbar.items = [TabItem(i) for i in self.children if i != view and i.dock == CENTER]
        # update mru list
        folder = view.path
        self.mru.children[2:] = [i for i in self.mru.children[2:] if i.path != folder.path]
        n = PathItem(folder)
        delattr(n, 'children')
        self.mru.children.insert(2, n)
        del self.mru.children[22:] # remember last 20 folders
        try:
            for i in range(9):
                n = self.mru.children[i+2]
                n.name = '*%d: %s' % (i+1, n.basename)
        except IndexError:
            pass

    def open(self, path):
        leaf = None
        if not isinstance(path, Path):
            path = Path(path)
        if not path.browsable:
            leaf = path
            path = path.parent
        try:
            target = self.focus
            if target and (modifiers() & (CONTROL | MBUTTON)) == 0:
                target.go(path)
            else:
                target = ExplorerView(self)
                try:
                    target.go(path)
                except:
                    target.close()
                    raise
            if leaf:
                target.select(leaf)
        except:
            self.notify('"%s" を開けません。' % path, WARNING)
            raise

    def notify(self, message, mode = LOG):
        if mode != LOG:
            beep(mode)
        if mode == ERROR:
            alert(message, 'エラー', mode)
        elif self.notification_mode >= mode or self.notification_time + 200 < now():
            self.statusbar.value = message
            self.notification_mode = mode
            self.notification_time = now()

#=============================================================================

class Editor(TextBox):
    def __init__(self, *args, **kwds):
        super(Editor, self).__init__(*args, **kwds)
        self.keymap = user.editor_keymap(self)

    def select_basename(self):
        value = self.value
        dot = value.rfind('.')
        if dot > 0:
            ln = len(value)
            basename = (0, dot)
            if self.selection == basename:
                self.selection = (dot+1, ln)
            else:
                self.selection = basename

    def select_extension(self):
        value = self.value
        dot = value.rfind('.')
        if dot > 0:
            ln = len(value)
            extension = (dot+1, ln)
            if self.selection == extension:
                self.selection = (0, dot)
            else:
                self.selection = extension

    def delete_before(self):
        selection = self.selection
        pos = max(selection[0], selection[1])
        self.value = self.value[pos:]
        self.selection = 0

    def delete_after(self):
        selection = self.selection
        pos = min(selection[0], selection[1])
        self.value = self.value[:pos]
        self.selection = pos


def read_settings():
    settings = SafeConfigParser()
    settings.read(os.path.join(config.PATH_USER, 'settings.ini'))
    for i in settings.items('config'):
        setattr(config, i[0].upper(), i[1])
    return settings

#=============================================================================
# main

error.init()
settings = read_settings()
w = settings.get('Window', 'default_width', 640)
h = settings.get('Window', 'default_height', 480)
config.EDITOR = Editor
config.KEYMAP = user.global_keymap()
System.fork(lambda: Form(size=(w, h)))
