// commctrl.cpp

#include "stdafx.h"
#include "commctrl.hpp"
#include "drawing.hpp"
#include "math.hpp"

#pragma comment(lib, "comctl32")

using namespace brownie;

//==================================================================================================================================================================
// ListView

void brownie::ListView_Adjust(HWND hwnd) throw()
{
	HWND header = ListView_GetHeader(hwnd);
	if (!header || !IsWindowVisible(header))
		return;

	const int columns = Header_GetItemCount(header);
	if (columns <= 0)
		return;
	const int rows = ListView_GetItemCount(hwnd);
	if (rows <= 0)
		return;

	NMHEADER notify = { header, ::GetDlgCtrlID(header), HDN_DIVIDERDBLCLICK, 0, 0 };
	for (int i = 0; i < columns; ++i)
	{
		notify.iItem = i;
		::SendMessage(hwnd, WM_NOTIFY, notify.hdr.idFrom, (LPARAM)&notify);
	}
}

void brownie::ListView_SelectItem_SVSI(HWND hwnd, int index, UINT svsi) throw()
{
	if (svsi & SVSI_DESELECTOTHERS)
		ListView_SetItemState(hwnd, -1, 0, LVIS_FOCUSED | LVIS_SELECTED);
	UINT lvState = LVIS_SELECTED;
	if (svsi & SVSI_FOCUSED)
		lvState |= LVIS_FOCUSED;
	ListView_SetItemState(hwnd, index, lvState, lvState);
	if (svsi & SVSI_ENSUREVISIBLE)
		ListView_EnsureVisible(hwnd, index, FALSE);
}

int brownie::ListView_GetItemText(HWND hwnd, int row, int column, WCHAR buf[], size_t bufsize)
{
	LVITEM item = { LVIF_TEXT };
	item.iItem = row;
	item.iSubItem = column;
	item.pszText = buf;
	item.cchTextMax = (int)bufsize;
	return (int)::SendMessage(hwnd, LVM_GETITEMTEXT, (WPARAM)(row), (LPARAM)&item);
}

int brownie::ListView_GetColumnCount(HWND hwnd) throw()
{
	return Header_GetItemCount(ListView_GetHeader(hwnd));
}

HRESULT brownie::ListView_GetSortedColumn(HWND hwnd, int* column, bool* ascending)
{
	ASSERT(column);
	ASSERT(ascending);
	*column = -1;
	HWND header = ListView_GetHeader(hwnd);
	if (!::IsWindow(header))
		return E_INVALIDARG;
	const int headerCount = Header_GetItemCount(header);
	*column = ListView_GetSelectedColumn(hwnd);
	*ascending = false;
	// SelectedColumn  Header ̊֌WȂĂꍇ̂
	for (int i = 0; i < headerCount; ++i)
	{
		HDITEM info = { HDI_FORMAT };
		Header_GetItem(header, i, &info);
		if (i == *column)
		{
			if ((info.fmt & (HDF_SORTUP | HDF_SORTDOWN)) == 0)
			{	// SelectedColumn Ȃ̂Ƀ\[g}[NtĂȂI
				*column = -1; // invalid
				return E_UNEXPECTED;
			}
			*ascending = ((info.fmt & HDF_SORTDOWN) == HDF_SORTDOWN);
		}
		else if (info.fmt & (HDF_SORTUP | HDF_SORTDOWN))
		{	// SelectedColumn ȊOɃ\[g}[NtĂI
			*column = -1; // invalid
			return E_UNEXPECTED;
		}
	}
	return S_OK;
}

HRESULT brownie::ListView_SetSortedColumn(HWND hwnd, int column, bool ascending)
{
	if (column < 0)
		return E_INVALIDARG;

	HWND hwndOwner = ::GetParent(hwnd);
	HWND header = ListView_GetHeader(hwnd);
	if (!::IsWindow(header))
		return E_INVALIDARG;
	if (column >= Header_GetItemCount(header))
		return E_INVALIDARG;
	for (int i = 0; i < 10; ++i) // ςȂƂNĖ[vɂȂ̂h߁B
	{
		NMLISTVIEW notify = { hwnd, ::GetDlgCtrlID(hwnd), LVN_COLUMNCLICK, -1, column };
		::SendMessage(hwndOwner, WM_NOTIFY, notify.hdr.idFrom, (LPARAM)&notify);
		int c; bool a;
		if (SUCCEEDED(ListView_GetSortedColumn(hwnd, &c, &a)) && c == column && a == ascending)
			break;
	}
	return S_OK;
}

bool brownie::ListView_HitTestColumn(HWND hwnd, int column, int x)
{
	HeaderH header = ListView_GetHeader(hwnd);
	if (header && header.Visible)
	{
		RECT rc;
		header.GetItemRect(column, &rc);
		int xx = x + GetScrollPos(hwnd, SB_HORZ);
		if (rc.left < xx && xx < rc.right)
			return true;
	}
	return false;
}

namespace
{
	static UINT ListView_CalcText(DRAWTEXTPARAMS* params, int fmt, bool first)
	{
		UINT style = DT_VCENTER | DT_SINGLELINE | DT_END_ELLIPSIS | DT_NOPREFIX;

		switch (fmt & LVCFMT_JUSTIFYMASK)
		{
		case LVCFMT_LEFT:
			style |= DT_LEFT;
			params->iLeftMargin = (first ? 2 : 8);
			break;
		case LVCFMT_RIGHT:
			style |= DT_RIGHT;
			params->iLeftMargin  = 2;
			params->iRightMargin = 6;
			break;
		default:
			style |= DT_CENTER;
			params->iLeftMargin = 2;
			params->iRightMargin = 2;
			break;
		}

		return style;
	}

	static void ListView_DrawText(HDC dc, PWSTR text, RECT* rc, int fmt)
	{
		DRAWTEXTPARAMS params = { sizeof(DRAWTEXTPARAMS) };
		UINT style = ListView_CalcText(&params, fmt, false);
		DrawTextEx(dc, text, -1, rc, style, &params);
	}

	static void ListView_AdjustFocusRect(HDC dc, PWSTR text, int ln, RECT* rcFocus, UINT style, DRAWTEXTPARAMS* params)
	{
		int bottom = rcFocus->bottom;
		DrawTextEx(dc, text, ln, rcFocus, style | DT_CALCRECT, params);
		rcFocus->right += 4;
		rcFocus->bottom = bottom;
	}
}

inline bool RectInRect(const RECT& outer, const RECT& inner)
{
	if (!PtInRect(&outer, RECT_NW(inner)))
		return false;
	POINT se = RECT_SE(inner);
	se.x -= 1; se.y -= 1;
	return PtInRect(&outer, se) != 0;
}

inline static bool ListView_IsRenaming(HWND hwnd, RECT rcBounds)
{
	HWND edit = ListView_GetEditControl(hwnd);
	if (!edit)
		return false;
	RECT rcEdit;
	ClientToScreen(hwnd, &rcBounds);
	::GetWindowRect(edit, &rcEdit);
	return math::min(rcBounds.bottom - rcEdit.top, rcEdit.bottom - rcBounds.top) > RECT_H(rcBounds) / 2;
}

static bool IsDragDrop(HWND hwnd, const RECT& rcFocus, const RECT& rcIcon, int nameColumnRight)
{
	if (IsKeyPressed(VK_LBUTTON) || IsKeyPressed(VK_RBUTTON))
	{
		POINT cursor;
		::GetCursorPos(&cursor);
		::ScreenToClient(hwnd, &cursor);
		RECT rc = { rcIcon.left, rcFocus.top, rcFocus.right + 2, rcIcon.bottom };
		rc.right = math::min<int>(rc.right, nameColumnRight);
		return ::PtInRect(&rc, cursor) != 0;
	}
	return false;
}

static int SortedColumnRGB(int c)
{
    if (c < 128)
        return 255 - ((255 - c) * 31 / 32);
    else
        return c * 31 / 32;
}

static COLORREF SortedColumnColor(COLORREF c)
{
	int r = SortedColumnRGB(GetRValue(c));
	int g = SortedColumnRGB(GetGValue(c));
	int b = SortedColumnRGB(GetBValue(c));
	return RGB(r, g, b);
}

void brownie::ListView_BeginDraw(HWND hwnd, LVDRAWDESC& desc) throw()
{
	desc.SizeColumn = -1;

	HWND header = ListView_GetHeader(hwnd);
	if (header)
	{
		int scroll = GetScrollPos(hwnd, SB_HORZ);

		RECT rcName;
		Header_GetItemRect(header, 0, &rcName);
		desc.NameColumnRight = rcName.right - scroll;

		int index = ListView_GetSelectedColumn(hwnd);
		if (index >= 0)
		{	// XN[ĂꍇɑΉB
			RECT rcSorted;
			Header_GetItemRect(header, index, &rcSorted);
			desc.SortedColumnLeft = rcSorted.left - scroll;
			desc.SortedColumnRight = rcSorted.right - scroll;
		}
		else
		{	// G[H
			desc.SortedColumnLeft = desc.SortedColumnRight = 0;
		}

		// ĕ`̈悪ACeɔĂȂƁAI[ih[[`Ă΂ȂB
		// wiF݂̂œhԂ̂邽߂ɁAIɍĕ`̈g傷B
		RECT rcUpdate;
		::GetUpdateRect(hwnd, &rcUpdate, FALSE);
		int count = Header_GetItemCount(header);
		if (count > 0)
		{
			int	rightmost;
			int* order = (int*) alloca(count * sizeof(int));
			if (Header_GetOrderArray(header, count, order))
				rightmost = order[count - 1];
			else
				rightmost = count - 1;

			RECT rc;
			Header_GetItemRect(header, rightmost, &rc);
			if (rcUpdate.left > rc.right - 1)
			{
				rcUpdate.left = rc.right - 1;
				::InvalidateRect(hwnd, &rcUpdate, FALSE);
			}

			// TCYJ MB, GB ŕ\邩H
			static PCWSTR STR_SIZE = _W("Size");
			if (LISTVIEW_AUTOUNIT && STR_SIZE)
			{
				HDITEM item = { HDI_TEXT };
				WCHAR text[MAX_PATH];
				item.pszText = text;
				item.cchTextMax = lengthof(text);
				for (int i = 0; i < count; ++i)
				{
					if (Header_GetItem(header, i, &item) &&
						wcscmp(text, STR_SIZE) == 0)
					{
						desc.SizeColumn = i;
						break;
					}
				}
			}
		}
	}

	// ̃EBhEɔ킳Ăꍇ́AhbOł͂ȂB
	desc.DragDropNow = desc.DragDrop;
	if (desc.DragDropNow)
	{
		POINT cursor;
		::GetCursorPos(&cursor);
		if (hwnd != WindowFromPoint(cursor))
			desc.DragDropNow = false;
	}

	desc.HasFocus = (hwnd == GetFocus());
	desc.ImageList = ListView_GetImageList(hwnd, LVSIL_SMALL);
	for (int i = 0; i < 2; ++i)
		desc.BackgroundColor[i+2] = SortedColumnColor(desc.BackgroundColor[i]);
}

void brownie::ListView_DrawItem(const DRAWITEMSTRUCT* draw, const LVDRAWDESC& desc) throw()
{
	RECT	rcLabel, rcFocus, rcIcon, rcBounds, rcClient;
	WCHAR	text[MAX_PATH];
	bool	drawFocus = ((draw->itemState & ODS_FOCUS) != 0);
	HWND	hwnd = draw->hwndItem;
	HDC		dc = draw->hDC;
	UINT	itemState = draw->itemState;
	DRAWTEXTPARAMS params = { sizeof(DRAWTEXTPARAMS) };

	SetBkMode(dc, TRANSPARENT);

	// 
	LVCOLUMN	column = { LVCF_FMT };
	ListView_GetColumn(hwnd, 0, &column);

	// TCY
	GetClientRect(hwnd, &rcClient);
	ListView_GetItemRect(hwnd, draw->itemID, &rcBounds, LVIR_BOUNDS);
	ListView_GetItemRect(hwnd, draw->itemID, &rcLabel, LVIR_LABEL);
	rcBounds.right = math::min<int>(rcBounds.right, desc.NameColumnRight);
	rcLabel.right = math::min<int>(rcLabel.right, desc.NameColumnRight);
	SetRect(&rcFocus, rcLabel.left, rcBounds.top, rcLabel.right, rcBounds.bottom);
	rcBounds.left = desc.SortedColumnLeft;
	rcBounds.right = desc.SortedColumnRight;

	// wi
	int odd = draw->itemID % 2;
	COLORREF color = desc.BackgroundColor[odd];
	COLORREF sorted = desc.BackgroundColor[odd + 2];
	FillRect(dc, RECT_FROM_LTRB(rcClient.left, rcBounds.top, rcBounds.left, rcBounds.bottom), color);
	FillRect(dc, RECT_FROM_LTRB(rcBounds.left, rcBounds.top, rcBounds.right, rcBounds.bottom), sorted);
	FillRect(dc, RECT_FROM_LTRB(rcBounds.right, rcBounds.top, rcClient.right, rcBounds.bottom), color);

	// Ȍ擾
	LVITEM item = { LVIF_TEXT | LVIF_IMAGE | LVIF_STATE };
	item.iItem = draw->itemID;
	item.iSubItem = 0;
	item.pszText = text;
	item.cchTextMax = lengthof(text);
	item.stateMask = (UINT)-1;
	ListView_GetItem(hwnd, &item);

	if (desc.ImageList && item.iImage >= 0)
		ListView_GetItemRect(hwnd, draw->itemID, &rcIcon, LVIR_ICON);
	else
		SetRect(&rcIcon, rcLabel.left, rcLabel.top, rcLabel.left, rcLabel.bottom);

	// ÕeLXg
	int		syscolor_back;
	int		syscolor_text;
	int		ln = lstrlen(text);
	UINT	style = ListView_CalcText(&params, column.fmt, true);
	ListView_AdjustFocusRect(dc, text, ln, &rcFocus, style, &params);
	rcFocus.right = math::min<int>(rcFocus.right, desc.NameColumnRight);
	if (desc.DragDropNow && IsDragDrop(hwnd, rcFocus, rcIcon, desc.NameColumnRight))
	{
		syscolor_back = COLOR_HIGHLIGHT;
		syscolor_text = COLOR_HIGHLIGHTTEXT;
		itemState |= ODS_SELECTED;
	}
	else if (itemState & ODS_SELECTED)
	{
		if (ListView_IsRenaming(hwnd, rcBounds))
		{
			syscolor_back = syscolor_text = COLOR_WINDOW;
			drawFocus = false;
		}
		else if (desc.HasFocus)
		{
			syscolor_back = COLOR_HIGHLIGHT;
			syscolor_text = COLOR_HIGHLIGHTTEXT;
		}
		else
		{
			syscolor_back = COLOR_3DFACE;
			syscolor_text = COLOR_WINDOWTEXT;
			drawFocus = false;
		}
	}
	else
	{
		syscolor_back = COLOR_WINDOW;
		syscolor_text = COLOR_WINDOWTEXT;
	}

	if (syscolor_text != COLOR_WINDOW)
	{
		if (syscolor_back != COLOR_WINDOW)
			FillRect(dc, rcFocus, GetSysColor(syscolor_back));
		SetTextColor(dc, GetSysColor(syscolor_text));
		DrawTextEx(dc, text, ln, &rcLabel, style, &params);
		if (syscolor_text != COLOR_WINDOWTEXT)
			SetTextColor(dc, GetSysColor(COLOR_WINDOWTEXT));
	}

	// ACR
	if (desc.ImageList && item.iImage >= 0)
	{
		UINT style = ((itemState & ODS_SELECTED) ? ILD_FOCUS : ILD_NORMAL) | ILD_TRANSPARENT;
		if (item.stateMask & LVIS_OVERLAYMASK)
			style |= (ILD_OVERLAYMASK & item.state);
		if (ListView_GetItemState(hwnd, draw->itemID, LVIS_CUT))
			ImageList_DrawEx(desc.ImageList, item.iImage, dc, rcIcon.left, rcIcon.top, 0, 0, CLR_DEFAULT, color, style);
		else
			ImageList_Draw(desc.ImageList, item.iImage, dc, rcIcon.left, rcIcon.top, style);
	}

	// ̗̃eLXg`
	int ncolumn = ListView_GetColumnCount(hwnd);
	for (int i = 1; i < ncolumn; i++)
	{
		RECT rc;
		ListView_GetColumn(hwnd, i, &column);
		ListView_GetSubItemRect(hwnd, draw->itemID, i, LVIR_LABEL, &rc);
		ListView_GetItemText(hwnd, draw->itemID, i, text, lengthof(text));

		if (i == desc.SizeColumn)
		{
			PWSTR s = text;
			PWSTR t = text;
			while (*t)
			{
				if (*t != L',')
					*s++ = *t;
				t++;
			}
			*s = L'\0';
			if (s > text + 2 && s[-2] == L'K' && s[-1] == L'B')
			{
				int kB = _wtoi(text);
				if (kB > 1000 * 1024)
					swprintf(text, lengthof(text), L"%.2f GB", kB / (1024.0 * 1024.0));
				else if (kB > 1000)
					swprintf(text, lengthof(text), L"%.2f MB", kB / 1024.0);
			}
		}
		ListView_DrawText(dc, text, &rc, column.fmt);
	}

	if (drawFocus)
		DrawFocusRect(dc, &rcFocus);
}

//==================================================================================================================================================================
// TreeView

void* brownie::TreeView_GetItemData(HWND hwnd, HTREEITEM item) throw()
{
	if (!item)
		return null;
	TVITEM tvi = { TVIF_PARAM };
	tvi.hItem = item;
	if (::SendMessage(hwnd, TVM_GETITEM, 0, (LPARAM)&tvi))
		return (void*)tvi.lParam;
	return null;
}

HTREEITEM brownie::TreeView_Insert(HWND hwnd, HTREEITEM parent, const TVITEM& item)
{
	TVINSERTSTRUCT tvis = { parent };
	tvis.item = item;
	return (HTREEITEM)::SendMessage(hwnd, TVM_INSERTITEM, 0, (LPARAM)&tvis);
}

HTREEITEM brownie::TreeView_Insert(HWND hwnd, HTREEITEM parent, void* data, PCWSTR name, int icon) throw()
{
	TVITEM item = { TVIF_PARAM | TVIF_CHILDREN | TVIF_TEXT | TVIF_IMAGE | TVIF_SELECTEDIMAGE };
	item.lParam    = (LPARAM)data;
	item.cChildren = I_CHILDRENCALLBACK;
	item.pszText   = const_cast<PWSTR>(name);
	item.iImage = item.iSelectedImage = icon;
	return TreeView_Insert(hwnd, parent, item);
}

HTREEITEM brownie::TreeView_SelectItemByPos(HWND hwnd, int x, int y)
{
	TVHITTESTINFO hit = { x, y };
	if (HTREEITEM hItem = TreeView_HitTest(hwnd, &hit))
	{
		switch (hit.flags)
		{
		case TVHT_ONITEM:
		case TVHT_ONITEMICON:
		case TVHT_ONITEMLABEL:
		case TVHT_ONITEMSTATEICON:
			TreeView_SelectItem(hwnd, hItem);
			return hItem;
		}
	}
	return null;
}

//==================================================================================================================================================================
// ToolBar

void brownie::ToolBar_InsertItem(HWND hwnd, int image, BYTE style, BYTE state, INT_PTR text)
{
	TBBUTTON btn = { 0 };
	btn.iBitmap = image;
	btn.fsStyle = style;
	btn.fsState = state;
	btn.iString = text;
	::SendMessage(hwnd, TB_INSERTBUTTON, 0, (LPARAM)&btn);
}

int brownie::ToolBar_GetItemCount(HWND hwnd)
{
	return (int)::SendMessage(hwnd, TB_BUTTONCOUNT, 0, 0);
}

UINT brownie::ToolBar_GetItemState(HWND hwnd, int nID) throw()
{
	return (UINT)::SendMessage(hwnd, TB_GETSTATE, nID, 0);
}

bool brownie::ToolBar_SetItemState(HWND hwnd, int nID, UINT state) throw()
{
	return 0 != ::SendMessage(hwnd, TB_SETSTATE, nID, state);
}

int brownie::ToolBar_GetHotItem(HWND hwnd) throw()
{
	return (int)::SendMessage(hwnd, TB_GETHOTITEM, 0, 0);
}

int brownie::ToolBar_SetHotItem(HWND hwnd, int index) throw()
{
	return (int)::SendMessage(hwnd, TB_SETHOTITEM, index, 0);
}

bool brownie::ToolBar_GetItemRect(HWND hwnd, int index, RECT* rc) throw()
{
	return ::SendMessage(hwnd, TB_GETITEMRECT, index, (LPARAM)rc) != 0;
}

SIZE brownie::ToolBar_GetItemSize(HWND hwnd)
{
	LRESULT lResult = ::SendMessage(hwnd, TB_GETBUTTONSIZE, 0, 0);
	SIZE sz = { GET_XY_LPARAM(lResult) };
	return sz;
}

SIZE brownie::ToolBar_GetDefaultSize(HWND hwnd)
{
	RECT rc = { 0 };
	int count = ToolBar_GetItemCount(hwnd);
	::SendMessage(hwnd, TB_GETITEMRECT, count-1, (LPARAM)&rc);
	SIZE sz = { rc.right, rc.bottom };
	return sz;
}

void brownie::ToolBar_AutoSize(HWND hwnd)
{
	DWORD style = (DWORD)::GetWindowLong(hwnd, GWL_STYLE);
	if (style & CCS_NORESIZE)
	{
		SIZE szBtn = ToolBar_GetItemSize(hwnd);
		szBtn.cy += TOOLBAR_MARGIN;
		SIZE szCur = GetClientSize(hwnd);
		if (style & CCS_VERT)
			szCur.cx = math::max(szCur.cx, szBtn.cx);
		else
			szCur.cy = math::max(szCur.cy, szBtn.cy);
		SetClientSize(hwnd, SIZE_WH(szCur));
	}
	else
	{
		::SendMessage(hwnd, TB_AUTOSIZE, 0, 0);
	}
}

int brownie::ToolBar_HitTest(HWND hwnd, POINT pt) throw()
{
	return (int)::SendMessage(hwnd, TB_HITTEST, 0, (LPARAM)&pt);
}

int brownie::ToolBar_HitTest(HWND hwnd, int x, int y) throw()
{
	POINT pt = { x, y };
	return ToolBar_HitTest(hwnd, pt);
}

void brownie::ToolBar_InsertItems(HWND hwnd, size_t count, const TBBUTTON buttons[]) throw()
{
	::SendMessage(hwnd, TB_ADDBUTTONS, count, (LPARAM)buttons);
}

//==================================================================================================================================================================
// ReBar

ReBar_Band::ReBar_Band(DWORD rbbim)
{
	memset(this, 0, sizeof(REBARBANDINFO));
	if (WINDOWS_VERSION < WINDOWS_VISTA)
		cbSize = offsetof(REBARBANDINFO, rcChevronLocation);
	else
		cbSize = sizeof(REBARBANDINFO);
	fMask = rbbim;
}

BOOL brownie::ReBar_GetBarInfo(HWND hwnd, REBARINFO* bar)
{
	return (BOOL)::SendMessage(hwnd, RB_GETBARINFO, 0, (LPARAM)bar);
}

void brownie::ReBar_SetBarInfo(HWND hwnd, const REBARINFO* bar)
{
	VERIFY(::SendMessage(hwnd, RB_SETBARINFO, 0, (LPARAM)bar));
}

void brownie::ReBar_ShowBand(HWND hwnd, int index, bool show)
{
	VERIFY(::SendMessage(hwnd, RB_SHOWBAND, index, show));
}

int brownie::ReBar_GetBandCount(HWND hwnd)
{
	return (int)::SendMessage(hwnd, RB_GETBANDCOUNT, 0, 0L);
}

bool brownie::ReBar_GetBandInfo(HWND hwnd, int index, REBARBANDINFO* band)
{
	return 0 != ::SendMessage(hwnd, RB_GETBANDINFO, index, (LPARAM)band);
}

void brownie::ReBar_SetBandInfo(HWND hwnd, int index, const REBARBANDINFO* band)
{
	VERIFY(::SendMessage(hwnd, RB_SETBANDINFO, index, (LPARAM)band));
}

int brownie::ReBar_FindBand(HWND hwnd, HWND target)
{
	ReBar_Band band(RBBIM_CHILD);
	int count = ReBar_GetBandCount(hwnd);
	for (int i = 0; i < count; ++i)
	{
		if (ReBar_GetBandInfo(hwnd, i, &band) && band.hwndChild == target)
			return i;
	}
	return -1;
}

void brownie::ReBar_InsertBand(HWND hwnd, const REBARBANDINFO* band, int index)
{
	VERIFY(::SendMessage(hwnd, RB_INSERTBAND, index, (LPARAM)band));
}

void brownie::ReBar_InsertBand(HWND hwnd, HWND child, SIZE size, int index)
{
	bool locked = ReBar_GetLocked(hwnd);
	ReBar_Band band(RBBIM_STYLE | RBBIM_CHILD | RBBIM_IDEALSIZE | RBBIM_CHILDSIZE | RBBIM_SIZE);
	band.fStyle = RBBS_BREAK | (locked ? RBBS_NOGRIPPER : RBBS_GRIPPERALWAYS);
	band.hwndChild  = child;
	band.cxIdeal    = size.cx;
	band.cxMinChild = 0;
	band.cyMinChild = size.cy;
	band.cx         = size.cx;
	ReBar_InsertBand(hwnd, &band, index);
}

bool brownie::ReBar_DeleteBand(HWND hwnd, int index)
{
	return 0 != ::SendMessage(hwnd, RB_DELETEBAND, index, 0L);
}

bool brownie::ReBar_DeleteBand(HWND hwnd, HWND child)
{
	int index = ReBar_FindBand(hwnd, child);
	if (index < 0)
		return false;
	return ReBar_DeleteBand(hwnd, index);
}

bool brownie::ReBar_SyncBands(HWND hwnd)
{
	bool changed = false;
	ReBar_Band band(RBBIM_CHILD | RBBIM_STYLE);
	const int count = ReBar_GetBandCount(hwnd);
	for (int i = 0; i < count; ++i)
	{
		if (!ReBar_GetBandInfo(hwnd, i, &band))
			continue;
		bool visible = IsVisible(band.hwndChild);
		if (visible != band.Visible)
		{
			changed = true;
			ReBar_ShowBand(hwnd, i, visible);
		}
	}
	return changed;
}

const bool ReBar_DefaultLocked = true;

bool brownie::ReBar_GetLocked(HWND hwnd)
{
	ReBar_Band band(RBBIM_STYLE);
	if (ReBar_GetBandInfo(hwnd, 0, &band))
		return band.Locked;
	return ReBar_DefaultLocked;
}

void brownie::ReBar_SetLocked(HWND hwnd, bool value)
{
	// X^C̕ύXł́ǍԂ𒲐ĂȂB
	// ̂߁A폜ǍŒǉȂB
	int count = ReBar_GetBandCount(hwnd);
	if (count == 0)
		return;

	typedef std::vector<ReBar_Band> Bands;
	Bands bands;
	bands.reserve(count);
	LockRedraw pause(hwnd);
	for (int i = 0; i < count; i++)
	{
		ReBar_Band band(RBBIM_STYLE | RBBIM_CHILD | RBBIM_IDEALSIZE | RBBIM_CHILDSIZE | RBBIM_SIZE);
		if (ReBar_GetBandInfo(hwnd, 0, &band))
		{
			if (bands.empty() && band.Locked == value)
				return; // ύX̕KvȂ
			band.Locked = value;
			bands.push_back(band);
		}
		ReBar_DeleteBand(hwnd, 0);
	}
	// ǉȂ
	ASSERT(ReBar_GetBandCount(hwnd) == 0);
	count = (int)bands.size();
	for (int i = 0; i < count; i++)
		ReBar_InsertBand(hwnd, &bands[i], i);
}

//==================================================================================================================================================================
// ToolTip

int brownie::ToolTip_GetCount(HWND hwnd) throw()
{
	return (int)::SendMessage(hwnd, TTM_GETTOOLCOUNT, 0, 0L);
}

void brownie::ToolTip_Add(HWND hwnd, const TOOLINFO* info) throw()
{
	TOOLINFO query = { sizeof(TOOLINFO) };
	query.uId = info->uId;
	query.hwnd = info->hwnd;
	if (::SendMessage(hwnd, TTM_GETTOOLINFO, 0, (LPARAM)&query))
		::SendMessage(hwnd, TTM_SETTOOLINFO, 0, (LPARAM)info);
	else
		VERIFY(::SendMessage(hwnd, TTM_ADDTOOL, 0, (LPARAM)info) );
}

void brownie::ToolTip_Add(HWND hwnd, HWND owner, UINT nID, PCWSTR text, const RECT& rc) throw()
{
	TOOLINFO info = { sizeof(TOOLINFO) };
	info.uId = nID;
	info.uFlags = TTF_SUBCLASS;
	info.hwnd = owner;
	info.lpszText = const_cast<PWSTR>(text);
	info.rect = rc;
	ToolTip_Add(hwnd, &info);
}

void brownie::ToolTip_Remove(HWND hwnd, HWND owner, UINT nID) throw()
{
	TOOLINFO info = { sizeof(TOOLINFO) };
	info.uId = nID;
	info.hwnd = owner;
	::SendMessage(hwnd, TTM_DELTOOL, 0, (LPARAM)&info);
}

//==================================================================================================================================================================
// Header

int brownie::Header_HitTest(HWND hwnd, int x, int y)
{
	HDHITTESTINFO hit = { x, y };
	return (int)::SendMessage(hwnd, HDM_HITTEST, 0, (LPARAM)&hit);
}

HRESULT brownie::Header_GetSortedColumn(HWND hwnd, int* column, bool* ascending) throw()
{
	HDITEM item = { HDI_FORMAT };
	int n = Header_GetItemCount(hwnd);
	for (int i = 0; i < n; ++i)
	{
		if (Header_GetItem(hwnd, i, &item) && (item.fmt & (HDF_SORTUP | HDF_SORTDOWN)))
		{
			*column = i;
			*ascending = ((item.fmt & HDF_SORTDOWN) == HDF_SORTDOWN);
			return S_OK;
		}
	}
	return E_FAIL;
}

HRESULT brownie::Header_SetSortedColumn(HWND hwnd, int column, bool ascending) throw()
{
	HRESULT hr = E_FAIL;
	HDITEM item = { HDI_FORMAT };
	int n = Header_GetItemCount(hwnd);
	for (int i = 0; i < n; ++i)
	{
		if (!Header_GetItem(hwnd, i, &item))
			continue;
		if (i == column)
		{
			item.fmt &= ~(HDF_SORTUP | HDF_SORTDOWN);
			item.fmt |= (ascending ? HDF_SORTDOWN : HDF_SORTUP);
			Header_SetItem(hwnd, i, &item);
			hr = S_OK;
		}
		else if (item.fmt & (HDF_SORTUP | HDF_SORTDOWN))
		{
			item.fmt &= ~(HDF_SORTUP | HDF_SORTDOWN);
			Header_SetItem(hwnd, i, &item);
		}
	}
	return hr;
}

//==================================================================================================================================================================
// StatusBar

bool brownie::StatusBar_GetItemRect(HWND hwnd, int part, RECT* rc)
{
	std::vector<int> widths(part+1);
	int n = (int)::SendMessage(hwnd, SB_GETPARTS, part, (LPARAM)&widths[0]);
	if (n < part)
		return false;
	GetClientRect(hwnd, rc);
	rc->left = (part > 0 ? widths[part-1] : 0);
	if (widths[part] != -1)
		rc->right = widths[part];
	return true;
}

int brownie::StatusBar_HitTest(HWND hwnd, int x) throw()
{
	if (int n = (int)::SendMessage(hwnd, SB_GETPARTS, 0, 0))
	{
		std::vector<int> widths(n);
		::SendMessage(hwnd, SB_GETPARTS, n, (LPARAM)&widths[0]);
		for (int i = 0; i < n; ++i)
		{
			if (widths[i] == -1 || x < widths[i])
				return i;
		}
	}
	return -1;
}
