// config.cpp
#pragma once

#include "stdafx.h"
#include "config.hpp"
#include "system.hpp"

using namespace brownie;

namespace
{
	const WCHAR RELPATH_BIN[]			= /* PATH_ROOT */ L"bin";
	const WCHAR	RELPATH_USER[]			= /* PATH_ROOT */ L"usr";
	const WCHAR	RELPATH_FAVORITE[]		= /* PATH_USER */ L"favorite";
	const WCHAR	RELPATH_TOOLS[]			= /* PATH_ROOT */ L"tools";
	const WCHAR	RELPATH_FOLDER_STATE[]	= /* PATH_USER */ L"settings.dat";
}

int			brownie::WINDOWS_VERSION = 0;
WCHAR		brownie::PATH_ROOT[MAX_PATH];
WCHAR		brownie::PATH_BIN[MAX_PATH];
WCHAR		brownie::PATH_USER[MAX_PATH];
WCHAR		brownie::PATH_FAVORITE[MAX_PATH];
WCHAR		brownie::PATH_TOOLS[MAX_PATH];
WCHAR		brownie::PATH_FOLDER_STATE[MAX_PATH];
WCHAR		brownie::PATH_SOUND_NAVIGATION[MAX_PATH] = L"AppEvents\\Schemes\\Apps\\Explorer\\Navigating\\.Current";
int			brownie::GESTURE_NAPTIME = 100; // [ms] ȏ㎞Ԃoꍇ͐VȃWFX`Ƃ݂ȂiȂǁj
int			brownie::GESTURE_TIMEOUT = 500; // [ms] ȏ~ĂƁAWFX`LZB
// XXX: ZƁA܂ɃhbOȂBWindowsŃ^CAEgp[^Ă̂H
int			brownie::GESTURE_MAXLEN = 1;
PyObject*	brownie::EDITOR;
COLORREF	brownie::LISTVIEW_BACKGROUND = ::GetSysColor(COLOR_WINDOW);
int			brownie::LISTVIEW_AUTOADJUST = 200;
bool		brownie::LISTVIEW_AUTOUNIT = false;
bool		brownie::LISTVIEW_LOOP = true;
bool		brownie::LISTVIEW_WHEELSELECT = false;
bool		brownie::RENAME_EXTENSION = true;
int			brownie::TAB_MAXLEN = 200;

namespace
{
#define CONFIG_BOOL(name)												\
	object	get_##name()			{ return Boolean::newobj(name); }	\
	void	set_##name(PyObject* o)	{ name = Boolean::value(o); }

#define CONFIG_CONST_INT(name)											\
	object	get_##name()			{ return Integer::newobj(name); }

#define CONFIG_INT(name)												\
	CONFIG_CONST_INT(name)												\
	void	set_##name(PyObject* o)	{ name = cast<int>(o); }

#define CONFIG_ENUM(type, name)											\
	CONFIG_CONST_INT(name)												\
	void	set_##name(PyObject* o)	{ name = (type)cast<int>(o); }

#define CONFIG_CONST_STR(name)											\
	object	get_##name()			{ return Unicode::newobj(name); }

#define CONFIG_STR(name)												\
	object	get_##name()			{ return Unicode::newobj(name); }	\
	void	set_##name(PyObject* o)	{ wcscpy_s(name, str(Unicode::newobj(o))); }

#define CONFIG_OBJECT(name)												\
	object	get_##name()			{ return object(name); }			\
	void	set_##name(PyObject* o)	{ incref(o); decref(name); name = o; }

#define CONFIG_COLORREF(name)											\
	object	get_##name()			{ return FromCOLORREF(name); }		\
	void	set_##name(PyObject* o)	{ name = ToCOLORREF(o); }

	object FromCOLORREF(COLORREF c)
	{
		return build("(iii)", GetRValue(c), GetGValue(c), GetBValue(c));
	}

	COLORREF ToCOLORREF(PyObject* o)
	{
		if (is<Null>(o))
			return ::GetSysColor(COLOR_WINDOW);
		int r = getitem<int>(o, 0);
		int g = getitem<int>(o, 1);
		int b = getitem<int>(o, 2);
		return RGB(r, g, b);
	}

	class Config : public Object
	{
	public:
		~Config();

	public: // python
		static PCSTR __name__() throw()	{ return "brownie.Config"; }
		static PyGetSetDef* __member__() throw();

		CONFIG_CONST_INT(VERSION);
		CONFIG_CONST_INT(WINDOWS_VERSION);
		CONFIG_CONST_STR(PATH_ROOT);
		CONFIG_CONST_STR(PATH_BIN);
		CONFIG_CONST_STR(PATH_USER);
		CONFIG_CONST_STR(PATH_FAVORITE);
		CONFIG_CONST_STR(PATH_FOLDER_STATE);
		CONFIG_CONST_STR(PATH_TOOLS);
		CONFIG_STR(PATH_SOUND_NAVIGATION);

		CONFIG_OBJECT(EDITOR)
		CONFIG_INT(GESTURE_MAXLEN)
		CONFIG_INT(LISTVIEW_AUTOADJUST)
		CONFIG_BOOL(LISTVIEW_AUTOUNIT)
		CONFIG_COLORREF(LISTVIEW_BACKGROUND)
		CONFIG_BOOL(LISTVIEW_LOOP)
		CONFIG_BOOL(LISTVIEW_WHEELSELECT)
		CONFIG_BOOL(RENAME_EXTENSION)
		CONFIG_INT(TAB_MAXLEN)

		object	get_KEYMAP()			{ return System::get_keymap(); }
		void	set_KEYMAP(PyObject* o)	{ System::set_keymap(o); }
	};
}

PYTHON_MEMBERS(Config)
	PYTHON_MEMBER(VERSION				, GET )
	PYTHON_MEMBER(WINDOWS_VERSION		, GET )
	PYTHON_MEMBER(PATH_ROOT				, GET )
	PYTHON_MEMBER(PATH_BIN				, GET )
	PYTHON_MEMBER(PATH_USER				, GET )
	PYTHON_MEMBER(PATH_FAVORITE			, GET )
	PYTHON_MEMBER(PATH_FOLDER_STATE		, GET )
	PYTHON_MEMBER(PATH_TOOLS			, GET )
	PYTHON_MEMBER(PATH_SOUND_NAVIGATION	, BOTH)
	PYTHON_MEMBER(GESTURE_MAXLEN		, BOTH)
	PYTHON_MEMBER(EDITOR				, BOTH)
	PYTHON_MEMBER(LISTVIEW_AUTOADJUST	, BOTH)
	PYTHON_MEMBER(LISTVIEW_AUTOUNIT		, BOTH)
	PYTHON_MEMBER(LISTVIEW_BACKGROUND	, BOTH)
	PYTHON_MEMBER(LISTVIEW_LOOP			, BOTH)
	PYTHON_MEMBER(LISTVIEW_WHEELSELECT	, BOTH)
	PYTHON_MEMBER(KEYMAP				, BOTH)
	PYTHON_MEMBER(RENAME_EXTENSION		, BOTH)
	PYTHON_MEMBER(TAB_MAXLEN			, BOTH)
PYTHON_END()

void brownie::ConfigInit()
{
	// init WINDOWS_VERSION
	OSVERSIONINFOEX info = { sizeof(OSVERSIONINFOEX) };
	::GetVersionEx((OSVERSIONINFO*)&info);
	WINDOWS_VERSION = ((info.dwMajorVersion    << 24) +
					   (info.dwMinorVersion    << 16) +
					   (info.wServicePackMajor <<  8) +
					   (info.wServicePackMinor));

	// - PATH_ROOT
	//     - PATH_BIN
	//     - PATH_USER
	//         - PATH_FAVORITE
	//         - PATH_FOLDER_STATE
	//     - PATH_TOOLS

	::GetModuleFileName(null, PATH_ROOT, MAX_PATH);
	::PathRemoveFileSpec(PATH_ROOT);

	::PathCombine(PATH_BIN			, PATH_ROOT, RELPATH_BIN);
	::PathCombine(PATH_USER			, PATH_ROOT, RELPATH_USER);
	::PathCombine(PATH_FAVORITE		, PATH_USER, RELPATH_FAVORITE);
	::PathCombine(PATH_FOLDER_STATE	, PATH_USER, RELPATH_FOLDER_STATE);
	::PathCombine(PATH_TOOLS		, PATH_ROOT, RELPATH_TOOLS);

	::SetCurrentDirectory(PATH_ROOT);
}

#include <shellapi.h>

namespace
{
	class Trash : public Object
	{
	public: // python
		static PCSTR __name__() throw()	{ return "brownie.Trash"; }
		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();

		void clear()
		{
			py::Unlock unlock;
			SHEmptyRecycleBin(System::GetWindow(), null, 0);
		}
		object get_format()
		{
			WCHAR buf[32] = L"";
			{
				py::Unlock unlock;
				SHQUERYRBINFO info = { sizeof(SHQUERYRBINFO) };
				if (SUCCEEDED(SHQueryRecycleBin(null, &info)) && info.i64Size > 0)
					StrFormatByteSize64(info.i64Size, buf, lengthof(buf));
			}
			return Unicode::newobj(buf);
		}
		object get_size()
		{
			SHQUERYRBINFO info = { sizeof(SHQUERYRBINFO) };
			{
				py::Unlock unlock;
				SHQueryRecycleBin(null, &info);
			}
			return Numeric::newobj(info.i64Size);
		}
	};
}

PYTHON_METHODS(Trash)
	PYTHON_METHOD(clear)
PYTHON_END()

PYTHON_MEMBERS(Trash)
	PYTHON_MEMBER(format	, GET)
	PYTHON_MEMBER(size		, GET)
PYTHON_END()

void brownie::config_export(Module* m)
{
	PyType_Ready(typeof<Config>());
	m->def("config", newobj<Config>()->self);

	PyType_Ready(typeof<Trash>());
	m->def("trash", newobj<Trash>()->self);
}

//==================================================================================================================================================================

Config::~Config()
{
	System::set_keymap(null);
	decref(EDITOR);
}
