// dialog.cpp

#include "stdafx.h"
#include "dialog.hpp"
#include "system.hpp"
#include "path.hpp"

using namespace brownie;

//==================================================================================================================================================================

LRESULT Dialog::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_ENDSESSION:
		if (!wParam)
			break;
		// fall through
	case WM_CLOSE:
		::EnableWindow(::GetAncestor(m_owner, GA_ROOT), true);
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

bool Dialog::OnPress(const MSG& msg) throw()
{
	switch (msg.message)
	{
	case WM_KEYDOWN:
		switch (msg.wParam)
		{
		case VK_RETURN:
			if (m_result)
			{
				py::Lock lock;
				m_value = null;
				std::vector<object> value;

				foreach (object o, m_result)
					value.push_back(parse_result(o));
				else
					m_value = parse_result(m_result);

				if (!m_value)
				{
					tuple t = Tuple::newobj(value.size());
					for (size_t i = 0; i < value.size(); ++i)
						t->attach(i, value[i].detach());
					value.clear();
					m_value = t;
				}
			}
			Close();
			break;
		case VK_ESCAPE:
			Close();
			break;
		}
		break;
	}
	return __super::OnPress(msg);
}

void Dialog::QueryLayout(ILayout& layout) const
{
	SIZE sz = { DIALOG_MARGIN, DIALOG_MARGIN };
	layout.Margin = sz;
	__super::QueryLayout(layout);
}

object Dialog::parse_result(PyObject* o)
{
	return cast<Window>(o)->get_value();
}

void Dialog::__init__(Tuple* args, Dict* kw) throw(...)
{
	m_owner = __init__parse(args, kw, CENTER);
	if (!m_owner)
		m_owner = System::GetWindow();
	create(DS_MODALFRAME | WS_POPUP | WS_THICKFRAME | WS_CLIPCHILDREN | WS_CAPTION | WS_SYSMENU,
		WS_EX_TOOLWINDOW | WS_EX_DLGMODALFRAME | WS_EX_WINDOWEDGE | WS_EX_CONTROLPARENT, 0, ::GetAncestor(m_owner, GA_ROOT));
	if (HMENU sysmenu = ::GetSystemMenu(m_hwnd, false))
	{
		::RemoveMenu(sysmenu, SC_MAXIMIZE, MF_BYCOMMAND);
		::RemoveMenu(sysmenu, SC_MINIMIZE, MF_BYCOMMAND);
		::RemoveMenu(sysmenu, SC_RESTORE, MF_BYCOMMAND);
	}
}

object Dialog::run()
{
	if (m_hwnd)
	{
		HWND	root = ::GetAncestor(m_owner, GA_ROOT);

		py::Unlock unlock;

		// ̏Ԃł͂܂\
		CenterWindow(m_hwnd, m_owner);
		if (m_owner)
			::EnableWindow(root, false);
		set_Visible(true);
		SetFocus();

		while (IsWindow())
		{
			PumpMessage();
		}

		if (m_owner)
		{	// WM_CLOSEŏς̂͂AÔ߁B
			::EnableWindow(root, true);
			::SetActiveWindow(root);
			::SetFocus(m_owner);
		}
	}

	return m_value;
}

PYTHON_MEMBERS(Dialog)
	PYTHON_MEMBER(value	, VAR)
	PYTHON_MEMBER(result, VAR)
PYTHON_END()

PYTHON_METHODS(Dialog)
	PYTHON_METHOD(run)
PYTHON_END()

//==================================================================================================================================================================

HRESULT FileDialog::Run(IShellItem** pp)
{
	if (!pp)
		return E_POINTER;
	*pp = null;

	HRESULT hr;

	ref<IFileDialog>	dlg;
	if FAILED(hr = dlg.newobj(CLSID_FileOpenDialog))
		return hr;

	ref<IFileDialogEvents>	events;
	DWORD					cookie = 0;
//	FileDialogEvents_CreateInstance(&events);
	dlg->Advise(events, &cookie);

/*
	DWORD dwFlags = 0;
	hr = dlg->GetOptions(&dwFlags);
	hr = dlg->SetOptions(dwFlags | FOS_FORCEFILESYSTEM);
	hr = dlg->SetFileTypes(ARRAYSIZE(c_rgSaveTypes), c_rgSaveTypes);
	hr = dlg->SetFileTypeIndex(INDEX_BROWNIE);
	hr = dlg->SetDefaultExtension(L"brownie");
*/

	hr = dlg->Show(System::GetWindow());
	if (hr == S_OK)
		hr = dlg->GetResult(pp);

	if (cookie)
		dlg->Unadvise(cookie);

	return hr;
}

object FileDialog::run()
{
	HRESULT hr;
	ref<IShellItem>	item;

	{
		py::Unlock unlock;
		hr = Run(&item);
	}

	if FAILED(hr)
		throw WindowsError("CLSID_FileOpenDialog not found");

	return item ? Path::newobj(item) : null;
}

PYTHON_METHODS(FileDialog)
	PYTHON_METHOD(run)
PYTHON_END()
