// dragdrop.cpp

#include "stdafx.h"
#include "dragdrop.hpp"

using namespace brownie;

DropTarget::DropTarget() : m_owner(null)
{
}

DropTarget::~DropTarget()
{
	Revoke();
}

void DropTarget::Register(Window* w)
{
	ASSERT(!m_owner);
	if SUCCEEDED(::RegisterDragDrop(w->Handle, this))
		m_owner = w;
}

void DropTarget::Revoke()
{
	if (m_owner)
	{
		::RevokeDragDrop(m_owner->Handle);
		m_owner = null;
	}
}

STDMETHODIMP DropTarget::DragEnter(IDataObject* data, DWORD keys, POINTL pt, DWORD* effect)
{
	ASSERT(!m_data);
	ASSERT(m_owner);

	if (!m_owner)
		return E_UNEXPECTED;

	m_data = data;
	SetRectEmpty(&m_bounds);
	if (!m_helper)
		m_helper.newobj(CLSID_DragDropHelper); // XXX: XbhɂЂƂŏ\H
	m_helper->DragEnter(m_owner->Handle, data, (POINT*)&pt, *effect);
	return S_OK;
}

STDMETHODIMP DropTarget::DragOver(DWORD keys, POINTL pt, DWORD* effect)
{
	ASSERT(m_data);
	ASSERT(m_owner);

	if (!m_data || !m_owner)
	{
		if (effect)
			*effect = DROPEFFECT_NONE;
		return E_UNEXPECTED;
	}

	HRESULT hr = S_OK;
	POINT ptClient = { pt.x, pt.y };
	m_owner->ScreenToClient(&ptClient);
	if (!PtInRect(&m_bounds, ptClient))
	{
		ref<IDropTarget> target;
		int dx = ::GetSystemMetrics(SM_CXDRAG);
		int dy = ::GetSystemMetrics(SM_CYDRAG);
		SetRect(&m_bounds, ptClient.x - dx, ptClient.y - dy, ptClient.x + dx, ptClient.y + dy);
		m_owner->OnDragDrop(ptClient, &m_bounds, &target);
		if (m_target != target)
		{
			if (m_target)
			{	// ؂ւ̂ŁAȑOtargetB
				m_target->DragLeave();
				m_target = null;
			}
			if (target)
			{
				DWORD e = DROPEFFECT_COPY | DROPEFFECT_MOVE | DROPEFFECT_LINK;
				if SUCCEEDED(target->DragEnter(m_data, keys, pt, &e))
					m_target = target; // Enter󂯓ꂽꍇ́Aȍ~target֑B
			}
		}
	}

	if (m_target)
		hr = m_target->DragOver(keys, pt, effect);
	else if (effect)
		*effect = DROPEFFECT_NONE;
	m_helper->DragOver((POINT*)&pt, *effect);

	return hr;
}

STDMETHODIMP DropTarget::DragLeave()
{
	HRESULT hr = S_OK;
	if (m_target)
	{
		hr = m_target->DragLeave();
		m_target = null;
	}
	m_data = null;
	m_helper->DragLeave();
	return hr;
}

STDMETHODIMP DropTarget::Drop(IDataObject* data, DWORD keys, POINTL pt, DWORD* effect)
{
	HRESULT hr = S_OK;
	if (m_target)
	{
		hr = m_target->Drop(data, keys, pt, effect);
		m_target = null;
	}
	else if (effect)
		*effect = DROPEFFECT_NONE;
	m_data = null;
	m_helper->Drop(data, (POINT*)&pt, *effect);
	return hr;
}
