// fileop.cpp

#include "stdafx.h"
#include "shell.hpp"
#include "fileop.hpp"
#include "stream.hpp"
#include "system.hpp"
#include "xp/xp_class.hpp"

using namespace brownie;

namespace
{
	ref<IShellItemArray> ShellItemArrayCreate(size_t num, PCWSTR items[])
	{
		return null;
	}
}

//==================================================================================================================================================================
// File operations

HRESULT brownie::FileRename(IFileOperation* op, IShellItem* item, PCWSTR name)
{
	return op->RenameItem(item, name, null);
}

HRESULT brownie::FileRename(IFileOperation* op, IUnknown* items, PCWSTR name)
{
	return op->RenameItems(items, name);
}

HRESULT brownie::FileRename(IFileOperation* op, PCWSTR item, PCWSTR name)
{
	HRESULT hr;
	ref<IFileOperationXp> xp;
	if SUCCEEDED(op->QueryInterface(&xp))
		return xp->RenameItem(item, name);
	ref<IShellItem> path;
	if FAILED(hr = ItemCreate(&path, item))
		return hr;
	return FileRename(op, path, name);
}

HRESULT brownie::FileRename(IFileOperation* op, size_t num, PCWSTR items[], PCWSTR name);

HRESULT brownie::FileMove(IFileOperation* op, IShellItem* item, IShellItem* folder, PCWSTR name)
{
	return op->MoveItem(item, folder, name, null);
}

HRESULT brownie::FileMove(IFileOperation* op, IUnknown* items, IShellItem* folder)
{
	return op->MoveItems(items, folder);
}

namespace
{
	HRESULT CreateShellItems(PCWSTR item, PCWSTR path, IShellItem** src, IShellItem** dir, PCWSTR* name)
	{
		HRESULT hr;
		if FAILED(hr = ItemCreate(src, item))
			return hr;
		*name = PathFindFileName(path);
		WCHAR folder[MAX_PATH];
		wcsncpy_s(folder, path, *name - path);
		return ItemCreate(dir, folder);
	}
}

HRESULT brownie::FileMove(IFileOperation* op, PCWSTR item, PCWSTR path)
{
	HRESULT hr;
	ref<IFileOperationXp> xp;
	if SUCCEEDED(op->QueryInterface(&xp))
		return xp->MoveItem(item, path);
	ref<IShellItem> src, dir;
	PCWSTR name;
	if FAILED(hr = CreateShellItems(item, path, &src, &dir, &name))
		return hr;
	return FileMove(op, src, dir, name);
}

HRESULT brownie::FileMove(IFileOperation* op, size_t num, PCWSTR items[], PCWSTR folder);

HRESULT brownie::FileCopy(IFileOperation* op, IShellItem* item, IShellItem* folder, PCWSTR name)
{
	return op->CopyItem(item, folder, name, null);
}

HRESULT brownie::FileCopy(IFileOperation* op, IUnknown* items, IShellItem* folder)
{
	return op->CopyItems(items, folder);
}

HRESULT brownie::FileCopy(IFileOperation* op, PCWSTR item, PCWSTR path)
{
	HRESULT hr;
	ref<IFileOperationXp> xp;
	if SUCCEEDED(op->QueryInterface(&xp))
		return xp->CopyItem(item, path);
	ref<IShellItem> src, dir;
	PCWSTR name;
	if FAILED(hr = CreateShellItems(item, path, &src, &dir, &name))
		return hr;
	return FileCopy(op, src, dir, name);
}

HRESULT brownie::FileCopy(IFileOperation* op, size_t num, PCWSTR items[], PCWSTR folder);

HRESULT brownie::FileDelete(IFileOperation* op, IShellItem* item)
{
	return op->DeleteItem(item, null);
}

HRESULT brownie::FileDelete(IFileOperation* op, IUnknown* items)
{
	return op->DeleteItems(items);
}

HRESULT brownie::FileDelete(IFileOperation* op, PCWSTR item)
{
	HRESULT hr;
	ref<IFileOperationXp> xp;
	if SUCCEEDED(op->QueryInterface(&xp))
		return xp->DeleteItem(item);
	ref<IShellItem> path;
	if FAILED(hr = ItemCreate(&path, item))
		return hr;
	return FileDelete(op, path);
}

HRESULT brownie::FileDelete(IFileOperation* op, size_t num, PCWSTR items[])
{
	ref<IFileOperationXp> xp;
	if SUCCEEDED(op->QueryInterface(&xp))
		return xp->DeleteItems(num, items);
	else if (ref<IShellItemArray> arr = ShellItemArrayCreate(num, items))
		return op->DeleteItems(arr);
	else
		return E_FAIL;
}


HRESULT brownie::FileNew(IFileOperation* op, IShellItem* folder, PCWSTR name) throw()
{
	// (folder, attr, name, pszTemplateName, sink);
	return op->NewItem(folder, 0, name, null, null);
}

HRESULT brownie::FileNew(IFileOperation* op, PCWSTR path) throw()
{
	HRESULT hr;
	ref<IFileOperationXp> xp;
	if SUCCEEDED(op->QueryInterface(&xp))
		return xp->NewItem(path);

	ref<IShellItem> item;
	PCWSTR name = PathFindFileName(path);
	WCHAR dir[MAX_PATH];
	wcsncpy_s(dir, path, name - path);
	if FAILED(hr = ItemCreate(&item, dir))
		return hr;
	return FileNew(op, item, name);
}

//==================================================================================================================================================================
// FileOperation

void FileOperation_::rename(Tuple* args)
{
	ensure();
	m_op->RenameItem(null, L"", null);
}

void FileOperation_::move(Tuple* args)
{
	ensure();
	m_op->MoveItem(null, null, L"", null);
}

void FileOperation_::copy(Tuple* args)
{
	ensure();
	m_op->CopyItem(null, null, L"", null);
}

void FileOperation_::delete_(Tuple* args)
{
	ensure();
	m_op->DeleteItem(null, null);
}

void FileOperation_::new_(Tuple* args)
{
	ensure();
	m_op->NewItem(null, 0, L"", null, null);
}

void FileOperation_::run()
{
	if (m_op)
		m_op->PerformOperations();
}

void FileOperation_::ensure()
{
	if (!m_op)
	{
		if FAILED(m_op.newobj(CLSID_FileOperation))
			throw OSError("failed to create CLSID_FileOperation");
	}
}

//==================================================================================================================================================================

PYTHON_METHODS(FileOperation_)
	PYTHON_METHOD(rename)
	PYTHON_METHOD(move)
	PYTHON_METHOD(copy)
	PYTHON_METHOD_(delete, delete_)
	PYTHON_METHOD_(new   , new_)
	PYTHON_METHOD(run)
PYTHON_END()

//================================================================================

namespace
{
	static bool IsCopy()
	{
		static const UINT CF_PREFERREDDROPEFFECT = RegisterClipboardFormat(CFSTR_PREFERREDDROPEFFECT);
		HGLOBAL hDropEffect = ::GetClipboardData(CF_PREFERREDDROPEFFECT);
		DWORD* pdw = reinterpret_cast<DWORD*>(GlobalLock(hDropEffect));
		ASSERT(pdw);
		DWORD dwEffect = (pdw ? *pdw : DROPEFFECT_COPY);
		GlobalUnlock(hDropEffect);
		return (dwEffect & DROPEFFECT_COPY) != 0;
	}

	HRESULT FileCutOrCopy(PCTSTR srcSingle, DWORD dwDropEffect)
	{
		if(!srcSingle)
			return E_INVALIDARG;
		size_t len = lstrlen(srcSingle);
		if(len == 0)
			return E_INVALIDARG;

		//CF_HDROP쐬
		DROPFILES dropfiles = { sizeof(DROPFILES), 0, 0, true, TRUE };
		HGLOBAL hDrop;
		try
		{
			ref<IStream> stream = StreamCreate(&hDrop);
			StreamWrite(stream, &dropfiles, sizeof(DROPFILES));
			StreamWrite(stream, srcSingle , (len+1)*sizeof(WCHAR));
			StreamWrite(stream, L"\0"  , sizeof(WCHAR));
		}
		catch (HRESULT hr)
		{
			::GlobalFree(hDrop);
			return hr;
		}

		// Preferred DropEffect쐬
		HGLOBAL hDropEffect = GlobalAlloc(GMEM_MOVEABLE, sizeof(DWORD));
		DWORD* pdw = static_cast<DWORD*>(GlobalLock(hDropEffect));
		*pdw = dwDropEffect;
		GlobalUnlock(hDropEffect);

		// Nbv{[hɃf[^[Zbg
		HWND hwnd = System::GetWindow();
		UINT CF_PREFERREDDROPEFFECT = RegisterClipboardFormat(CFSTR_PREFERREDDROPEFFECT);
		if(!::OpenClipboard(hwnd))
		{
			::GlobalFree(hDrop);
			::GlobalFree(hDropEffect);
			return HRESULT_FROM_WIN32(GetLastError());
		}
		::EmptyClipboard();
		::SetClipboardData(CF_HDROP, hDrop);
		::SetClipboardData(CF_PREFERREDDROPEFFECT, hDropEffect);
		::CloseClipboard();
		return S_OK;
	}
}

/*
HRESULT brownie::FilePaste(PCTSTR dst)
{
	HWND hwnd = GetForm();
	if(::IsClipboardFormatAvailable(CF_HDROP)
	&& ::PathIsDirectory(dst)
	&& ::OpenClipboard(hwnd))
	{
		if(HDROP hDrop = (HDROP)::GetClipboardData(CF_HDROP))
		{
			bool copy = IsCopy();

			StringBuffer srcfiles;
			const UINT count = ::DragQueryFile(hDrop, 0xFFFFFFFF, NULL, 0);
			for(UINT i = 0; i < count; ++i)
			{
				WCHAR path[MAX_PATH];
				::DragQueryFile(hDrop, i, path, MAX_PATH);
				srcfiles.append(path, str::length(path)+1);
			}

			if(!copy)
				::EmptyClipboard();
			::CloseClipboard();

			if(!srcfiles.empty())
			{
				srcfiles.push_back('\0');
				return DoFileOperation(srcfiles, dst, (copy ? FO_COPY : FO_MOVE), FOF_ALLOWUNDO, hwnd);
			}
			return S_OK;
		}
		::CloseClipboard();
	}
	return HRESULT_FROM_WIN32(GetLastError());
}
*/

HRESULT brownie::FileCut(PCWSTR src)
{
	return FileCutOrCopy(src, DROPEFFECT_MOVE);
}

HRESULT brownie::FileCopy(PCWSTR src)
{
	return FileCutOrCopy(src, DROPEFFECT_COPY | DROPEFFECT_LINK);
}

//================================================================================

namespace
{
	static bool IsEqualClassName(HWND hwnd, PCTSTR clsname)
	{
		if(!hwnd)
			return false;
		WCHAR buffer[MAX_PATH] = L"";
		::GetClassName(hwnd, buffer, MAX_PATH);
		return wcscmp(buffer, clsname) == 0;
	}
	static BOOL CALLBACK EnumFileDialog(HWND hwnd, LPARAM lParam)
	{
		HWND hComboBox = NULL;
		if(!::IsWindowVisible(hwnd))
			return TRUE;
		if(IsEqualClassName(hwnd, L"#32770")
		&& IsEqualClassName(::GetDlgItem(hwnd, 0x00000461), L"SHELLDLL_DefView")
		&& IsEqualClassName(hComboBox = ::GetDlgItem(hwnd, 0x0000047C), L"ComboBoxEx32"))
		{
			// gbvx t@C_CAO
			*(HWND*)lParam = hComboBox;
			return FALSE;
		}

		// t@C_CAOł͂Ȃ̂Ŏq{
		if(HWND hwndChild = FindWindowEx(hwnd, null, L"#32770", null))
		{
			if(IsEqualClassName(::GetDlgItem(hwndChild, 0x00000461), L"SHELLDLL_DefView")
			&& IsEqualClassName(hComboBox = ::GetDlgItem(hwndChild, 0x0000047C), L"ComboBoxEx32"))
			{
				*(HWND*)lParam = hComboBox;
				return FALSE;
			}
		}
        return TRUE;
	}
	static HWND FindPathEditOfFileDialog()
	{
		HWND hwnd = null;
		EnumWindows(EnumFileDialog, (LPARAM)&hwnd);
		return hwnd;
	}
}

HRESULT brownie::FileDialogSetPath(PCWSTR path)
{
	HWND hComboBox = FindPathEditOfFileDialog();
	if(!hComboBox)
		return E_FAIL;

	WCHAR orig[MAX_PATH];
	::SendMessage(hComboBox, WM_GETTEXT, MAX_PATH, (LPARAM)orig);
	::SendMessage(hComboBox, WM_SETTEXT, 0, (LPARAM)path);
	::PostMessage(hComboBox, WM_KEYDOWN, VK_RETURN, 0);
	HWND dlg = ::GetAncestor(hComboBox, GA_ROOT);
	::BringWindowToTop(dlg);
	::SetActiveWindow(dlg);
	::SendMessage(hComboBox, WM_SETTEXT, 0, (LPARAM)orig);
	return S_OK;
}
