// folderview.cpp

#include "stdafx.h"
#include "folderview.hpp"
#include "drawing.hpp"
#include "shell.hpp"
#include "node.hpp"
#include "path.hpp"
#include "xp/xp_class.hpp"

using namespace brownie;

namespace
{
	const bool m_ShowAllFiles = false;
	const UINT WM_GETDRAGIMAGE = RegisterWindowMessage(DI_GETDRAGIMAGE);

	// XXX: CX^XƂɂЂƂAXbhƂɂЂƂ̕B
	static LVDRAWDESC theDrawDesc;

	const UINT	TIMER_NOTIFY	= 'NTFY';
	const UINT	DELAY_NOTIFY	= 50;	// ms

	const UINT	TIMER_ADJUST	= 'ADJT';
	const UINT	DELAY_ADJUST	= 200;	// ms
}

//==================================================================================================================================================================

namespace
{
	enum SHIDM
	{
		SHIDM_GROUPING	= 30209,
		SHIDM_NAME		= 30210, // ȉ͂̃JŃ\[g
		SHIDM_SIZE		= 30211,
		SHIDM_TYPE		= 30212,
		SHIDM_MTIME		= 30213,
		SHIDM_CTIME		= 30214,
		SHIDM_ATIME		= 30215,
	};

	static inline bool OwnerDrawEnabled() throw()
	{
		return LISTVIEW_BACKGROUND != ::GetSysColor(COLOR_WINDOW);
	}

	static ref<IExplorerBrowser> ExplorerCreate(IServiceProvider* owner, const RECT& bounds, const FOLDERSETTINGS& settings, DWORD* outCookie)
	{
		ref<IExplorerBrowser> browser;
		ref<IExplorerBrowserEvents> events;
		HWND hwnd = Hwnd::From(owner);
		DWORD cookie = 0;

		if FAILED(browser.newobj(CLSID_ExplorerBrowser))
			return null;

		browser->SetOptions(EBO_NOWRAPPERWINDOW);

		if SUCCEEDED(IUnknown_SetSite(browser, owner))
		{
			if SUCCEEDED(owner->QueryInterface(&events))
				browser->Advise(events, &cookie);
			if SUCCEEDED(browser->Initialize(hwnd, &bounds, &settings))
			{
				*outCookie = cookie;
				browser->SetPropertyBag(PATH_FOLDER_STATE);
				return browser;
			}
			// on error
			if (events)
				browser->Unadvise(cookie);
			IUnknown_SetSite(browser, null);
		}
		return null;
	}

	void ExplorerDestroy(IExplorerBrowser* browser, DWORD cookie)
	{
		browser->Unadvise(cookie);
		IUnknown_SetSite(browser, null);
		browser->Destroy();
	}
}

//==================================================================================================================================================================

bool FolderView::OnPress(const MSG& msg) throw()
{
	if (!m_browser || !m_listview)
		return false;

	ref<IShellView> view;
	if (m_listview.IsParentOf(msg.hwnd)
	&& SUCCEEDED(GetCurrentView(&view))
	&& view->TranslateAccelerator(const_cast<MSG*>(&msg)) == S_OK)
	{	// m_listview ̎qɑ΂郁bZ[WGCCł̖OύX
		return true;
	}
	return __super::OnPress(msg);
}

bool FolderView::OnGesture(const MSG& msg)
{
	// Xgr[Ăꍇׂ͂Ď󂯓B
	if (!m_listview.IsWindow())
		return true;

	// OҏWpEditł͎󂯓ȂB
	if (msg.hwnd == m_listview.GetEdit())
		return false;

	// wb_ł͍_uNbN͎󂯓ȂB
	if (msg.message == WM_LBUTTONDBLCLK && msg.hwnd == m_listview.GetHeader())
		return false;

	// NbN݂̂̃WFX`͎󂯓ȂB
	if (msg.message == WM_LBUTTONDOWN || msg.message == WM_LBUTTONUP)
		return false;

	LVHITTESTINFO hit = { GET_XY_LPARAM(msg.lParam) };
	if (msg.hwnd != m_listview)
		MapWindowPoints(msg.hwnd, m_listview, &hit.pt, 1);

	if (msg.message == WM_LBUTTONDBLCLK && m_listview.HitTestColumn(0, hit.pt.x))
		return false;	// OJ͍NbNWFX`֎~B

	int index = m_listview.HitTest(&hit);

	// ENbN̏ꍇAɑIς݂̍ڂJnꍇ́AhbOƂ݂ȂB
	if (msg.message == WM_RBUTTONDOWN &&
		index >= 0 &&
		(hit.flags & LVHT_ONITEM) &&
		ListView_GetSelectedCount(m_listview) > 0 &&
		ListView_GetItemState(m_listview, index, LVIS_SELECTED) == LVIS_SELECTED)
	{
		return false;
	}

	// ACe̍ANbN̓WFX`ȂB
	if (index >= 0 && (MSG2MK(msg) & (MK_LBUTTON | MK_MBUTTON)))
		return false;

	return __super::OnGesture(msg);
}

UINT FolderView::GetMode()
{
	UINT mode;
	ref<IFolderView> view;
	if (m_listview &&
		SUCCEEDED(GetCurrentView(&view)) &&
		SUCCEEDED(view->GetCurrentViewMode(&mode)))
		return mode;
	else
		return (UINT)-1;
}

void FolderView::ListView_UpdateOwnerDraw()
{
	UINT mode = GetMode();
	if (mode != (UINT)-1)
	{
		bool ownerDraw = HasStyle(m_listview, LVS_OWNERDRAWFIXED);
		if (mode == FVM_DETAILS)
		{
			if (!ownerDraw && OwnerDrawEnabled())
			{
				m_listview.ModifyStyle(0, LVS_OWNERDRAWFIXED);
				m_listview.Invalidate(); // ͍ĕ`悾ł悢B
			}
		}
		else
		{
			if (ownerDraw)
			{
				m_listview.ModifyStyle(LVS_OWNERDRAWFIXED, 0);
				Refresh(); // ͍Ő񂪕KvB
			}
		}
	}
}

LRESULT FolderView::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_UPDATEUISTATE:
		__super::OnMessage(msg, wParam, lParam);
		__super::ListView_OnMessage(msg, wParam, lParam);
		ListView_UpdateOwnerDraw();
		return 0;
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		if (m_listview)
		{
			MSG m = { m_listview, msg, wParam, lParam };
			ref<IShellView> view;
			if (SUCCEEDED(GetCurrentView(&view)) && view->TranslateAccelerator(&m) == S_OK)
				return 0;
			return m_listview.SendMessage(msg, wParam, lParam);
		}
		break;
	case WM_VSCROLL:
	case WM_HSCROLL:
		return m_listview.SendMessage(msg, wParam, lParam);
	case SB_SETTEXT:
		OnNotify((PCWSTR)lParam);
		return true;
	case WM_CONTEXTMENU:
		if (!IsParentOf((HWND)wParam))
			return m_defview.SendMessage(msg, wParam, lParam);
		break;
	case WM_TIMER:
		if (OnTimer((UINT)wParam))
			return 0;
		break;
	case WM_DESTROY:
		m_listview.Detach();
		m_defview.Detach();
		if (m_browser)
		{
			ref<IExplorerBrowser> browser(m_browser);
			DWORD cookie = m_cookie;
			m_browser = null;
			m_cookie = 0;
			ExplorerDestroy(browser, cookie);
		}
		SetCurrentFolder(null);
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

void FolderView::OnNotify(PCWSTR message)
{
	ssize_t ln = len(message);
	m_notification.resize(ln+1);
	wcscpy_s(&m_notification[0], ln+1, message);
	// ZԂɉxĂ΂邱Ƃ邽߁A኱xB
	SetTimer(TIMER_NOTIFY, DELAY_NOTIFY);
}

bool FolderView::OnTimer(UINT wID)
{
	KillTimer(wID);
	switch (wID)
	{
	case TIMER_NOTIFY:
	{
		py::Lock lock;
		try
		{
			if (object on_notify = getattr(self, "on_notify"))
			{
				if (m_notification.size() <= 1)
				{	// ̏ꍇ
					m_notification.resize(64);
					int s = m_listview.GetSelectedCount();
					int c = m_listview.GetItemCount();
					int ln;
					if (s > 0)
						ln = swprintf_s(&m_notification[0], m_notification.size(), L"%d / %d ̃IuWFNgI", s, c);
					else
						ln = swprintf_s(&m_notification[0], m_notification.size(), L"%d ̃IuWFNg", c);
					m_notification.resize(ln);
				}
				apply(on_notify, "ui", &m_notification[0], LOG);
			}
		}
		catch (Exception&)
		{
			Exception::clear();
		}
		return true;
	}
	case TIMER_ADJUST:
		Adjust();
		return true;
	default:
		return false;
	}
}

namespace
{
	HRESULT IFolderView_GetPath(IFolderView* view, int index, WCHAR path[MAX_PATH])
	{
		HRESULT hr;
		ref<IShellFolder> folder;
		ILPtr item;
		SUCCEEDED(hr = view->GetFolder(IID_PPV_ARGS(&folder))) &&
		SUCCEEDED(hr = view->Item(index, &item)) &&
		SUCCEEDED(hr = ILDisplayName(folder, item, SHGDN_FORPARSING, path));
		return hr;
	}
}

void FolderView::DefView_OnBeginLabelEdit(NMLVDISPINFO* info)
{
	Hwnd edit = m_listview.GetEdit();
	if (WINDOWS_VERSION < WINDOWS_VISTA && !RENAME_EXTENSION)
	{	// XP̂݁BVista ł͊gqIȂ̂ftHgB
		// XXX: Vista ɂāAzip t@CȂǂ̈͂ǂȂĂH
		WCHAR path[MAX_PATH];
		ref<IFolderView> view;
		if (SUCCEEDED(GetCurrentView(&view))
		&& SUCCEEDED(IFolderView_GetPath(view, info->item.iItem, path))
		&& !PathIsDirectory(path))
		{
			WCHAR name[MAX_PATH];
			edit.GetText(name, MAX_PATH);
			PCWSTR ext = ::PathFindExtension(name);
			int base = (int)(ext - name);
			if (base > 0)
				Edit_SetSelection(edit, 0, base);
		}
	}
	SubclassSingleLineEdit(edit);
}

LRESULT FolderView::DefView_OnMessage(Hwnd& hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_NOTIFY:
		switch (((NMHDR*)lParam)->code)
		{
		case LVN_INSERTITEM:
		case LVN_DELETEITEM:
			if (HasStyle(m_listview, LVS_OWNERDRAWFIXED))
			{	// ꕔĕ`悳Ȃ̂ŁAŜ𖳌B
				m_listview.Invalidate();
			}
			break;
		case LVN_BEGINLABELEDIT:
			DefView_OnBeginLabelEdit((NMLVDISPINFO*)lParam);
			break;
		}
		break;
	case WM_DRAWITEM:
		if (DefView_OnDrawItem((DRAWITEMSTRUCT*)lParam))
			return true;
		break;
/*	case WM_MEASUREITEM:
	{
		MEASUREITEMSTRUCT* measure = (MEASUREITEMSTRUCT*)lParam;
		if (measure->CtlType == ODT_LISTVIEW)
		{
			::DefSubclassProc(hwnd, msg, wParam, lParam);
			measure->itemHeight = 50; // itemHeight ̂ݔfBWidth͂߁B
			return true;
		}
		break;
	}
*/	case WM_CONTEXTMENU:
		if (m_listview == (HWND)wParam)
		{
			if SUCCEEDED(PopupContextMenu(GET_XY_LPARAM(lParam)))
				return 0;
		}
		else
			m_listview.SelectItem(-1);
		break;
	}
	return ::DefSubclassProc(hwnd, msg, wParam, lParam);
}

bool FolderView::DefView_OnDrawItem(DRAWITEMSTRUCT* draw)
{
	if (draw->CtlType != ODT_LISTVIEW || draw->hwndItem != m_listview)
		return false;

	ListView_DrawItem(draw, theDrawDesc);
	return true;
}

namespace
{
	bool ListView_Loop(ListViewH w, UINT dir, UINT vk)
	{
		if (!LISTVIEW_LOOP)
			return false;
		int focus = w.GetNextItem(-1, LVNI_FOCUSED);
		if (focus < 0)
			return false;
		int next = w.GetNextItem(focus, dir);
		if (next >= 0 && next != focus)
			return false;
		w.SendMessage(WM_KEYDOWN, vk, 0);
		return true;
	}
}

LRESULT FolderView::ListView_DefProcNoOwnerDraw(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	// HACK: I[ih[ƁAxȊÕNbNAACẽNbNƂ݂ȂĂ܂B
	// uI[ih[~߂邱ƂŁAI[ih[Ɠ肪g悤ɂB
	DWORD	style = GetStyle(hwnd);
	bool	ownerdraw = ((style & LVS_OWNERDRAWFIXED) != 0);
	__try
	{
		if (ownerdraw)
			SetStyle(hwnd, style & ~LVS_OWNERDRAWFIXED);
		return ::DefSubclassProc(hwnd, msg, wParam, lParam);
	}
	__finally
	{
		// OҏWn܂Ăꍇ́ALVS_OWNERDRAWFIXED ɖ߂ȂB
		// ҏWLZĂ܂߁B
		if (ownerdraw && !ListView_GetEditControl(hwnd))
			SetStyle(hwnd, GetStyle(hwnd) | LVS_OWNERDRAWFIXED);
	}
}

LRESULT FolderView::ListView_OnMessage(ListViewH& hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_PAINT:
	case WM_PRINTCLIENT:
		if (HasStyle(hwnd, LVS_OWNERDRAWFIXED))
		{
			theDrawDesc.BackgroundColor[0] = ::GetSysColor(COLOR_WINDOW);
			theDrawDesc.BackgroundColor[1] = LISTVIEW_BACKGROUND;
			ListView_BeginDraw(hwnd, theDrawDesc);
		}
		break;
	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
		theDrawDesc.DragDrop = (hwnd.HitTest(GET_XY_LPARAM(lParam)) >= 0);
		return ListView_DefProcNoOwnerDraw(hwnd, msg, wParam, lParam);
	case WM_LBUTTONDBLCLK:
	case WM_RBUTTONDBLCLK:
	case WM_XBUTTONDOWN:
	case WM_XBUTTONDBLCLK:
		theDrawDesc.DragDrop = false;
		return ListView_DefProcNoOwnerDraw(hwnd, msg, wParam, lParam);
	case LVM_HITTEST:
		if (hwnd.GetEdit())
			break; // OҏW̏ꍇ͂̂܂܁BȂҏWLZ邽߁B
		return ListView_DefProcNoOwnerDraw(hwnd, msg, wParam, lParam);
	case WM_MBUTTONDOWN:
	{
		theDrawDesc.DragDrop = false;
		int index = hwnd.HitTest(GET_XY_LPARAM(lParam));
		if (index >= 0)
		{
			if (GET_KEYSTATE_WPARAM(wParam) & MK_CONTROL)
			{
				hwnd.SetItemState(index, LVIS_FOCUSED | LVIS_SELECTED, LVIS_FOCUSED | LVIS_SELECTED);
				hwnd.EnsureVisible(index);
			}
			else
				hwnd.SelectItem(index);
		}
		return 0;
	}
	case WM_MBUTTONUP:
	{
		theDrawDesc.DragDrop = false;
		int index = hwnd.HitTest(GET_XY_LPARAM(lParam));
		if (index >= 0 && hwnd.GetItemState(index, LVIS_SELECTED))
			Press(VK_RETURN, GET_KEYSTATE_WPARAM(wParam) | MK_MBUTTON);
		return 0;
	}
	case WM_LBUTTONUP:
	case WM_RBUTTONUP:
		theDrawDesc.DragDrop = false;
		break;
	case WM_KILLFOCUS:
		theDrawDesc.DragDrop = true;
		break;
	case WM_KEYDOWN:
		switch (wParam)
		{
		case VK_UP:
			if (ListView_Loop(hwnd, LVNI_ABOVE, VK_END))
				return 0;
			break;
		case VK_DOWN:
			if (ListView_Loop(hwnd, LVNI_BELOW, VK_HOME))
				return 0;
			break;
		}
	case WM_MOUSEWHEEL:
		if (LISTVIEW_WHEELSELECT && GetMode() == FVM_DETAILS)
		{
			POINT pt = { GET_XY_LPARAM(lParam) };
			m_listview.ScreenToClient(&pt);
			if (m_listview.HitTestColumn(0, pt.x))
			{
				WPARAM vk = GET_WHEEL_DELTA_WPARAM(wParam) > 0 ? VK_UP : VK_DOWN;
				m_listview.SendMessage(WM_KEYDOWN, vk, 0);
				return 0;
			}
		}
		break;
	default:
		if (msg == WM_GETDRAGIMAGE)
			return ListView_DefProcNoOwnerDraw(hwnd, msg, wParam, lParam);
		break;
	}
	__super::ListView_OnMessage(msg, wParam, lParam);
	return ::DefSubclassProc(hwnd, msg, wParam, lParam);
}

void FolderView::QueryLayout(ILayout& layout) const
{
	if (m_defview)
		layout.Insert(m_defview, CENTER);
	__super::QueryLayout(layout);
}

void FolderView::OnDragDrop(POINT pt, RECT* rc, IDropTarget** pp) throw()
{
	ref<IShellFolder> parent;
	const ITEMIDLIST* leaf;
	if SUCCEEDED(ILParentFolder(&parent, m_item, &leaf))
		parent->GetUIObjectOf(null, 1, &leaf, __uuidof(IDropTarget), null, (void**)pp);
}

RECT FolderView::get_CenterArea() const throw()
{
	if (m_defview)
	{	// ȑÕr[ꍇ́ATCYۂB
		RECT bounds;
		::GetWindowRect(m_defview, &bounds);
		ScreenToClient(&bounds);
		return bounds;
	}
	else
	{	// ꍇ́AvZB
		return __super::get_CenterArea();
	}
}

HRESULT FolderView::GetCurrentView(REFINTF pp) const throw()
{
	if (!m_browser)
	{
		*pp.pp = null;
		return E_UNEXPECTED;
	}
	return m_browser->GetCurrentView(pp.iid, pp.pp);
}

HRESULT FolderView::GetCurrentFolder(REFINTF pp) const throw()
{
	HRESULT hr;
	ref<IFolderView> view;
	if FAILED(hr = GetCurrentView(&view))
		return hr;
	return view->GetFolder(pp.iid, pp.pp);
}

void FolderView::SetCurrentFolder(const ITEMIDLIST* folder)
{
	if (folder)
		m_item.attach(ILClone(folder));
	else
		m_item.clear();
	{
		py::Lock lock;
		m_path = null;
	}
}

/// @param	x	X in screen coordinate.
/// @param	y	Y in screen coordinate.
HRESULT FolderView::PopupContextMenu(int x, int y) throw()
{
	// obNOEhł̃ReNXgj[́AftHg̐U镑ɔCB
	// Oŕ\ꍇAẽj[oƂ܂ł͂ł邪A
	// VK쐬AIɖOύXԂɂ邱Ƃ߁B
	// j[ɍڂǉ̂łȂAOŕ\闘_ɂ͖B
	//
	// ACẽReLXgj[́AOŕ\B
	// ZIPtH_Ńj[\ȂہiXPˑHj邽߁B

	// TODO: CL[Ńj[ς
	// UINT mods = GetModifiers();
	// if ((mods & MK_KEYS) == 0)
	// 	return E_ABORT;

	HRESULT hr;
	POINT	ptScreen = { x, y };

	if (ptScreen.x == -1 && ptScreen.y == -1)
	{	// j[L[̏ꍇ̓
		int index = m_listview.GetNextItem(-1, LVNI_SELECTED);
		if (index < 0)
			return E_FAIL;	// obNOEh̏ꍇ

		// IACe̍ɕ\B
		RECT rc = { 0, 0, 0, 0 };
		m_listview.GetItemRect(index, &rc, LVIR_BOUNDS);
		ptScreen.x = rc.left;
		ptScreen.y = rc.bottom;
		m_listview.ClientToScreen(&ptScreen);
	}
	else
	{
		LVHITTESTINFO hit = { ptScreen.x, ptScreen.y };
		m_listview.ScreenToClient(&hit.pt);
		int index = m_listview.HitTest(&hit);
		if (index < 0)
		{
			m_listview.SelectItem(-1);
			return E_FAIL;	// obNOEh̏ꍇ
		}
	}

	ref<IShellView> view;
	if FAILED(hr = GetCurrentView(&view))
		return hr;
	hr = IShellView_Popup(view, ptScreen, this);
	switch (hr)
	{
	case S_DEFAULT:
	{
		ref<IFolderView>	fv;
		ILPtr				item;
		LVHITTESTINFO hit = { ptScreen.x, ptScreen.y };
		m_listview.ScreenToClient(&hit.pt);
		int index = m_listview.HitTest(&hit);
		if (index < 0)
			index = m_listview.GetNextItem(-1, LVNI_FOCUSED);
		if (index < 0)
			index = m_listview.GetNextItem(-1, LVNI_SELECTED);
		if (index >= 0 &&
			SUCCEEDED(view->QueryInterface(&fv)) &&
			SUCCEEDED(fv->Item(index, &item)))
		{
			BrowseObject(item, SBSP_SAMEBROWSER | SBSP_RELATIVE);
		}
		break;
	}
	case S_RENAME:	// OύX(rename)́AUIƎɒ񋟂Ȃ΂ȂȂB
		Rename();
		break;
	}
	return hr;
}

HRESULT FolderView::Rename() throw()
{
	// tH[JX̍ڂIĂȂꍇAŏ̑IڂtH[JXB
	// ڂ̃l[̏ꍇɁAɔ铮邽߁B
	if (m_listview)
	{
		int focus = m_listview.GetNextItem(-1, LVNI_ALL | LVNI_FOCUSED);
		if (focus >= 0 && !m_listview.GetItemState(focus, LVIS_SELECTED))
		{
			int selected = m_listview.GetNextItem(-1, LVNI_SELECTED);
			if (selected >= 0)
				m_listview.SetItemState(selected, LVIS_FOCUSED, LVIS_FOCUSED);
		}
	}

	ref<IFolderView2> view;
	if SUCCEEDED(GetCurrentView(&view))
		return view->DoRename();
	else
		return Press(VK_F2, 0); // view->SelectItem(*, SVSI_EDIT);
}

HRESULT FolderView::Refresh() throw()
{
	HRESULT hr;

	ref<IShellView> view;
	if FAILED(hr = GetCurrentView(&view))
		return hr;
	ref<IFolderView> fv;
	view->QueryInterface(&fv);

	// Ԃ̕ۑ
	ref<IExplorerBrowserXp> xp;
	if SUCCEEDED(view->QueryInterface(&xp))
		xp->SaveViewState();

	// IԂ̕ۑ
	CIDA*	selection = CIDACreate(view, SVGIO_SELECTION);
	ILPtr	focus;
	if (!selection && fv)
	{
		int	index;
		if SUCCEEDED(fv->GetFocusedItem(&index))
			fv->Item(index, &focus);
	}

	// \[gԂ̕ۑ
	int		column;
	bool	ascending;
	if FAILED(m_listview.GetSortedColumn(&column, &ascending))
	{
		column = 0;
		ascending = true;
	}

	// XV
	hr = view->Refresh();

	// \[gԂ̕
	m_listview.SetSortedColumn(column, ascending);

	// IԂ̕
	if (selection)
	{
		UINT svsi = SVSI_ENSUREVISIBLE | SVSI_SELECT | SVSI_FOCUSED | SVSI_DESELECTOTHERS;
		size_t count = CIDACount(selection);
		for (size_t i = 0; i < count; ++i)
		{
			if SUCCEEDED(view->SelectItem(CIDAChild(selection, i), svsi))
				svsi &= ~(SVSI_FOCUSED | SVSI_ENSUREVISIBLE | SVSI_DESELECTOTHERS);
		}
		CIDAFree(selection);
	}
	else if (focus)
	{
		view->SelectItem(focus, SVSI_ENSUREVISIBLE | SVSI_FOCUSED | SVSI_DESELECTOTHERS);
	}

	return hr;
}

CIDA* FolderView::ExecuteSelection(IShellView* view) throw()
{
	CIDA* items = CIDACreate(view, SVGIO_SELECTION | SVGIO_FLAG_VIEWORDER);
	if (!items)
		return null;

	{
		py::Lock lock;
		try
		{
			if (object on_execute = getattr(self, "on_execute"))
			{
				ref<IShellFolder> folder;
				ref<IFolderView> folderview;
				if SUCCEEDED(view->QueryInterface(&folderview))
					folderview->GetFolder(IID_PPV_ARGS(&folder));
				ref<Paths> r = Paths::newobj(items, folder);
				if (cast<bool>(apply(on_execute, "O", r->self)))
				{
					CIDAFree(items);
					return null;
				}
			}
		}
		catch (Exception&)
		{
			Exception::notify();
		}
	}

	return items;
}

bool FolderView::get_AutoArrange() const
{
	ref<IShellView> view;
	FOLDERSETTINGS settings;
	if (SUCCEEDED(GetCurrentView(&view)) && SUCCEEDED(view->GetCurrentInfo(&settings)))
		return (settings.fFlags & FWF_AUTOARRANGE) != 0;
	return false;
}

void FolderView::set_AutoArrange(bool value)
{
	ref<IShellView> view;
	FOLDERSETTINGS settings;
	if (SUCCEEDED(GetCurrentView(&view)) && SUCCEEDED(view->GetCurrentInfo(&settings)))
	{
		settings.fFlags |= FWF_AUTOARRANGE;
//		view->SetCurrentInfo(&settings);
	}
}

void FolderView::set_Grouping(bool value)
{
	// ListView_EnableGroupView() ͎gȂBShellView Ƃ̈ѐĂ܂߁B
	if (this->Grouping != value)
		m_defview.PostCommand(SHIDM_GROUPING);
}

//==================================================================================================================================================================
// IServiceProvider

//struct __declspec(uuid("4C96BE40-915C-11CF-99D3-00AA004AE837")) ITopLevelBrowser;
//#define SID_STopLevelBrowser		__uuidof(ITopLevelBrowser)

STDMETHODIMP FolderView::QueryService(REFGUID guidService, REFIID riid, void **ppv)
{
	HRESULT hr = E_NOINTERFACE;
	if (guidService == SID_SExplorerBrowserFrame ||
		guidService == SID_SShellBrowser)
	{
		hr = QueryInterface(riid, ppv);
	}
	else
	{
	// SID_STopLevelBrowser
	// IID_IFolderTypeModifier
	// IID_IBrowserProgressAggregator
	// F80C2137-5829-4CE9-9F81-A95E159DD8D5
		*ppv = null;
	}
	return hr;
}

//==================================================================================================================================================================
// IExplorerBrowserEvents

STDMETHODIMP FolderView::OnNavigationPending(const ITEMIDLIST* folder)
{
	// Returning failure from this will cancel the navigation. 
	return S_OK;
}

STDMETHODIMP FolderView::OnViewCreated(IShellView* view)
{
	// Called once the view window has been created. Do any last minute modifcations to the view here before it is shown (set view modes, folder flags, etc...) 
	theDrawDesc.DragDrop = false;
	if (HWND hwnd = Hwnd::From(view))
	{
		if (m_defview != hwnd)
		{
			m_defview.Attach(hwnd, this);
			if (m_listview.Attach(::FindWindowEx(m_defview, null, WC_LISTVIEW, null), this))
			{
				m_listview.DisableIME();
				ListView_UpdateOwnerDraw();
				if (HFONT font = GetFont())
					m_listview.SetFont(font);
			}
		}
	}
	return S_OK;
}

STDMETHODIMP FolderView::OnNavigationComplete(const ITEMIDLIST* folder)
{
	// Called once the navigation has succeeded (after OnViewCreated). 
	SetCurrentFolder(folder);
	Event("on_navigate");
	return S_OK;
}

STDMETHODIMP FolderView::OnNavigationFailed(const ITEMIDLIST* folder)
{
	// Called if a navigation failed, despite the call to IShellBrowser::BrowseObject succeeding. 
	return S_OK;
}

//==================================================================================================================================================================
// IOleWindow

STDMETHODIMP FolderView::GetWindow(HWND* phwnd)
{
	*phwnd = m_hwnd;
	return S_OK;
}

//==================================================================================================================================================================
// ICommDlgBrowser

// handle double click and ENTER key if needed
STDMETHODIMP FolderView::OnDefaultCommand(IShellView* view)
{
	CIDA* items = ExecuteSelection(view);
	if (!items)
		return S_OK;

	HRESULT hr = S_FALSE;
	if (ILPtr item = ILCreate(items, 0))
		if (ILIsFolder(item))
			hr = BrowseObject(item, SBSP_SAMEBROWSER | SBSP_ABSOLUTE);
	CIDAFree(items);
	if (hr == S_OK)
		return S_OK;

	return S_FALSE;
}

STDMETHODIMP FolderView::OnStateChange(IShellView* view, ULONG uChange)
{	//handle selection, rename, focus if needed
	switch (uChange)
	{
	case CDBOSC_SETFOCUS:
	case CDBOSC_KILLFOCUS:
	case CDBOSC_SELCHANGE:
	case CDBOSC_RENAME:
	case CDBOSC_STATECHANGE:
		break;
	}
	return E_NOTIMPL;
}

STDMETHODIMP FolderView::IncludeObject(IShellView* view, const ITEMIDLIST* pidl)
{	// \t@C̃tB^OB
	// S_OK: show, S_FALSE: hide
	return S_OK;

/*
	// Bt@CB́AGetViewFlags()ł͌ʂ߁AŌʂɃtB^O
	if (!m_pCurrentFolder || !m_pCurrentEntry)
		return S_OK;
	// tB^.
	if (m_PatternMask)
	{
		WCHAR name[MAX_PATH];
		if SUCCEEDED(ILDisplayName(m_pCurrentFolder, pidl, SHGDN_NORMAL | SHGDN_INFOLDER, name, MAX_PATH))
		{
			if (!PatternEquals(m_PatternMask.str(), name))
			{
				SFGAOF flags = SFGAO_FOLDER;
				if SUCCEEDED(m_pCurrentFolder->GetAttributesOf(1, &pidl, &flags))
				{	// p^[}b`ȂAtH_̂݉BB
					if ((flags & SFGAO_FOLDER) == 0)
						return S_FALSE;
				}
			}
		}
	}
	if (!theShellFlagState.fShowAllObjects)
		return S_OK; // łɃtB^Oς
	if (m_ShowAllFiles)
		return S_OK; // S\邩
	SFGAOF flags = SFGAO_GHOSTED;
	if SUCCEEDED(m_pCurrentFolder->GetAttributesOf(1, &pidl, &flags))
	{
		return (flags & SFGAO_GHOSTED) ? S_FALSE : S_OK;
	}
	else
	{	// Ȃs邱ƂBdȂ̂ŁAWin32APIoRŁB
		HRESULT result = S_OK;
		// SHGetRealIDL(m_pCurrentFolder, pidl, &fullpath)@ŝŁAOŘA邱Ƃɂ.
		ITEMIDLIST* fullpath = ILCombine(m_pCurrentEntry->ID, pidl);
		TCHAR path[MAX_PATH];
		if SUCCEEDED(ILGetPath(fullpath, path))
		{
			DWORD dwAttrs = GetFileAttributes(path);
			if (dwAttrs != (DWORD)-1 && (dwAttrs & FILE_ATTRIBUTE_HIDDEN))
				result = S_FALSE;
		}
		ILFree(fullpath);
		return result;
	}
*/
}

//==================================================================================================================================================================
// ICommDlgBrowser2

STDMETHODIMP FolderView::GetDefaultMenuText(IShellView* view, WCHAR* pszText, int cchMax)
{
	ref<IFolderView>	fv;
	ref<IShellFolder>	folder;
	ITEMIDLIST*			item = null;
	SFGAOF				flags = SFGAO_BROWSABLE | SFGAO_FOLDER;
	int					focus = m_listview.GetNextItem(-1, LVNI_FOCUSED);

	if (focus >= 0 &&
		SUCCEEDED(view->QueryInterface(&fv)) &&
		SUCCEEDED(fv->GetFolder(IID_PPV_ARGS(&folder))) &&
		SUCCEEDED(fv->Item(focus, &item)) &&
		SUCCEEDED(folder->GetAttributesOf(1, (const ITEMIDLIST**)&item, &flags)) &&
		(flags & (SFGAO_BROWSABLE | SFGAO_FOLDER)))
	{
		wcscpy_s(pszText, cchMax, L"ړ(&G)");
	}
	else
	{
		pszText[0] = 0;
	}
	ILFree(item);
	return S_OK;
}

STDMETHODIMP FolderView::GetViewFlags(DWORD* pdwFlags)
{	// ׂẴt@C\ׂۂ̖₢킹B
	// CDB2GVF_SHOWALLFILES: Bt@C\B
	// 0: Bt@C\ȂB
	if (m_ShowAllFiles)
		*pdwFlags = CDB2GVF_SHOWALLFILES;
	else
		*pdwFlags = 0;
	return S_OK;
}

//==================================================================================================================================================================
// ICommDlgBrowser3

//==================================================================================================================================================================
// IShellBrowser

STDMETHODIMP FolderView::GetControlWindow(UINT id, HWND* pHwnd)
{
	if (!pHwnd)
		return E_POINTER;
	switch(id)
	{
	case FCW_STATUS:
		*pHwnd = m_hwnd;
		return S_OK;
	}
	return E_NOTIMPL;
}

STDMETHODIMP FolderView::SendControlMsg(UINT id, UINT msg, WPARAM wParam, LPARAM lParam, LRESULT* lResult)
{
	if (id != FCW_STATUS || msg != SB_SETTEXT)
		return E_NOTIMPL;
	OnNotify((PCWSTR)lParam);
	if (lResult)
		*lResult = TRUE;
	return S_OK;
}

STDMETHODIMP FolderView::QueryActiveShellView(IShellView** pp)
{
	return GetCurrentView(pp);
}

STDMETHODIMP FolderView::BrowseObject(const ITEMIDLIST* pidl, UINT wFlags)
{
	// wFlags
	//	ǂ̃uEUŊJH
	//		SBSP_DEFBROWSER  : tH_IvV̐ݒɂ
	//		SBSP_SAMEBROWSER : 
	//		SBSP_NEWBROWSER  : V
	//	VJꍇ̃IvV
	//		SBSP_DEFMODE : Ɠic[Hj
	//		SBSP_OPENMODE : ʏ탂[hic[ȂHj
	//		SBSP_EXPLOREMODE : GNXv[[hic[Hj
	//		SBSP_HELPMODE : wv
	//		SBSP_NOTRANSFERHIST : 𕡐Ȃ
	//	pidl ̉
	//		SBSP_ABSOLUTE
	//		SBSP_RELATIVE
	//	s
	//SBSP_ALLOW_AUTONAVIGATE
	//SBSP_KEEPSAMETEMPLATE
	//SBSP_KEEPWORDWHEELTEXT
	//SBSP_ACTIVATE_NOFOCUS
	//SBSP_CALLERUNTRUSTED
	//SBSP_TRUSTFIRSTDOWNLOAD
	//SBSP_UNTRUSTEDFORDOWNLOAD
	//SBSP_NOAUTOSELECT
	//SBSP_WRITENOHISTORY
	//SBSP_TRUSTEDFORACTIVEX
	//SBSP_FEEDNAVIGATION
	//SBSP_REDIRECT
	//SBSP_INITIATEDBYHLINKFRAME

	if (!::IsWindow(m_hwnd))
		return E_UNEXPECTED;

	HRESULT hr;
	if (m_browser)
	{
		hr = m_browser->BrowseToIDList(pidl, wFlags);
	}
	else
	{	// ĂяȍꍇB
		DWORD cookie;

		if (!pidl && (wFlags & SBSP_ABSOLUTE) == 0)
			return E_INVALIDARG;

		FolderState state;
		if FAILED(hr = FolderStateLoad(pidl, &state, null))
			return hr;

		ref<IExplorerBrowser> browser = ExplorerCreate(this, CenterArea, state, &cookie);
		if (!browser)
			return E_FAIL;

		if FAILED(hr = browser->BrowseToIDList(pidl, wFlags))
		{
			ExplorerDestroy(browser, cookie);
			return hr;
		}

		// 쐬ɐ̂ŁAoϐXVB
		m_browser = browser;
		m_cookie = cookie;
	}

	if (hr == S_UPDATED)
		OnAdjust();
	return S_OK;
}

void FolderView::OnAdjust()
{
	if (TIMER_ADJUST > 0)
		SetTimer(TIMER_ADJUST, LISTVIEW_AUTOADJUST);
	else if (TIMER_ADJUST == 0)
		Adjust();
}

//==================================================================================================================================================================

int FolderView::select(PyObject* o, UINT svsi)
{
	if (is<Path>(o))
	{	// single item
		ref<IShellView> view;
		if SUCCEEDED(GetCurrentView(&view))
		{
			view->SelectItem(ILFindLastID(cast<Path>(o)->ID), svsi); 
			return 1;
		}
	}
	else if (is<Paths>(o))
	{	// fast path for Paths
		Paths* paths = cast<Paths>(o);
		ref<IShellView> view;
		ssize_t ln;
		const CIDA* cida;
		if (SUCCEEDED(GetCurrentView(&view))
		&& (ln = paths->__len__()) > 0
		&& (cida = paths->GetCIDA()) != null)
		{
			view->SelectItem(CIDAChild(cida, 0), svsi); 
			svsi &= ~(SVSI_FOCUSED | SVSI_ENSUREVISIBLE | SVSI_DESELECTOTHERS);
			for (ssize_t i = 1; i < ln; ++i)
				view->SelectItem(CIDAChild(cida, i), SVSI_SELECT);
			return (int)ln;
		}
	}
	return __super::select(o, svsi);
}

void FolderView::cut()		{ press('X', MK_CONTROL); }
void FolderView::copy()		{ press('C', MK_CONTROL); }
void FolderView::paste()	{ press('V', MK_CONTROL); }
void FolderView::delete_()
{
	if (!m_item)
		return;
	WCHAR dir[MAX_PATH];
	if SUCCEEDED(ILPath(m_item, dir))
	{
		if (PathIsUNC(dir))
		{
			bury(); // lbg[NpX
			return;
		}
		else if (PathStripToRoot(dir))
		{
			switch (GetDriveType(dir))
			{
			case DRIVE_REMOVABLE:
			case DRIVE_REMOTE:
			case DRIVE_CDROM:
			case DRIVE_RAMDISK:
				bury(); // S~gȂꍇ́AS폜ɂB
				return;
			}
		}
	}
	press(VK_DELETE, 0);
}
void FolderView::bury()		{ press(VK_DELETE, MK_SHIFT); }
void FolderView::undo()		{ press('Z', MK_CONTROL); }
void FolderView::rename()	{ py::Unlock unlock; Rename(); }
void FolderView::refresh()	{ py::Unlock unlock; Refresh(); }

namespace
{
	void veirfy_filename(PCWSTR filename)
	{
		if (wcspbrk(filename, STR_ERR_PATHCHAR))
			throw IOError(_("Error.PathChar"));
	}
}

void FolderView::new_(Tuple* args)
{
	if (!m_item)
		return;

	PyObject*	o_name = null;
	PyObject*	o_extention = null;
	brownie::parse(args, "|OO", &o_name, &o_extention);
	unicode		u_name = cast<unicode>(o_name);
	unicode		u_extention = cast<unicode>(o_extention);
	PCWSTR		s_name = str(u_name);
	PCWSTR		s_extention = str(u_extention);

	if (empty(s_name))
	{
		static PCWSTR STR_NEWFOLDER = _W("NewFolder");
		u_name = null;
		s_name = STR_NEWFOLDER;
	}
	else
		veirfy_filename(s_name);

	HRESULT hr;
	WCHAR dir[MAX_PATH];
	WCHAR path[MAX_PATH];
	{
		py::Unlock unlock;
		if SUCCEEDED(hr = ILPath(m_item, dir))
		{
			PathCombine(path, dir, s_name);
			if (!empty(s_extention))
			{
				if (s_extention[0] != L'.')
					wcscat_s(path, L".");
				wcscat_s(path, s_extention);
			}
			if (PathFileExists(path))
				PathMakeUniqueName(path, MAX_PATH, null, s_name, dir);

			if (empty(s_extention) && *PathFindExtension(s_name) != L'.')
				hr = HRESULT_FROM_WIN32(SHCreateDirectory(m_hwnd, path));
			else
			{
				ref<IFileOperation> op;
				SUCCEEDED(hr = op.newobj(CLSID_FileOperation)) &&
				SUCCEEDED(hr = FileNew(op, path)) &&
				SUCCEEDED(hr = op->PerformOperations());
			}
		}
	}

	if FAILED(hr)
		throw WindowsError("Cannot create new file or directory");

	py::Unlock unlock;

	ref<IFolderView> view;
	if SUCCEEDED(GetCurrentView(&view))
	{
		int		c = 0;
		int		index = -1;

		view->SelectItem(-1, SVSI_DESELECTOTHERS);

		// 쐬Ă300msقǂ悤B
		PumpMessage();
		Sleep(200);
		PumpMessage();

		// Xgr[ƂđIB
		// Ȃ IShellView.SelectItem() ƁAĂ܂߁B
		LVFINDINFO find = { LVFI_STRING, PathFindFileName(path) };
		for (; c < 20 && (index = ListView_FindItem(m_listview, -1, &find)) < 0; ++c)
		{
			TRACE(L"%2d: %s", c, path);
			Sleep(50);
			PumpMessage();
		}

		if (index >= 0)
		{
			UINT svsi = SVSI_ENSUREVISIBLE | SVSI_SELECT | SVSI_FOCUSED | SVSI_DESELECTOTHERS;
			if (!u_name)
				svsi |= SVSI_EDIT; // O^Ȃꍇ́AҏWԂɂB
			view->SelectItem(index, svsi);
		}
	}
}

void FolderView::go(PyObject* value)
{
	HRESULT hr = E_INVALIDARG;
	if (is<Path>(value))
	{
		const ITEMIDLIST* item = cast<Path>(value)->ID;
		py::Unlock unlock;
		hr = BrowseObject(item, SBSP_SAMEBROWSER | SBSP_ABSOLUTE);
	}
	else if (is<Integer>(value))
	{
		if (m_browser)
		{
			int	mv = Integer::value(value);
			py::Unlock unlock;
			if (mv == FOLDER_PARENT)
				hr = BrowseObject(null, SBSP_SAMEBROWSER | SBSP_PARENT);
			else
			{
				ref<IExplorerBrowserXp> xp;
				if SUCCEEDED(m_browser->QueryInterface(&xp))
				{
					hr = xp->BrowseToHistory(mv);
					if (hr == S_UPDATED)
						OnAdjust();
				}
				else if (mv < 0)
					hr = BrowseObject(null, SBSP_SAMEBROWSER | SBSP_NAVIGATEBACK);
				else if (mv > 0)
					hr = BrowseObject(null, SBSP_SAMEBROWSER | SBSP_NAVIGATEFORWARD);
			}
		}
	}
	else if (unicode path = Unicode::newobj(value))
	{
		PCWSTR s = str(path);
		py::Unlock unlock;
		if (ILPtr item = ILCreate(s))
			hr = BrowseObject(item, SBSP_SAMEBROWSER | SBSP_ABSOLUTE);
	}

	if FAILED(hr)
		throw IOError(_("Error.FileNotFound"));
}

object FolderView::parse(PyObject* value)
{
	if (!is<Integer>(value))
		return Path::newobj(value);

	if (m_browser)
	{
		int	mv = Integer::value(value);
		if (mv == FOLDER_PARENT)
		{
			if (ITEMIDLIST* parent = ILCloneParent(m_item))
				return Path::newobj_nocopy(parent);
		}
		else
		{
			ref<IExplorerBrowserXp> xp;
			if SUCCEEDED(m_browser->QueryInterface(&xp))
			{
				ref<IShellItem> item;
				if SUCCEEDED(xp->GetHistory(&item, mv))
					return Path::newobj(item);
			}
		}
	}
	return null;
}

object FolderView::get_items(STATUS status)
{
	CIDA* items = null;
	ref<IFolderView> view;
	ref<IShellFolder> folder;

	if SUCCEEDED(GetCurrentView(&view))
	{
		py::Unlock unlock;

		view->GetFolder(IID_PPV_ARGS(&folder));

		switch (status)
		{
		case FOCUSED:
			items = CIDACreate(view, SVGIO_SELECTION);
			break;
		case SELECTED:
			items = CIDACreate(view, SVGIO_SELECTION | SVGIO_FLAG_VIEWORDER);
			break;
		case CHECKED:
			items = CIDACreate(view, SVGIO_CHECKED | SVGIO_FLAG_VIEWORDER);
			break;
		default:
			items = CIDACreate(view, SVGIO_ALLVIEW | SVGIO_FLAG_VIEWORDER);
			break;
		}
	}

	if (!items)
		return null;

	ref<Paths> r = Paths::newobj(items, folder);
	CIDAFree(items);
	return object(r->self);
}

object FolderView::get_path()
{
	if (!m_path)
		m_path = Path::newobj(m_item);
	return m_path;
}

object FolderView::mode(Tuple* args, Dict* kw)
{
	if (!m_browser)
		return null;

	HRESULT hr;
	PCSTR kwds[] = { "mode", "size", null };
	int mode = -1;
	int size = -1;
	brownie::parse(args, kw, "|ii", kwds, &mode, &size);
	if (size < -1)
		size = -1;

	ref<IFolderView2> view2;
	ref<IFolderView> view;
	int prev_mode = 0;
	int prev_size = -1;

	if SUCCEEDED(GetCurrentView(&view2))
	{
		view2->GetViewModeAndIconSize((FOLDERVIEWMODE*)&prev_mode, &prev_size);
		if (mode < 0)
			mode = prev_mode;
		hr = view2->SetViewModeAndIconSize((FOLDERVIEWMODE)mode, size);
	}
	else if SUCCEEDED(GetCurrentView(&view))
	{
		view->GetCurrentViewMode((UINT*)&prev_mode);
		if (mode < 0)
			mode = prev_mode;
		switch (mode)
		{
		case FVM_THUMBNAIL:
		case FVM_ICON:
		case FVM_SMALLICON:
		case FVM_LIST:
			if (size > 32)
				mode = FVM_THUMBNAIL;
			else if (size > 16)
				mode = FVM_ICON;
			else if (size >= 0)
				mode = FVM_LIST;
			break;
		}
		hr = view->SetCurrentViewMode(mode);
	}
	else
		hr = E_NOTIMPL;

	if SUCCEEDED(hr)
		return build("(ii)", prev_mode, prev_size);
	else
		throw ValueError("Invalide explorer view mode");
}

namespace
{
	list EnumHistory(IExplorerBrowser* browser, int dir)
	{
		ASSERT(browser);

		ref<IExplorerBrowserXp> xp;
		if FAILED(browser->QueryInterface(&xp))
			return null;
		list folders = List::newobj(0);
		ref<IEnumShellItems> items;
		if SUCCEEDED(xp->EnumHistory(&items, dir))
		{
			ref<IShellItem> item;
			while (items->Next(1, &item, null) == S_OK)
			{
				folders->append(Path::newobj(item)->self);
				item = null;
			}
		}
		return folders;
	}
}

object FolderView::history(PyObject* value)
{
	if (!m_browser)
		return null;

	switch (Integer::value(value))
	{
	case FOLDER_BACK:
		return EnumHistory(m_browser, -1);
	case FOLDER_FORWARD:
		return EnumHistory(m_browser, +1);
	default:
		throw ValueError("Invalid argument for ExplorereView.parse()");
	}
}

//============================================================================

PYTHON_METHODS(FolderView)
	PYTHON_METHOD(bury)
	PYTHON_METHOD_(delete, delete_)
	PYTHON_METHOD(copy)
	PYTHON_METHOD(cut)
	PYTHON_METHOD(go)
	PYTHON_METHOD(history)
	PYTHON_METHOD(mode)
	PYTHON_METHOD_(new, new_)
	PYTHON_METHOD(parse)
	PYTHON_METHOD(paste)
	PYTHON_METHOD(refresh)
	PYTHON_METHOD(rename)
	PYTHON_METHOD(undo)
PYTHON_END()

PYTHON_MEMBERS(FolderView)
	PYTHON_MEMBER(path		, GET)
PYTHON_END()
