// headerbar.cpp

#include "stdafx.h"
#include "headerbar.hpp"
#include <vssym32.h>

using namespace brownie;

//==================================================================================================================================================================

LRESULT HeaderBar::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_CREATE:
		if (LRESULT lResult = __super::OnMessage(msg, wParam, lParam))
			return lResult;
		m_theme.Create(m_hwnd, VSCLASS_HEADER);
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

namespace
{
	const int TEXT_MARGIN = 4;
	const int THEME_PART[] =
	{
		HP_HEADERITEM,
		HP_HEADERITEM,
		HP_HEADERITEMLEFT,
		HP_HEADERITEMRIGHT,
		0,
	};
}

void HeaderBar::OnDraw(HDC dc, const RECT& rcPaint)
{
	RECT		rc = ClientArea;
	int			part = THEME_PART[Dock];
	const Item*	items = GetItems();
	ssize_t		count = GetItemCount();

	// 
	rc.left = items[count].offset;
	rc.right += 16; // E[͂͂ݏo悤
	m_theme.Background(dc, part, HIS_NORMAL, &rc);
	if (count <= 0)
		return;

	Icons*		icons = GetIcons();
	HIMAGELIST	shImageList = SHGetImageList(icons);
	HFONT		font = GetFont();
	HFONT		old = SelectFont(dc, font);
	for (ssize_t i = 0; i < count; ++i)
	{
		rc.left = items[i].offset;
		rc.right = items[i+1].offset;

		int		mode;
		STATUS	status;
		const int modes[] = { HIS_NORMAL, HIS_HOT, HIS_PRESSED, HIS_NORMAL, HIS_NORMAL, HIS_HOT };
		GetStatus(&mode, &status, i, modes);

		m_theme.Background(dc, part, mode, &rc);
		InflateRect(&rc, -TEXT_MARGIN, 0);
		if (items[i].icon >= 0)
		{
			SIZE	szIcon = IconsSize(icons, shImageList, items[i].icon, 0);
			IconsDraw(dc, items[i].icon, rc.left, (rc.top + rc.bottom - szIcon.cy) / 2, status, icons, shImageList);
			rc.left += szIcon.cx + TEXT_MARGIN;
		}
		m_theme.Text(dc, part, mode, &rc, str(items[i].name), DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX, status);
	}
	SelectFont(dc, old);
}

void HeaderBar::OnLayout()
{
	__super::OnLayout();

	bool		modified = false;
	ssize_t		count = GetItemCount();

	if (count > 0)
	{
		Item*	items = GetItems();
		Icons*	icons = GetIcons();
		SIZE	szIcon = IconsSize(icons, null, 16, 16);
		HFONT	font = GetFont();
		HDC		dc = GetDC();
		int		offset = 0;
		for (ssize_t i = 0; i <= count; ++i)
		{
			if (items[i].offset != offset)
			{
				items[i].offset = offset;
				modified = true;
			}
			int w = TEXT_MARGIN * 2;
			if (items[i].icon >= 0)
				w += szIcon.cx + TEXT_MARGIN;
			w += FontSize(font, str(items[i].name), dc).cx;
			offset += w;
		}
		ReleaseDC(dc);
	}
	if (modified)
		Invalidate();
}
