// hwnd.cpp

#include "stdafx.h"
#include "hwnd.hpp"
#include "class.hpp"

using namespace brownie;

//==================================================================================================================================================================

int MapWindowRect(HWND hwndFrom, HWND hwndTo, RECT* pRect)
{
	return ::MapWindowPoints(hwndFrom, hwndTo, (POINT*)pRect, 2);
}

void SetRedraw(HWND hwnd, bool redraw) throw()
{
	SendMessage(hwnd, WM_SETREDRAW, (WPARAM)redraw, 0);
}

void CenterWindow(HWND hwnd, HWND outer) throw()
{
	if (!outer)
		outer = ::GetParent(hwnd);
	RECT rcSelf;
	RECT rcParent;

	::GetWindowRect(hwnd, &rcSelf);

	if (!outer || !::IsWindowVisible(outer) || IsIconic(outer))
		::SystemParametersInfo(SPI_GETWORKAREA, NULL, &rcParent, NULL);
	else
		::GetWindowRect(outer, &rcParent);

	::SetWindowPos(hwnd, null,
		(rcParent.left + rcParent.right) / 2 - RECT_W(rcSelf) / 2,
		(rcParent.top + rcParent.bottom) / 2 - RECT_H(rcSelf) / 2,
		0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}

WNDPROC SetWindowProc(HWND hwnd, WNDPROC wndproc)
{
#ifdef _WIN64
	return (WNDPROC)::SetWindowLongPtr(hwnd, GWL_WNDPROC, LONG_PTR(wndproc));
#else
	return (WNDPROC)(LONG_PTR)::SetWindowLongPtr(hwnd, GWL_WNDPROC, (LONG)(LONG_PTR)wndproc);
#endif
}

SIZE GetClientSize(HWND hwnd) throw()
{
	RECT rc;
	::GetClientRect(hwnd, &rc);
	SIZE sz = { rc.right, rc.bottom };
	return sz;
}

void SetClientSize(HWND hwnd, int w, int h) throw()
{
	// HACK: 1񂾂ł͓K؂ȃTCYɂȂȂꍇ邽߁A2sB
	for (int i = 0; i < 2; ++i)
	{
		RECT rcWindow, rcClient;
		::GetWindowRect(hwnd, &rcWindow);
		::GetClientRect(hwnd, &rcClient);
		int dx = w - RECT_W(rcClient);
		int dy = h - RECT_H(rcClient);
		if (dx != 0 || dy != 0)
			::SetWindowPos(hwnd, null, 0, 0,
				RECT_W(rcWindow) + dx, RECT_H(rcWindow) + dy,
				SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
	}
}

bool HasFocus(HWND hwnd) throw()
{
	if (!::IsWindow(hwnd))
		return false;
	HWND focus = ::GetFocus();
	return focus && (hwnd == focus) || ::IsChild(hwnd, focus);
}

void PumpMessage() throw()
{
	ASSERT(!py::HasLock());
	MSG msg;
	while (::PeekMessage(&msg, null, 0, 0, PM_REMOVE))
	{
		::TranslateMessage(&msg);
		::DispatchMessage(&msg);
	}
}

void InitCommonControls(UINT flags) throw()
{
	INITCOMMONCONTROLSEX cc = { sizeof(INITCOMMONCONTROLSEX), flags };
	::InitCommonControlsEx(&cc);
}

UINT GetButtonState() throw()
{
	UINT keys = 0;
	if (IsKeyPressed(VK_LBUTTON))	keys |= MK_LBUTTON;
	if (IsKeyPressed(VK_RBUTTON))	keys |= MK_RBUTTON;
	if (IsKeyPressed(VK_MBUTTON))	keys |= MK_MBUTTON;
	if (IsKeyPressed(VK_XBUTTON1))	keys |= MK_XBUTTON1;
	if (IsKeyPressed(VK_XBUTTON2))	keys |= MK_XBUTTON2;
	return keys;
}

UINT GetModifiers() throw()
{
	UINT keys = GetButtonState();
	if (IsKeyPressed(VK_CONTROL))	keys |= MK_CONTROL;
	if (IsKeyPressed(VK_SHIFT))		keys |= MK_SHIFT;
	if (IsKeyPressed(VK_MENU))		keys |= MK_ALTKEY;
	return keys;
}

void SetKeyState(UINT vkey, UINT mods)
{
	BYTE keys[256];
	GetKeyboardState(keys);
	BYTE vkeyControl = ((mods & MK_CONTROL) ? 0x80 : 0x00);
	BYTE vkeyShift   = ((mods & MK_SHIFT) ? 0x80 : 0x00);
	BYTE vkeyAlt     = ((mods & MK_ALTKEY) ? 0x80 : 0x00);
	BYTE vkeyLButton = ((mods & MK_LBUTTON) ? 0x80 : 0x00);
	BYTE vkeyRButton = ((mods & MK_RBUTTON) ? 0x80 : 0x00);
	BYTE vkeyMButton = ((mods & MK_MBUTTON) ? 0x80 : 0x00);
	BYTE vkeyXButton = ((mods & MK_XBUTTON1) ? 0x80 : 0x00);
	BYTE vkeyYButton = ((mods & MK_XBUTTON2) ? 0x80 : 0x00);
	keys[VK_CONTROL ] = vkeyControl;
	keys[VK_LCONTROL] = vkeyControl;
	keys[VK_RCONTROL] = vkeyControl;
	keys[VK_SHIFT ] = vkeyShift;
	keys[VK_LSHIFT] = vkeyShift;
	keys[VK_RSHIFT] = vkeyShift;
	keys[VK_MENU ] = vkeyAlt;
	keys[VK_LMENU] = vkeyAlt;
	keys[VK_RMENU] = vkeyAlt;
	keys[VK_LBUTTON ] = vkeyLButton;
	keys[VK_RBUTTON ] = vkeyRButton;
	keys[VK_MBUTTON ] = vkeyMButton;
	keys[VK_XBUTTON1] = vkeyXButton;
	keys[VK_XBUTTON2] = vkeyYButton;
	if (vkey)
		keys[vkey] = 0x80;
	SetKeyboardState(keys);
}

namespace
{
	inline void RestoreKey(BYTE* keys, UINT vkey)
	{
		keys[vkey] = (BYTE)(((UINT)GetAsyncKeyState(vkey) & 0x8000U) >> 8);
	}
}

void ResetKeyState(UINT vkey)
{
	BYTE keys[256];
	GetKeyboardState(keys);
	RestoreKey(keys, VK_CONTROL);
	RestoreKey(keys, VK_LCONTROL);
	RestoreKey(keys, VK_RCONTROL);
	RestoreKey(keys, VK_SHIFT);
	RestoreKey(keys, VK_LSHIFT);
	RestoreKey(keys, VK_RSHIFT);
	RestoreKey(keys, VK_MENU);
	RestoreKey(keys, VK_LMENU);
	RestoreKey(keys, VK_RMENU);
	RestoreKey(keys, VK_LBUTTON);
	RestoreKey(keys, VK_RBUTTON);
	RestoreKey(keys, VK_MBUTTON);
	RestoreKey(keys, VK_XBUTTON1);
	RestoreKey(keys, VK_XBUTTON2);
	if (vkey)
		RestoreKey(keys, vkey);
	SetKeyboardState(keys);
}

void BoardcastMessage(UINT msg, WPARAM wParam, LPARAM lParam) throw()
{
	struct Param
	{
		UINT	msg;
		WPARAM	wParam;
		LPARAM	lParam;

		static BOOL CALLBACK Do(HWND hwnd, LPARAM lParam)
		{
			Param* m = (Param*)lParam;
			if (m->msg != WM_DESTROY)
			{
				::PostMessage(hwnd, m->msg, m->wParam, m->lParam);
				::EnumChildWindows(hwnd, &Param::Do, lParam);
			}
			else
				::DestroyWindow(hwnd);
			return TRUE;
		}
	};

	Param m = { msg, wParam, lParam };
	::EnumThreadWindows(::GetCurrentThreadId(), &Param::Do, (LPARAM)&m);
}

//==================================================================================================================================================================

HWND Hwnd::From(IOleWindow* w) throw()
{
	HWND hwnd = null;
	if (!w || FAILED(w->GetWindow(&hwnd)))
		return null;
	return hwnd;
}

HWND Hwnd::From(IUnknown* p) throw()
{
	ref<IOleWindow> w;
	if (!p || FAILED(p->QueryInterface(&w)))
		return null;
	return From(w);
}

void Hwnd::GetPos(RECT* bounds) const throw()
{
	::GetWindowRect(m_hwnd, bounds);
	if (HWND hwndParent = ::GetAncestor(m_hwnd, GA_PARENT))
		::MapWindowRect(null, hwndParent, bounds);
}

void Hwnd::ModifyStyle(DWORD dwRemove, DWORD dwAdd, DWORD dwRemoveEx, DWORD dwAddEx) throw()
{
	if (dwRemove || dwAdd)
	{
		DWORD dwStyle = ::GetWindowLong(m_hwnd, GWL_STYLE);
		DWORD dwNewStyle = (dwStyle & ~dwRemove) | dwAdd;
		if (dwStyle != dwNewStyle)
			::SetWindowLong(m_hwnd, GWL_STYLE, dwNewStyle);
	}
	if (dwRemoveEx || dwAddEx)
	{
		DWORD dwStyleEx = ::GetWindowLong(m_hwnd, GWL_EXSTYLE);
		DWORD dwNewStyleEx = (dwStyleEx & ~dwRemoveEx) | dwAddEx;
		if (dwStyleEx != dwNewStyleEx)
			::SetWindowLong(m_hwnd, GWL_EXSTYLE, dwNewStyleEx);
	}
	SetPos(0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED | SWP_DRAWFRAME);
}

HWND Hwnd::Create(PCWSTR className, HWND parent, DWORD style, DWORD styleEx,
				  int x, int y, int w, int h) throw()
{
	ASSERT(!m_hwnd);
	m_hwnd = ::CreateWindowEx(styleEx, className, null, style, x, y, w, h, parent, null, GetModuleHandle(null), null);
	ASSERT(m_hwnd);
	return m_hwnd;
}

void Hwnd::Activate() throw()
{
	if (::IsIconic(m_hwnd))
		::ShowWindow(m_hwnd, SW_RESTORE);
	if (!::IsWindowVisible(m_hwnd))
		::ShowWindow(m_hwnd, SW_SHOW);

	HWND hwndForeground = GetForegroundWindow();
	if (hwndForeground == m_hwnd)
		return;
	DWORD fg = GetWindowThreadProcessId(hwndForeground, null);
	DWORD me = GetWindowThreadProcessId(m_hwnd, null);
	AttachThreadInput(me, fg, TRUE);

	DWORD current, zero = 0;
	SystemParametersInfo(SPI_GETFOREGROUNDLOCKTIMEOUT, 0, &current, 0);
	SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT, 0, &zero, 0);
	SetForegroundWindow(m_hwnd);
	SystemParametersInfo(SPI_SETFOREGROUNDLOCKTIMEOUT, 0, &current, 0);

	AttachThreadInput(me, fg, FALSE);
}

void Hwnd::TrackMouseEvent(DWORD tme, DWORD time) throw()
{
	TRACKMOUSEEVENT e = { sizeof(TRACKMOUSEEVENT), tme, m_hwnd, time };
	::TrackMouseEvent(&e);
}
