// icons.cpp

#include "stdafx.h"
#include "exception.hpp"
#include "icons.hpp"
#include "drawing.hpp"
#include "math.hpp"

using namespace brownie;

//==================================================================================================================================================================

Icons::Icons()
{
	for (size_t i = 0; i < 3; ++i)
		m_handle[i] = null;
	m_shImageList = null;
}

Icons::~Icons()
{
	for (size_t i = 0; i < 3; ++i)
		if (m_handle[i])
			::ImageList_Destroy(m_handle[i]);
}

namespace
{
	static HIMAGELIST LoadImageListOfPostfix(PCWSTR path, PCWSTR postfix, PCWSTR extension)
	{
		WCHAR buffer[MAX_PATH];
		wcscpy_s(buffer, path);
		::PathRemoveExtension(buffer);
		wcscat_s(buffer, postfix);
		wcscat_s(buffer, extension);
		return ImageListLoad(buffer);
	}
}

void Icons::__init__(Tuple* args, Dict* kw)
{
	PyObject* o = null;
	parse(args, "O", &o);
	unicode u = cast<unicode>(o);
	PCWSTR path = str(u);

	m_handle[0] = ImageListLoad(path);
	if (!m_handle[0])
		throw IOError("Cannot create Icons");

	PCWSTR extension = ::PathFindExtension(path);
	m_handle[1] = LoadImageListOfPostfix(path, L"-hot", extension);
	m_handle[2] = LoadImageListOfPostfix(path, L"-disable", extension);
}

HIMAGELIST Icons::get_Handle(DWORD status) const throw()
{
	if ((status & DISABLED) && m_handle[2])
		return m_handle[2];
	if ((status & HOT) && m_handle[1])
		return m_handle[1];
	return m_handle[0];
}

int Icons::get_Count() const throw()
{
	return ImageList_GetImageCount(m_handle[0]);
}

SIZE Icons::get_Size() const throw()
{
	SIZE sz = { 0, 0 };
	ImageList_GetIconSize(m_handle[0], (int*)&sz.cx, (int*)&sz.cy);
	return sz;
}

HICON Icons::Extract(int index)
{
	if (index < 0)
		return null;
	return ImageList_GetIcon(m_handle[0], index & 0xFFFF, ILD_NORMAL);
}

//==================================================================================================================================================================

object Icons::get_count()	{ return Integer::newobj(get_Count()); }
object Icons::get_size()	{ return Tuple::newobj(get_Size()); }

//==================================================================================================================================================================

PYTHON_MEMBERS(Icons)
	PYTHON_MEMBER(count	, GET)
	PYTHON_MEMBER(size	, GET)
PYTHON_END()

//==================================================================================================================================================================

HIMAGELIST brownie::SHGetImageList(Icons* icons) throw()
{
	return SHGetImageList(icons ? icons->Size.cy : 16);
}

bool brownie::IconsDraw(HDC dc, int icon, int x, int y, DWORD status, Icons* icons, HIMAGELIST shImageList) throw()
{
	if (icon < 0)
		return false;

	IMAGELISTDRAWPARAMS params = { sizeof(IMAGELISTDRAWPARAMS) };

	if (IsShellIcon(icon))
	{
		params.himl = shImageList;
		params.i = LOWORD(icon);
	}
	else if (icons)
	{
		params.himl = icons->Handle[status];
		params.i = icon;
	}
	else
		return false;

	if (!params.himl)
		return false;

	params.hdcDst = dc;
	params.x = x;
	params.y = y;
	params.rgbFg = CLR_DEFAULT;
	params.rgbBk = CLR_NONE;
	return ImageList_DrawIndirect(&params) != 0;
}

SIZE brownie::IconsSize(Icons* icons, HIMAGELIST shImageList, int icon, int minWidth) throw()
{
	SIZE r = { minWidth, 0 };
	if (icons)
	{
		SIZE s = icons->Size;
		if (icon >= 0 && !IsShellIcon(icon))
			r.cy = math::max(r.cy, s.cy);
		r.cx = math::max(r.cx, s.cx);
	}
	if (shImageList)
	{
		int cx, cy;
		if (ImageList_GetIconSize(shImageList, &cx, &cy))
		{
			if (IsShellIcon(icon))
				r.cy = math::max<long>(r.cy, cy);
			r.cx = math::max<long>(r.cx, cx);
		}
	}
	return r;
}
