// itemid.cpp

#include "stdafx.h"
#include "alert.hpp"
#include "config.hpp"
#include "math.hpp"
#include "shell.hpp"
#include "stream.hpp"
#include "system.hpp"

using namespace brownie;

//==================================================================================================================================================================

IStream& operator >> (IStream& stream, const ITEMIDLIST*& v) throw(HRESULT)
{
	UINT32 size;
	stream >> size;
	ILPtr p(size);
	StreamRead(&stream, p, size);
	v = p.detach();
	return stream;
}

IStream& operator << (IStream& stream, const ITEMIDLIST* v) throw(HRESULT)
{
	UINT32 size = ILGetSize(v);
	stream << size;
	StreamWrite(&stream, v, size);
	return stream;
}

ITEMIDLIST* brownie::ILCreate(PCWSTR path_, DWORD* pdwAttr) throw()
{
	if (empty(path_))
		return null;

	WCHAR path[MAX_PATH];
	PathCanonicalize(path, path_);

	// SHILCreateFromPath  / Ȃ߁AOŕϊB
	for (size_t i = 0; path[i]; ++i)
		if (path[i] == L'/')
			path[i] = L'\\';

	// TOOD: Ǝ̃pX߂sȂ΁AŁB

	// ITEMIDLIST 쐬B
	ITEMIDLIST* pidl = null;
	DWORD dwReserved = (pdwAttr ? *pdwAttr : 0);
	::SHILCreateFromPath(path, &pidl, pdwAttr);
	if (pidl)
        return pidl;

	// sB[NAĂ܂̂ŁAłxZbg
	if (pdwAttr)
		*pdwAttr = dwReserved;

	// ":mem:pid" `Ȃ΁ALoRł̎󂯓nB
	size_t ln = len(path);
	if (ln < 4 || path[0] != L':')
		return null;
	PCWSTR nextColon = wcschr(path+1, L':');
	if (!nextColon)
		return null;
	HANDLE mem = (HANDLE)(INT_PTR)cast<int>(path+1);
	DWORD pid = cast<DWORD>(nextColon+1);
	pidl = ILCreate(mem, pid);
	if (pidl && pdwAttr && *pdwAttr != 0)
	{
		if (ILIsRoot(pidl))
		{
			*pdwAttr &= SFGAO_FOLDER | SFGAO_BROWSABLE | SFGAO_HASSUBFOLDER;
		}
		else
		{
			const ITEMIDLIST* leaf = null;
			ref<IShellFolder> parent;
			if SUCCEEDED(ILParentFolder(&parent, pidl, &leaf))
				parent->GetAttributesOf(1, &leaf, pdwAttr);
			else
				*pdwAttr &= SFGAO_FOLDER | SFGAO_BROWSABLE;
		}
	}
	return pidl;
}

ITEMIDLIST* brownie::ILCreate(IShellFolder* folder, PCWSTR relname) throw()
{
	if (!folder || !relname)
		return null;
	ULONG eaten = 0;
	ITEMIDLIST* leaf = null;
	folder->ParseDisplayName(System::GetWindow(), null, const_cast<PWSTR>(relname), &eaten, &leaf, null);
	return leaf;
}

ITEMIDLIST* brownie::ILCreate(int csidl) throw()
{
	if (csidl == CSIDL_MYDOCUMENTS)
		return ILCreate(FOLDER_DOCUMENT);
	ITEMIDLIST* item;
	if FAILED(SHGetSpecialFolderLocation(System::GetWindow(), csidl, &item))
		return null;
	return item;
}

ITEMIDLIST*	brownie::ILCreate(const CIDA* items, size_t index) throw()
{
	const ITEMIDLIST* leaf = CIDAChild(items, index);
	if (!leaf)
		return null;
	const ITEMIDLIST* parent = CIDAParent(items);
	return ILCombine(parent, leaf);
}

ITEMIDLIST* brownie::ILCreate(IShellItem* item) throw()
{
	if (!item)
		return null;
	ref<IPersistIDList> persist;
	ITEMIDLIST* pidl;
	if (FAILED(item->QueryInterface(&persist)) || FAILED(persist->GetIDList(&pidl)))
		return null;
	return pidl;
}

ITEMIDLIST* brownie::ILCreate(HANDLE mem, DWORD pid) throw()
{
	ITEMIDLIST* pidl = null;
	if (ITEMIDLIST* pidlShared = (ITEMIDLIST*)SHLockShared(mem, pid))
	{
		if (!IsBadReadPtr(pidlShared, 1))
			pidl = ILClone(pidlShared);
		SHUnlockShared(pidlShared);
		SHFreeShared(mem, pid);
	}
	return pidl;
}

ITEMIDLIST* brownie::ILCreate(HWND explorer)
{
	DWORD pid = GetCurrentProcessId();
	HANDLE mem = (HANDLE)SendMessage(explorer, WM_USER+12, pid, 0);
	return ILCreate(mem, pid);
}

ITEMIDLIST* brownie::ILCloneParent(const ITEMIDLIST* item) throw()
{
	if (!item || ILIsRoot(item))
		return null;
	ITEMIDLIST* leaf = ILFindLastID(item);
	if (item == leaf)
		return ILCreate(CSIDL_DESKTOP);
	size_t size = ((BYTE*)leaf - (BYTE*)item);
	BYTE* parent = (BYTE*)CoTaskMemAlloc(size + sizeof(USHORT));
	memcpy(parent, item, size);
	*(USHORT*)(parent + size) = 0;
	return (ITEMIDLIST*)parent;
}

bool brownie::ILIsFolder(const ITEMIDLIST* item) throw()
{
	if (!item)
		return false;
	if (ILIsRoot(item))
		return true;
	ref<IShellFolder> folder;
	const ITEMIDLIST* leaf;
	if FAILED(ILParentFolder(&folder, item, &leaf))
		return false;
	SFGAOF flags = SFGAO_FOLDER;
	if FAILED(folder->GetAttributesOf(1, &leaf, &flags))
		return false;
	return (flags & SFGAO_FOLDER) != 0;
}

HRESULT brownie::ILInfo(const ITEMIDLIST* item, WIN32_FIND_DATA* info)
{
	HRESULT hr;
	WCHAR path[MAX_PATH];
	if SUCCEEDED(ILPath(item, path))
	{
		if (HANDLE handle = FindFirstFile(path, info))
		{
			FindClose(handle);
			return S_OK;
		}
	}

	ref<IShellFolder> folder;
	const ITEMIDLIST* leaf;
	if FAILED(hr = ILParentFolder(&folder, item, &leaf))
		return hr;
	return SHGetDataFromIDList(folder, leaf, SHGDFIL_FINDDATA, info, sizeof(WIN32_FIND_DATA));
}

HRESULT brownie::ILParentFolder(IShellFolder** pp, const ITEMIDLIST* item, const ITEMIDLIST** leaf) throw()
{
	return SHBindToParent(item, __uuidof(IShellFolder), (void**)pp, leaf);
}

HRESULT brownie::ILFolder(IShellFolder** pp, const ITEMIDLIST* item) throw()
{
	HRESULT hr;

	if (!item)
		return E_POINTER;
	if (ILIsRoot(item))
		return SHGetDesktopFolder(pp);

	ref<IShellFolder> parent;
	const ITEMIDLIST* leaf;
	if FAILED(hr = ILParentFolder(&parent, item, &leaf))
		return hr;

	return parent->BindToObject(leaf, null, __uuidof(IShellFolder), (void**)pp);
}

HRESULT brownie::ILContextMenu(IContextMenu** pp, const ITEMIDLIST* item) throw()
{
	HRESULT hr;
	ref<IShellFolder> parent;
	const ITEMIDLIST* leaf = null;
	if FAILED(hr = ILParentFolder(&parent, item, &leaf))
		return hr;
	return parent->GetUIObjectOf(System::GetWindow(), 1, &leaf, __uuidof(IContextMenu), null, (void**)pp);
}

HRESULT brownie::ILDisplayName(IShellFolder* folder, const ITEMIDLIST* leaf, DWORD shgdn, WCHAR name[MAX_PATH])
{
	STRRET strret = { 0 };
	HRESULT hr = folder->GetDisplayNameOf(leaf, shgdn, &strret);
	if FAILED(hr)
		return hr;
	return StrRetToBuf(&strret, leaf, name, MAX_PATH);
}

DWORD_PTR brownie::ILFileInfo(const ITEMIDLIST* pidl, SHFILEINFO* info, DWORD flags, DWORD dwFileAttr)
{
	return ::SHGetFileInfo((PCWSTR)pidl, dwFileAttr, info, sizeof(SHFILEINFO), flags | SHGFI_PIDL);
}

int brownie::ILIconIndex(const ITEMIDLIST* pidl)
{
	SHFILEINFO  file;
	if (!ILFileInfo(pidl, &file, SHGFI_SYSICONINDEX | SHGFI_SMALLICON))
		return I_IMAGENONE;
	return file.iIcon;
}

namespace
{
	HRESULT HresultFromShellExecute(HINSTANCE hInstance, PCWSTR path, PCWSTR verb, HWND hwnd)
	{
		if (hInstance > (HINSTANCE)32)
			return S_OK;
		if (hwnd)
		{	// hwnd w肳ꂽꍇ̂݃G[\
			WCHAR buf[1024] = L"";
			switch ((INT_PTR)hInstance)
			{
			case SE_ERR_FNF:
			case SE_ERR_PNF:
				wsprintf(buf, L"\"%s\" ܂", path);
				break;
			case SE_ERR_ACCESSDENIED:
			case SE_ERR_SHARE:
				wsprintf(buf, L"\"%s\" ɃANZXł܂", path);
				break;
			case SE_ERR_OOM: // Out of Memory
				wcscpy_s(buf, L"sĂ邽ߎsł܂ł");
				break;
			case SE_ERR_DLLNOTFOUND:
				wsprintf(buf, L"\"%s\" s邽߂ɕKvDLL܂ł", path);
				break;
			case SE_ERR_ASSOCINCOMPLETE:
			case SE_ERR_NOASSOC:
				wsprintf(buf, L"\"%s\" ɑ΂ ANV \"%s\" ֘AtĂ܂", path, verb);
				break;
			case SE_ERR_DDETIMEOUT:
			case SE_ERR_DDEFAIL:
			case SE_ERR_DDEBUSY:
				wsprintf(buf, L"DDEG[ \"%s\" sł܂ł", path);
				break;
			default:
				wsprintf(buf, L"%s sł܂ł", path);
				break;
			}
			if (!empty(buf))
				Alert(buf, null, MB_OK | MB_ICONERROR);
		}
		return HRESULT_FROM_WIN32(GetLastError());
	}

	HRESULT DoPathExecute(PCWSTR path, PCWSTR verb, PCWSTR args, PCWSTR dir, HWND hwnd)
	{
		ASSERT(!empty(path));

		// path
		WCHAR quoted[MAX_PATH+2];
		// exe
		PCWSTR exe = null;
		WCHAR exepath[MAX_PATH];
		// dir
		WCHAR dirpath[MAX_PATH];
		// verb
		WCHAR verbbuf[MAX_PATH] = L"";
		if (!empty(verb))
		{
			ExpandEnvironmentStrings(verb, verbbuf, MAX_PATH);
			verb = verbbuf;
			if (verb[0] == L'.')
			{	// verb = gq
				if FAILED(brownie::RegGetAssocExe(verb, exepath))
					return E_INVALIDARG; // ֘AtꂽEXE擾łȂ
				exe = exepath;
			}
			else if (*PathFindExtension(verb) == L'.')
			{	// verb = EXEt@Cł͂Ȃ
				exe = verb;
			}
		}

		if (empty(dir))
		{	// [LOfBNǵANt@CƓ
			wcscpy_s(dirpath, path);
			PathRemoveFileSpec(dirpath);
			dir = dirpath;
		}

		if (hwnd)
			hwnd = ::GetAncestor(hwnd, GA_ROOTOWNER);

		HINSTANCE hInstance;
		if (exe)
		{
			if (wcschr(path, L' '))
			{
				size_t ln = len(path);
				quoted[0] = L'\"';
				wcscpy_s(quoted+1, MAX_PATH - 1, path);
				quoted[ln+1] = L'\"';
				quoted[ln+2] = L'\0';
				path = quoted;
			}
#ifdef NOTIMPL
			if (args)
				...;
#endif
			hInstance = ::ShellExecute(hwnd, null, exe, path, dir, SW_SHOWNORMAL);
		}
		else
		{
			hInstance = ::ShellExecute(hwnd, verb, path, args, dir, SW_SHOWNORMAL);
		}
		return HresultFromShellExecute(hInstance, path, verb, hwnd);
	}

	HRESULT DoILExecute(const ITEMIDLIST* pidl, PCWSTR path, PCWSTR verb, PCWSTR args, PCWSTR dir, HWND hwnd)
	{
		SHELLEXECUTEINFO info = { sizeof(SHELLEXECUTEINFO), SEE_MASK_IDLIST, ::GetAncestor(hwnd, GA_ROOTOWNER) };
		if (!hwnd)
			info.fMask |= SEE_MASK_FLAG_NO_UI;
		info.nShow = SW_SHOWNORMAL;
		info.lpIDList = (ITEMIDLIST*)pidl;
		info.lpDirectory = dir;
		info.lpVerb = verb;
		info.lpParameters = args;
		WCHAR dirpath[MAX_PATH];
		if (empty(dir) && !empty(path))
		{	// [LOfBNǵANt@CƓ
			wcscpy_s(dirpath, path);
			PathRemoveFileSpec(dirpath);
			info.lpDirectory = dirpath;
		}
		if (ShellExecuteEx(&info))
			return S_OK;
		return HresultFromShellExecute(info.hInstApp, path, verb, hwnd);
	}

	bool str_replace(WCHAR buf[], size_t bufsize, PCWSTR from, PCWSTR to)
	{
		PWSTR pos = wcsstr(buf, from);
		if (!pos)
			return false;
		size_t ln = len(pos);
		ssize_t from_ln = len(from);
		ssize_t to_ln = len(to);
		memmove(pos + to_ln, pos + from_ln, (ln+1) * sizeof(WCHAR));
		memcpy(pos, to, to_ln * sizeof(WCHAR));
		return true;
	}

	bool split_path(PCWSTR* src, WCHAR dst[MAX_PATH])
	{
		if (!src || !*src)
			return false;
		while (iswspace(**src)) { (*src)++; }

		PCWSTR end;
		switch (**src)
		{
		case L'\0':
			return false;
		case L'"':
			end = wcschr((*src) + 1, L'"');
			if (end)
			{
				wcsncpy_s(dst, MAX_PATH, *src, end - *src + 1);
				*src = end + 1;
			}
			else
			{	// NH[g~XAΏB
				wcscpy_s(dst, MAX_PATH, *src);
				wcscat_s(dst, MAX_PATH, L"\"");
				*src = null;
			}
			return true;
		default:
			end = *src;
			while (*end && !iswspace(*end)) { end++; }
			wcsncpy_s(dst, MAX_PATH, *src, end - *src);
			*src = end;
			return true;
		}
	}
}

HRESULT brownie::ILExecute(const ITEMIDLIST* pidl, PCWSTR verb, PCWSTR args, PCWSTR dir)
{
	if (!pidl)
		return E_INVALIDARG;

	HWND hwnd = System::GetWindow();

	WCHAR path[MAX_PATH] = L"";

	if SUCCEEDED(ILPath(pidl, path))
	{
		if (!empty(verb))
		{	// pX擾łȂꍇ́AǂsłȂB
			HRESULT hr;
			if SUCCEEDED(hr = DoPathExecute(path, verb, args, dir, null))
				return S_OK;
			else
				verb = null; // ̓ł͎słȂB
		}

		// V[gJbg擾
		ref<IShellLink> link;
		ILPtr			target;
		if SUCCEEDED(LinkResolve(path, &target, &link))
		{
			const size_t BUFSIZE = 65536;
			WCHAR arguments[BUFSIZE];

			if (SUCCEEDED(link->GetArguments(arguments, BUFSIZE)) && !empty(arguments))
			{
				// dir
				bool used = str_replace(arguments, BUFSIZE, L"<dir>", dir);

				if (wcsstr(arguments, L"<all>"))
				{	// all
					if (empty(args))
						return S_FALSE;
					str_replace(arguments, BUFSIZE, L"<all>", args);
				}
				else if (wcsstr(arguments, L"<each>"))
				{	// each
					if (empty(args))
						return S_FALSE;

					WCHAR arg[MAX_PATH];
					PCWSTR a = args;
					while (split_path(&a, arg))
					{
						WCHAR tmp[BUFSIZE];
						wcscpy_s(tmp, arguments);
						if (str_replace(tmp, BUFSIZE, L"<each>", arg))
							DoILExecute(target, path, verb, tmp, dir, hwnd);
					}
					return S_OK;
				}
				else if (!used && !empty(args))
				{	// wȂBɒǉB
					if (!empty(arguments))
						wcscat_s(arguments, L" ");
					wcscat_s(arguments, args);
				}

				return DoILExecute(target, path, verb, arguments, dir, hwnd);
			}
		}
	}

	return DoILExecute(pidl, path, verb, args, dir, hwnd);
}

namespace
{
	inline bool IsNull(const ITEMIDLIST* item) throw()
	{
		return !item || item->mkid.cb <= 2;
	}
}

/// ITEMIDLIST rA-1, 0, +1 ̂ÂꂩԂ.
int brownie::ILCompare(const ITEMIDLIST* lhs, const ITEMIDLIST* rhs) throw()
{
	for (;;)
	{
		if (lhs == rhs)	return  0;
		if (IsNull(lhs))
		{
			if (IsNull(rhs))
				return 0;
			else
				return -1;
		}
		else if (IsNull(rhs))
			return 1;

		UINT lhsSize = lhs->mkid.cb - 2;
		UINT rhsSize = rhs->mkid.cb - 2;
		UINT size = math::min(lhsSize, rhsSize);

		if (int r = memcmp(lhs->mkid.abID, rhs->mkid.abID, size))
			return r;

		if (lhsSize < rhsSize) return -1;
		if (lhsSize > rhsSize) return  1;

		lhs = ILGetNext(lhs);
		rhs = ILGetNext(rhs);
	}
}

//==================================================================================================================================================================

CIDA* brownie::CIDACreate(IDataObject* data) throw()
{
	if (!data)
		return null;

	FORMATETC FORMAT_IDLIST = { (CLIPFORMAT)::RegisterClipboardFormat(CFSTR_SHELLIDLIST), null, DVASPECT_CONTENT, -1, TYMED_HGLOBAL };
	STGMEDIUM medium;
	if FAILED(data->GetData(&FORMAT_IDLIST, &medium))
		return null;
	CIDA* src = (CIDA*)::GlobalLock(medium.hGlobal);
	size_t size = CIDASize(src);
	CIDA* dst = (CIDA*)::CoTaskMemAlloc(size);
	memcpy(dst, src, size);
	::GlobalUnlock(medium.hGlobal);
	::GlobalFree(medium.hGlobal);
	return dst;
}

CIDA* brownie::CIDACreate(IShellView* view, int svgio) throw()
{
	if (!view)
		return null;

	// FIXME: XPł SVGIO_CHECKED ȂƂB

	ref<IDataObject> data;
	if FAILED(view->GetItemObject(svgio, IID_PPV_ARGS(&data)))
		return null;

	return CIDACreate(data);
}

CIDA* brownie::CIDACreate(IFolderView* view, int svgio) throw()
{
	if (!view)
		return null;

	// FIXME: XPł SVGIO_CHECKED ȂƂB

	ref<IDataObject> data;
	if FAILED(view->Items(svgio, IID_PPV_ARGS(&data)))
		return null;

	return CIDACreate(data);
}

HGLOBAL brownie::CIDACreate(const ITEMIDLIST* pidl)
{
	const ITEMIDLIST* leaf = ILFindLastID(pidl);
	if (!leaf)
		return NULL;
	size_t pidlSize   = ILGetSize(pidl);
	size_t headerSize = sizeof(UINT)*3;
	size_t parentSize = (BYTE*)leaf - (BYTE*)pidl;
	size_t leafSize   = pidlSize - parentSize - 2;
	size_t totalSize  = headerSize + parentSize + leafSize + 4;

	ASSERT(ILGetSize(leaf) == leafSize + 2);
	ASSERT(pidlSize + 2 + headerSize == totalSize);

	HGLOBAL hGlobal = ::GlobalAlloc(GMEM_MOVEABLE, totalSize);
	CIDA* p = (CIDA*)::GlobalLock(hGlobal);
	memset(p, 0, totalSize);
	p->cidl = 1;
	p->aoffset[0] = (UINT) headerSize;
	p->aoffset[1] = (UINT) headerSize + parentSize + 2;
	// [ (WORD)0 Ȃ̂ŁARs[Ȃimemset()łłɃ[߂Ăj
	memcpy(((BYTE*)p) + p->aoffset[0], pidl, parentSize);
	memcpy(((BYTE*)p) + p->aoffset[1], leaf, leafSize);
	::GlobalUnlock(hGlobal);
	return hGlobal;
}

CIDA* brownie::CIDAClone(const CIDA* items) throw()
{
	size_t size = CIDASize(items);
	CIDA* clone = (CIDA*)::CoTaskMemAlloc(size);
	memcpy(clone, items, size);
	return clone;
}

const ITEMIDLIST* brownie::CIDAParent(const CIDA* items) throw()
{
	if (!items)
		return null;
	return (const ITEMIDLIST*)(((BYTE*)items) + items->aoffset[0]);
}

const ITEMIDLIST* brownie::CIDAChild(const CIDA* items, size_t index) throw()
{
	if (!items || index >= CIDACount(items))
		return null;
	return (const ITEMIDLIST*)(((BYTE*)items) + items->aoffset[index+1]);
}

size_t brownie::CIDASize(const CIDA* items)
{
	ASSERT(items);
	return items->aoffset[items->cidl] + ILGetSize(CIDAChild(items, CIDACount(items)-1));
}

//==================================================================================================================================================================
