﻿// listview.cpp

#include "stdafx.h"
#include "listview.hpp"
#include "node.hpp"
#include "string.hpp"

using namespace brownie;

//==================================================================================================================================================================

void AbstractListView::__init__(Tuple* args, Dict* kw)
{
	abstract(typeof<AbstractListView>(), self);
	m_adjustRequested = false;
	__super::__init__(args, kw);
	DisableIME();
}

void AbstractListView::ListView_OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_UPDATEUISTATE:
		if (m_adjustRequested)
		{
			ListViewH v = ListView_Handle();
			ListView_Adjust(v);
			m_adjustRequested = false;
		}
//		ListView_UpdateOwnerDraw();
		break;
	}
}

void AbstractListView::Adjust()
{
	m_adjustRequested = true;
	Update(true);
}

void AbstractListView::adjust()
{
	Adjust();
}

object AbstractListView::get_selection()
{
	// TODO: proxyオブジェクトを返し、__len__() が呼び出されている限りは
	// ListView_GetSelectedCount() の値を返す。内容を取得される段階になって
	// self.items(SELECTED) で初期化する。
	return get_items(SELECTED);
}

void AbstractListView::set_selection(PyObject* o)
{
	select(o, SVSI_SELECT);
}

object AbstractListView::select(Tuple* args)
{
	int selected = 0;
	ssize_t ln = len(args);
	if (ln > 0)
	{
		selected += select(args->item[0], SVSI_SELECT | SVSI_FOCUSED | SVSI_ENSUREVISIBLE | SVSI_DESELECTOTHERS);
		for (ssize_t i = 1; i < ln; ++i)
			selected += select(args->item[i], SVSI_SELECT);
	}
	return cast<integer>(selected);
}

int AbstractListView::select(PyObject* o, UINT svsi)
{
	ListViewH v = ListView_Handle();
	if (is<Null>(o))
	{	// select none
		v.SetItemState(-1, 0, LVIS_SELECTED);
		return 0;
	}
	else if (is<Integer>(o))
	{	// by index
		int index = cast<int>(o);
		if (index < 0)
		{	// select all
			// Ctrl+A は、「隠しファイルがある」という警告が出てウザい。
			// そのため、代わりに直接 ListViewH で全選択する。
			v.SetItemState(-1, LVIS_SELECTED, LVIS_SELECTED);
			return v.GetItemCount();
		}
		else
		{
			ListView_SelectItem_SVSI(v, index, svsi);
			return 1;
		}
	}
	else if (is<Unicode>(o) || is<String>(o))
	{	// by name or wildcard
		unicode u = cast<unicode>(o);
		PCWSTR pattern = str(u);

		bool (*match)(PCWSTR pattern, PCWSTR text_);
		if (StrPBrk(pattern, L"*?"))
			match = MatchPatternI;
		else
			match = MatchContainsI;

		int selected = 0;
		int ln = ListView_GetItemCount(v);
		for (int i = 0; i < ln; ++i)
		{
			WCHAR text_[MAX_PATH];
			if (!v.GetItemText(i, 0, text_, MAX_PATH))
				continue;
			if (!match(pattern, text_))
				continue;
			ListView_SelectItem_SVSI(v, i, svsi);
			svsi &= ~(SVSI_FOCUSED | SVSI_ENSUREVISIBLE | SVSI_DESELECTOTHERS);
			++selected;
		}
		return selected;
	}
	else if (is_callable(o))
	{	// by filter (callable)
		return select_by_callable(o, svsi);
	}
	else if (object match = getattr(o, "match"))
	{	// by re object
		if (is_callable(match))
			return select_by_callable(match, svsi);
	}

	// for generic iterables
	int selected = 0;
	foreach (object i, o)
	{
		selected += select(i, svsi);
		svsi &= ~(SVSI_FOCUSED | SVSI_ENSUREVISIBLE | SVSI_DESELECTOTHERS);
	}
	else
		throw ValueError("argument is not selectable");
	return selected;
}

int AbstractListView::select_by_callable(PyObject* callable, UINT svsi)
{
	ListViewH v = ListView_Handle();
	int selected = 0;
	int ln = v.GetItemCount();
	for (int i = 0; i < ln; ++i)
	{
		WCHAR text[MAX_PATH];
		if (!v.GetItemText(i, 0, text, MAX_PATH))
			continue;
		if (is<Null>(apply(callable, "u", text)))
			continue;
		ListView_SelectItem_SVSI(v, i, svsi);
		svsi &= ~(SVSI_FOCUSED | SVSI_ENSUREVISIBLE | SVSI_DESELECTOTHERS);
		++selected;
	}
	return selected;
}

void AbstractListView::set_FullRowSelect(bool value)
{
	if (ListViewH v = ListView_Handle())
		v.SetExtendedStyle(value ? LVS_EX_FULLROWSELECT : 0, LVS_EX_FULLROWSELECT);
}

void AbstractListView::set_GridLine(bool value)
{
	if (ListViewH v = ListView_Handle())
		v.SetExtendedStyle(value ? LVS_EX_GRIDLINES : 0, LVS_EX_GRIDLINES);
}

bool AbstractListView::get_Grouping() const
{
	if (ListViewH v = ListView_Handle())
		return v.IsGroupViewEnabled();
	return false;
}

void AbstractListView::set_Grouping(bool value)
{
	if (ListViewH v = ListView_Handle())
		return v.EnableGroupView(value);
}

object AbstractListView::get_grouping()
{
	return Boolean::newobj(this->Grouping);
}

void AbstractListView::set_grouping(PyObject* value)
{
	this->Grouping = cast<bool>(value);
}

PYTHON_METHODS(AbstractListView)
	PYTHON_METHOD(adjust)
	PYTHON_METHOD(select)
PYTHON_END()

PYTHON_MEMBERS(AbstractListView)
	PYTHON_MEMBER(grouping	, BOTH)
	PYTHON_MEMBER(selection	, BOTH)
PYTHON_END()

//==================================================================================================================================================================

void ListView::__init__(Tuple* args, Dict* kw)
{
	HWND parent = __init__parse(args, kw, CENTER);
	InitCommonControls(ICC_LISTVIEW_CLASSES);
	create(WC_LISTVIEW,
		WS_VSCROLL | WS_HSCROLL | LVS_REPORT | LVS_SHAREIMAGELISTS,
		WS_EX_CLIENTEDGE, parent);
	DisableIME();
}

Icons* ListView::GetIcons() const throw()
{
	return cast<Icons>(m_icons);
}

void ListView::SetIcons(Icons* value) throw()
{
	if (m_icons == value)
		return;
	if (value)
	{
		SendMessage(LVM_SETIMAGELIST, LVSIL_SMALL, (LPARAM)value->Handle[NORMAL]);
	}
	else
	{
		SendMessage(LVM_SETIMAGELIST, LVSIL_SMALL);
	}
	m_icons = value;
	NotifyResize();
}

PyObject* ListView::GetItem(int index) const
{
//	return (PyObject*)TreeView_GetItemData(m_hwnd, item);
	return null;
}

void ListView::Clear() throw()
{
	if (!m_items)
		return;

	LockRedraw pause(m_hwnd);
	DeleteAllItems();

	py::Lock lock;
	m_items = null;
}

LRESULT ListView::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_KILLFOCUS:
//		OnKillFocus();
		break;
	case WM_LBUTTONDBLCLK:
		if (OnLButtonDblClk(wParam, lParam))
			return 0;
		break;
	case WM_MBUTTONDOWN:
		break;
	case WM_MBUTTONUP:
		if (OnMButtonUp(wParam, lParam))
			return 0;
		break;
	case WM_MBUTTONDBLCLK:
		break;
	case WM_RBUTTONDOWN:
		break;
	case WM_CONTEXTMENU:
		if (OnContextMenu(lParam))
			return 0;
		break;
	case WM_DESTROY:
		Clear();
		break;
	}

	__super::ListView_OnMessage(msg, wParam, lParam);
	return __super::OnMessage(msg, wParam, lParam);
}

LRESULT	ListView::OnMessage(Window* from, NMHDR& nm)
{
	switch (nm.code)
	{
	case LVN_DELETEITEM:
	case LVN_BEGINDRAG:
	case LVN_BEGINRDRAG:
	case LVN_KEYDOWN:
	case LVN_BEGINLABELEDIT:
	case LVN_ENDLABELEDIT:
		break;
	}
	return __super::OnMessage(from, nm);
}

bool ListView::OnGesture(const MSG& msg)
{
//	ノードの名前変更中をハンドル
//			if (IsChild(msg.hwnd))
//				return true;

	return __super::OnGesture(msg);
}

bool ListView::OnLButtonDblClk(WPARAM wParam, LPARAM lParam)
{
	LVHITTESTINFO hit = { GET_XY_LPARAM(lParam) };
	int index = HitTest(&hit);
	if (index >= 0)
	{
	//	ExecuteItem(index, VK_LBUTTON);
		return true;
	}
	return false;
}

bool ListView::OnMButtonUp(WPARAM wParam, LPARAM lParam)
{
	LVHITTESTINFO hit = { GET_XY_LPARAM(lParam) };
	int index = HitTest(&hit);
	if (index >= 0)
	{
	//	SelectDropTarget(index);
	//	SetKeyState(VK_MBUTTON, GetModifiers() | MK_MBUTTON);
	//	{
	//		py::Lock lock;
	//		node_execute(item);
	//	}
	//	ResetKeyState(VK_MBUTTON);
	//	SelectDropTarget(null);
		return true;
	}
	return false;
}

bool ListView::OnContextMenu(LPARAM lParam)
{
	POINT ptScreen = { GET_XY_LPARAM(lParam) };
	int index;
	if (ptScreen.x == -1 || ptScreen.y == -1)
	{
		index = GetNextItem(-1, LVNI_FOCUSED);
		if (index < 0)
			return false;
		RECT rc;
		GetItemRect(index, &rc, true);
		POINT pt = { rc.left, rc.bottom };
		ClientToScreen(&pt);
		ptScreen = pt;
	}
	else
	{
		LVHITTESTINFO hit = { ptScreen.x, ptScreen.y };
		ScreenToClient(&hit.pt);
		index = HitTest(&hit);
	}

	PyObject* node = GetItem(index);
	if (!node)
		return false;

	return SUCCEEDED(NodePopupMenu(node, ptScreen, cast<Icons>(m_icons)));
}

//==================================================================================================================================================================

void ListView::clear()
{
	py::Unlock unlock;
	Clear();
}

object ListView::get_items(STATUS status)
{
	switch (status)
	{
	case FOCUSED:
		break;
	case SELECTED:
		break;
	case CHECKED:
		break;
	default:
		break;
	}
	return m_items;
}

void ListView::set_columns(PyObject* value)
{
	//clear_columns();
//	ListView_DeleteColumn(__hwnd__, 0);

	ssize_t ln = len(value);
	if (ln <= 0)
		return;

	// TODO: プロキシオブジェクトの作成
	m_columns = Tuple::newobj(value);

	if (value)
	{
		LVCOLUMN column = { LVCF_TEXT | LVCF_SUBITEM | LVCF_WIDTH };
		column.cx = 100;
		foreach (object o, value)
		{
			unicode u = Unicode::newobj(o);
			column.pszText = const_cast<PWSTR>(str(u));
			InsertColumn(column);
			column.iSubItem++;
		}
	}
}

void ListView::set_icons(PyObject* value)
{
	SetIcons(cast<Icons>(value));
}

void ListView::set_items(PyObject* value)
{
	if (m_items == value)
		return;

	clear();

	// TODO: プロキシオブジェクトの作成
	m_items = Tuple::newobj(value);
	ssize_t ln = len(m_items);
	if (ln <= 0)
		return;

	ssize_t columns = len(m_columns);
	if (columns <= 0)
		return;

	foreach (object i, m_items)
	{
		// first item
		int index;
		{
			object v = (columns == 1 ? i : getitem(i, 0));
			LVITEM item = { LVIF_TEXT, INT_MAX };
			unicode	name = node_get_name(v);
			item.pszText = const_cast<PWSTR>(str(name));
			index = InsertItem(item);
		}
		for (ssize_t c = 1; c < columns; c++)
		{
			object v = getitem(i, c);
			LVITEM item = { LVIF_TEXT, index, (int)c };
			unicode	name = node_get_name(v);
			item.pszText = const_cast<PWSTR>(str(name));
			ListView_SetItem(m_hwnd, &item);
		}
	}

	Adjust();
}

//==================================================================================================================================================================

PYTHON_MEMBERS(ListView)
	PYTHON_MEMBER(columns	, SET)
	PYTHON_MEMBER(icons		, SET)
	PYTHON_MEMBER(items		, SET)
PYTHON_END()

//==================================================================================================================================================================
