// menu.cpp

#include "stdafx.h"
#include "icons.hpp"
#include "math.hpp"
#include "node.hpp"
#include "system.hpp"
#include "window.hpp"

using namespace brownie;

namespace
{
	typedef HandleT<HMENU	, DestroyMenu>		Menu;

	inline BOOL AppendMenu(HMENU menu, const MENUITEMINFO& info, UINT pos = -1) throw()
	{
		return ::InsertMenuItem(menu, pos, TRUE, &info);
	}

	// MENUINFO
	//	.dwContextHelpID = IDiTuj[Ô݁j
	//	.dwMenuData = PyObject*
	// MENUITEMINFO
	//	.wID = ID
	//	.dwItemData = PyObject*
	// Nodes
	//	Nodes[ID-1] = PyObject*

	struct MenuStack
	{
		size_t				top;	// length of theMenuNodes
		HMENU				handle;
		ref<IContextMenu>	menu;
	};

	bool					thePopupRunning = false;
	HHOOK					theMenuHook;
	Font					theMenuFont;
	std::vector<MenuStack>	theMenuStack;
	std::vector<HMENU>		theSubMenuStack;
	std::vector<object>		theMenuNodes;
	Icons*					theMenuIcons;
	HIMAGELIST				theSHImageList;
	const UINT	RESERVED_FIRST = 0x7000;
	const UINT	RESERVED_NUM = 1024;
	const UINT	DUMMY_MASK = 0x8000;

	UINT DockToTPM(DIRECTION dock)
	{
		switch (dock)
		{
		case CENTER:return TPM_CENTERALIGN | TPM_VCENTERALIGN;
		case SOUTH:	return TPM_LEFTALIGN   | TPM_BOTTOMALIGN;
		case WEST:	return TPM_RIGHTALIGN  | TPM_TOPALIGN;
		case EAST:	return TPM_LEFTALIGN   | TPM_TOPALIGN;
		default:	return TPM_LEFTALIGN   | TPM_TOPALIGN;
		}
	}

	// [Ԓl, Ԓl+RESERVED_NUM) ͈̔͂ID͗\񂳂̂ŁARɎgpłB
	UINT enter_menu(HMENU handle, IContextMenu* menu)
	{
		MenuStack stack;
		stack.top = theMenuNodes.size();
		stack.handle = handle;
		stack.menu = menu;
		theMenuStack.push_back(stack);
		return RESERVED_FIRST + RESERVED_NUM * (theMenuStack.size() - 1);
	}

	// ܂ɖȊOɔj邱Ƃ邽߁AɑΏB
	void leave_menu(HMENU handle)
	{
		while (!theMenuStack.empty())
		{
			MenuStack stack = theMenuStack.back();
			if (stack.handle != handle && ::IsMenu(stack.handle))
				break;

			theMenuStack.pop_back();
			// FIXME: Ƃ肠ΏBB
		//	theMenuNodes.resize(stack.top);

			if (stack.handle == handle)
				break;
		}

		if (theMenuStack.empty())
			theMenuNodes.clear();

		if (Window* w = Window::From(System::GetWindow()))
			w->event("on_notify", "ui", L"", LOG);
		ASSERT(!theMenuStack.empty() || theMenuNodes.empty());
	}

	inline UINT EnterMenu(HMENU handle, IContextMenu* menu)
	{
		return enter_menu(handle, menu);
	}

	void LeaveMenu(HMENU handle)
	{
		if (theMenuStack.back().top != theMenuNodes.size())
		{
			py::Lock lock;
			leave_menu(handle);
		}
		else
		{
			theMenuStack.pop_back();
			ASSERT(!theMenuStack.empty() || theMenuNodes.empty());
		}
	}

	PyObject* FindNode(UINT wID) throw()
	{
		wID &= ~DUMMY_MASK;
		if (wID == 0 || wID >= RESERVED_FIRST)
			return null;
		if (wID > theMenuNodes.size())
		{
			ASSERT(!"should not happen");
			return null;
		}
		return theMenuNodes[wID-1];
	}

	PyObject* GetNode(UINT wID, DWORD_PTR dwItemData) throw()
	{
		if (dwItemData && (DWORD_PTR)FindNode(wID) == dwItemData)
			return (PyObject*)dwItemData;
		else
			return null;
	}

	PyObject* GetNode(HMENU menu, UINT index) throw()
	{
		MENUITEMINFO item = { sizeof(MENUITEMINFO), MIIM_DATA | MIIM_ID };
		::GetMenuItemInfo(menu, index, TRUE, &item);
		return GetNode(item.wID, item.dwItemData);
	}

	void AppendDummyMenu(HMENU handle, UINT uID, PyObject* o)
	{
		MENUITEMINFO dummy = { sizeof(MENUITEMINFO), MIIM_STRING | MIIM_ID | MIIM_DATA | MIIM_STATE | MIIM_FTYPE };
		dummy.wID = uID | DUMMY_MASK;
		dummy.dwItemData = (DWORD_PTR)o;
		dummy.fState = MFS_DEFAULT;
		dummy.fType = MFT_OWNERDRAW;
		dummy.dwTypeData = L"";
		AppendMenu(handle, dummy);
	}

	UINT register_node(PyObject* o) throw()
	{
		theMenuNodes.push_back(object(o));
		return (UINT)theMenuNodes.size();
	}

	void menu_get_name(WCHAR dst[], PyObject* o)
	{
		unicode	name = node_get_name(o);
		if (!name)
		{
			dst[0] = L'\0';
			return;
		}

		const size_t dstlen = MAX_PATH;
		size_t i = 0;
		PCWSTR src = str(name);
		for (; *src && i < dstlen-1; ++src, ++dst, ++i)
		{
			switch (*src)
			{
			case L'*':
				*dst = L'&';
				continue;
			case L'&':
				if (i > dstlen-2)
					break;
				*dst = *src;
				++dst; ++i;
				break;
			}
			*dst = *src;
		}
		*dst = L'\0';
	}
}

void Menu_Reset()
{
	NONCLIENTMETRICS metrics;
	GetNonClientMetrics(&metrics);
	theMenuFont.Attach(::CreateFontIndirect(&metrics.lfMenuFont));
}

namespace
{
	const int MENU_ICON_MIN_W   = 16;	// ACR̕
	const int MENU_ICON_SPACE_W = 2;	// ACR̍Ě
	const int MENU_ICON_SPACE_H = 2;	// ACȐ㉺̌
	const int MENU_ITEM_PADDING = 2;	// ACRꍇ̍ڂ̌
	const int MENU_TEXT_INDENT  = 2;	// eLXg̍̒ǉ̌
	const int MENU_TEXT_SPACE   = 16;	// eLXg̉Ě ̂Ԃ

	SIZE MenuMeasureItem(PCWSTR name, int icon) throw()
	{
		if (!theMenuFont)
			Menu_Reset();

		if (icon >= 0 && !theSHImageList && IsShellIcon(icon))
		{	// VFACRĂȂ΁AōsB
			theSHImageList = SHGetImageList(theMenuIcons);
		}

		SIZE sz;
		if (empty(name) && icon < 0)
		{
			sz.cx = sz.cy = 0;
		}
		else
		{
			DC dc = CreateCompatibleDC(null);
			SIZE szText = FontSize(theMenuFont, name, dc);
			SIZE szIcon = IconsSize(theMenuIcons, theSHImageList, icon, MENU_ICON_MIN_W);
			sz.cx = szText.cx + szIcon.cx + MENU_ICON_SPACE_W * 2 + MENU_TEXT_INDENT + MENU_TEXT_SPACE;
			sz.cy = math::max<int>(szText.cy + MENU_ITEM_PADDING, szIcon.cy + MENU_ICON_SPACE_H * 2);
		}
		return sz;
	}

	void MenuDrawText(HDC dc, RECT& rc, PCTSTR text, COLORREF color)
	{
		SetTextColor(dc, color);
		int nTab = -1;
		int count = ::lstrlen(text);
		for (int i = 0; i < count; i++)
		{
			if (text[i] == L'\t')
			{
				nTab = i;
				break;
			}
		}
		DrawText(dc, text, nTab, &rc, DT_LEFT | DT_SINGLELINE | DT_VCENTER);
		if (nTab >= 0)
			DrawText(dc, &text[nTab + 1], -1, &rc, DT_RIGHT | DT_SINGLELINE | DT_VCENTER);
	}

	void MenuDrawMark(HDC dc, bool selected, bool checked, bool disabled, const RECT& rc)
	{
		if (!checked)
			return;

		const int nColor = (disabled ? COLOR_BTNSHADOW : (selected ? COLOR_MENU : COLOR_MENUTEXT));
		const COLORREF clrCheck = ::GetSysColor(nColor);
		const int w = rc.right - rc.left;
		const int h = rc.bottom - rc.top;
		RECT rcSource = { 0, 0, w, h };

		// set colors
		const COLORREF clrBlack = RGB(0, 0, 0);
		const COLORREF clrWhite = RGB(255, 255, 255);
		COLORREF clrTextOld = SetTextColor(dc, clrBlack);
		COLORREF clrBkOld = SetBkColor(dc, clrWhite);

		// create source image
		DC dcCheck = CreateCompatibleDC(dc);
		Bitmap bmpCheck = CreateCompatibleBitmap(dc, w, h);
		HBITMAP oldCheck = SelectBitmap(dcCheck, bmpCheck);
		FillRect(dcCheck, rcSource, clrCheck);

		// create mask
		DC dcMask = CreateCompatibleDC(dc);
		Bitmap bmpMask = CreateBitmap(w, h, 1, 1, null);
		HBITMAP oldMask = SelectBitmap(dcMask, bmpMask);
		DrawFrameControl(dcMask, &rcSource, DFC_MENU, DFCS_MENUCHECK);

		// draw the checkmark transparently
		const DWORD ROP_DSno = 0x00BB0226L;
		const DWORD ROP_DSa  = 0x008800C6L;
		const DWORD ROP_DSo  = 0x00EE0086L;
		const DWORD ROP_DSna = 0x00220326L;

		// draw checkmark - special case black and white colors
		if (clrCheck == clrWhite)
		{
			BitBlt(dc, rc.left, rc.top, w, h, dcMask  , 0, 0, ROP_DSno);
			BitBlt(dc, rc.left, rc.top, w, h, dcCheck, 0, 0, ROP_DSa);
		}
		else
		{
			if (clrCheck != clrBlack)
				BitBlt(dcCheck, 0, 0, w, h, dcMask, 0, 0, ROP_DSna);
			BitBlt(dc, rc.left, rc.top, w, h, dcMask  , 0, 0, ROP_DSa);
			BitBlt(dc, rc.left, rc.top, w, h, dcCheck, 0, 0, ROP_DSo);
		}

		// restore all
		SetTextColor(dc, clrTextOld);			
		SetBkColor(dc, clrBkOld);
		SelectBitmap(dcCheck, oldCheck);
		SelectBitmap(dcMask, oldMask);
	}

	void MenuDrawItem(HDC dc, const RECT& rc, DWORD status, PCWSTR name, int icon) throw()
	{
		if (empty(name) && icon < 0)
			return;

		SIZE szIcon = IconsSize(theMenuIcons, theSHImageList, icon, MENU_ICON_MIN_W);

		const int iconw = szIcon.cx + MENU_ICON_SPACE_W * 2;
		const bool selected = (status & SELECTED) != 0;
		const bool disabled = (status & DISABLED) != 0;
		const bool checked  = (status & CHECKED)  != 0;

		RECT rcIcon = { rc.left, rc.top, rc.left + iconw, rc.bottom };
		RECT rcText = { rc.left + iconw, rc.top, rc.right, rc.bottom };

		int menuTextColorIndex;
		if (selected)
		{
			FillRect(dc, rc, GetSysColor(COLOR_HIGHLIGHT));
			menuTextColorIndex = COLOR_HIGHLIGHTTEXT;
		}
		else
		{
			FillRect(dc, rc, GetSysColor(COLOR_MENU));
			menuTextColorIndex = COLOR_MENUTEXT;
		}
		if (empty(name))
		{	// separator
			int y = (rc.top + rc.bottom) / 2;
			DrawLine(dc, rc.left, y, rc.right, y);
		}
		else
		{
			rcText.left += MENU_TEXT_INDENT;
			rcText.right -= MENU_TEXT_SPACE;
			SetBkMode(dc, TRANSPARENT);
			COLORREF color = GetSysColor(disabled ? COLOR_GRAYTEXT : menuTextColorIndex);
			MenuDrawText(dc, rcText, name, color);
			if (icon >= 0)
			{
				IconsDraw(dc, icon,
					rcIcon.left + MENU_ICON_SPACE_W,
					(rcIcon.top + rcIcon.bottom - szIcon.cy) / 2,
					status, theMenuIcons, theSHImageList);
			}
			else
			{
				szIcon.cy = math::min(szIcon.cx, RECT_H(rcIcon) - MENU_ICON_SPACE_H * 2);
				rcIcon.left  += MENU_ICON_SPACE_W;
				rcIcon.right  = rcIcon.left + szIcon.cx;
				rcIcon.top    = (rcIcon.top + rcIcon.bottom - szIcon.cy) / 2;
				rcIcon.bottom = rcIcon.top + szIcon.cy;
				MenuDrawMark(dc, selected, checked, disabled, rcIcon);
			}
		}
	}

	inline DWORD OdsToStatus(DWORD ods)
	{
		DWORD status = 0;
		if (ods & ODS_FOCUS)	status |= FOCUSED;
		if (ods & ODS_CHECKED)	status |= CHECKED;
		if (ods & ODS_HOTLIGHT)	status |= HOT;
		if (ods & ODS_SELECTED)	status |= SELECTED;
		if (ods & (ODS_GRAYED | ODS_DISABLED))
			status |= DISABLED;
		return status;
	}

	bool OnMeasureItem(MEASUREITEMSTRUCT* measure)
	{
		if (measure->CtlType != ODT_MENU)
			return false;

		if (measure->itemID & DUMMY_MASK)
		{	// ẽ_uNbN̑̃_~[̃ftHgACeB`悷Kv͂ȂB
			measure->itemWidth  = measure->itemHeight = 0;
			return true;
		}

		if (PyObject* node = GetNode(measure->itemID, measure->itemData))
		{
			py::Lock lock;
			try
			{
				WCHAR name[MAX_PATH];
				menu_get_name(name, node);
				int icon = node_get_icon(node);
				SIZE sz = MenuMeasureItem(name, icon);
				measure->itemWidth  = sz.cx;
				measure->itemHeight = sz.cy;
				return true;
			}
			catch (Exception&)
			{
				Exception::clear();
			}
		}
		return false;
	}

	bool OnDrawItem(DRAWITEMSTRUCT* draw)
	{
		if (draw->CtlType != ODT_MENU)
			return false;

		if (draw->itemID & DUMMY_MASK)
		{	// ẽ_uNbN̑̃_~[̃ftHgACeB`悷Kv͂ȂB
			return true;
		}

		if (PyObject* node = GetNode(draw->itemID, draw->itemData))
		{
			py::Lock lock;
			try
			{
				WCHAR name[MAX_PATH];
				menu_get_name(name, node);
				int icon = node_get_icon(node);
				MenuDrawItem(draw->hDC, draw->rcItem, OdsToStatus(draw->itemState), name, icon);
				return true;
			}
			catch (Exception&)
			{
				Exception::clear();
			}
		}
		return false;
	}
}

namespace
{
	// j[𖄂߂Błɕ\ׂʒuԂB
	void fill_menu(HMENU menu, Iterator* iter, RECT* rc = null) throw(...)
	{
		if (!menu || !iter)
			return;
		while (object o = iter->next())
		{
			WCHAR namebuf[MAX_PATH];
			MENUITEMINFO item = { sizeof(MENUITEMINFO) };

			if (is<Null>(o))
			{
				item.fMask |= MIIM_FTYPE;
				item.fType |= MFT_SEPARATOR;
			}
			else
			{
				menu_get_name(namebuf, o);

				item.fMask |= MIIM_STRING | MIIM_ID | MIIM_DATA | MIIM_STATE | MIIM_FTYPE;
				item.fType |= MFT_OWNERDRAW; // {IɂׂẴACe̓I[ih[
				item.wID = register_node(o);
				item.dwTypeData = namebuf;
				item.dwItemData = (ULONG_PTR)(PyObject*)o;

				if (node_has_children(o))
				{
					item.fMask |= MIIM_SUBMENU;
					item.hSubMenu = ::CreateMenu();
					MENUINFO info = { sizeof(MENUINFO), MIM_MENUDATA | MIM_HELPID };
					info.dwMenuData = item.dwItemData;
					info.dwContextHelpID = item.wID;
					::SetMenuInfo(item.hSubMenu, &info);
				}

				if (!node_enabled(o) || (!item.hSubMenu && !node_executable(o)))
					item.fState |= MFS_DISABLED;
				if (node_checked(o))
					item.fState |= MFS_CHECKED;
			}

			if (!AppendMenu(menu, item) && item.hSubMenu)
				::DestroyMenu(item.hSubMenu); // }Ɏŝō폜B
		}
	}

	HRESULT GetCommandString(IContextMenu* menu, UINT uID, WCHAR name[MAX_PATH], UINT gcsW, UINT gcsA) throw()
	{
		HRESULT hr;
		if SUCCEEDED(menu->GetCommandString(uID, gcsW, null, (LPSTR)name, MAX_PATH))
			return S_OK;
		CHAR nameA[MAX_PATH];
		if FAILED(hr = menu->GetCommandString(uID, gcsA, null, nameA, MAX_PATH))
			return hr;
		MultiByteToWideChar(CP_ACP, 0, nameA, -1, name, MAX_PATH);
		return S_OK;
	}

	bool IsRenameCommand(IContextMenu* menu, UINT uID) throw()
	{
		WCHAR name[MAX_PATH];
		return SUCCEEDED(GetCommandString(menu, uID, name, GCS_VERBW, GCS_VERBA))
			&& _wcsicmp(name, L"rename") == 0;
	}

	HRESULT Invoke(IContextMenu* menu, UINT id, UINT first, UINT last, HWND hwnd) throw()
	{
		ASSERT(menu);
		if (id == 0)
		{	// LZ
			return S_FALSE;
		}
		else if (first <= id && id <= last)
		{	// VFj[
			// j[SɕɎsB
			EndMenu();
			PumpMessage();

			UINT offset = id - first;
			if (offset < RESERVED_NUM)
			{
				if (IsRenameCommand(menu, offset))
				{
					return S_RENAME;
				}
				else
				{
					CMINVOKECOMMANDINFO info = { sizeof(CMINVOKECOMMANDINFO) };
					info.hwnd   = hwnd;
					info.lpVerb = MAKEINTRESOURCEA(offset);
					info.nShow  = SW_SHOWNORMAL;
					menu->InvokeCommand(&info);
					return S_OK;
				}
			}
		}
		return E_FAIL;
	}

	UINT GetMenuItemType(HMENU handle, INT pos) throw()
	{
		MENUITEMINFO info = { sizeof(MENUITEMINFO), MIIM_FTYPE };
		if (GetMenuItemInfo(handle, pos, TRUE, &info))
			return info.fType;
		else
			return 0;
	}

	HRESULT IContextMenu_AppendItems(IContextMenu* menu, HMENU handle, UINT first, UINT last, RECT* rc = null)
	{
		HRESULT hr;
		if (!menu)
			return E_POINTER;

		// sǉB
		int length = GetMenuItemCount(handle);
		if (length >= 0)
		{
			MENUITEMINFO info = { sizeof(MENUITEMINFO), MIIM_FTYPE };
			if (GetMenuItemInfo(handle, 0, TRUE, &info))
			{
				info.fType |= MFT_MENUBREAK;
				SetMenuItemInfo(handle, 0, TRUE, &info);
			}
			else
			{
				length = 0; // unknown error
			}
		}

		// CMF_CANRENAME : drag-and-drop ̏ꍇ͎gȂׂ
		// CMF_EXTENDEDVERBS : SHIFT+RClick ̏ꍇɊgR}h\ׂ
		if FAILED(hr = menu->QueryContextMenu(handle, 0, first, last, CMF_CANRENAME | CMF_EXPLORE))
			return hr;

		// Zp[^̒Oɉs}邽߁A菜B
		if (length > 0)
		{
			int menubreak = GetMenuItemCount(handle) - length;
			if (GetMenuItemType(handle, menubreak) & MFT_MENUBREAK)
				if (GetMenuItemType(handle, menubreak-1) & MFT_SEPARATOR)
					DeleteMenu(handle, menubreak-1, MF_BYPOSITION);
		}

		return S_OK;
	}

	int estimate_width(HMENU menu, HWND hwnd)
	{
		UINT sum = 48; // FIXME: magic number
		UINT width = 0;

		int count = ::GetMenuItemCount(menu);
		for (int i = 0; i < count; ++i)
		{
			MENUITEMINFO info = { sizeof(MENUITEMINFO), MIIM_DATA | MIIM_ID | MIIM_FTYPE };
			if (GetMenuItemInfo(menu, i, true, &info))
			{
				if (info.fType & MFT_MENUBREAK)
				{
					sum += width;
					width = 0;
				}

				if (info.wID & DUMMY_MASK)
					continue; // fast path for dummy item

				if (PyObject* node = GetNode(info.wID, info.dwItemData))
				{	// python item
					try
					{
						WCHAR name[MAX_PATH];
						menu_get_name(name, node);
						int icon = node_get_icon(node);
						SIZE sz = MenuMeasureItem(name, icon);
						width = math::max<UINT>(width, sz.cx);
					}
					catch (Exception&) { Exception::clear(); }
				}
				else
				{	// shell item
					MEASUREITEMSTRUCT measure;
					measure.CtlID = 0;
					measure.CtlType = ODT_MENU;
					measure.itemData = info.dwItemData;
					measure.itemID = info.wID;
					measure.itemWidth = measure.itemHeight = 0;
					::SendMessage(hwnd, WM_MEASUREITEM, 0, (LPARAM)&measure);
					width = math::max(width, measure.itemWidth);
				}
			}
		}

		return sum + width;
	}

	inline UINT TPM_GETVALIGN(UINT tpm)
	{
		return tpm & (TPM_TOPALIGN | TPM_VCENTERALIGN | TPM_BOTTOMALIGN);
	}

	UINT adjust_position(HMENU menu, HWND hwnd, POINT* pt, UINT tpm)
	{
		// Pɍ̓_Ŕ肷̂AłҒʂ̈ʒuɂȂ悤B
		HMONITOR monitor = MonitorFromPoint(*pt, MONITOR_DEFAULTTONEAREST);
		MONITORINFO info = { sizeof(MONITORINFO) };
		if (GetMonitorInfo(monitor, &info))
		{
			int width = estimate_width(menu, hwnd);
			if (width > RECT_W(info.rcWork))
				return TPM_GETVALIGN(tpm) | TPM_LEFTALIGN; // LBǂ悤ȂB
			if (pt->x < info.rcWork.left)
				pt->x = info.rcWork.left;
			if (pt->x + width > info.rcWork.right)
			{
				pt->x = info.rcWork.right - 1;
				return TPM_GETVALIGN(tpm) | TPM_RIGHTALIGN;
			}
		}
		return tpm;
	}

	HRESULT menu_popup(Iterator* iter, IContextMenu* menu, POINT pt, HWND hwnd, UINT tpm = TPM_TOPALIGN | TPM_LEFTALIGN) throw()
	{
		if (!iter && !menu)
			return S_FALSE;

		Menu handle(::CreatePopupMenu());

//		{	[hXj[H
//			MENUINFO info = { sizeof(MENUINFO), MIM_STYLE };
//			GetMenuInfo(handle, &info);
//			info.dwStyle |= MNS_MODELESS | MNS_DRAGDROP;
//			SetMenuInfo(handle, &info);
//		}

		INT first = enter_menu(handle, menu);
		INT last = first + RESERVED_NUM - 1;

		HRESULT hr = E_FAIL;

		try
		{
			RECT rc = { 0 };
			fill_menu(handle, iter, &rc);

			PyObject* node;

			{
				py::Unlock unlock;
				IContextMenu_AppendItems(menu, handle, first, last, &rc);
			}

			tpm = adjust_position(handle, hwnd, &pt, tpm);

			{
				py::Unlock unlock;
				INT id = ::TrackPopupMenuEx(handle, tpm | TPM_RETURNCMD | TPM_RECURSE, pt.x, pt.y, hwnd, null);
				node = FindNode(id);
				if (!node && menu)
					hr = Invoke(menu, id, first, last, hwnd);
			}

			if (node)
			{
				node_execute(node);
				hr = S_OK;
			}
		}
		catch (Exception&)
		{
			// ł̓bZ[W{bNXoȂSB
			// FIXME: ł̓G[ʒmłȂȂĂ܂B
			Exception::clear();
			hr = E_FAIL;
		}

		leave_menu(handle);
		return hr;
	}
}

namespace
{
	void OnUninitMenuPopup(HMENU handle) throw()
	{
		std::vector<HMENU>::iterator i = std::find(theSubMenuStack.begin(), theSubMenuStack.end(), handle);
		if (i != theSubMenuStack.end())
			theSubMenuStack.erase(i);
	}

	void OnInitMenuPopup(HMENU handle, UINT index) throw()
	{
		theSubMenuStack.push_back(handle);

		MENUINFO info = { sizeof(MENUINFO), MIM_MENUDATA | MIM_HELPID };
		::GetMenuInfo(handle, &info);
		PyObject* o = GetNode(info.dwContextHelpID, info.dwMenuData);
		if (!o)
		{	// ɏĂB
			return;
		}

		{
			py::Lock lock;
			iterator iter = node_enum_children(o);
			fill_menu(handle, iter);

			if (GetMenuItemCount(handle) == 0)
			{
				MENUITEMINFO item = { sizeof(MENUITEMINFO), MIIM_STRING | MIIM_ID | MIIM_STATE };
				item.fState = MFS_GRAYED | MFS_DISABLED;
				item.dwTypeData = L"(empty)";
				AppendMenu(handle, item);
			}

			if (node_executable(o))
			{	// R}hƎq̗j[
				AppendDummyMenu(handle, info.dwContextHelpID, o);
			}
		}

		info.fMask = MIM_HELPID;
		info.dwContextHelpID = 0; // ɌĂ΂ꂽƂ̂ݏB
		::SetMenuInfo(handle, &info);
	}

	void OnMenuRButtonUp(HMENU handle, UINT index) throw()
	{
		if (PyObject* o = GetNode(handle, index))
		{
			py::Lock lock;
			ref<IContextMenu> menu;
			iterator iter = node_enum_menu(o, &menu);
			if (iter || menu)
			{
				POINT pt;
				::GetCursorPos(&pt);
				if (menu_popup(iter, menu, pt, System::GetWindow()) == S_OK)
				{	// qj[ꍇ́Aej[B
					// XXX: ĂsقǂH
					EndMenu();
				}
			}
		}
	}

	bool OnMenuChar(UINT type, UINT ch, HMENU hMenu, LRESULT& lResult)
	{
		if (type == MF_POPUP && ch == ' ')
		{
			int count = ::GetMenuItemCount(hMenu);
			for (int i = 0; i < count; ++i)
			{
				if (GetMenuState(hMenu, i, MF_BYPOSITION) & MFS_HILITE)
				{
					RECT rc;
					::GetMenuItemRect(null, hMenu, i, &rc);
					POINT pt = { RECT_CENTER(rc) };
					DoubleClick(pt);
					// MNC_IGNORE MNC_CLOSE MNC_EXECUTE MNC_SELECT
					lResult = MAKELRESULT(i, MNC_SELECT);
					return true;
				}
			}
		}
		return false;
	}

	bool IsMenuWindow(HWND hwnd)
	{
		WCHAR clsname[7];
		return ::GetClassName(hwnd, clsname, 7) && wcscmp(L"#32768", clsname) == 0;
	}

	PyObject* MenuItemFromPoint(HMENU handle, POINT pt, HMENU* outMenu, int* outIndex)
	{
		int index = ::MenuItemFromPoint(null, handle, pt);
		if (index >= 0)
		{
			*outMenu = handle;
			*outIndex = index;
			return GetNode(handle, index);
		}
		return null;
	}

	PyObject* MenuItemFromPoint(POINT pt, HMENU* outMenu, int* outIndex)
	{
		if (!IsMenuWindow(WindowFromPoint(pt)))
			return null;

		for (ssize_t i = theSubMenuStack.size() - 1; i >= 0; --i)
		{
			if (PyObject* r = MenuItemFromPoint(theSubMenuStack[i], pt, outMenu, outIndex))
				return r;
		}
		
		for (ssize_t i = theMenuStack.size() - 1; i >= 0; --i)
		{
			if (PyObject* r = MenuItemFromPoint(theMenuStack[i].handle, pt, outMenu, outIndex))
				return r;
		}
		return null;
	}

	LRESULT CALLBACK MenuMouseHook(int code, WPARAM wParam, LPARAM lParam)
	{
		if (code == HC_ACTION)
		{
			MSLLHOOKSTRUCT* info = (MSLLHOOKSTRUCT*)lParam;
			HMENU	menu;
			int		index;

			switch (wParam)
			{
			case WM_RBUTTONUP:
				// Tuj[JĂƁA̐ej[ł WM_MENURBUTTONUP 
				// ȂȂB̂߁AOŃCxg𔭐B
				if (MenuItemFromPoint(info->pt, &menu, &index))
					PostMessage(System::GetWindow(), WM_MENURCLICK, index, (LPARAM)menu);
				break;

			case WM_MBUTTONUP:
				// j[ACe𒆃NbNŋNł悤ɂB
				if (PyObject* node = MenuItemFromPoint(info->pt, &menu, &index))
				{
					EndMenu();
					NodeExecute(node, VK_MBUTTON, MK_MBUTTON);
				}
				break;
			}
		}
		return ::CallNextHookEx(theMenuHook, code, wParam, lParam);
	}
}

namespace
{
	inline bool IsMenuMessage(UINT msg)
	{
		switch (msg)
		{
		case WM_MENUDRAG:
		case WM_MENUGETOBJECT:
		case WM_MENUSELECT:
		case WM_INITMENUPOPUP:
		case WM_UNINITMENUPOPUP:
		case WM_MENUCHAR:
		case WM_MENURBUTTONUP:
		case WM_MEASUREITEM:
		case WM_DRAWITEM:
			return true;
		default:
			return false;
		}
	}
}

bool Menu_OnMessage(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam, LRESULT& lResult) throw()
{
	if (!theMenuStack.empty() && IsMenuMessage(msg))
	{
		for (ssize_t i = theMenuStack.size() - 1; i >= 0; --i)
		{
			Window* w = Window::From(hwnd);
			if (IContextMenu* menu = theMenuStack[i].menu)
			{
				// Iڂ̃wveLXgʒmB
				if (w && msg == WM_MENUSELECT)
				{
					WCHAR	helptext[MAX_PATH] = L"";
					if (!(HIWORD(wParam) & MF_POPUP))
					{
						UINT	id = LOWORD(wParam);
						UINT	first = RESERVED_FIRST + RESERVED_NUM * i;
						UINT	offset = id - first;
						if (offset < RESERVED_NUM)
							GetCommandString(menu, offset, helptext, GCS_HELPTEXTW, GCS_HELPTEXTA);
					}
					w->Event("on_notify", "ui", helptext, LOG);
				}

				// IContextMenu ɏB
				ref<IContextMenu3> menu3;
				ref<IContextMenu2> menu2;
				if SUCCEEDED(menu->QueryInterface(&menu3))
				{
					if (menu3->HandleMenuMsg2(msg, wParam, lParam, &lResult) == S_OK)
						return true;
				}
				else if SUCCEEDED(menu->QueryInterface(&menu2))
				{
					if (menu3->HandleMenuMsg(msg, wParam, lParam) == S_OK)
						return true;
				}
			}
		}
	}

	switch (msg)
	{
///	MNS_DRAGDROP
//	case WM_MENUDRAG:
//	case WM_MENUGETOBJECT:
	case WM_INITMENUPOPUP:
		OnInitMenuPopup((HMENU)wParam, LOWORD(lParam));
		return true;
	case WM_UNINITMENUPOPUP:
		OnUninitMenuPopup((HMENU)wParam);
		break;
	case WM_MENUCHAR:
		return OnMenuChar(HIWORD(wParam), LOWORD(wParam), (HMENU)lParam, lResult);
	case WM_MENURBUTTONUP:
		return true; // ̃bZ[W͎gȂBɁgB
	case WM_MENURCLICK:
		OnMenuRButtonUp((HMENU)lParam, (UINT)wParam);
		return true;
	case WM_MEASUREITEM:
		if (wParam == 0) // by menu
			return 0 != (lResult = OnMeasureItem((MEASUREITEMSTRUCT*)lParam));
		break;
	case WM_DRAWITEM:
		if (wParam == 0) // by menu
			return 0 != (lResult = OnDrawItem((DRAWITEMSTRUCT*)lParam));
		break;
	}
	return false;
}

PyObject* brownie::popup(PyObject*, PyObject* args, PyObject* kw) throw()
{
	try
	{
		POINT		pt;
		DIRECTION	dock = NORTH;
		PyObject*	node = null;
		PyObject*	o_icons = null;
		PyObject*	where = null;
		PCSTR kwds[] = { "node", "icons", "where", null };
		parse((Tuple*)args, (Dict*)kw, "O|OO", kwds, &node, &o_icons, &where);

		Icons*		icons = cast<Icons>(o_icons);

		if (is<Window>(where))
		{
			Window* w = cast<Window>(where);
			RECT rc;
			::GetWindowRect(w->Handle, &rc);
			pt.x = RECT_CENTER_W(rc);
			pt.y = RECT_CENTER_H(rc);
			dock = CENTER;
		}
		else if (where)
			pt = cast<POINT>(where);
		else
			pt = ::GetCursorPos();

		if (S_OK == node_popup_children(node, pt, dock, icons))
			Py_RETURN_TRUE;
		else
			Py_RETURN_FALSE;
	}
	catch (Exception&)
	{
		return null;
	}
}

HRESULT brownie::node_popup(Iterator* iter, IContextMenu* menu, POINT pt, DIRECTION dock, Icons* icons) throw()
{
	HRESULT	hr;

	ASSERT(!theMenuHook);

	thePopupRunning = true;
	theSubMenuStack.clear();
	if (icons)
	{
		theMenuIcons = icons;
		theSHImageList = SHGetImageList(icons);
	}
	theMenuHook = ::SetWindowsHookEx(WH_MOUSE, MenuMouseHook, ::GetModuleHandle(null), GetCurrentThreadId());

	hr = menu_popup(iter, menu, pt, System::GetWindow(), DockToTPM(dock));

	::UnhookWindowsHookEx(theMenuHook);
	theMenuHook = null;
	theMenuIcons = null;
	theSHImageList = null;
	theSubMenuStack.clear();

	// j[̑ŃWFX`\̂邽߁A
	// j[̓}EX{^ׂė̂҂B
	// A^CAEg5bB
	// FIXME: uWFX`}vƂړI͓ɂ炸A
	// GestureNXƂ͕قȂBZ\H
	DWORD timeout = ::GetTickCount() + 5 * 1000;
	while (GetButtonState() != 0 && timeout < ::GetTickCount())
	{
		MSG msg;
		::GetMessage(&msg, 0, 0, 0);
		if (msg.message < WM_MOUSEFIRST && WM_MOUSELAST < msg.message)
		{
			::TranslateMessage(&msg);
			::DispatchMessage(&msg);
		}
	}

	thePopupRunning = false;

	return hr;
}

HRESULT brownie::node_popup_menu(PyObject* node, POINT pt, Icons* icons) throw()
{
	if (is<Null>(node))
		return E_INVALIDARG;

	ref<IContextMenu> menu;
	iterator iter = node_enum_menu(node, &menu);
	if (!iter && !menu)
		return E_NOTSUPPORT;
	return node_popup(iter, menu, pt, NORTH, icons);
}

HRESULT brownie::node_popup_children(PyObject* node, POINT pt, DIRECTION dock, Icons* icons) throw()
{
	if (is<Null>(node))
		return E_INVALIDARG;

	iterator iter = node_enum_children(node);
	if (!iter)
		return E_NOTSUPPORT;
	return node_popup(iter, null, pt, dock, icons);
}

HRESULT brownie::NodePopupMenu(PyObject* node, POINT pt, Icons* icons) throw()
{
	if (is<Null>(node))
		return E_INVALIDARG;
	py::Lock lock;
	return node_popup_menu(node, pt, icons);
}

HRESULT brownie::NodePopupChildren(PyObject* node, const RECT& rc, DIRECTION dock, Icons* icons) throw()
{
	if (is<Null>(node))
		return E_INVALIDARG;

	POINT pt;
	switch (dock)
	{
	case SOUTH:	pt = RECT_NW(rc); break;
	case WEST:	pt = RECT_NE(rc); break;
	case EAST:	pt = RECT_NW(rc); break;
	default:	pt = RECT_SW(rc); break;
	}

	py::Lock lock;
	return node_popup_children(node, pt, dock, icons);
}

// j[[v܂̓EBhËړETCY̏ꍇ͐^.
bool IsLocalLoop() throw()
{
	if (thePopupRunning)
		return true;
	GUITHREADINFO info = { sizeof(GUITHREADINFO) };
	return ::GetGUIThreadInfo(::GetCurrentThreadId(), &info) && (info.flags & (GUI_INMENUMODE | GUI_INMOVESIZE)) != 0;
}

HRESULT brownie::IShellView_Popup(IShellView* view, POINT pt, ICommDlgBrowser2* browser) throw()
{
	const UINT	ID_DEFAULT = 0xFFFF;

	HRESULT hr;
	ref<IContextMenu> menu;
	if FAILED(hr = view->GetItemObject(SVGIO_SELECTION, IID_PPV_ARGS(&menu)))
		return S_FALSE;

	Menu handle(::CreatePopupMenu());
	INT first = EnterMenu(handle, menu);
	INT last = first + RESERVED_NUM - 1;
	IContextMenu_AppendItems(menu, handle, first, last);

	WCHAR text[MAX_PATH];
	hr = browser->GetDefaultMenuText(view, text, MAX_PATH);
	if (hr == S_OK && !empty(text))
	{
		MENUITEMINFO item = { sizeof(MENUITEMINFO) };

		// 2Ԗڂ̓Zp[^
		item.fMask = MIIM_FTYPE;
		item.fType = MFT_SEPARATOR;
		InsertMenuItem(handle, 0, true, &item);

		// 1Ԗڂ̓ftHg
		item.fMask = MIIM_STRING | MIIM_ID;
		item.wID = ID_DEFAULT;
		item.dwTypeData = text;
		InsertMenuItem(handle, 0, true, &item);

		// ftHgw MFS_DEFAULT ł͂ȂASetMenuDefauItem ōsB
		SetMenuDefaultItem(handle, 0, true);
	}

	HWND	hwnd = System::GetWindow();
	INT		id = ::TrackPopupMenuEx(handle, TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RETURNCMD, pt.x, pt.y, hwnd, null);
	hr = Invoke(menu, id, first, last, hwnd);
	LeaveMenu(handle);

	return id == ID_DEFAULT ? S_DEFAULT : hr;
}
