// rebar.cpp

#include "stdafx.h"
#include "drawing.hpp"
#include "rebar.hpp"
#include "node.hpp"

using namespace brownie;

//==================================================================================================================================================================

void ReBar::__init__(Tuple* args, Dict* kw)
{
	HWND parent = __init__parse(args, kw, NORTH);
	InitCommonControls(ICC_COOL_CLASSES);
	create(REBARCLASSNAME,
		RBS_DBLCLKTOGGLE | RBS_VARHEIGHT | RBS_BANDBORDERS | CCS_NODIVIDER | CCS_NOPARENTALIGN | DockToCCS(Dock),
		WS_EX_COMPOSITED, parent);
	DisableIME();
	REBARINFO info = { sizeof(REBARINFO) };
	SetBarInfo(&info);
	OnReset();
}

LRESULT ReBar::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_PARENTNOTIFY:
		switch (LOWORD(wParam))
		{
		case WM_CREATE:
		case WM_DESTROY:
			OnChildCreate(LOWORD(wParam) == WM_CREATE, (HWND)lParam);
			break;
		}
		break;
	case WM_SETTINGCHANGE:
		OnReset();
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

void ReBar::OnChildResize(Window* child)
{
	int index = FindBand(child->Handle);
	if (index >= 0)
	{
		SIZE sz = child->DefaultSize;
		bool visible = IsVisible(child->Handle);
		ReBar_Band band(RBBIM_CHILDSIZE | RBBIM_IDEALSIZE);
		if (band.cxIdeal != (UINT)sz.cx ||
		   band.cxMinChild != (UINT)sz.cx ||
		   band.cyMinChild != (UINT)sz.cy ||
		   band.Visible != visible)
		{
			band.cxIdeal = sz.cx;
			band.cxMinChild = sz.cx;
			band.cyMinChild = sz.cy;
			band.Visible = visible;
			SetBandInfo(index, &band);
			Update(true);
		}
	}
}

void ReBar::OnReset() throw()
{
	SetFont(GetStockFont(DEFAULT_GUI_FONT));
}

LRESULT ReBar::OnMessage(Window* from, NMHDR& nm)
{
	switch (nm.code)
	{
	case RBN_HEIGHTCHANGE:
		from->Update(true);
		break;
	}
	return __super::OnMessage(from, nm);
}

void ReBar::OnChildCreate(bool create, Hwnd child) throw()
{
	if (child.Parent == m_hwnd && HasStyle(child, WS_CHILD))
	{
		if (create)
		{
			SIZE size;
			if (Window* w = From(child))
				size = w->DefaultSize;
			else
				size.cx = size.cy = 32; // Ƃ肠
			child.ModifyStyle(0, 0, WS_EX_COMPOSITED, WS_EX_TRANSPARENT);
			InsertBand(child, size);
		}
		else
		{
			DeleteBand(child);
		}
	}
}

void ReBar::OnLayout()
{
	SyncBands();
//	__super::OnLayout();
}

//==================================================================================================================================================================

PYTHON_MEMBERS(ReBar)
	PYTHON_MEMBER(locked, BOTH)
PYTHON_END()

//==================================================================================================================================================================

object ReBar::get_locked()
{
	return Boolean::newobj(GetLocked());
}

void ReBar::set_locked(PyObject* value)
{
	SetLocked(cast<bool>(value));
}


#if 0

public: // IPersistMessage
	void LoadFromMessage(const message& msg)
	{
		m_locked = msg["fixed"];
		if (message archive = msg["bands"])
		{
			// łɒǉꂽohёւAVǉقyȂ̂ŁB
			typedef std::vector<ReBar_Band> Bands;
			Bands bands;
			ReBar_Band band(RBBIM_SIZE | RBBIM_STYLE | RBBIM_CHILD | RBBIM_CHILDSIZE);
			for (int i = 0; GetBandInfo(i, &band); i++)
			{
				bands.push_back(band);
			}
			size_t size = bands.size();
			for (size_t i = 0; i < size; i++)
			{
				DeleteBand(0);
				if (message msg = archive[i])
				{
					string name = msg["name"];
					if (name.empty())
						break;
					UINT32 width     = msg["width"]     | 0;
					bool   linebreak = msg["linebreak"] | true;
					bool   visible   = msg["visible"]   | true;
					for (size_t j = 0; j < size; j++)
					{
						TCHAR wndName[MAX_PATH];
						::GetWindowText(bands[j].hwndChild, wndName, MAX_PATH);
						if (name == wndName)
						{
							bands[j].cx = width;
							bands[j].fStyle = RBBS_GRIPPERALWAYS | (linebreak ? RBBS_BREAK : 0) | (visible ? 0 : RBBS_HIDDEN);
							if (i != j)
								std::swap(bands[i], bands[j]);
							break;
						}
					}
				}
			}
			ASSERT(GetBandCount() == 0);
			for (size_t i = 0; i < size; ++i)
			{
				InsertBand(i, &bands[i]);
				ref<IWindow> w;
				if SUCCEEDED(QueryInterfaceInWindow(bands[i].hwndChild, &w))
					w->Visible = bands[i].Visible;
			}
			if (m_locked)
			{	// ohЂƂȂꍇRBBS_NOGRIPPERݒ肷ƁAIRBBS_GRIPPERALWAYSĂ܂B
				// h߁AInsertBand()SetBandInfo(~RBBS_NOGRIPPER)ʁXɍsB
				UpdateFixedStyle();
			}
		}
	}
	message SaveToMessage()
	{
		message msg;
		msg["fixed"] = m_locked;
		message archive;
		ReBar_Band band(RBBIM_SIZE | RBBIM_STYLE | RBBIM_CHILD);
		for (int i = 0; GetBandInfo(i, &band); i++)
		{
			TCHAR name[MAX_PATH];
			::GetWindowText(band.hwndChild, name, MAX_PATH);
			message msg;
			msg["name"]      = name;
			msg["width"]     = (UINT32)band.cx;
			msg["linebreak"] = band.Break;
			msg["visible"]   = band.Visible;
			archive[i] = msg;
		}
		msg["bands"]   = archive;
		return msg;
	}
};

#endif
