// shell.cpp

#include "stdafx.h"
#include "shell.hpp"
#include "hwnd.hpp"
#include "stream.hpp"
#include "icons.hpp"

using namespace brownie;

//==================================================================================================================================================================

namespace
{
	#define INVALID_FUNCTION_POINTER	((void*)(UINT_PTR)-1)

	// ̊֐͔rsȂėǂB
	// ʂɂ͂Ȃ邪Aǂ|C^̂ŁB
	static bool DynamicLoadVoidPtr(void** fn, PCWSTR module, PCSTR name, int index = 0) throw()
	{
		ASSERT(fn);
		if (*fn == INVALID_FUNCTION_POINTER)
			return false;
		if (*fn)
			return true;
		HINSTANCE hDLL = brownie::DllLoad(module);
		*fn = ::GetProcAddress(hDLL, name);
		if (!*fn && index > 0)
			*fn = ::GetProcAddress(hDLL, (PCSTR)(INT_PTR)index);
		if (*fn)
			return true;
		*fn = INVALID_FUNCTION_POINTER;
		return false;
	}

	template < typename T >
	inline bool DynamicLoad(T** fn, PCWSTR module, PCSTR name, int index = 0) throw()
	{
		return DynamicLoadVoidPtr((void**)fn, module, name, index);
	}
}

HRESULT brownie::ItemCreate(IShellItem** pp, PCWSTR path)
{
	static HRESULT (__stdcall *fn)(PCWSTR, IBindCtx*, REFIID, void**) = null;
	if (DynamicLoad(&fn, L"shell32.dll", "SHCreateItemFromParsingName"))
		return fn(path, null, IID_PPV_ARGS(pp));

	return ItemCreate(pp, ILPtr(ILCreate(path)));
}

HRESULT brownie::ItemCreate(IShellItem** pp, IShellFolder* folder, const ITEMIDLIST* parent, const ITEMIDLIST* leaf)
{
	static HRESULT (__stdcall *fn)(PCIDLIST_ABSOLUTE, IShellFolder*, PCUITEMID_CHILD, REFIID, void**) = null;
	if (DynamicLoad(&fn, L"shell32.dll", "SHCreateItemWithParent"))
		return fn(parent, folder, leaf, IID_PPV_ARGS(pp));

	return SHCreateShellItem(parent, folder, leaf, pp);
}

HRESULT brownie::ItemCreate(IShellItem** pp, const ITEMIDLIST* item)
{
	static HRESULT (__stdcall *fn)(PCIDLIST_ABSOLUTE, REFIID, void**) = null;
	if (DynamicLoad(&fn, L"shell32.dll", "SHCreateItemFromIDList"))
		return fn(item, IID_PPV_ARGS(pp));

	HRESULT hr;
	ref<IShellFolder> folder;
	const ITEMIDLIST* leaf = null;
	if FAILED(hr = ILParentFolder(&folder, item, &leaf))
		return hr;
	return SHCreateShellItem(ILPtr(ILCloneParent(item)), folder, leaf, pp);
}

HRESULT brownie::ItemCreate(IShellItem** pp, IShellItem* parent, PCWSTR name)
{
	static HRESULT (__stdcall *fn)(IShellItem*, PCWSTR, IBindCtx*, REFIID, void**) = null;
	if (DynamicLoad(&fn, L"shell32.dll", "SHCreateItemFromRelativeName"))
		return fn(parent, name, null, IID_PPV_ARGS(pp));

	ref<IShellFolder> folder;
	if SUCCEEDED(parent->BindToHandler(null, BHID_SFObject, IID_PPV_ARGS(&folder)))
		if (ILPtr item = ILCreate(parent))
			if (ILPtr leaf = ILCreate(folder, name))
				return ItemCreate(pp, folder, item, leaf);
	
	return E_FAIL;
}

//==================================================================================================================================================================

HRESULT brownie::HDropToString(HDROP hDrop, IStream* stream, PCWSTR sep)
{
	if (!hDrop || !stream)
		return E_INVALIDARG;

	try
	{
		size_t sepln = len(sep);
		int count = DragQueryFile(hDrop, 0xFFFFFFFF, NULL, 0);
		bool first = true;
		for(int i = 0; i < count; ++i)
		{
			WCHAR path[MAX_PATH];
			int ln = DragQueryFile(hDrop, i, path, MAX_PATH);
			if (ln > 0)
			{
				if (first)
					first = false;
				else
					StreamWriteT(stream, sep, sepln);
				StreamWriteT(stream, path, ln);
			}
		}
		StreamWriteT(stream, L"\0", 1);
		return S_OK;
	}
	catch (HRESULT hr)
	{
		return hr;
	}
}

//==================================================================================================================================================================

HRESULT brownie::LinkResolve(PCWSTR src, ITEMIDLIST** dst, IShellLink** pp) throw()
{
	HRESULT hr;
	ref<IShellLink> link;
	ref<IPersistFile> persist;

	SUCCEEDED(hr = link.newobj(CLSID_ShellLink)) &&
	SUCCEEDED(hr = link->QueryInterface(&persist)) &&
	SUCCEEDED(hr = persist->Load(src, STGM_READ | STGM_SHARE_DENY_NONE)) &&
	SUCCEEDED(hr = link->GetIDList(dst));

	if (pp)
		link.copyto(pp);
	return hr;
}

HRESULT brownie::LinkResolve(PCWSTR src, WCHAR dst[MAX_PATH], IShellLink** pp) throw()
{
	HRESULT hr;
	ref<IShellLink> link;
	ref<IPersistFile> persist;

	SUCCEEDED(hr = link.newobj(CLSID_ShellLink)) &&
	SUCCEEDED(hr = link->QueryInterface(&persist)) &&
	SUCCEEDED(hr = persist->Load(src, STGM_READ | STGM_SHARE_DENY_NONE)) &&
	SUCCEEDED(hr = link->GetPath(dst, MAX_PATH, null, 0));

	if (pp)
		link.copyto(pp);
	return hr;
}

HRESULT brownie::LinkCreate(const ITEMIDLIST* item, PCWSTR dst, IShellLink** pp) throw()
{
	HRESULT hr;
	ref<IShellLink> link;
	ref<IPersistFile> persist;

	// t@Cɑ΂ăV[gJbg쐬ꍇ́A
	// [LOfBNgA܂ސetH_ɐݒ肷B
	WCHAR dir[MAX_PATH] = L"";
	if (SUCCEEDED(ILPath(item, dir)) && !PathIsDirectory(dir))
		PathRemoveFileSpec(dir);

	SUCCEEDED(hr = link.newobj(CLSID_ShellLink)) &&
	SUCCEEDED(hr = link->SetIDList(item)) &&
	(empty(dir) || SUCCEEDED(hr = link->SetWorkingDirectory(dir))) &&
	SUCCEEDED(hr = link->QueryInterface(&persist)) &&
	SUCCEEDED(hr = persist->Save(dst, STGM_WRITE | STGM_CREATE | STGM_SHARE_EXCLUSIVE));

	if (pp)
		link.copyto(pp);
	return hr;
}

//==================================================================================================================================================================

#if 1
void brownie::SHObserveBegin(HWND hwnd, UINT msg)	{ }
void brownie::SHObserveEnd()	{ }
HRESULT brownie::SHObserveNotify(WPARAM wParam, LPARAM lParam)	{ return S_OK; }
#else

namespace
{
	const DWORD SHCNE_GENERIC_CREATE = SHCNE_CREATE | SHCNE_MKDIR | SHCNE_DRIVEADD | SHCNE_MEDIAINSERTED;
	const DWORD SHCNE_GENERIC_DELETE = SHCNE_DELETE | SHCNE_RMDIR | SHCNE_DRIVEREMOVED | SHCNE_MEDIAREMOVED;
	const DWORD SHCNE_GENERIC_RENAME = SHCNE_RENAMEFOLDER | SHCNE_RENAMEITEM;
	const DWORD SHCNE_GENERIC_EVENTS = SHCNE_GENERIC_CREATE | SHCNE_GENERIC_DELETE | SHCNE_GENERIC_RENAME;

	static void OnSHCreate(const ITEMIDLIST* pidl)
	{
		TRACE(L"SHCNE_GENERIC_CREATE");
	}
	static void OnSHDelete(const ITEMIDLIST* pidl)
	{
		TRACE(L"SHCNE_GENERIC_DELETE");
	}
	static void OnSHRename(const ITEMIDLIST* pidlOld, const ITEMIDLIST* pidlNew)
	{
		TRACE(L"SHCNE_GENERIC_RENAME");
	}

	ULONG		theSystemNotify;
}

void brownie::SHObserveBegin(HWND hwnd, UINT msg)
{
	ASSERT(theSystemNotify == 0);
	ILPtr desktop;
	if SUCCEEDED(::SHGetSpecialFolderLocation(hwnd, CSIDL_DESKTOP, &desktop))
	{
		const DWORD SHCNRF_SOURCE = SHCNRF_NewDelivery | SHCNRF_ShellLevel;
		SHChangeNotifyEntry entry = { desktop, true };
		theSystemNotify = ::SHChangeNotifyRegister(hwnd, SHCNRF_SOURCE, SHCNE_GENERIC_EVENTS, msg, 1, &entry);
	}
}

void brownie::SHObserveEnd()
{
	if (theSystemNotify)
	{
		::SHChangeNotifyDeregister(theSystemNotify);
		theSystemNotify = 0;
	}
}

HRESULT brownie::SHObserveNotify(WPARAM wParam, LPARAM lParam)
{
	LONG what = 0;
	ITEMIDLIST** pidls = null;
	HANDLE hLock = ::SHChangeNotification_Lock((HANDLE)wParam, (DWORD)lParam, &pidls, &what);
	if (!hLock)
		return E_FAIL;
	__try
	{
		if (what & SHCNE_GENERIC_CREATE)
			OnSHCreate(pidls[0]);
		if (what & SHCNE_GENERIC_DELETE)
			OnSHDelete(pidls[0]);
		if (what & SHCNE_GENERIC_RENAME)
			OnSHRename(pidls[0], pidls[1]);
	}
	__finally
	{
		::SHChangeNotification_Unlock(hLock);
	}
	return S_OK;
}

#endif

//==================================================================================================================================================================

namespace
{
	UINT __stdcall PrivateExtractIconsStab(PCWSTR lpszFile, int nIconIndex, int cxIcon, int cyIcon, HICON *phicon, UINT *piconid, UINT nIcons, UINT flags)
	{
		for (UINT i = 0; i < nIcons; ++i)
			piconid[i] = 0;
		if (cxIcon > 16 || cyIcon > 16)
			return ExtractIconEx(lpszFile, nIconIndex, phicon, null, nIcons);
		else
			return ExtractIconEx(lpszFile, nIconIndex, null, phicon, nIcons);
	}
}

HICON brownie::IconLoad(PCWSTR filename, int index, int w, int h)
{
	static UINT (__stdcall *fn)(PCWSTR lpszFile, int nIconIndex, int cxIcon, int cyIcon, HICON *phicon, UINT *piconid, UINT nIcons, UINT flags) = null;
	if (!fn)
	{
		if (!DynamicLoad(&fn, L"user32.dll", "PrivateExtractIconsW"))
			fn = PrivateExtractIconsStab;
	}
	HICON hIcon;
	UINT id;
	if (fn(filename, index, w, h, &hIcon, &id, 1, 0) == 1)
		return hIcon;
	else
		return null;
}

// SHIL_LARGE          0   // normally 32x32
// SHIL_SMALL          1   // normally 16x16
// SHIL_EXTRALARGE     2
// SHIL_SYSSMALL       3   // like SHIL_SMALL, but tracks system small icon metric correctly
// SHIL_JUMBO          4   // normally 256x256

namespace
{
	int GetIconIndex(PCWSTR path)
	{
		if (!path)
			return 0;
		SHFILEINFO  file;
		if (!SHGetFileInfo(path, 0, &file, sizeof(file), SHGFI_SYSICONINDEX | SHGFI_SMALLICON))
			return 0;
		return file.iIcon;
	}

	int GetIconIndex(int csidl)
	{
		ITEMIDLIST* pidl = null;
		if FAILED(SHGetSpecialFolderLocation(null, csidl, &pidl))
			return 0;
		int ret = ILIconIndex(pidl);
		ILDelete(pidl);
		return ret;
	}

	int GetDriveIconIndex(UINT type)
	{
		DWORD drives = GetLogicalDrives();
		// FDD(A:)㏑Ă܂̂ŁA2(C:)n߂
		for (int i = 2; i < 26; ++i)
		{
			if (drives & (1 << i))
			{
				WCHAR path[] = L"*:\\";
				path[0] = (WCHAR)(L'A' + i);
				if (GetDriveType(path) == type)
					return GetIconIndex(path);
			}
		}
		return 0;
	}
/*
	int ExpGetDummyFile(PCWSTR leaf)
	{
		TCHAR path[MAX_PATH];
		::PathCombine(path, PathOf(Var), leaf);
		if (!::PathFileExists(path))
			brownie::NewDocument(path);
		return GetIconIndex(path);
	}
*/
	int IconID_Index(int ID)
	{
		switch (ID)
		{
		case 0: // `̃t@C
		case 1: // W̕ 
			return GetIconIndex(L"unknown");
		case 2: // W EXE, COM t@C
			return GetIconIndex(L"unknown.exe");
		case 3: // Ԃ̃tH_ 
			// XXX: exeuĂpX́AʂȃACRw肳ĂȂƉ肷
			return GetIconIndex(PATH_BIN);
		case 4: // JĂtH_ 
			break;
		case 5: // Floppy Drive (5.25)
		case 6: // Floppy Drive (3.5) 
			return GetIconIndex(L"A:\\");
		case 7: // [ouhCu
			return GetDriveIconIndex(DRIVE_REMOVABLE);
		case 8: // n[hhCu 
			return GetDriveIconIndex(DRIVE_FIXED);
		case 9: // lbg[NhCu 
			return GetDriveIconIndex(DRIVE_REMOTE);
		case 10: // lbg[NhCuiItCj 
			break;
		case 11: // CD Drive 
			return GetDriveIconIndex(DRIVE_CDROM);
		case 12: // z RAM hCu 
			return GetDriveIconIndex(DRIVE_RAMDISK);
		case 13: // lbg[NS 
			return GetIconIndex(CSIDL_NETWORK);
		case 17: // lbg[NRs[^ 
			break;
		case 18: // [NO[v 
			break;
		case 19: // vOiX^[gj[j 
			return GetIconIndex(CSIDL_STARTMENU);
		case 20: // ŋߎgt@CiX^[gj[j 
			return GetIconIndex(CSIDL_RECENT);
		case 21: // ݒiX^[gj[j 
		case 22: // iX^[gj[j 
		case 23: // wviX^[gj[j 
		case 24: // t@Cw肵ĎsiX^[gj[j 
		case 25: // TXyhiX^[gj[j 
		case 26: // hbLOXe[ViX^[gj[j 
		case 27: // Windows ̏IiX^[gj[j 
		case 28: // VFAtH_pACR 
		case 29: // V[gJbg̏ 
			break;
		case 34: // fXNgbv
			return GetIconIndex(CSIDL_DESKTOP);
		case 36: // vOO[v 
		case 40: // y CD 
		case 44: // OIt 
		default:
			break;
		}
		return -1;
	}

	void OverrideIcons(IImageList* imagelist)
	{		
		HUSKEY hKey;
		if (ERROR_SUCCESS == SHRegOpenUSKey(L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Icons", KEY_READ, null, &hKey, TRUE))
		{
			const DWORD LEN = 512;
			WCHAR cchName[LEN], cchValue[LEN];
			DWORD nName, nValue, nType;
			for (DWORD index = 0;
				nName = LEN, nValue = LEN,
				ERROR_SUCCESS == SHRegEnumUSValue(hKey, index, cchName, &nName, &nType, cchValue, &nValue, SHREGENUM_HKLM);
				++index)
			{
				// 'Shell Icons' format:
				// name: <icon id>
				// value: "<resource path>,<icon index>"
				if (PWSTR comma = wcsrchr(cchValue, L','))
				{
					int iconID = _wtoi(cchName);
					int resID = _wtoi(comma+1);
					*comma = L'\0';
					int iconIndex = IconID_Index(iconID);
					if (iconIndex >= 0)
					{
						if (!imagelist)
							continue;
						int w, h;
						if FAILED(imagelist->GetIconSize(&w, &h))
							continue;
						if (HICON hIcon = IconLoad(cchValue, resID, w, h))
						{
							int ret;
							imagelist->ReplaceIcon(iconIndex, hIcon, &ret);
							::DestroyIcon(hIcon);
						}
					}
				}
			}
			SHRegCloseUSKey(hKey);
		}
    }
}

HIMAGELIST brownie::SHGetImageList(int size)
{
	static ref<IImageList> imagelists[SHIL_LAST+1];

	int what;
	if (size < 32)
		what = SHIL_SMALL;
	else if (size < 48)
		what = SHIL_LARGE;
	else if (WINDOWS_VERSION < WINDOWS_VISTA || size < 256)
		what = SHIL_EXTRALARGE;
	else
		what = SHIL_JUMBO;

	ref<IImageList>& image = imagelists[what];

	if (!image &&
		SUCCEEDED(SHGetImageList(what, IID_PPV_ARGS(&image))) &&
		WINDOWS_VERSION < WINDOWS_VISTA)
	{
		COLORREF c; // gȂ null ɂ segfault Ă܂B
		image->SetBkColor(CLR_NONE, &c);
		OverrideIcons(image); // TODO: VISTAłŒׂB܂̂ƂXP̂݁B
	}

	if (image)
	{
		// MSDN says: The IImageList pointer type can be cast as an HIMAGELIST as needed.
		return (HIMAGELIST)image.ptr;
	}

	return null;
}

BOOL brownie::AwareDPI() throw()
{
	static BOOL (*fn)() = null;
	if (DynamicLoad(&fn, L"user32.dll", "SetProcessDPIAware"))
		return fn();
	return FALSE;
}
