// statusbar.cpp

#include "stdafx.h"
#include "statusbar.hpp"
#include <vssym32.h>

using namespace brownie;

//==================================================================================================================================================================

void StatusBar::__init__(Tuple* args, Dict* kw)
{
	__super::__init__(args, kw, SOUTH, 0, 0);
}

LRESULT StatusBar::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_NCHITTEST:
	{
		POINT pt = { GET_XY_LPARAM(lParam) };
		ScreenToClient(&pt);
		if (pt.x >= GetItem(GetItemCount())->offset)
			return HTBOTTOMRIGHT;
		break;
	}
	case WM_CREATE:
		if (LRESULT lResult = __super::OnMessage(msg, wParam, lParam))
			return lResult;
		m_themeSB.Create(m_hwnd, VSCLASS_STATUS);
		m_themeTB.Create(m_hwnd, VSCLASS_TOOLBAR);
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

PCWSTR StatusBar::get_Value() const throw()
{
	return static_cast<Unicode*>(m_value.ptr)->str();
}

namespace
{
	const int	TEXT_MARGIN = 4;
	const int	GRIPPER_WIDTH = 24;
}

void StatusBar::OnLayout()
{
	__super::OnLayout();

	bool		modified = false;
	ssize_t		count = GetItemCount();

	if (count > 0)
	{
		Item*	items = GetItems();
		Icons*	icons = GetIcons();
		SIZE	szIcon = IconsSize(icons, null, 16, 16);
		HFONT	font = GetFont();

		HDC		dc = GetDC();

		std::vector<int> width(count+1);
		width[0] = 0;
		for (ssize_t i = 0; i < count; ++i)
		{
			int w = TEXT_MARGIN * 2;
			if (items[i].icon >= 0)
				w += szIcon.cx + TEXT_MARGIN;
			w += FontSize(font, str(items[i].name), dc).cx;
			width[i+1] = width[i] + w;
		}
		ReleaseDC(dc);

		int	offset = ClientSize.cx - GRIPPER_WIDTH - width[count];
		for (ssize_t i = 0; i <= count; ++i)
		{
			if (items[i].offset != width[i] + offset)
			{
				items[i].offset = width[i] + offset;
				modified = true;
			}
		}
	}
	if (modified)
		Invalidate();
}

void StatusBar::OnDraw(HDC dc, const RECT& rcPaint)
{
	RECT		rc = ClientArea;
	const Item*	items = GetItems();
	ssize_t		count = GetItemCount();

	HFONT		font = GetFont();
	HFONT		old = SelectFont(dc, font);

	m_themeSB.Edge(dc, SP_PANE, 0, &rc, BDR_SUNKENOUTER, BF_TOP, &rc);

	// E[
	rc.left = items[count].offset;
	m_themeSB.Background(dc, SP_GRIPPER, 0, &rc);

	// [
	rc.left = 0;
	rc.right = items[0].offset;
	FillRect(dc, rc, GetSysColorBrush(COLOR_3DFACE));
	InflateRect(&rc, -TEXT_MARGIN, 0);
	m_themeSB.Text(dc, SP_PANE, 0, &rc, get_Value(), DT_LEFT | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX);

	if (count <= 0)
		return;

	Icons*		icons = GetIcons();
	HIMAGELIST	shImageList = SHGetImageList(icons);
	for (ssize_t i = 0; i < count; ++i)
	{
		rc.left = items[i].offset;
		rc.right = items[i+1].offset;

		int		mode;
		STATUS	status;
		const int modes[] = { TS_NORMAL, TS_HOT, TS_PRESSED, TS_DISABLED, TS_CHECKED, TS_HOTCHECKED };
		GetStatus(&mode, &status, i, modes);

		m_themeTB.Background(dc, TP_BUTTON, mode, &rc);
		InflateRect(&rc, -TEXT_MARGIN, 0);

		if (items[i].icon >= 0)
		{
			SIZE	szIcon = IconsSize(icons, shImageList, items[i].icon, 0);
			IconsDraw(dc, items[i].icon, rc.left, (rc.top + rc.bottom - szIcon.cy) / 2, status, icons, shImageList);
			rc.left += szIcon.cx + TEXT_MARGIN;
		}

		m_themeSB.Text(dc, SP_PANE, 0, &rc, str(items[i].name), DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX, status);
	}
	SelectFont(dc, old);
}

//==================================================================================================================================================================

void StatusBar::set_value(PyObject* value)
{
	m_value = cast<unicode>(value);
	RECT rc = ClientArea;
	rc.left = 0;
	rc.right = GetItem(0)->offset;
	Invalidate(rc);
}

//==================================================================================================================================================================

PYTHON_MEMBERS(StatusBar)
	PYTHON_MEMBER(value		, SET)
PYTHON_END()
