// string.cpp

#include "stdafx.h"
#include "config.hpp"
#include "string.hpp"
#include <map>
using namespace brownie;

namespace
{
	inline bool eq(WCHAR lhs, WCHAR rhs) throw()	{ return ChrCmpI(lhs, rhs) == 0; }
	inline PCWSTR inc(PCWSTR s) throw()				{ return s + 1; }
}

bool brownie::MatchContainsI(PCWSTR pattern, PCWSTR text) throw()
{
	return StrStrI(text, pattern) != null;
}

bool brownie::MatchPatternI(PCWSTR pattern, PCWSTR text) throw()
{
	PCWSTR cp = 0;
	PCWSTR mp = 0;
	PCWSTR text_ = text;

	while (*text_ && *pattern != ';' && *pattern != '*')
	{
		if (!eq(*pattern, *text_) && (*pattern != '?'))
			return false;
		pattern = inc(pattern);
		text_ = inc(text_);
	}
	if (*pattern != ';')
	{
		while (*text_)
		{
			if (*pattern == '*')
			{
				pattern = inc(pattern);
				if (!*pattern || *pattern == ';')
					return true;
				mp = pattern;
				cp = inc(text_);
			}
			else if (eq(*pattern, *text_) || (*pattern == '?'))
			{
				pattern = inc(pattern);
				text_ = inc(text_);
			}
			else
			{
				pattern = mp;
				text_ = cp;
				cp = inc(cp);
			}
		}
		while (*pattern == '*')
		{
			pattern = inc(pattern);
		}
	}
	if (*pattern == ';')
	{
		return true;
	}
	if (*pattern)
	{
		mp = pattern;
		while ((*mp) && (*mp != ';'))
		{
			mp = inc(mp);
		}
		if (*mp == ';')
		{
			pattern = inc(mp);
			return MatchPatternI(pattern, text);
		}
	}

	return !*pattern;
}

namespace
{
	inline bool CharIsPrintable(UINT c)
	{
		return c > 127 || isprint(c);
	}

	class Map
	{
	private:
		struct Value
		{
			PCSTR	utf8;
			PCWSTR	utf16;
		};
		typedef std::map<PCSTR, Value, Less>	map_t;
		mutable map_t	m_map;

	public:
		Map()
		{
			WCHAR path[MAX_PATH];
			PathCombine(path, PATH_ROOT, L"res\\message.txt");
			FILE* f;
			if (_wfopen_s(&f, path, L"rt") == 0)
			{
				CHAR line[1024];
				while (fgets(line, lengthof(line), f))
				{
					PSTR key = line;
					while (isspace(*key)) { ++key; }
					switch (*key)
					{
					case '\0':
					case '#':
						break; // skip brank or comment line
					}
					if (PSTR s = strchr(key, '\t'))
					{
						// trim head
						while (!CharIsPrintable(*s))
							*s++ = '\0';
						// trim tail
						PSTR e = s + len(s) - 1;
						while (e > s && !CharIsPrintable(*e))
							*e-- = '\0';
						Value value = { _strdup(s), null };
						m_map.insert(map_t::value_type(_strdup(key), value));
					}
				}
				fclose(f);
			}
		}
		~Map()
		{
			for (map_t::iterator i = m_map.begin(); i != m_map.end(); ++i)
			{
				free(const_cast<PSTR>(i->first));
				free(const_cast<PSTR>(i->second.utf8));
				free(const_cast<PWSTR>(i->second.utf16));
			}
		}

		PCSTR UTF8(PCSTR name) const throw()
		{
			map_t::iterator i = m_map.find(name);
			if (i == m_map.end())
			{
#ifdef _DEBUG
				if (name[len(name)-1] != '_')
				{
					TRACE(L"UTF8: %S", name);
					ASSERT(!"message.txt");
				}
#endif
				return name;
			}
			return i->second.utf8;
		}

		PCWSTR UTF16(PCSTR name) const throw()
		{
			map_t::iterator i = m_map.find(name);
			if (i == m_map.end())
				return null;
			if (!i->second.utf16)
			{
				ssize_t ln = len(i->second.utf8) + 1;
				PWSTR buffer = (PWSTR)malloc(sizeof(WCHAR) * ln);
				MultiByteToWideChar(CP_UTF8, 0, i->second.utf8, -1, buffer, (int)ln);
				i->second.utf16 = buffer;
			}
			return i->second.utf16;
		}
	};

	const Map& GetMap() throw()
	{
		static Map theMap;
		return theMap;
	}
}

PCSTR  brownie::_(PCSTR name) throw()
{
	return GetMap().UTF8(name);
}

PCSTR brownie::_(PCSTR category, PCSTR name) throw()
{
	CHAR buffer[MAX_PATH];
	sprintf_s(buffer, "%s.%s", category, name);
	PCSTR r = GetMap().UTF8(buffer);
	return r == buffer ? null : r;
}

PCWSTR brownie::_W(PCSTR name) throw()
{
	return GetMap().UTF16(name);
}
