// tabbar.cpp

#include "stdafx.h"
#include "tabbar.hpp"
#include <vssym32.h>

using namespace brownie;

//==================================================================================================================================================================

void TabBar::__init__(Tuple* args, Dict* kw)
{
	m_selection = -1;
	__super::__init__(args, kw);
}

SIZE TabBar::get_DefaultSize() const throw()
{
	SIZE	sz = __super::get_DefaultSize();
	switch (Dock)
	{
	case NORTH:
	case SOUTH:
		sz.cx = TAB_MAXLEN;
		break;
	case EAST:
	case WEST:
		sz.cy = TAB_MAXLEN;
		break;
	}
	return sz;
}

void TabBar::SetSelectedItem(ssize_t index) throw()
{
	if (m_selection == index)
		return;
	RECT rc;
	if (m_selection >= 0)
	{
		GetItemRect(m_selection, &rc);
		Invalidate(rc);
	}
	const Item* item = GetItem(index);
	if (item && item->enabled)
	{
		m_selection = index;
		GetItemRect(m_selection, &rc);
		Invalidate(rc);
	}
	else
		m_selection = -1;
}

LRESULT TabBar::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_CREATE:
		if (LRESULT lResult = __super::OnMessage(msg, wParam, lParam))
			return lResult;
		m_theme.Create(m_hwnd, VSCLASS_TAB);
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

namespace
{
	const int TEXT_MARGIN = 4;
	const int THEME_PART[] =
	{
		TABP_TABITEM,
		TABP_TABITEMBOTHEDGE,
		TABP_TABITEMLEFTEDGE,
		TABP_TABITEMRIGHTEDGE,
		0,
	};
}

void TabBar::OnDraw(HDC dc, const RECT& rcPaint)
{
	int			part = THEME_PART[Dock];
	const Item*	items = GetItems();
	ssize_t		count = GetItemCount();
	RECT		rc = ClientArea;
	rc.bottom -= 1;

	HFONT	font = GetFont();
	HFONT	old = SelectFont(dc, font);
	for (ssize_t i = 0; i < count; ++i)
	{
		rc.left = items[i].offset;
		rc.right = items[i+1].offset + 2;

		int		mode;
		STATUS	status;
		if (i == m_selection)
		{
			mode = TTIS_SELECTED;
			status = SELECTED;
		}
		else
		{
			const int modes[] = { TTIS_NORMAL, TTIS_HOT, TTIS_SELECTED, TTIS_DISABLED, TTIS_FOCUSED, TTIS_FOCUSED };
			GetStatus(&mode, &status, i, modes);
		}

		m_theme.Background(dc, part, mode, &rc);
		InflateRect(&rc, -TEXT_MARGIN, 0);
		m_theme.Text(dc, part, mode, &rc, str(items[i].name), DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX, status);
	}
	SelectFont(dc, old);
}

void TabBar::OnLayout()
{
	__super::OnLayout();

	ssize_t		count = GetItemCount();

	if (count > 0)
	{
		SIZE sz = ClientSize;
		int		offset = 0;
		Item*	items = GetItems();
		HDC		dc = GetDC();
		for (ssize_t i = 0; i <= count; ++i)
		{
			items[i].offset = offset;
			if (TAB_MAXLEN * count < sz.cx)
				offset += TAB_MAXLEN;
			else
				offset = (int)((i+1) * sz.cx / count);
		}
		ReleaseDC(dc);
	}
	Invalidate();
}

//==================================================================================================================================================================

object TabBar::get_selection()
{
	return Integer::newobj((int)GetSelectedItem());
}

void TabBar::set_selection(PyObject* value)
{
	SetSelectedItem(cast<ssize_t>(value));
}

//==================================================================================================================================================================

PYTHON_MEMBERS(TabBar)
	PYTHON_MEMBER(selection, BOTH)
PYTHON_END()

//==================================================================================================================================================================

/*
	case WM_MOUSEMOVE:
		if (m_dragTab >= 0 && (GET_KEYSTATE_WPARAM(wParam) & MK_LBUTTON))
		{
			int index = PointToIndex(GET_XY_LPARAM(lParam));
			if (index >= 0 && m_dragTab != index)
			{
				Tab::Swap(m_tabs[m_dragTab], m_tabs[index]);
				m_dragTab = index;
				m_dragSwap = true;
				if (Window* target = cast<Window>(m_target))
					target->Update(true);
				Invalidate();
			}
		}
		else
			m_dragTab = -1;
		return 0;
	case WM_LBUTTONDOWN:
		m_dragTab = -1;
		m_dragSwap = false;
		if ((GET_KEYSTATE_WPARAM(wParam) & MK_BUTTONS) == MK_LBUTTON)
		{	// {^̂قɂ͉ĂȂ
			m_dragTab = PointToIndex(GET_XY_LPARAM(lParam));
		}
		return 0;
	case WM_LBUTTONUP:
		if (!m_dragSwap)
		{
			int index = PointToIndex(GET_XY_LPARAM(lParam));
			if (index >= 0 && index == m_dragTab)
				::SetFocus(m_tabs[index].hwnd);
		}
		m_dragTab = -1;
		m_dragSwap = false;
		return 0;
	}
*/
