// textview.cpp

#include "stdafx.h"
#include "drawing.hpp"
#include "textview.hpp"
#include "path.hpp"

using namespace brownie;

namespace
{
	HFONT GetDefaultFont(int size)
	{
		HFONT font = (HFONT)GetStockFont(DEFAULT_GUI_FONT);
		LOGFONT info;
		GetObject(font, sizeof(info), &info);
		info.lfHeight = -size;
		return CreateFontIndirect(&info);
	}
}

//==================================================================================================================================================================

void TextBox::__init__(Tuple* args, Dict* kw)
{
	HWND parent = __init__parse(args, kw, CENTER);
	if (m_hwnd)
	{
		create(WC_EDIT, 0, 0, parent);
	}
	else
	{
		create(WC_EDIT, ES_AUTOHSCROLL, WS_EX_CLIENTEDGE, parent);
		SetFont(GetDefaultFont(20));
	}
}

object TextBox::get_selection()
{
	int beg = 0, end = 0;
	Edit_GetSelection(m_hwnd, &beg, &end);
	return build("(ii)", beg, end);
}

void TextBox::set_selection(PyObject* o)
{
	int beg, end;

	if (is<Integer>(o))
	{	// ʒû
		beg = end = cast<int>(o);
	}
	else
	{	// JnƏI
		beg = getitem<int>(o, 0);
		end = getitem<int>(o, 1);
	}

	Edit_SetSelection(m_hwnd, beg, end);
}

void TextBox::cut() throw()			{ PostMessage(WM_CUT, 0, 0); }
void TextBox::copy() throw()		{ PostMessage(WM_COPY, 0, 0); }
void TextBox::paste() throw()		{ PostMessage(WM_PASTE, 0, 0); }
void TextBox::delete_() throw()		{ press(VK_DELETE, 0); }
void TextBox::undo() throw()		{ press('Z', MK_CONTROL); }

void TextBox::limit(PyObject* o)
{
	Path* path = cast<Path>(o);
	ref<IShellFolder> folder;
	if SUCCEEDED(path->GetObject(BHID_SFObject, &folder))
		SHLimitInputEdit(m_hwnd, folder);
}

//==================================================================================================================================================================

PYTHON_METHODS(TextBox)
	PYTHON_METHOD(cut)
	PYTHON_METHOD(copy)
	PYTHON_METHOD(paste)
	PYTHON_METHOD_(delete, delete_)
	PYTHON_METHOD(limit)
	PYTHON_METHOD(undo)
PYTHON_END()

PYTHON_MEMBERS(TextBox)
	PYTHON_MEMBER(selection, BOTH)
PYTHON_END()

//==================================================================================================================================================================

void TextView::__init__(Tuple* args, Dict* kw)
{
	HWND parent = __init__parse(args, kw, CENTER);
	create(WC_EDIT, ES_MULTILINE | WS_HSCROLL, 0, parent);
}

//==================================================================================================================================================================

void Label::__init__(Tuple* args, Dict* kw)
{
	HWND parent = __init__parse(args, kw, CENTER);
	create(WC_STATIC, 0, 0, parent);
}

//==================================================================================================================================================================

void brownie::SubclassSingleLineEdit(HWND hwnd)
{
	py::Lock lock;
	try
	{
		if (is_callable(EDITOR))
		{
			dict	kw = Dict::newobj();
			kw->item["__handle"] = Integer::newobj((int)(INT_PTR)hwnd);
			apply(EDITOR, null, kw);
		}
	}
	catch (Exception&)
	{
		Exception::notify();
	}
}
