// thread.cpp

#include "stdafx.h"
#include "math.hpp"
#include "thread.hpp"
#include "window.hpp"

using namespace brownie;

 //==================================================================================================================================================================

namespace
{
	void SetINPUT(INPUT* in, const POINT* pt, UINT mk, bool down = false)
	{
		in->type = INPUT_MOUSE;
		if (pt)
		{
			in->mi.dx = ((DWORD)pt->x * 0xFFFF / ::GetSystemMetrics(SM_CXSCREEN));
			in->mi.dy = ((DWORD)pt->y * 0xFFFF / ::GetSystemMetrics(SM_CYSCREEN));
			in->mi.dwFlags = MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MOVE;
		}
		switch (mk)
		{
		case MK_LBUTTON:
			in->mi.dwFlags |= (down ? MOUSEEVENTF_LEFTDOWN : MOUSEEVENTF_LEFTUP);
			break;
		case MK_RBUTTON:
			in->mi.dwFlags |= (down ? MOUSEEVENTF_RIGHTDOWN : MOUSEEVENTF_RIGHTUP);
			break;
		case MK_MBUTTON:
			in->mi.dwFlags |= (down ? MOUSEEVENTF_MIDDLEDOWN : MOUSEEVENTF_MIDDLEUP);
			break;
		case MK_XBUTTON1:
			in->mi.dwFlags |= (down ? MOUSEEVENTF_XDOWN : MOUSEEVENTF_XUP);
			in->mi.mouseData = XBUTTON1;
			break;
		case MK_XBUTTON2:
			in->mi.dwFlags |= (down ? MOUSEEVENTF_XDOWN : MOUSEEVENTF_XUP);
			in->mi.mouseData = XBUTTON2;
			break;
		}

	}

	class Gesture
	{
	private:
		Window*		m_window;
		HWND		m_hwnd;
		INT			m_vk; // n{^VK\
		INT			m_mk; // n{^MK\
		POINT		m_startpos;
		POINT		m_lastpos;
		DWORD		m_starttime;
		DWORD		m_lasttime;
		std::vector<DIRECTION>	m_dirs;
		bool		m_done;	// ȂƂ1̃WFX`sĂtrueB

	public:
		Gesture(Thread* owner, Window* window, const MSG& msg);
		bool	Next();

		unicode	status() const;

	private:
		bool	Invoke(const MSG& msg) throw();
		void	Restart(const MSG* msg = null) throw();
		void	Terminate(const MSG& msg);
		bool	OnMove();
		bool	GetMessageWithTimeout(MSG* msg);
	};
}

//==================================================================================================================================================================

void DoubleClick(POINT ptClick)
{
	int w = math::max(1, ::GetSystemMetrics(SM_CXSCREEN));
	int h = math::max(1, ::GetSystemMetrics(SM_CYSCREEN));
	POINT pt; ::GetCursorPos(&pt);
	INPUT in[6] = { 0 };
	for (int i = 0; i < 6; ++i) { in[i].type = INPUT_MOUSE; }
	in[0].mi.dx = 65535 * ptClick.x / w;
	in[0].mi.dy = 65535 * ptClick.y / h;
	in[0].mi.dwFlags = MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MOVE | MOUSEEVENTF_LEFTUP;
	in[1].mi.dwFlags = in[3].mi.dwFlags = MOUSEEVENTF_LEFTDOWN;
	in[2].mi.dwFlags = in[4].mi.dwFlags = MOUSEEVENTF_LEFTUP;
	in[5].mi.dx = 65535 * pt.x / w;
	in[5].mi.dy = 65535 * pt.y / h;
	in[5].mi.dwFlags = MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_MOVE;
	::SendInput(6, in, sizeof(INPUT));
}

//==================================================================================================================================================================

void Thread::DragDrop(UINT mk, POINT from, POINT to) throw()
{
	INPUT in[3] = { 0 };

	// ̈ʒuŗBɂ郁bZ[W̓]}B
	m_mouseState = MouseDisabled;
	SetINPUT(&in[0], null, mk, false);
	::SendInput(1, in, sizeof(INPUT));
	PumpMessage();
	// hbÕG~[VB
	m_mouseState = MouseGesture;
	SetINPUT(&in[0], &from, mk, true);	// ŏ̈ʒuŃNbN
	SetINPUT(&in[1], &to, 0);			// ̈ʒuֈړ
	::SendInput(2, in, sizeof(INPUT));
	PumpMessage();
}

void Thread::Update() throw()
{
	m_needsUpdate = true;
}

void Thread::MessageLoop(HWND hwnd) throw()
{
	m_hwnd = hwnd;
	__try
	{
		if (!::IsWindowVisible(hwnd))
			::ShowWindow(hwnd, SW_SHOW);

		for (;;)
		{
			PumpMessage();
			if (m_needsUpdate)
			{	// bZ[WL[ɂȂ̂ŁAWM_UPDATEUISTATE |XgB
				ASSERT(!py::HasLock());
				BoardcastMessage(WM_UPDATEUISTATE);
				m_needsUpdate = false;
			}
			else if (!::IsWindow(hwnd))
			{	// PumpMessageセ̂܂WaitMessageĂ͂ȂȂB
				// hwnd ͊ɔjς݂ł\B
				break;
			}
			else
			{
				ASSERT(!py::HasLock());
				::WaitMessage();
			}
		}
	}
	__finally
	{
		m_hwnd = null;
	}
}

/// }EXzC[J[\̃EBhEɑ.
void Thread::RedirectMouseWheel(MSG& msg)
{
	// WM_MOUSEWHEEL/WM_MOUSEHWHEEL ̃}EXʒu̓XN[WnȂ̂ŁACKvȂ.
	POINT pt = { GET_XY_LPARAM(msg.lParam) };
	HWND hwnd = ::WindowFromPoint(pt);
	if (hwnd && ::GetWindowThreadProcessId(hwnd, null) == ::GetCurrentThreadId() && ::IsWindowEnabled(hwnd))
		m_hwndLastWheel = msg.hwnd = hwnd;
	else if (::IsWindow(m_hwndLastWheel) && ::IsWindowEnabled(m_hwndLastWheel))
		msg.hwnd = m_hwndLastWheel;
	else
		m_hwndLastWheel = null;
}

extern bool IsLocalLoop() throw();

namespace
{
	Window* WindowFromHwnd(HWND hwnd) throw()
	{
		for (; hwnd; hwnd = ::GetAncestor(hwnd, GA_PARENT))
		{
			if (Window* window = Window::From(hwnd))
				return window;
		}
		return null;
	}

	Window* FindTargetWindow(HWND hwnd) throw()
	{
		if (!::IsWindowEnabled(hwnd) || IsLocalLoop())
			return null;
		return WindowFromHwnd(hwnd);
	}

	inline void CancelMessage(MSG& msg) throw()
	{
		// bZ[WnĥŁAbZ[Ŵ𖳌KvB
		// ߂lCallNextHookEx()Ă΂ȂȂǂ̕@ł͂ȂA
		// bZ[WύXāuȂbZ[WvɂĂ܂̂ǂ悤B
		msg.message = WM_NULL;
	}

	inline UINT MK2VK(UINT mk)
	{
		switch (mk)
		{
		case MK_LBUTTON : return VK_BUTTON_L;
		case MK_RBUTTON : return VK_BUTTON_R;
		case MK_MBUTTON : return VK_BUTTON_M;
		case MK_XBUTTON1: return VK_BUTTON_X;
		case MK_XBUTTON2: return VK_BUTTON_Y;
		default         : ASSERT(0); return 0;
		}
	}

#define VECTOR_TO_ARRAY(v)		((v).empty() ? null : &(v)[0]), (v).size()

	static void notify(PyObject* on_notify, Unicode* message)
	{
		try
		{
			apply(on_notify, "Oi", message, LOG);
		}
		catch (Exception&)
		{
			Exception::clear();
		}
	}
}

// WM_*BUTTONDOWN or WM_*BUTTONDBLCLK
// true ԂƃbZ[W̓LZ
bool Thread::OnMouseDown(const MSG& msg) throw()
{
	switch (m_mouseState)
	{
	case MouseDisabled:
		return true;
	case MouseGesture:
		return false;
	default:
		break;
	}

	// WFX`Jn邩H
	if (GESTURE_MAXLEN <= 0)
		return false;	// WFX`ŜɂȂĂB
	Window* w = FindTargetWindow(msg.hwnd);
	if (!w || !w->OnGesture(msg))
		return false;	// WFX`͎󂯓ȂB

	// tB[hobN
	object on_notify;
	{
		py::Lock lock;
		on_notify = getattr(w->self, "on_notify");
	}

	// WFX`[v
	TRACE(L"Gesture.Start");
	m_mouseState = MouseGesture;
	Gesture gesture(this, w, msg);
	while (gesture.Next())
	{
		// WFX`̃tB[hobN
		if (on_notify)
		{
			py::Lock lock;
			notify(on_notify, gesture.status());
		}
	}
	// WFX`IBbZ[WB
	if (on_notify)
	{
		py::Lock lock;
		notify(on_notify, Unicode::newobj(L""));
		on_notify = null; // fXgN^ɔCĂ͂ȂȂB
	}
	m_mouseState = MouseNormal;
	TRACE(L"Gesture.Stop");
	return true;
}

// WM_*BUTTONUP or WM_MOUSEWHEEL
// true ԂƃbZ[W̓LZ
bool Thread::OnMouseUp(const MSG& msg) throw()
{
	return m_mouseState == MouseDisabled;
}

void Thread::OnMessage(MSG& msg) throw()
{
	switch (msg.message)
	{
	case WM_LBUTTONUP:
	case WM_RBUTTONUP:
	case WM_MBUTTONUP:
	case WM_XBUTTONUP:
		if (OnMouseUp(msg))
			CancelMessage(msg);
		else
			Update();
		break;
	case WM_LBUTTONDOWN: case WM_LBUTTONDBLCLK:
	case WM_RBUTTONDOWN: case WM_RBUTTONDBLCLK:
	case WM_MBUTTONDOWN: case WM_MBUTTONDBLCLK:
	case WM_XBUTTONDOWN: case WM_XBUTTONDBLCLK:
		if (OnMouseDown(msg))
			CancelMessage(msg);
		else
			Update();
		break;
	case WM_MOUSEWHEEL:
	case WM_MOUSEHWHEEL:
		if (OnMouseUp(msg)) // P̃zC[WFX`
			CancelMessage(msg);
		else
		{
			RedirectMouseWheel(msg);
			Update();
		}
		break;
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		Update();
		if (Window* w = FindTargetWindow(msg.hwnd))
			if (w->OnPress(msg))
				CancelMessage(msg);
		break;
	case WM_CHAR:
	case WM_ACTIVATEAPP:
	case WM_CLOSE:
	case WM_COMMAND:
	case WM_MENUCOMMAND:
		Update();
		break;
//	case WM_MOUSEMOVE:
//		break;
	}

	// QS_PAINT͊܂߂ȂBWM_PAINT͍Ăт֖߂ĂȂ߁B
	if (m_needsUpdate && HIWORD(GetQueueStatus(QS_ALLINPUT & ~QS_PAINT)) == 0)
	{	// bZ[WL[ɂȂ̂ŁAWM_UPDATEUISTATE |XgB
		m_needsUpdate = false;
		ASSERT(!py::HasLock());
		BoardcastMessage(WM_UPDATEUISTATE);
	}
}

LRESULT CALLBACK Thread::HookMSG(int code, WPARAM wParam, LPARAM lParam) throw()
{
	Thread* me = Get();
	if (code == HC_ACTION && wParam == PM_REMOVE)
		me->OnMessage(*(MSG*)lParam);
	return ::CallNextHookEx(me->m_hookMSG, code, wParam, lParam);
}

//==================================================================================================================================================================
namespace
{
	inline INT MSG2VK2(UINT msg, WPARAM wParam) throw()
	{
		switch (msg)
		{
		case WM_LBUTTONDOWN: case WM_LBUTTONUP:	return VK_BUTTON_L;
		case WM_RBUTTONDOWN: case WM_RBUTTONUP:	return VK_BUTTON_R;
		case WM_MBUTTONDOWN: case WM_MBUTTONUP:	return VK_BUTTON_M;
		case WM_XBUTTONDOWN: case WM_XBUTTONUP:	return GET_XBUTTON(wParam, VK_BUTTON_X, VK_BUTTON_Y);
		case WM_LBUTTONDBLCLK:	return VK_BUTTON_LL;
		case WM_RBUTTONDBLCLK:	return VK_BUTTON_RR;
		case WM_MBUTTONDBLCLK:	return VK_BUTTON_MM;
		case WM_XBUTTONDBLCLK:	return GET_XBUTTON(wParam, VK_BUTTON_XX, VK_BUTTON_YY);
		case WM_MOUSEWHEEL:		return GET_WHEEL(wParam, VK_WHEEL_UP, VK_WHEEL_DOWN);
		case WM_MOUSEHWHEEL:	return GET_WHEEL(wParam, VK_WHEEL_LEFT, VK_WHEEL_RIGHT);
		default:				return 0;
		}
	}
}

Gesture::Gesture(Thread* owner, Window* window, const MSG& msg) :
	m_window(window), m_hwnd(msg.hwnd), m_done(false)
{
	ASSERT(m_window);
	ASSERT(m_hwnd);
	m_vk = MSG2VK2(msg.message, msg.wParam);
	m_mk = MSG2MK(msg);
	Restart(&msg);
}

void Gesture::Restart(const MSG* msg) throw()
{
	if (msg)
	{
		m_startpos.x = GET_X_LPARAM(msg->lParam);
		m_startpos.y = GET_Y_LPARAM(msg->lParam);
		::ClientToScreen(msg->hwnd, &m_startpos);
	}
	else
	{
		m_done = true;
		::GetCursorPos(&m_startpos);
		if (!::IsWindow(m_hwnd))
			m_hwnd = ::WindowFromPoint(m_startpos);
		if (!::IsWindow(m_hwnd))
			m_hwnd = m_window->Handle;
	}
	m_lastpos = m_startpos;
	m_starttime = m_lasttime = ::GetTickCount();
	m_dirs.clear();	// ׂẴP[Xŋ̃nYAÔ߁B
	m_window->SetCapture();
}

bool Gesture::Next()
{
	if (m_vk == 0)
	{
		m_window->ReleaseCapture();
		::GetCursorPos(&m_lastpos);
		Thread::Get()->DragDrop(m_mk, m_startpos, m_lastpos);
		return false;
	}

	MSG msg = { 0 }; // Ăяoɂ̓[邱ƁI
	while (GetMessageWithTimeout(&msg))
	{
		if (::GetCapture() != m_window->Handle)
			return false;

		switch (msg.message)
		{
		case WM_MOUSEMOVE:
			if (OnMove())
				return true;	// continue
			break;
		case WM_LBUTTONUP:
		case WM_RBUTTONUP:
		case WM_MBUTTONUP:
		case WM_XBUTTONUP:
		case WM_LBUTTONDBLCLK:
		case WM_RBUTTONDBLCLK:
		case WM_MBUTTONDBLCLK:
		case WM_XBUTTONDBLCLK:
			if (MSG2MK(msg) == m_mk)
			{	// n{^ꂽB
				Terminate(msg);
				return false;	// finish
			}
			else
			{	// n{^ȊOꂽB
				// fall down
			}
		case WM_MOUSEWHEEL:
		case WM_MOUSEHWHEEL:
			m_window->ReleaseCapture();
			Invoke(msg);
			if (!GetButtonState())
				return false;
			Restart();
			return true;
		default:
			::TranslateMessage(&msg);
			::DispatchMessage(&msg);
			break;
		}
	}

	// ^CAEg

	// {^ĂȂ΁Â܂܏I
	if (!IsKeyPressed(MK2VK(m_mk)))
		return false;

	// WFX`̑߂AhbOĂƂɂB
	m_vk = 0;
	return true;
}

void Gesture::Terminate(const MSG& msg)
{
	m_window->ReleaseCapture();

	if (!m_window->Gesture(m_vk, VECTOR_TO_ARRAY(m_dirs)) && !m_done) 
	{
		// TODO: }EXG~[VɏH
		// WFX`R}ĥ͖ŁAʏ̃NbNG~[gB
		POINT pt;
		::GetCursorPos(&pt);
		::ScreenToClient(m_hwnd, &pt);
		LPARAM lParamClient = MAKELPARAM(pt.x, pt.y);
		// eNjbN
		// WM_RBUTTONUP bZ[WL[ɓꂽAWM_RBUTTONDOWN 𒼐SendMessageB
		// ۂɎ󂯎鏇Ԃ́ARDOWN => RUP => CONTEXTMENUB
		// SendMessage()̑OɃ|XgĂKv̂́A
		// EhbOLȃ^CṽRg[iXgAc[Ȃǁj
		// WM_RBUTTONDOWN AĂȂȂ邽߁B
		// Ȃ݂ɁAWM_RBUTTONDOWN PostMessageƁA}EXWFX`ƗŖ[vɂȂB
		switch (msg.message)
		{
		case WM_LBUTTONUP:
			::PostMessage(m_hwnd, WM_LBUTTONUP  , msg.wParam, lParamClient);
			::SendMessage(m_hwnd, WM_LBUTTONDOWN, msg.wParam | MK_LBUTTON, lParamClient);
			break;
		case WM_RBUTTONUP:
			::PostMessage(m_hwnd, WM_RBUTTONUP  , msg.wParam, lParamClient);
			::SendMessage(m_hwnd, WM_RBUTTONDOWN, msg.wParam | MK_RBUTTON, lParamClient);
			break;
		case WM_MBUTTONUP:
			::PostMessage(m_hwnd, WM_MBUTTONUP  , msg.wParam, lParamClient);
			::SendMessage(m_hwnd, WM_MBUTTONDOWN, msg.wParam | MK_MBUTTON, lParamClient);
			break;
		case WM_XBUTTONUP:
			::PostMessage(m_hwnd, WM_XBUTTONUP  , msg.wParam, lParamClient);
			::SendMessage(m_hwnd, WM_XBUTTONDOWN, msg.wParam | GET_XBUTTON(msg.wParam, MK_XBUTTON1, MK_XBUTTON2), lParamClient);
			break;
		}
	}
}

bool Gesture::Invoke(const MSG& msg) throw()
{
	UINT vkey = MSG2VK(msg);
	if (vkey == 0)
		return false;
	TRACE(L"Gesture.Invoke (0x%x)", vkey);
	return m_window->Gesture(vkey, null, 0);
}

bool Gesture::OnMove()
{
	POINT pt;
	::GetCursorPos(&pt);

	int dx = pt.x-m_lastpos.x, dy = pt.y-m_lastpos.y;
	if (abs(dx) < ::GetSystemMetrics(SM_CXDRAG) && abs(dy) < ::GetSystemMetrics(SM_CYDRAG))
		return false; // s
	m_lastpos = pt;

	static const DIRECTION DIR[2][2] =
	{
		{ NORTH, WEST },
		{ EAST, SOUTH }
	};
	DIRECTION dir = DIR[dy+dx>0 ? 1 : 0][dy-dx>0 ? 1 : 0];

	DWORD now = ::GetTickCount();
	if (m_dirs.empty() || m_dirs.back() != dir || (now - m_lasttime) > (DWORD)GESTURE_NAPTIME)
		m_dirs.push_back(dir);	// VWFX`
	m_lasttime = now;
	return true;
}

bool Gesture::GetMessageWithTimeout(MSG* msg)
{
	size_t ln = m_dirs.size();

	//  GESTURE_MAXLEN ȏ̑́AWFX`ł͂ȂƔ肳B
	// AxԂoĂȂƁAhbOɈڍsȂ悤Ȃ̂ŁA
	// ȂƂ GESTURE_TIMEOUT ̎Ԃo߂܂ł͑҂B
	if (ln > (size_t)GESTURE_MAXLEN
	&& ::GetTickCount() - m_starttime > (DWORD)GESTURE_TIMEOUT)
		return false;

	if (ln == 0)
	{
		::GetMessage(msg, null, 0, 0);
		m_lasttime = ::GetTickCount();
		return true;
	}

	switch (msg->message)
	{
	case WM_MOUSEMOVE:
	case WM_TIMER:
	case WM_PAINT:
	case WM_UPDATEUISTATE:
		break;	// ^C}ZbgȂB
	default:
		m_lasttime = ::GetTickCount();
		break;
	}

	while (!::PeekMessage(msg, null, 0, 0, PM_REMOVE))
	{
		// Î~Ă GESTURE_TIMEOUT ̎Ԃo߂ƁAWFX`ł͂ȂƔ肳B
		if (::GetTickCount() - m_lasttime > (DWORD)GESTURE_TIMEOUT)
			return false;
		::Sleep(1);
	}
	return true;
}

namespace
{
	PCWSTR VK2NAME(UINT vk)
	{
		switch (vk)
		{
		case VK_BUTTON_L:		return L"NbN";
		case VK_BUTTON_R:		return L"ENbN";
		case VK_BUTTON_M:		return L"NbN";
		case VK_BUTTON_X:		return L"4NbN";
		case VK_BUTTON_Y:		return L"5NbN";
		case VK_BUTTON_LL:		return L"_uNbN";
		case VK_BUTTON_RR:		return L"E_uNbN";
		case VK_BUTTON_MM:		return L"_uNbN";
		case VK_BUTTON_XX:		return L"4_uNbN";
		case VK_BUTTON_YY:		return L"5_uNbN";
		case VK_WHEEL_UP:		return L"zC[";
		case VK_WHEEL_DOWN:		return L"zC[";
		case VK_WHEEL_LEFT:		return L"zC[";
		case VK_WHEEL_RIGHT:	return L"zC[E";
		default:				return null;
		}
	}
}

unicode Gesture::status() const
{
	const PCWSTR FORMAT = L"[%s] ";
	PCWSTR	name = VK2NAME(m_vk);
	if (!name)
		return Unicode::newobj(L"");
	ssize_t	ln = len(name);
	size_t	bufln = m_dirs.size() + ln + wcslen(FORMAT);
	PWSTR	buf;
	unicode	r = Unicode::newobj(&buf, bufln);
	swprintf_s(buf, bufln, FORMAT, name);
	buf += ln + 3;
	PCWSTR DIRS = L"";
	for (size_t i = 0; i < m_dirs.size(); ++i)
		*buf++ = DIRS[m_dirs[i]];
	*buf = 0;
	return r;
}

/*
#include "class.hpp"

namespace brownie
{
	class Thread : public Object
	{
	private:
		object	m_name;

	public: // python
		static PCSTR __name__() throw()					{ return "brownie.Thread"; }
		void __init__(group=None, target=None, name=None, args=(), kwargs={});
		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();

	public:
		void	start();
		void	run();
		void	join(Tuple* args);	// ([timeout])
		object	getName();
		void	setName(PyObject* o);
		object	isAlive();

		static object	activeCount();
		static object	currentThread();
		static object	enumerate();	// list of Thread
	};
}
*/
