// toolbar.cpp

#include "stdafx.h"
#include "toolbar.hpp"
#include "node.hpp"
#include "path.hpp"

using namespace brownie;

namespace
{
	const int ID_COMMAND	= 1000;
}

//==================================================================================================================================================================

void ToolBar::__init__(Tuple* args, Dict* kw)
{
	HWND parent = __init__parse(args, kw, NORTH);
	InitCommonControls(ICC_BAR_CLASSES);
	create(TOOLBARCLASSNAME,
		TBSTYLE_FLAT | TBSTYLE_LIST | TBSTYLE_TOOLTIPS | CCS_NODIVIDER | CCS_NOPARENTALIGN | CCS_NORESIZE | DockToCCS(Dock),
		0, parent);
	DisableIME();
	SendMessage(CCM_SETVERSION, 6, 0);
	SendMessage(TB_BUTTONSTRUCTSIZE, sizeof(TBBUTTON), 0);
	SendMessage(TB_SETEXTENDEDSTYLE, 0, TBSTYLE_EX_DRAWDDARROWS);
	SendMessage(TB_SETINDENT, TOOLBAR_INDENT, 0);

	// [̃{^ dwData ݒłȂoOiHj邽߁A_~[̃{^}B
	InsertItem(I_IMAGENONE, BTNS_SEP, TBSTATE_HIDDEN);
	AutoSize();
	m_drop.Register(this);
}

void ToolBar::set_Dock(DIRECTION value)
{
	ModifyStyle(CCS_TOP | CCS_BOTTOM | CCS_LEFT | CCS_RIGHT, DockToCCS(value));
	__super::set_Dock(value);
}

SIZE ToolBar::get_DefaultSize() const throw()
{
	return GetDefaultSize();
}

Icons* ToolBar::GetIcons() const throw()
{
	return cast<Icons>(m_icons);
}

void ToolBar::SetIcons(Icons* value) throw()
{
	if (m_icons == value)
		return;
	if (value)
	{
		SendMessage(TB_SETIMAGELIST			, 0, (LPARAM)value->Handle[NORMAL]);
		SendMessage(TB_SETIMAGELIST			, 1, (LPARAM)SHGetImageList(value));
		SendMessage(TB_SETDISABLEDIMAGELIST	, 0, (LPARAM)value->Handle[DISABLED]);
		SendMessage(TB_SETHOTIMAGELIST		, 0, (LPARAM)value->Handle[HOT]);
	}
	else
	{
		SendMessage(TB_SETIMAGELIST			, 0, NULL);
		SendMessage(TB_SETIMAGELIST			, 1, NULL);
		SendMessage(TB_SETDISABLEDIMAGELIST	, 0, NULL);
		SendMessage(TB_SETHOTIMAGELIST		, 0, NULL);
	}
	m_icons = value;
	AutoSize();
	NotifyResize();
}

PyObject* ToolBar::GetItem(ssize_t index) const throw()
{
	if (index < 1)
		return null;

	TBBUTTON btn;
	if (!SendMessage(TB_GETBUTTON, index, (LPARAM)&btn))
		return null; // error
	return (PyObject*)btn.dwData;
}

void ToolBar::Clear() throw()
{
	if (!m_items)
		return;
	py::Lock lock;
	clear();
}

bool ToolBar_IsPressed(HWND hwnd, UINT nID) throw()
{
	return 0 != ::SendMessage(hwnd, TB_ISBUTTONPRESSED, nID, 0);
}

bool ToolBar::OnMouseDown(UINT msg, WPARAM wParam, LPARAM lParam)
{
	POINT pt = { GET_XY_LPARAM(lParam) };
	if (msg == WM_MOUSEWHEEL)
		ScreenToClient(&pt);

	int index = HitTest(pt);
	if (PyObject* item = GetItem(index))
	{
		if (ToolBar_IsPressed(m_hwnd, index + ID_COMMAND))
		{
			// j[\BLZB
			return false;
		}
		else
		{
			py::Lock lock;
			if ((msg == WM_MOUSEWHEEL || !node_executable(item)) && node_has_children(item))
			{
				RECT rc;
				GetItemRect(index, &rc);
				// 1. {^ꂽԂɂ
				// FIXME: t{^̏ꍇɁAE[PRESSԂɂɂ́H
				SendMessage(TB_PRESSBUTTON, index + ID_COMMAND, MAKELONG(true, 0));
				// 2. ʂɔf
				UpdateWindow(m_hwnd);
				// 3. j[̕\Aj[VitF[hCȂǁj
				node_popup_children(item, ClientToScreen(RECT_SW(rc)), this->Dock, cast<Icons>(m_icons));
				// 4. łSendł͂ȂPostĝ|CgB
				//    xNbNƂɃj[悤ɂ邽߁B
				PostMessage(TB_PRESSBUTTON, index + ID_COMMAND, MAKELONG(false, 0));
				return true;
			}
		}
	}
	return false;
}

LRESULT ToolBar::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_LBUTTONDBLCLK:
	case WM_LBUTTONDOWN:
	case WM_MOUSEWHEEL:
		if (OnMouseDown(msg, wParam, lParam))
			return 0;
		break;
	case WM_MBUTTONUP:
		if (PyObject* o = GetItem(HitTest(GET_XY_LPARAM(lParam))))
			NodeExecute(o, VK_MBUTTON, MK_MBUTTON);
		break;
	case WM_CONTEXTMENU:
		if (OnContextMenu(lParam))
			return 0;
		break;
	case WM_CLOSE:
	case WM_DESTROY:
		m_drop.Revoke();
		Clear();
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

LRESULT	ToolBar::OnMessage(Window* from, NMHDR& nm)
{
	switch (nm.code)
	{
	case WM_COMMAND:
		NodeExecute(GetItem(nm.idFrom - ID_COMMAND));
		return 0;
	case TBN_DROPDOWN:
		DropDown(((NMTOOLBAR&)nm).iItem - ID_COMMAND);
		return TBDDRET_DEFAULT;
	}
	return __super::OnMessage(from, nm);
}

bool ToolBar::OnGesture(const MSG& msg)
{
	UINT vk = MSG2VK(msg);
	if (vk == VK_LBUTTON || vk == VK_MBUTTON)
	{
		POINT pt = { GET_XY_LPARAM(msg.lParam) };
		if (msg.hwnd != m_hwnd)
			MapWindowPoints(msg.hwnd, m_hwnd, &pt, 1);
		if (HitTest(pt) >= 0)
			return false;
	}
	return __super::OnGesture(msg);
}

void ToolBar::DropDown(int index)
{
	if (PyObject* node = GetItem(index))
		NodePopupChildren(this, node, index);
}

bool ToolBar::OnContextMenu(LPARAM lParam)
{
	int		index;
	POINT	ptScreen;

	if (lParam == 0xFFFFFFFF)
	{
		RECT rc;
		index = GetHotItem();
		if (index <= 0 || !GetItemRect(index, &rc))
			return false;
		ptScreen = ClientToScreen(RECT_SW(rc));
	}
	else
	{
		ptScreen.x = GET_X_LPARAM(lParam);
		ptScreen.y = GET_Y_LPARAM(lParam);
		POINT ptClient = ScreenToClient(ptScreen);
		index = HitTest(ptClient);
		if (index < 0)
			return false;
	}

	return SUCCEEDED(NodePopupMenu(GetItem(index), ptScreen, cast<Icons>(m_icons)));
}

namespace
{
	class DropTargetWrapper : public Unknown< implements<IDropTarget> >
	{
	private:
		ref<IDropTarget>	m_target;
		ToolBar*			m_owner;
		int					m_index;

	public:
		DropTargetWrapper(IDropTarget* target, ToolBar* owner, int index) : 
			m_target(target), m_owner(owner), m_index(index)
		{ }
	public: // IDropTarget
		STDMETHODIMP DragEnter(IDataObject* data, DWORD keys, POINTL pt, DWORD* effect)
		{
			HRESULT hr;
			if SUCCEEDED(hr = m_target->DragEnter(data, keys, pt, effect))
			{
				m_owner->SetHotItem(m_index);
//				m_owner->DropDown(m_index);
			}
			return hr;
		}
		STDMETHODIMP DragOver(DWORD keys, POINTL pt, DWORD* effect)
		{
			return m_target->DragOver(keys, pt, effect);
		}
		STDMETHODIMP DragLeave()
		{
			HRESULT hr = m_target->DragLeave();
			m_owner->SetHotItem(-1);
			return hr;
		}
		STDMETHODIMP Drop(IDataObject* data, DWORD keys, POINTL pt, DWORD* effect)
		{
			HRESULT hr = m_target->Drop(data, keys, pt, effect);
			m_owner->SetHotItem(-1);
			return hr;
		}
	};
}

void ToolBar::OnDragDrop(POINT pt, RECT* rc, IDropTarget** pp) throw()
{
	int index = HitTest(pt);
	if (index > 0)
	{
		if (rc)
			GetItemRect(index, rc);
		PyObject* o = GetItem(index);
		if (is<Path>(o))
		{
			Path* path = cast<Path>(o);
			ref<IDropTarget> target;
			if SUCCEEDED(path->GetObject(BHID_SFUIObject, &target))
				*pp = new DropTargetWrapper(target, this, index);
		}
	}
}

void ToolBar::OnUpdate()
{
	LockRedraw pause(m_hwnd);
	py::Lock lock;
	try
	{
		int index = 0;
		foreach (object i, m_items)
		{
			++index;
			if (is<Null>(i))
				continue;

			UINT curState = GetItemState(index + ID_COMMAND);
			UINT newState = (curState & ~(TBSTATE_ENABLED | TBSTATE_CHECKED));
			if (node_enabled(i)) { newState |= TBSTATE_ENABLED; }
			if (node_checked(i)) { newState |= TBSTATE_CHECKED; }
			if (newState != curState)
				SetItemState(index + ID_COMMAND, newState);
		}
	}
	catch (Exception&)
	{
		// `̂тɌĂ΂̂ŁAw^ɕ񍐃_CAOô͕|B
		Exception::clear();
	}
	__super::OnUpdate();
}

//==================================================================================================================================================================

void ToolBar::set_icons(PyObject* value)
{
	SetIcons(cast<Icons>(value));
}

void ToolBar::set_items(PyObject* value)
{
	if (m_items == value)
		return;

	clear();

	// TODO: vLVIuWFNg̍쐬
	m_items = Tuple::newobj(value);
	ssize_t ln = len(m_items);
	if (ln <= 0)
		return;

	std::vector<TBBUTTON>	buttons;
	std::vector<unicode>	names;
	buttons.reserve(ln);
	names.reserve(ln);
	bool useShellIcon = false;

	foreach (object o, m_items)
	{
		TBBUTTON btn = { 0 };
		if (is<Null>(o))
		{	// separator
			btn.fsStyle = BTNS_SEP;
		}
		else
		{
			unicode	name     = node_get_name(o);
			bool	children = node_has_children(o);
			bool	executable  = node_executable(o);
			bool	enabled = node_enabled(o);
			bool	checked = node_checked(o);
			btn.iBitmap   = node_get_icon(o);
			btn.fsStyle   = BTNS_AUTOSIZE;
			btn.fsState   = 0;
			btn.dwData    = (DWORD_PTR)o.detach();
			btn.idCommand = (int)(ID_COMMAND + buttons.size() + 1);

			if (enabled && (children || executable)) { btn.fsState |= TBSTATE_ENABLED; }
			if (checked) { btn.fsState |= TBSTATE_CHECKED; }
			if (name)
			{
				btn.fsStyle |= BTNS_SHOWTEXT;
				btn.iString = (INT_PTR)str(name);
				// Ԃ΂̂ړIB
				// buttons[n].iString != names[n] ł\ȂB
				names.push_back(name);
			}

			// BTNS_WHOLEDROPDOWN ͎gȂBɁ`悳Ă܂߁B
			// ʏ{^Ƃēo^AOŃhbv_EǂfB
			if (children && executable)
				btn.fsStyle |= BTNS_DROPDOWN; // {^{q
			else
				btn.fsStyle |= BTNS_BUTTON; // qȂ or q̂
		
			if (IsShellIcon(btn.iBitmap))
				useShellIcon = true;
		}
		buttons.push_back(btn);
	}

	if (!buttons.empty())
	{
		if (useShellIcon && SendMessage(TB_GETIMAGELIST, 0, 0) == 0)
			SendMessage(TB_SETIMAGELIST, 1, (LPARAM)SHGetImageList(cast<Icons>(m_icons)));
		InsertItems(buttons.size(), &buttons[0]);
		NotifyResize();
	}
}

void ToolBar::clear() throw()
{
	LockRedraw pause(m_hwnd);
	for (int i = GetItemCount()-1; i > 0; --i)
	{
		try
		{
			decref(GetItem(i));
		}
		catch (Exception&)
		{
			Exception::clear();
		}
		SendMessage(TB_DELETEBUTTON, i, 0);
	}
	ASSERT(GetItemCount() == 1);
}

//==================================================================================================================================================================

PYTHON_MEMBERS(ToolBar)
	PYTHON_MEMBER(icons	, SET)
	PYTHON_MEMBER(items	, SET)
PYTHON_END()

//==================================================================================================================================================================
