// treeview.cpp

#include "stdafx.h"
#include "node.hpp"
#include "treeview.hpp"

using namespace brownie;

namespace
{
	const DWORD WS_TREEVIEW		= WS_VSCROLL | WS_HSCROLL | TVS_HASBUTTONS | TVS_HASLINES | TVS_SHOWSELALWAYS | TVS_EDITLABELS;
	const DWORD WS_TREEVIEW_EX	= WS_EX_CLIENTEDGE;
}

//==================================================================================================================================================================

void TreeView::__init__(Tuple* args, Dict* kw)
{
	HWND parent = __init__parse(args, kw, CENTER);
	InitCommonControls(ICC_TREEVIEW_CLASSES);
	create(WC_TREEVIEW, WS_TREEVIEW, WS_TREEVIEW_EX, parent);
	DisableIME();
}

Icons* TreeView::GetIcons() const throw()
{
	return cast<Icons>(m_icons);
}

void TreeView::SetIcons(Icons* value) throw()
{
	if (m_icons == value)
		return;
	// SHGetImageList(value)
	HIMAGELIST hImageList = (value ? value->Handle[NORMAL] : null);
	SetImageList(hImageList, TVSIL_NORMAL);
	m_icons = value;
	NotifyResize();
}

PyObject* TreeView::GetItem(HTREEITEM item) const
{
	return (PyObject*)GetItemData(item);
}

void TreeView::Clear() throw()
{
	if (!m_items)
		return;

	LockRedraw pause(m_hwnd);
	DeleteAllItems();

	py::Lock lock;
	m_items = null;
}

LRESULT TreeView::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_KILLFOCUS:
		if (!IsChild(m_hwnd, (HWND)wParam))
			EndEditLabelNow(false);
		break;
	case WM_LBUTTONDBLCLK:
		if (OnClickExecute(wParam, lParam, VK_LBUTTON))
			return 0;
		break;
	case WM_MBUTTONUP:
		if (OnClickExecute(wParam, lParam, VK_MBUTTON))
			return 0;
		break;
	case WM_RBUTTONDOWN:
	case WM_MBUTTONDOWN:
		SelectItem(GET_XY_LPARAM(lParam));
		return 0;
	case WM_RBUTTONUP:
		// c[r[ł͂ȂAftHgiWM_CONTEXTMENU𔭐j
		return ::DefWindowProc(m_hwnd, msg, wParam, lParam);
	case WM_CONTEXTMENU:
		if (OnContextMenu(lParam))
			return 0;
		break;
	case WM_DESTROY:
		Clear();
		break;
	}
	return __super::OnMessage(msg, wParam, lParam);
}

LRESULT	TreeView::OnMessage(Window* from, NMHDR& nm)
{
	switch (nm.code)
	{
	case TVN_DELETEITEM:
		OnDeleteItem((NMTREEVIEW&)nm);
		break;
	case TVN_ITEMEXPANDING:
		OnExpanding((NMTREEVIEW&)nm);
		return 0;
	case TVN_ITEMEXPANDED:
		OnExpanded((NMTREEVIEW&)nm);
		break;
	case TVN_GETDISPINFO:
		OnGetDispInfo((NMTVDISPINFO&)nm);
		return 0;
	case TVN_BEGINDRAG:
	case TVN_BEGINRDRAG:
	//	OnDrag(...);
		break;
	case TVN_KEYDOWN:
		switch (((NMTVKEYDOWN&)nm).wVKey)
		{
		case VK_RETURN:
			if (HTREEITEM item = GetSelection())
				if (PyObject* o = GetItem(item))
					NodeExecute(o, VK_RETURN);
			return 1; // non-zero CN^T[`Ɋ܂܂Ȃ
		case VK_SPACE:
			Expand(GetSelection(), TVE_TOGGLE);
			return 1;
		default:
			break;
		}
		break;
	case TVN_SELCHANGED:
//		final.OnSelChanged(nm.itemNew.item, (ParamType)nm.itemNew.lParam);
		//NMTREEVIEW* nm = (NMTREEVIEW&)nm;
		break;
	case TVN_BEGINLABELEDIT:
//		OnBeginLabelEdit();
		break;
	case TVN_ENDLABELEDIT:
//		OnEndLabelEdit();
		break;
	case NM_RETURN:
		// ftHǵur[v炷v̂ŁALZB
		return 1;
	}
	return __super::OnMessage(from, nm);
}

void TreeView::OnGetDispInfo(NMTVDISPINFO& disp)
{
	PyObject* node = (PyObject*)disp.item.lParam;
	py::Lock lock;
	if (disp.item.mask & TVIF_CHILDREN)
		disp.item.cChildren = node_has_children(node);
	disp.item.mask |= TVIF_DI_SETITEM;
}

void TreeView::OnDeleteItem(NMTREEVIEW& nm)
{
	py::Lock lock;
	decref((PyObject*)nm.itemOld.lParam);
}

void TreeView::OnExpanding(NMTREEVIEW& nm)
{
	HTREEITEM item = nm.itemNew.hItem;
	PyObject* node = (PyObject*)nm.itemNew.lParam;

	if (!GetChildItem(item))
	{
		py::Lock lock;
		iterator iter = node_enum_children(node);
		while (object o = iter->next())
			insert(item, o);
	}

	if (!GetChildItem(item))
	{	// JĂ݂A͎qĂȂ
		TVITEM tvi = { TVIF_HANDLE | TVIF_CHILDREN };
		tvi.hItem = item;
		tvi.cChildren = 0;
		SetItem(tvi);
	}
}

void TreeView::OnExpanded(NMTREEVIEW& nm)
{
	nm.action, nm.itemNew;
}

bool TreeView::OnGesture(const MSG& msg)
{
//	m[h̖OύXnh
//	if (IsChild(msg.hwnd))
//		return true;

	TVHITTESTINFO hit = { GET_XY_LPARAM(msg.lParam) };
	if (msg.hwnd != m_hwnd)
		MapWindowPoints(msg.hwnd, m_hwnd, &hit.pt, 1);
	if (HitTest(&hit))
		return false;

	return __super::OnGesture(msg);
}

bool TreeView::OnClickExecute(WPARAM wParam, LPARAM lParam, UINT vk)
{
	TVHITTESTINFO hit = { GET_XY_LPARAM(lParam) };
	if (HTREEITEM item = HitTest(&hit))
	{
		switch (hit.flags)
		{
		case TVHT_ONITEM:
		case TVHT_ONITEMICON:
		case TVHT_ONITEMLABEL:
		case TVHT_ONITEMSTATEICON:
			if (GetSelection() == item)
				if (PyObject* o = GetItem(item))
					NodeExecute(o, vk);
			return true;
		}
	}
	return false;
}

bool TreeView::OnContextMenu(LPARAM lParam)
{
	POINT ptScreen = { GET_XY_LPARAM(lParam) };
	HTREEITEM item;
	if (ptScreen.x == -1 || ptScreen.y == -1)
	{
		item = GetSelection();
		if (!item)
			return false;
		RECT rc;
		GetItemRect(item, &rc);
		POINT pt = { rc.left, rc.bottom };
		ClientToScreen(&pt);
		ptScreen = pt;
	}
	else
	{
		TVHITTESTINFO hit = { ptScreen.x, ptScreen.y };
		ScreenToClient(&hit.pt);
		item = HitTest(&hit);
		if (item != GetSelection())
			return false;
	}

	PyObject* node = GetItem(item);
	if (!node)
		return false;

	HRESULT hr = NodePopupMenu(node, ptScreen, cast<Icons>(m_icons));
	if (hr == S_RENAME)
		EditLabel(item);
	return SUCCEEDED(hr);
}

//==================================================================================================================================================================

void TreeView::set_icons(PyObject* value)
{
	SetIcons(cast<Icons>(value));
}

void TreeView::set_items(PyObject* value)
{
	if (m_items == value)
		return;

	clear();
	m_items = value;

	if (HTREEITEM item = insert(TVI_ROOT, value))
	{
		py::Unlock unlock;
		Expand(item, TVE_EXPAND);
		SelectItem(item);
	}
}

void TreeView::clear() throw()
{
	py::Unlock unlock;
	Clear();
}

HTREEITEM TreeView::insert(HTREEITEM parent, PyObject* o)
{
	unicode u = node_get_name(o);
	HTREEITEM h = InsertItem(parent, o, str(u), node_get_icon(o));
	if (h)
		incref(o);
	return h;
}

//==================================================================================================================================================================

PYTHON_MEMBERS(TreeView)
	PYTHON_MEMBER(icons	, SET)
	PYTHON_MEMBER(items	, SET)
PYTHON_END()

//==================================================================================================================================================================
