// window.cpp

#include "stdafx.h"
#include "window.hpp"
#include "system.hpp"

#include "commctrl.hpp"
#include "rebar.hpp"

using namespace brownie;

extern bool Menu_OnMessage(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam, LRESULT& lResult) throw();

namespace
{
	void ContextMenu_Append(HMENU menu, HWND hwnd, int id)
	{
		WCHAR name[MAX_PATH+4] = L"&N: ";
		PWSTR s = name + 4;
		if (::GetWindowText(hwnd, s, MAX_PATH) <= 0)
			::GetClassName(hwnd, s, MAX_PATH);
		int n = GetMenuItemCount(menu);
		if (n < 9)
		{
			name[1] = (WCHAR)(L'1' + n);
			s = name;
		}
		AppendMenu(menu, MF_STRING | (IsVisible(hwnd) ? MF_CHECKED : 0), id, s);
	}

	int ContextMenu_AppendReBar(HMENU menu, HWND hwnd, std::vector<HWND>& children)
	{
		int count = ReBar_GetBandCount(hwnd);
		for (int i = 0; i < count; ++i)
		{
			ReBar_Band band(RBBIM_CHILD);
			ReBar_GetBandInfo(hwnd, i, &band);
			children.push_back(band.hwndChild);
			ContextMenu_Append(menu, band.hwndChild, (int)children.size());
		}
		return count;
	}

	void OnContextMenu(Window* window, int x, int y)
	{
		POINT ptScreen = { x, y };

		std::vector<HWND> children;
		std::vector<HWND> rebar;
		LayoutEnumAll(window, children);

		HMENU popup = CreatePopupMenu();
		size_t count = children.size();
		for (size_t index = 0; index < count; ++index)
		{
			HWND h = children[index];
			Window* w = Window::From(h);
			if (!w || w->Dock == CENTER)
				continue;
			if (is<ReBar>(w->self))
			{
				ContextMenu_AppendReBar(popup, h, children);
				rebar.push_back(h);
			}
			else
				ContextMenu_Append(popup, h, (int)(index+1));
		}
		// bN
		const UINT ID_LOCK = 1000;
		bool locked = false;
		if (!rebar.empty())
		{
			static PCWSTR STR_LOCKBARS = _W("LockBars");
			locked = ReBar_GetLocked(rebar.front());
			AppendMenu(popup, MF_SEPARATOR, 0, 0);
			AppendMenu(popup, MF_STRING | (locked ? MF_CHECKED : 0), ID_LOCK, STR_LOCKBARS);
		}
		// j[
		UINT cmd = ::TrackPopupMenu(popup, TPM_RIGHTBUTTON | TPM_RETURNCMD, ptScreen.x, ptScreen.y, 0, window->Handle, 0);
		if (cmd == ID_LOCK)
		{
			for (size_t i = 0; i < rebar.size(); ++i)
				ReBar_SetLocked(rebar[i], !locked);
		}
		else if (0 < cmd && cmd <= children.size())
		{
			HWND h = children[cmd - 1];
			::ShowWindow(h, IsVisible(h) ? SW_HIDE : SW_SHOW);
		}
		::DestroyMenu(popup);
	}
}


namespace
{
	inline DIRECTION VerifyDirection(int value) throw()
	{
		switch (value)
		{
		case NORTH:
		case SOUTH:
		case WEST:
		case EAST:
			return (DIRECTION)value;
		default:
			return CENTER;
		}
	}
}

//==================================================================================================================================================================

Window::Window() throw() : m_rearrange(true), m_pressed(false), m_activated(false), m_dock(CENTER)
{
	m_size.cx = m_size.cy = CW_USEDEFAULT;
}

Window::~Window() throw()
{
	ASSERT(!m_hwnd);
}

void Window::__init__(Tuple* args, Dict* kw, DIRECTION dock, DWORD style, DWORD styleEx)
{
	HWND hwndParent = __init__parse(args, kw, dock);
	create(style, styleEx, CS_DBLCLKS, hwndParent);
}

void Window::__init__(Tuple* args, Dict* kw)
{
	__init__(args, kw, CENTER);
}

void Window::create(DWORD style, DWORD styleEx, DWORD classStyle, HWND parent)
{
	WNDCLASSEX wc = { sizeof(WNDCLASSEX) };
	wc.style = classStyle;
	wc.hCursor = ::LoadCursor(null, IDC_ARROW);
	wc.hbrBackground = ((styleEx & WS_EX_TRANSPARENT) ? null : (HBRUSH)(COLOR_3DFACE + 1));
	wc.lpfnWndProc = ::DefWindowProc;
	wc.hInstance = ::GetModuleHandle(null);
	create(wc, style, styleEx, parent);
}

void Window::create(PCWSTR cls, DWORD style, DWORD styleEx, HWND parent)
{
	WNDCLASSEX wc = { sizeof(WNDCLASSEX) };
	if (!::GetClassInfoEx(null, cls, &wc))
		throw WindowsError(cls);
	create(wc, style, styleEx, parent);
}

void Window::set_Dock(DIRECTION value) throw()
{
	m_dock = value;
	// dock̕ύX͍̂̂ăCAEgvȂB
	// Ae̍ăCAEgɂEBhETCYύX邱Ƃ߁A
	// ǂ͍ăCAEg\B
	Update(false);
	NotifyResize();
}

Window* Window::get_Parent() const throw()
{
	return From(GetAncestor(GA_PARENT));
}

Window* Window::get_Root() const throw()
{
	return From(GetAncestor(GA_ROOT));
}

namespace
{
	struct MESSAGE
	{
		UINT	msg;
		WPARAM	wParam;
		LPARAM	lParam;
		UINT	count;
	};

	BOOL CALLBACK SendMessageToChildrenCallback(HWND hwnd, LPARAM lParam)
	{
		MESSAGE* m = (MESSAGE*)lParam;
		if (Window::From(hwnd))
			::SendMessage(hwnd, m->msg, m->wParam, m->lParam);
		m->count++;
		return TRUE;
	}

	BOOL CALLBACK PostMessageToChildrenCallback(HWND hwnd, LPARAM lParam)
	{
		MESSAGE* m = (MESSAGE*)lParam;
		if (Window::From(hwnd))
			::PostMessage(hwnd, m->msg, m->wParam, m->lParam);
		m->count++;
		return TRUE;
	}
}

UINT Window::SendMessageToChildren(UINT msg, WPARAM wParam, LPARAM lParam) const throw()
{
	MESSAGE m = { msg, wParam, lParam };
	::EnumChildWindows(m_hwnd, SendMessageToChildrenCallback, (LPARAM)&m);
	return m.count;
}

UINT Window::PostMessageToChildren(UINT msg, WPARAM wParam, LPARAM lParam) const throw()
{
	MESSAGE m = { msg, wParam, lParam };
	::EnumChildWindows(m_hwnd, PostMessageToChildrenCallback, (LPARAM)&m);
	return m.count;
}

HWND Window::get_Focus() const throw()
{
	return m_focus;
}

void Window::set_Focus(HWND hwnd) throw()
{
	ASSERT(!hwnd || ::GetParent(hwnd) == m_hwnd);
	m_focus = hwnd;
	if (!m_activated && HasFocus(m_hwnd))
		 m_activated = true;
}

bool Window::Press(UINT vkey, UINT mods) throw()
{
	MSG msg = { m_hwnd, (mods & MK_ALTKEY) ? WM_SYSKEYDOWN : WM_KEYDOWN, vkey, 0 };
	SetKeyState(vkey, mods);
	__try
	{
		return (!m_pressed && OnPress(msg)) || (0 != SendMessage(msg.message, msg.wParam, msg.lParam));
	}
	__finally
	{
		ResetKeyState(vkey);
	}
}

bool Window::OnPress(const MSG& msg)
{
	if (m_pressed)
		return false; // avoid recursive call.
	if (m_keymap)
	{
		bool r = false;
		m_pressed = true;

		py::Lock lock;
		try
		{
			r = cast<bool>(apply(m_keymap, "Oii", self, (int)msg.wParam, GetModifiers()));
			m_pressed = false;
		}
		catch (Exception&)
		{
			m_pressed = false;
			Exception::notify();
		}
		if (r)
			return true;
	}
	// nhȂꍇ͐e֓]B
	if (Window* parent = Parent)
		return parent->OnPress(msg);
	else
		return false;
}

bool Window::OnGesture(const MSG& msg)
{
	if (m_gesture)
		return true;
	else if (Window* parent = Parent)
		return parent->OnGesture(msg);
	else
		return false;
}

bool Window::Gesture(UINT vkey, const DIRECTION dirs[], size_t ndirs) throw()
{
	if (m_gesture)
	{
		py::Lock lock;
		try
		{
			tuple tup = Tuple::newobj(ndirs);
			for (size_t i = 0; i < ndirs; ++i)
				tup->attach(i, Integer::newobj(dirs[i]).detach());
			if (cast<bool>(apply(m_gesture, "OiiO", self, vkey, GetModifiers(), tup)))
				return true;
		}
		catch (Exception&)
		{
			Exception::notify();
		}
	}
	// nhȂꍇ͐e֓]B
	if (Window* parent = Parent)
		return parent->Gesture(vkey, dirs, ndirs);
	else
		return false;
}

void Window::OnChildFocus(HWND child)
{
	while (child)
	{
		HWND parent = ::GetAncestor(child, GA_PARENT);
		if (parent == m_hwnd)
			break;
		ASSERT(0);
		child = parent;
	}

	this->Focus = child;
	if (HasFocus(m_hwnd))
	{
		if (!HasFocus(this->Focus))
			::SetFocus(this->Focus);
		if (Window* parent = Parent)
			parent->OnChildFocus(m_hwnd);
	}
}

void Window::OnChildResize(Window* child)
{
	// q̃TCYύXꂽ̂ŁAĔzuB
	Update(true);
}

void Window::NotifyResize()
{
	if (Window* parent = Parent)
		parent->OnChildResize(this);
}

LRESULT Window::OnMessage(UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
	case WM_UPDATEUISTATE:
		ASSERT(!py::HasLock());
		OnUpdate();
		return 0;
	case WM_SIZE:
		if (wParam == 0 || wParam == SIZE_MAXIMIZED || wParam == SIZE_RESTORED)
		{
			SIZE sz = { GET_XY_LPARAM(lParam) };
			if (m_size == sz)
				break;

			bool notify = false;
			if (this->Visible)
			{
				switch (m_dock)
				{
				case NORTH:
				case SOUTH:
					notify = (sz.cy != m_size.cy);
					break;
				case WEST:
				case EAST:
					notify = (sz.cx != m_size.cx);
					break;
				default:
					notify = true;
					break;
				}
			}
			m_size = sz;
			if (notify)
				NotifyResize();
			if (m_size == sz)
				Update(true);
		}
		break;
	case WM_NOTIFY:
		if (NMHDR* n = (NMHDR*)lParam)
		{
			ASSERT(wParam == n->idFrom);
			if (Window* w = Window::From(n->hwndFrom))
				return w->OnMessage(this, *n);
		}
		break;
	case WM_SETFOCUS:
		if (!::IsWindow(this->Focus))
			this->Focus = LayoutGetCenter(this);
		if (this->Focus)
			::SetFocus(this->Focus);
		if (Window* parent = Parent)
			parent->OnChildFocus(m_hwnd);
		break;
	case WM_PARENTNOTIFY:
		switch (LOWORD(wParam))
		{
		case WM_CREATE:
			Update(true);
			break;
		case WM_DESTROY:
			if (this->Focus == (HWND)lParam)
				this->Focus = null;
			Update(true);
			break;
		case WM_LBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:
		case WM_XBUTTONDOWN:
			if (Window* w = FromPoint(GET_XY_LPARAM(lParam)))
				if (w->Dock == CENTER)
					OnChildFocus(w->Handle);
			break;
		}
		break;
	case WM_MOUSEWHEEL:
	case WM_MOUSEHWHEEL:
		if (this->Focus)
		{
			Activate(GET_WHEEL_DELTA_WPARAM(wParam) < 0 ? EAST : WEST);
			return 0;
		}
		break;
	case WM_COMMAND:
		if (Window* w = Window::From((HWND)lParam))
		{	// from BUTTON or TOOLBAR
			NMHDR n = { (HWND)lParam, LOWORD(wParam), WM_COMMAND };
			return w->OnMessage(this, n);
		}
		else
		{	// from MENU or ACCEL
		}
		break;
	case WM_ERASEBKGND:
		if (HasStyleEx(m_hwnd, WS_EX_TRANSPARENT))
			return -1;
		break;
	case WM_SHOWWINDOW:
		if (lParam == 0) // by ShowWindow
		{
			BOOL shown = (BOOL)wParam;
			if (Window* parent = this->Parent)
			{
				NotifyResize();
				if (shown)
					Update(false);
			}
			else if (m_rearrange && shown)
			{	// ŏʃEBhE͕\OɃCAEgXVB
				UpdateLayout();
			}
		}
		break;
	case WM_ACTIVATEAPP:
		// CEBhEANeBuꂽƂGCs.
		if (!wParam && m_hwnd == System::GetWindow())
			py::CollectGarbage();
		break;
	case WM_SETTINGCHANGE:
		// WM_SETTINGCHANGE  bZ[WpEBhEɂ͑Ȃ̂ŁAʒmĂB
		// Xbh̗v͓Z߂̂ŁAʒmĂ\ȂB
		if (m_hwnd == System::GetWindow())
			System::OnSettingChange(wParam, lParam);
		PostMessageToChildren(msg, wParam, lParam);
		break;
	case WM_CREATE:
		if (LRESULT lResult = ::DefSubclassProc(m_hwnd, msg, wParam, lParam))
			return lResult; // error
		else
		{
			// ɑ͂ WM_SIZE ܂ő҂Ăǂ̂ł͂邪A
			// Ô߁A쐬ɂXVĂB
			// get_ClientSize() ́A܂gpłȂƂɒӁB
			m_size = GetClientSize(m_hwnd);
			return 0;
		}
	case WM_DESTROY:
		// tH[JX܂ܔjꂻȏꍇ́AetH[JXB
		// ǰŐetH[JX悤AxbZ[W𑗂B
		if (HasFocus(m_hwnd))
			if (Window* parent = Parent)
				parent->PostMessage(WM_PARENTFOCUS);
		break;
	case WM_PARENTFOCUS:
		SetFocus();
		return 0;
	case WM_ENDSESSION:
		if (!wParam)
			break;
		// fall through
	case WM_CLOSE:
		Event("on_close");
		if (!GetAncestor(GA_PARENT))
			ShowWindow(m_hwnd, SW_HIDE);
		SendMessageToChildren(msg, wParam, lParam);
		break;
	case WM_COPYDATA:
		if (COPYDATASTRUCT* data = (COPYDATASTRUCT*)lParam)
		{
			if (data->dwData == COPYDATA_MAGIC)
			{
				TRACE(L"%s (%d)", (PCWSTR)data->lpData, data->cbData);
				return TRUE;
			}
		}
		break;
	case WM_CONTEXTMENU:
		if (m_hwnd == System::GetWindow() && lParam != 0xFFFFFFFF)
		{
			POINT	ptScreen = { GET_XY_LPARAM(lParam) };
			POINT	ptClient = ScreenToClient(ptScreen);
			// ^Cgo[̉ENbNȂ悤ɁANCAg̈̏ꍇ̂݁B
			if (ptClient.x >= 0 && ptClient.y >= 0)
			{
				OnContextMenu(this, ptScreen.x, ptScreen.y);
				return 0;
			}
		}
		break;
	}

	LRESULT lResult = 0;
	if (Menu_OnMessage(m_hwnd, msg, wParam, lParam, lResult))
		return lResult;

	return ::DefSubclassProc(m_hwnd, msg, wParam, lParam);
}

LRESULT Window::OnMessage(Window* from, NMHDR& nm)
{
	if (nm.code != WM_COMMAND)
		return ::DefSubclassProc(from->m_hwnd, WM_NOTIFY, nm.idFrom, (LPARAM)&nm);
	else // WM_COMMAND͒ڌĂ΂B
		return 0;
}

void Window::OnUpdate()
{
	if (HWND hwnd = GetFocus())
	{
		while (HWND parent = ::GetAncestor(hwnd, GA_PARENT))
		{
			if (parent == m_hwnd)
			{
				this->Focus = hwnd;
				break;
			}
			hwnd = parent;
		}
	}
	if (m_rearrange && ::IsWindowVisible(m_hwnd))
		UpdateLayout();
	if (m_activated)
	{
		Event("on_activate");
		m_activated = false;
	}
}

void Window::OnLayout()
{
	RECT bounds = this->ClientArea;
	if (::IsRectEmpty(&bounds))
		return; // minimized?

	if (HWND hwnd = LayoutArrange(this, bounds))
		if (this->Focus != hwnd)
			OnChildFocus(hwnd);
}

void Window::QueryLayout(ILayout& layout) const
{
	// Ώۂ̎qEBhE
	for (HWND hwnd = ::GetWindow(m_hwnd, GW_CHILD); hwnd; hwnd = ::GetWindow(hwnd, GW_HWNDNEXT))
	{
		if (Window* child = Window::From(hwnd))
			layout.Insert(hwnd, child->Dock);
	}
}

Window* Window::FromPoint(int x, int y) const throw()
{
	POINT pt = { x, y };
	HWND hwnd = ChildWindowFromPointEx(m_hwnd, pt, CWP_SKIPINVISIBLE | CWP_SKIPDISABLED);
	return From(hwnd);
}

void Window::Activate(DIRECTION dir)
{
	if (dir == CENTER)
	{	// OɃtH[JXꂢ̂tH[JXB
		if (::IsWindow(this->Focus) && !HasFocus(this->Focus))
			::SetFocus(this->Focus);
		else
		{
			this->Focus = null;
			SetFocus();
		}
		return;
	}

	if (HWND hwnd = LayoutFocus(this, dir))
		OnChildFocus(hwnd);
}

SIZE Window::get_DefaultSize() const throw()
{
	SIZE s = { 0, 0 };
	switch (this->Dock)
	{
	case NORTH:
	case SOUTH:
		s.cy = this->ClientSize.cy;
		break;
	case EAST:
	case WEST:
		s.cx = this->ClientSize.cx;
		break;
	}
	return s;
}

void Window::Scroll(DIRECTION dir)
{
	switch (dir)
	{
	case NORTH:
		SendMessage(WM_VSCROLL, SB_PAGEUP);
		break;
	case SOUTH:
		SendMessage(WM_VSCROLL, SB_PAGEDOWN);
		break;
	case WEST:
		SendMessage(WM_HSCROLL, SB_PAGELEFT);
		break;
	case EAST:
		SendMessage(WM_HSCROLL, SB_PAGERIGHT);
		break;
	}
}

//==================================================================================================================================================================

object Window::get_parent() throw()	{ return object(Parent->self); }
object Window::get_root() throw()	{ return object(Root->self); }

object Window::get_name() throw()
{
	int ln = GetTextLength();
	if (ln < 0)
		return null;
	PWSTR buf;
	unicode r = Unicode::newobj(&buf, ln);
	GetText(buf, ln+1);
	return r;
}

void	Window::set_name(PyObject* value)		{ SetText(cast<unicode>(value)->str()); }
object	Window::get_visible() throw()			{ return cast<boolean>(get_Visible()); }
void	Window::set_visible(PyObject* value)	{ set_Visible(Boolean::value(value)); }
object	Window::get_location() throw()			{ return Tuple::newobj(get_Location()); }
void	Window::set_location(PyObject* value)	{ set_Location(cast<POINT>(value)); }
object	Window::get_dock() throw()				{ return cast<integer>(get_Dock()); }
void	Window::set_dock(PyObject* value)		{ set_Dock(VerifyDirection(cast<int>(value))); }
object	Window::get_value()						{ return get_name(); }
void	Window::set_value(PyObject* value)		{ set_name(value); }

object Window::get_children()
{
	std::vector<HWND> children;
	LayoutEnumAll(this, children);
	list r = List::newobj(0);
	for (std::vector<HWND>::const_iterator i = children.begin(); i != children.end(); ++i)
	{
		if (Window* w = From(*i))
			r->append(w->self);
	}
	return r;
}

object Window::get_size() throw()
{
	if (GetAncestor(GA_PARENT))
		return Tuple::newobj(get_Size());
	else
		return Tuple::newobj(get_ClientSize());
}
void Window::set_size(PyObject* value)
{
	if (GetAncestor(GA_PARENT))
		set_Size(cast<SIZE>(value));
	else
		set_ClientSize(cast<SIZE>(value));
}

object Window::get_focus()
{
	if (Window* w = From(this->Focus))
		return object(w->self);
	return null;
}

void Window::close() throw()
{
	Close();
}

void Window::move(Tuple* args)
{
	int x, y, w = -1, h = -1;
	const PCSTR formats[] = { "ii|ii", "(ii)|(ii)", "(iiii)", null };
	parse(args, formats, &x, &y, &w, &h);
	SetPos(x, y, w, h, (w >= 0 && h >= 0) ? 0 : SWP_NOSIZE);
}

void Window::resize(Tuple* args)
{
	set_size(args);
}

object Window::press(UINT vkey, UINT mods)
{
	if (255 < vkey)
		throw ValueError("Invalid vkey at Window.press()");

	bool r;
	{
		// Press() łǂꂾԂ邩͗\s\Ȃ̂ŁAPython bNĂB
		py::Unlock unlock;
		r = Press(vkey, mods);
	}

	return cast<boolean>(r);
}

object Window::press(Tuple* args)
{
	UINT vkey, mods = 0;
	parse(args, "i|i", &vkey, &mods);
	return press(vkey, mods);
}

void Window::activate(Tuple* args)
{
	DIRECTION dir = CENTER;
	parse(args, "|i", &dir);

	py::Unlock unlock;
	Activate(dir);
}

void Window::scroll(PyObject* o)
{
	return Scroll((DIRECTION)cast<int>(o));
}

//==================================================================================================================================================================

HWND Window::__init__parse(Tuple* args, Dict* kw, DIRECTION dock) throw()
{
	PCSTR kwds[] = { "parent", "size", "dock", "__handle", null };

	PyObject*	parent = null;
	PyObject*	size = null;
	HWND		hwnd = null;

	parse(args, kw, "|O!Oii", kwds, typeof<Window>(), &parent, &size, &dock, &hwnd);

	m_dock = VerifyDirection(dock);

	if (size)
		m_size = cast<SIZE>(size);

	if (::IsWindow(hwnd))
	{	// ̃EBhEɃA^b`B
		m_hwnd = hwnd;
	}

	return parent ? cast<Window>(parent)->m_hwnd : null;
}

void Window::OnDragDrop(POINT pt, RECT* rc, IDropTarget** pp) throw()
{
	*pp = null;
}

RECT Window::get_CenterArea() const throw()
{
	RECT bounds = this->ClientArea;
	if (::IsRectEmpty(&bounds))
		return bounds; // minimized?

	if (m_rearrange)
		const_cast<Window*>(this)->UpdateLayout(); // ɍĔzuv𔽉fĂB

	LayoutCenterArea(this, &bounds);
	return bounds;
}

void Window::UpdateLayout()
{
	// OnLayoutɍĂ m_rearrange=true ɂȂ邩Ȃ̂ŁA
	// =false ̑͌ďoɍsB
	OnLayout();
	m_rearrange = false;
}

//============================================================================

PYTHON_METHODS(Window)
	PYTHON_METHOD(activate)
	PYTHON_METHOD(close)
	PYTHON_METHOD(move)
	PYTHON_METHOD(press)
	PYTHON_METHOD(resize)
	PYTHON_METHOD(scroll)
PYTHON_END()

PYTHON_MEMBERS(Window)
	PYTHON_MEMBER(children		, GET )
	PYTHON_MEMBER(dock			, BOTH)
	PYTHON_MEMBER(focus			, GET )
	PYTHON_MEMBER(gesture		, VAR )
	PYTHON_MEMBER(keymap		, VAR )
	PYTHON_MEMBER(location		, BOTH)
	PYTHON_MEMBER(name			, BOTH)
	PYTHON_MEMBER(parent		, GET )
	PYTHON_MEMBER(root			, GET )
	PYTHON_MEMBER(size			, BOTH)
	PYTHON_MEMBER(value			, BOTH)
	PYTHON_MEMBER(visible		, BOTH)
PYTHON_END()
