// xp_filedialog.cpp

#include "stdafx.h"
#include "xp_class.hpp"
#include "hwnd.hpp"
#include "unknown.hpp"
#include "shell.hpp"
#include <commdlg.h>

#pragma comment(lib, "comdlg32")

using namespace brownie;

//==================================================================================================================================================================

namespace
{
	class XpFileDialog : public Unknown< implements<IModalWindow, IFileDialog> >
	{
	private:
		OPENFILENAME	m_ofn; 
		bool			m_isOpen;
		WCHAR			m_title[_MAX_FNAME];
		WCHAR			m_path[_MAX_PATH];
		WCHAR			m_extension[_MAX_EXT];

	public:
		XpFileDialog(bool isOpen);

	public: // IModalWindow
		STDMETHODIMP Show(HWND hwndParent);

	public: // IFileDialog
		STDMETHODIMP SetFileTypes(UINT cFileTypes, const COMDLG_FILTERSPEC* rgFilterSpec)	{ return E_NOTIMPL; }
		STDMETHODIMP SetFileTypeIndex(UINT iFileType)		{ return E_NOTIMPL; }
		STDMETHODIMP GetFileTypeIndex(UINT* piFileType)		{ return E_NOTIMPL; }
		STDMETHODIMP Advise(IFileDialogEvents* events, DWORD* cookie)	{ return E_NOTIMPL; }
		STDMETHODIMP Unadvise(DWORD cookie)					{ return E_NOTIMPL; }
		STDMETHODIMP SetOptions(DWORD fos)					{ return E_NOTIMPL; }
		STDMETHODIMP GetOptions(DWORD* pfos)				{ return E_NOTIMPL; }
		STDMETHODIMP SetDefaultFolder(IShellItem* item)		{ return E_NOTIMPL; }
		STDMETHODIMP SetFolder(IShellItem* item)			{ return E_NOTIMPL; }
		STDMETHODIMP GetFolder(IShellItem** item)			{ return E_NOTIMPL; }
		STDMETHODIMP GetCurrentSelection(IShellItem** item)	{ return E_NOTIMPL; }
		STDMETHODIMP SetFileName(PCWSTR name)				{ return E_NOTIMPL; }
		STDMETHODIMP GetFileName(LPWSTR* name)				{ return E_NOTIMPL; }
		STDMETHODIMP SetTitle(PCWSTR pszTitle)				{ return E_NOTIMPL; }
		STDMETHODIMP SetOkButtonLabel(PCWSTR label)		{ return E_NOTIMPL; }
		STDMETHODIMP SetFileNameLabel(PCWSTR label)		{ return E_NOTIMPL; }
		STDMETHODIMP GetResult(IShellItem** item);
		STDMETHODIMP AddPlace(IShellItem* item, FDAP fdap)	{ return E_NOTIMPL; }
		STDMETHODIMP SetDefaultExtension(PCWSTR extension);
		STDMETHODIMP Close(HRESULT hr)						{ return E_NOTIMPL; }
		STDMETHODIMP SetClientGuid(REFGUID guid)			{ return E_NOTIMPL; }
		STDMETHODIMP ClearClientData(void)					{ return E_NOTIMPL; }
		STDMETHODIMP SetFilter(IShellItemFilter* pFilter)	{ return E_NOTIMPL; }

	public: // IFileOpenDialog
		STDMETHODIMP GetResults(IShellItemArray** items)		{ return E_NOTIMPL; }
		STDMETHODIMP GetSelectedItems(IShellItemArray** items)	{ return E_NOTIMPL; }

	public: // IFileSaveDialog
		STDMETHODIMP SetSaveAsItem(IShellItem* item)		{ return E_NOTIMPL; }
		STDMETHODIMP SetProperties(IPropertyStore* store)	{ return E_NOTIMPL; }
		STDMETHODIMP SetCollectedProperties(IPropertyDescriptionList* pList, BOOL fAppendDefault){ return E_NOTIMPL; }
		STDMETHODIMP GetProperties(IPropertyStore** ppStore){ return E_NOTIMPL; }
		STDMETHODIMP ApplyProperties(IShellItem* item, IPropertyStore* store, HWND hwnd, IFileOperationProgressSink* pSink){ return E_NOTIMPL; }
	};
}

//==================================================================================================================================================================

UINT_PTR CALLBACK XpFileDialog_OnMessage(HWND hwnd, UINT msg, WPARAM wParam,LPARAM lParam)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		CenterWindow(::GetParent(hwnd)); // e擾̂|CgI
		break;
	}

	return 0;
}

XpFileDialog::XpFileDialog(bool isOpen) : m_isOpen(isOpen)
{
	memset(&m_ofn, 0, sizeof(m_ofn)); // initialize structure to 0/null
	m_path[0] = L'\0';
	m_title[0] = L'\0';

	m_ofn.lStructSize = sizeof(m_ofn);
	m_ofn.lpstrFile = m_path;
	m_ofn.nMaxFile = _MAX_PATH;
	m_ofn.lpstrFileTitle = m_title;
	m_ofn.nMaxFileTitle = _MAX_FNAME;
	m_ofn.Flags |= OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT | OFN_EXPLORER | OFN_ENABLEHOOK | OFN_ENABLESIZING;
	m_ofn.lpstrFilter = null; //  lpszFilter;
	m_ofn.hInstance = ::GetModuleHandle(null);
	m_ofn.lpfnHook = XpFileDialog_OnMessage;
//	m_ofn.hwndOwner = hWndParent;

	// setup initial file name
//	if (lpszFileName != null)
//		lstrcpyn(m_path, lpszFileName, _MAX_PATH);
}

STDMETHODIMP XpFileDialog::Show(HWND hwndParent)
{
//	ASSERT(m_ofn.Flags & OFN_ENABLEHOOK);
//	ASSERT(m_ofn.lpfnHook != null);	// can still be a user hook
//	ASSERT(m_ofn.Flags & OFN_EXPLORER);

	m_ofn.hwndOwner = hwndParent;
	if (!m_ofn.hwndOwner)
		m_ofn.hwndOwner = ::GetActiveWindow();

//	ASSERT(hwnd == null);
//	_AtlWinModule.AddCreateWndData(&m_thunk.cd, (CDialogImplBase*)this);
	BOOL r = (m_isOpen ? ::GetOpenFileName(&m_ofn) : ::GetSaveFileName(&m_ofn));
//	hwnd = null;

	return r ? S_OK : S_FALSE;
}

STDMETHODIMP XpFileDialog::SetDefaultExtension(PCWSTR extension)
{
	if (extension)
	{
		m_ofn.lpstrDefExt = m_extension;
		wcscpy_s(m_extension, extension);
	}
	else
		m_ofn.lpstrDefExt = null;
	return S_OK;
}

STDMETHODIMP XpFileDialog::GetResult(IShellItem** item)
{
	return ItemCreate(item, m_path);
}

//==================================================================================================================================================================

HRESULT brownie::XpFileOpenDialogCreate(REFINTF pp)
{
	return ref<XpFileDialog>::from(new XpFileDialog(true)).copyto(pp);
}

HRESULT brownie::XpFileSaveDialogCreate(REFINTF pp)
{
	return ref<XpFileDialog>::from(new XpFileDialog(false)).copyto(pp);
}
