// xp_knownfolder.cpp

#include "stdafx.h"
#include "xp_class.hpp"
#include "unknown.hpp"
#include "shell.hpp"

using namespace brownie;

//==================================================================================================================================================================

namespace
{
	class XpKnownFolder : public Unknown< implements<IKnownFolder> >
	{
	public: // IKnownFolder
		STDMETHODIMP GetId(KNOWNFOLDERID *pkfid);
		STDMETHODIMP GetCategory(KF_CATEGORY *pCategory);
		STDMETHODIMP GetShellItem(DWORD dwFlags, REFIID riid, void **ppv);
		STDMETHODIMP GetPath(DWORD dwFlags, LPWSTR *ppszPath);
		STDMETHODIMP SetPath(DWORD dwFlags, PCWSTR pszPath);
		STDMETHODIMP GetIDList(DWORD dwFlags, PIDLIST_ABSOLUTE *ppidl);
		STDMETHODIMP GetFolderType(FOLDERTYPEID *pftid);
		STDMETHODIMP GetRedirectionCapabilities(KF_REDIRECTION_CAPABILITIES *pCapabilities);
		STDMETHODIMP GetFolderDefinition(KNOWNFOLDER_DEFINITION *pKFD);
	};

	class XpKnownFolderManager : public Unknown< implements<IKnownFolderManager>, mixin<StaticLife> >
	{
	public: // IKnownFolderManager
		STDMETHODIMP FolderIdFromCsidl(int nCsidl, KNOWNFOLDERID *pfid)		{ return E_NOTIMPL; }
		STDMETHODIMP FolderIdToCsidl(REFKNOWNFOLDERID rfid, int *pnCsidl)	{ return E_NOTIMPL; }
		STDMETHODIMP GetFolderIds(KNOWNFOLDERID **ppKFId, UINT *pCount);
		STDMETHODIMP GetFolder(REFKNOWNFOLDERID rfid, IKnownFolder **ppkf);
		STDMETHODIMP GetFolderByName(PCWSTR pszCanonicalName, IKnownFolder **ppkf)				{ return E_NOTIMPL; }
		STDMETHODIMP RegisterFolder(REFKNOWNFOLDERID rfid, const KNOWNFOLDER_DEFINITION *pKFD)	{ return E_NOTIMPL; }
		STDMETHODIMP UnregisterFolder(REFKNOWNFOLDERID rfid)									{ return E_NOTIMPL; }
		STDMETHODIMP FindFolderFromPath(PCWSTR pszPath, FFFP_MODE mode, IKnownFolder **ppkf)	{ return E_NOTIMPL; }
		STDMETHODIMP FindFolderFromIDList(PCIDLIST_ABSOLUTE pidl, IKnownFolder **ppkf)			{ return E_NOTIMPL; }
		STDMETHODIMP Redirect(REFKNOWNFOLDERID rfid, HWND hwnd, KF_REDIRECT_FLAGS flags, PCWSTR pszTargetPath, UINT cFolders, const KNOWNFOLDERID *pExclusion, LPWSTR *ppszError)	{ return E_NOTIMPL; }
	};

	XpKnownFolderManager theXpKnownFolderManager;

	const struct
	{
		KNOWNFOLDERID	id;
		int				csidl;
	}
	KNOWNFOLDERID2CSIDL[] = 
	{
		{ FOLDERID_AddNewPrograms		, 	},
		{ FOLDERID_AdminTools			, CSIDL_ADMINTOOLS	},
		{ FOLDERID_AppUpdates			, 	},
		{ FOLDERID_CDBurning			, CSIDL_CDBURN_AREA	},
		{ FOLDERID_ChangeRemovePrograms	, 	},
		{ FOLDERID_CommonAdminTools		, CSIDL_COMMON_ADMINTOOLS	},
		{ FOLDERID_CommonOEMLinks		, 	},
		{ FOLDERID_CommonPrograms		, CSIDL_COMMON_PROGRAMS	},
		{ FOLDERID_CommonStartMenu		, CSIDL_COMMON_STARTMENU	},
		{ FOLDERID_CommonStartup		, CSIDL_COMMON_STARTUP	},
		{ FOLDERID_CommonTemplates		, CSIDL_COMMON_TEMPLATES	},
		{ FOLDERID_ComputerFolder		, CSIDL_DRIVES	},
		{ FOLDERID_ConflictFolder		, 	},
		{ FOLDERID_ConnectionsFolder	, CSIDL_CONNECTIONS	},
		{ FOLDERID_Contacts				, 	},
		{ FOLDERID_ControlPanelFolder	, CSIDL_CONTROLS	},
		{ FOLDERID_Cookies				, CSIDL_COOKIES	},
		{ FOLDERID_Desktop				, CSIDL_DESKTOP	},
		{ FOLDERID_Documents			, CSIDL_MYDOCUMENTS	},
		{ FOLDERID_Downloads			, 	},
		{ FOLDERID_Favorites			, CSIDL_FAVORITES	},
		{ FOLDERID_Fonts				, CSIDL_FONTS	},
		{ FOLDERID_Games				, 	},
		{ FOLDERID_GameTasks			, 	},
		{ FOLDERID_History				, CSIDL_HISTORY	},
		{ FOLDERID_InternetCache		, CSIDL_INTERNET_CACHE	},
		{ FOLDERID_InternetFolder		, CSIDL_INTERNET	},
		{ FOLDERID_Links				, 	},
		{ FOLDERID_LocalAppData			, CSIDL_LOCAL_APPDATA	},
		{ FOLDERID_LocalAppDataLow		, 	},
		{ FOLDERID_LocalizedResourcesDir, CSIDL_RESOURCES_LOCALIZED	},
		{ FOLDERID_Music				, CSIDL_MYMUSIC	},
		{ FOLDERID_NetHood				, CSIDL_NETHOOD	},
		{ FOLDERID_NetworkFolder		, CSIDL_NETWORK	},
		{ FOLDERID_OriginalImages		, 	},
		{ FOLDERID_PhotoAlbums			, 	},
		{ FOLDERID_Pictures				, CSIDL_MYPICTURES	},
		{ FOLDERID_Playlists			, 	},
		{ FOLDERID_PrintersFolder		, CSIDL_PRINTERS	},
		{ FOLDERID_PrintHood			, CSIDL_PRINTHOOD	},
		{ FOLDERID_Profile				, CSIDL_PROFILE	},
		{ FOLDERID_ProgramData			, 	},
		{ FOLDERID_ProgramFiles			, CSIDL_PROGRAM_FILES	},
		{ FOLDERID_ProgramFilesCommon	, CSIDL_PROGRAM_FILES_COMMON	},
		{ FOLDERID_ProgramFilesCommonX64, 	},
		{ FOLDERID_ProgramFilesCommonX86, 	},
		{ FOLDERID_ProgramFilesX64		, 	},
		{ FOLDERID_ProgramFilesX86		, 	},
		{ FOLDERID_Programs				, CSIDL_PROGRAMS	},
		{ FOLDERID_Public				, 	},
		{ FOLDERID_PublicDesktop		, CSIDL_COMMON_DESKTOPDIRECTORY	},
		{ FOLDERID_PublicDocuments		, CSIDL_COMMON_DOCUMENTS	},
		{ FOLDERID_PublicDownloads		, 	},
		{ FOLDERID_PublicGameTasks		, 	},
		{ FOLDERID_PublicMusic			, CSIDL_COMMON_MUSIC	},
		{ FOLDERID_PublicPictures		, CSIDL_COMMON_PICTURES	},
		{ FOLDERID_PublicVideos			, CSIDL_COMMON_VIDEO	},
		{ FOLDERID_QuickLaunch			, 	},
		{ FOLDERID_Recent				, CSIDL_RECENT	},
		{ FOLDERID_RecordedTV			, 	},
		{ FOLDERID_RecycleBinFolder		, CSIDL_BITBUCKET	},
		{ FOLDERID_ResourceDir			, CSIDL_RESOURCES	},
		{ FOLDERID_RoamingAppData		, 	},
		{ FOLDERID_SampleMusic			, 	},
		{ FOLDERID_SamplePictures		, 	},
		{ FOLDERID_SamplePlaylists		, 	},
		{ FOLDERID_SampleVideos			, 	},
		{ FOLDERID_SavedGames			, 	},
		{ FOLDERID_SavedSearches		, 	},
		{ FOLDERID_SEARCH_CSC			, 	},
		{ FOLDERID_SEARCH_MAPI			, 	},
		{ FOLDERID_SearchHome			, 	},
		{ FOLDERID_SendTo				, CSIDL_SENDTO	},
		{ FOLDERID_SidebarDefaultParts	, 	},
		{ FOLDERID_SidebarParts			, 	},
		{ FOLDERID_StartMenu			, CSIDL_STARTMENU	},
		{ FOLDERID_Startup				, CSIDL_STARTUP	},
		{ FOLDERID_SyncManagerFolder	, 	},
		{ FOLDERID_SyncResultsFolder	, 	},
		{ FOLDERID_SyncSetupFolder		, 	},
		{ FOLDERID_System				, CSIDL_SYSTEM	},
		{ FOLDERID_SystemX86			, 	},
		{ FOLDERID_Templates			, CSIDL_TEMPLATES	},
		{ FOLDERID_TreeProperties		, 	},
		{ FOLDERID_UserProfiles			, CSIDL_PERSONAL	},
		{ FOLDERID_UsersFiles			, 	},
		{ FOLDERID_Videos				, CSIDL_MYVIDEO	},
		{ FOLDERID_Windows				, CSIDL_WINDOWS	},
		{ GUID_NULL						, CSIDL_ALTSTARTUP	},
		{ GUID_NULL						, CSIDL_APPDATA				},
		{ GUID_NULL						, CSIDL_COMMON_ALTSTARTUP	},
		{ GUID_NULL						, CSIDL_COMMON_APPDATA		},
		{ GUID_NULL						, CSIDL_COMMON_FAVORITES	},
		{ GUID_NULL						, CSIDL_COMMON_OEM_LINKS	},
		{ GUID_NULL						, CSIDL_COMPUTERSNEARME		},
		{ GUID_NULL						, CSIDL_DESKTOPDIRECTORY	},
	};
}

//==================================================================================================================================================================

STDMETHODIMP XpKnownFolderManager::GetFolderIds(KNOWNFOLDERID **ppKFId, UINT *pCount)
{
	*pCount = lengthof(KNOWNFOLDERID2CSIDL);
	return E_NOTIMPL;
}

STDMETHODIMP XpKnownFolderManager::GetFolder(REFKNOWNFOLDERID rfid, IKnownFolder **ppkf)
{
	return E_NOTIMPL;
}

//==================================================================================================================================================================

HRESULT brownie::XpKnownFolderManagerCreate(REFINTF pp)
{
	return theXpKnownFolderManager.QueryInterface(pp);
}
