// xp_thumbcache.cpp

#include "stdafx.h"
#include "xp_class.hpp"
#include "unknown.hpp"
#include <thumbcache.h>

#pragma comment(lib, "comdlg32")

using namespace brownie;

//==================================================================================================================================================================

namespace
{
	class XpSharedBitmap : public Unknown< implements<ISharedBitmap> >
	{
	private:
		HBITMAP		m_handle;

	public:
		XpSharedBitmap();
		~XpSharedBitmap();
		STDMETHODIMP GetSharedBitmap(HBITMAP* phbm);
		STDMETHODIMP GetSize(SIZE* pSize);
		STDMETHODIMP GetFormat(WTS_ALPHATYPE* pat);
		STDMETHODIMP InitializeBitmap(HBITMAP hbm, WTS_ALPHATYPE wtsAT);
		STDMETHODIMP Detach(HBITMAP* phbm);
	};

	class XpLocalThumbnailCache : public Unknown< implements<IThumbnailCache> >
	{
	private:

	public: // IThumbnailCache
		STDMETHODIMP GetThumbnail(
			IShellItem* item, UINT requestedSize, WTS_FLAGS wts,
			ISharedBitmap** pp, WTS_CACHEFLAGS* outFlags, WTS_THUMBNAILID* outID);
        
        STDMETHODIMP GetThumbnailByID(
			WTS_THUMBNAILID thumbnailID, UINT requestedSize,
			ISharedBitmap** pp, WTS_CACHEFLAGS* outFlags);
	};
/*
	class XpThumbnailProvider : public Unknown< implements<IThumbnailProvider> >
	{
	public:
		STDMETHODIMP GetThumbnail(UINT cx, HBITMAP* phbmp, WTS_ALPHATYPE* pdwAlpha);
	};
*/
}

//==================================================================================================================================================================

/*
static IThumbnailCache*  theThumbnailCache;	// the thumbnail cache
WTS_THUMBNAILID thumbnailID;					// ID to retrieve cached bitmap
const UINT THUMBSIZE = 128;						// Size of thumbnail to cache
static IThumbnailCacheSetupHelper*  theThumbnailCacheHelper;
LPCWSTR aExtensions[2] = {L"TIF", L"JPEG"};

ISharedBitmap*  pBitmap	= null;
WTS_CACHEFLAGS cacheFlags		= 0; // result of the caching operation.
HBITMAP hBitmap					= null;
IShellItem* item			= null;
HRESULT hr						= S_OK; 
LPITEMIDLIST pidl				= null;

ZeroMemory(&thumbnailID.rgbKey[0], sizeof(thumbnailID.rgbKey));

theThumbnailCache->QueryInterface(IID_IThumbnailCacheSetupHelper, &theThumbnailCacheHelper);
theThumbnailCacheHelper->DeleteThumbnails(1, aExtensions);
SHCreateShellItem(null, null, pidl, &item);
theThumbnailCache->GetThumbnail(item, THUMBSIZE, WTS_INCACHEONLY, &pBitmap, &cacheFlags, &thumbnailID);
theThumbnailCache->GetThumbnail(item, THUMBSIZE, WTS_EXTRACT, &pBitmap, &cacheFlags, &thumbnailID);

	
ISharedBitmap* pBitmap = null;
HBITMAP hBitmap				 = null;
WTS_CACHEFLAGS cacheFlags;
hr = theThumbnailCache->GetThumbnailByID(thumbnailID, THUMBSIZE, &pBitmap, &cacheFlags);
*/

//==================================================================================================================================================================

XpSharedBitmap::XpSharedBitmap() : m_handle(null)
{
}

XpSharedBitmap::~XpSharedBitmap()
{
	if (m_handle)
		::DeleteObject(m_handle);
}

STDMETHODIMP XpSharedBitmap::GetSharedBitmap(HBITMAP* phbm)
{
	ASSERT(phbm);
	if (!phbm)
		return E_POINTER;
	*phbm = m_handle;
	return m_handle ? S_OK : E_HANDLE;
}

STDMETHODIMP XpSharedBitmap::GetSize(SIZE* pSize)
{
	ASSERT(pSize);
	if (!pSize)
		return E_POINTER;
	if (!m_handle)
	{
		pSize->cx = pSize->cy = 0;
		return E_HANDLE;
	}
	BITMAP info;
	::GetObject(m_handle, sizeof(BITMAP), &info);
	pSize->cx = info.bmWidth;
	pSize->cy = info.bmHeight;
	return S_OK;
}

STDMETHODIMP XpSharedBitmap::GetFormat(WTS_ALPHATYPE* pat)
{
	ASSERT(pat);
	if (!pat)
		return E_POINTER;
	if (!m_handle)
	{
		*pat = WTSAT_UNKNOWN;
		return E_HANDLE;
	}
	BITMAP info;
	::GetObject(m_handle, sizeof(BITMAP), &info);
	*pat = (info.bmBitsPixel >= 32 ? WTSAT_ARGB : WTSAT_RGB);
	return S_OK;
}

STDMETHODIMP XpSharedBitmap::InitializeBitmap(HBITMAP hbm, WTS_ALPHATYPE wtsAT)
{
	if (m_handle)
		return E_UNEXPECTED;
	m_handle = hbm;
	return S_OK;
}

STDMETHODIMP XpSharedBitmap::Detach(HBITMAP* phbm)
{
	ASSERT(phbm);
	if (!phbm)
		return E_POINTER;
	*phbm = m_handle;
	if (!m_handle)
		return E_HANDLE;
	m_handle = null;
	return S_OK;
}

//==================================================================================================================================================================

/*
WTS_EXTRACT
WTS_INCACHEONLY
WTS_FASTEXTRACT
WTS_SLOWRECLAIM
WTS_FORCEEXTRACTION
WTS_EXTRACTDONOTCACHE
*/

STDMETHODIMP XpLocalThumbnailCache::GetThumbnail(IShellItem* item, UINT requestedSize, WTS_FLAGS wts, ISharedBitmap** pp, WTS_CACHEFLAGS* outFlags, WTS_THUMBNAILID* outID)
{
	return E_NOTIMPL;
}
        
STDMETHODIMP XpLocalThumbnailCache::GetThumbnailByID(WTS_THUMBNAILID thumbnailID, UINT requestedSize, ISharedBitmap** pp, WTS_CACHEFLAGS* outFlags)
{
	return E_NOTIMPL;
}

//==================================================================================================================================================================

HRESULT brownie::XpSharedBitmapCreate(REFINTF pp)
{
	return newobj<XpSharedBitmap>(pp);
}

HRESULT brownie::XpLocalThumbnailCacheCreate(REFINTF pp)
{
	return newobj<XpLocalThumbnailCache>(pp);
}
