// xp_rstrtmgr.cpp

#include "stdafx.h"
#include <algorithm>
#include "ntddk.hpp"
#include "autoptr.hpp"
using namespace brownie;

namespace
{
	typedef NTSTATUS (NTAPI *TNtQueryObject)(
		IN  HANDLE						ObjectHandle OPTIONAL,
		IN  OBJECT_INFORMATION_CLASS	ObjectInformationClass,
		OUT PVOID						ObjectInformation,
		IN  ULONG						Length,
		OUT PULONG						ResultLength);

	typedef NTSTATUS (NTAPI *TNtQuerySystemInformation)(
		IN  SYSTEM_INFORMATION_CLASS	SystemInformationClass,
		OUT PVOID						SystemInformation,
		IN  ULONG						SystemInformationLength,
		OUT PULONG						ReturnLength OPTIONAL);

	const PCWSTR OBJECT_NAMES[] =
	{
		L"Adapter",
		L"Callback",
		L"Controller",
		L"DebugObject",
		L"Desktop",
		L"Device",
		L"Directory",
		L"Driver",
		L"Event",
		L"EventPair",
		L"File",
		L"FilterCommunicationPort",
		L"FilterConnectionPort",
		L"IoCompletion",
		L"Job",
		L"Key",
		L"KeyedEvent",
		L"Mutant",
		L"Port",
		L"Process",
		L"Profile",
		L"Section",
		L"Semaphore",
		L"SymbolicLink",
		L"Thread",
		L"Timer",
		L"Token",
		L"Type",
		L"WaitablePort",
		L"WindowStation",
		L"WmiGuid",
	};
}

NTSTATUS NTAPI NtQueryObject(
	IN  HANDLE						ObjectHandle OPTIONAL,
	IN  OBJECT_INFORMATION_CLASS	ObjectInformationClass,
	OUT PVOID						ObjectInformation,
	IN  ULONG						Length,
	OUT PULONG						ResultLength)
{
	static TNtQueryObject fn = (TNtQueryObject) GetProcAddress(GetModuleHandle(L"ntdll.dll"), "NtQueryObject");
	return fn(ObjectHandle, ObjectInformationClass, ObjectInformation, Length, ResultLength);
}

NTSTATUS NTAPI NtQuerySystemInformation(
	IN  SYSTEM_INFORMATION_CLASS	SystemInformationClass,
    OUT PVOID						SystemInformation,
    IN  ULONG						SystemInformationLength,
    OUT PULONG						ReturnLength OPTIONAL)
{
	static TNtQuerySystemInformation fn = (TNtQuerySystemInformation) GetProcAddress(GetModuleHandle(L"ntdll.dll"), "NtQuerySystemInformation");
	return fn(SystemInformationClass, SystemInformation, SystemInformationLength, ReturnLength);
}

namespace NT
{
	DWORD OBJECT_TYPE_ID[ObjectTypeMax];

	HRESULT Initialize() throw()
	{
		DWORD size = 0;
		AutoPtr<OBJECT_ALL_TYPES_INFORMATION> r;
retry:
		NTSTATUS code = NtQueryObject(NULL, ObjectAllTypesInformation, r, size, &size);
		switch (code)
		{
		case STATUS_SUCCESS:
			break;
		case STATUS_INFO_LENGTH_MISMATCH:
			r.realloc(size);
			goto retry;
		default:
			return HRESULT_FROM_WIN32(code);
		}

		OBJECT_TYPE_INFORMATION* o = r->TypeInformation;
		for (DWORD i = 0; i < r->NumberOfTypes; ++i)
		{
			size_t idx = std::lower_bound(OBJECT_NAMES, OBJECT_NAMES + lengthof(OBJECT_NAMES), o->TypeName, Less()) - OBJECT_NAMES;
			if (idx < ObjectTypeMax && OBJECT_NAMES[idx] == o->TypeName)
				OBJECT_TYPE_ID[idx] = i + 1;
			else
				ASSERT(!"Unknown NT Object");
			o = (OBJECT_TYPE_INFORMATION *)((BYTE*)o->TypeName.Buffer + ((o->TypeName.MaximumLength + 3) & ~3));
		}

#if _DEBUG
		for (DWORD i = 0; i < ObjectTypeMax; ++i)
			ASSERT(OBJECT_TYPE_ID[i]);
#endif

		return S_OK;
	}

	SYSTEM_HANDLE_INFORMATION* QuerySystemHandleInformation() throw()
	{
		DWORD size = 0;
		SYSTEM_HANDLE_INFORMATION head;
		switch (NtQuerySystemInformation(SystemHandleInformation, &head, sizeof(SYSTEM_HANDLE_INFORMATION), &size))
		{
		case STATUS_INFO_LENGTH_MISMATCH:
			break;
		default:
			return NULL;
		}

		size += 16 * sizeof(SYSTEM_HANDLE_INFORMATION); // ߂ɁB
		AutoPtr<SYSTEM_HANDLE_INFORMATION> r(size);

		for (;;)
		{
			switch (NtQuerySystemInformation(SystemHandleInformation, r, size, &size))
			{
			case STATUS_SUCCESS:
				return r.detach();
			case STATUS_INFO_LENGTH_MISMATCH:
				r.realloc(size);
				break;
			default:
				return NULL;
			}
		}
	}

	UNICODE_STRING* GetObjectName(HANDLE handle)
	{
		DWORD size = 0;
		UNICODE_STRING head;
		switch (NtQueryObject(handle, ObjectNameInformation, &head, sizeof(UNICODE_STRING), &size))
		{
		case STATUS_BUFFER_OVERFLOW:
		case STATUS_INFO_LENGTH_MISMATCH:
			break;
		default:
			return NULL;
		}

		size += 16 * sizeof(WCHAR); // ߂ɁB
		UNICODE_STRING* r = (UNICODE_STRING*) malloc(size);

		for (;;)
		{
			switch (NtQueryObject(handle, ObjectNameInformation, r, size, &size))
			{
			case STATUS_SUCCESS:
				return r;
			case STATUS_INFO_LENGTH_MISMATCH:
				r = (UNICODE_STRING*) realloc(r, size);
				break;
			default:
				free(r);
				return NULL;
			}
		}
	}

	UNICODE_STRING* GetObjectName(HANDLE process, const SYSTEM_HANDLE_TABLE_ENTRY_INFO& handle)
	{
		HANDLE dup;
		if (!DuplicateHandle(process, (HANDLE)handle.HandleValue, GetCurrentProcess(), &dup, 0, FALSE, DUPLICATE_SAME_ACCESS))
			return NULL;
		UNICODE_STRING* name = NT::GetObjectName(dup);
		CloseHandle(dup);
		return name;
	}
}
