// main.cpp

#include "stdafx.h"
#include "graphics.hpp"

#pragma comment(lib, "gdiplus")
#pragma comment(lib, "gdi32")
#pragma comment(lib, "comctl32")

using namespace brownie;

ULONG_PTR theToken;
Gdiplus::GdiplusStartupInput theStartupInput;

static Gdiplus::Color SysColor(int n)
{
	COLORREF c = ::GetSysColor(n);
	return Gdiplus::Color(GetRValue(c), GetGValue(c), GetBValue(c));
}

static void CopyBits(void* dst, int dststride, const void* src, int srcstride, size_t bpp, size_t width, size_t height)
{
	if (dststride >= 0 && dststride == srcstride)
	{
		memcpy(dst, src, srcstride * height);
	}
	else
	{
		BYTE* d = (BYTE*)dst;
		BYTE* s = (BYTE*)src;
		for (size_t i = 0; i < height; ++i)
		{
			memcpy(d, s, bpp * width);
			d += dststride;
			s += srcstride;
		}
	}
}

static HIMAGELIST LoadImageList(Gdiplus::Bitmap* image)
{
	int w = image->GetWidth();
	int h = image->GetHeight();

	HIMAGELIST imagelist = ImageList_Create(h, h, ILC_COLOR32, w/h, 20);
	ImageList_SetBkColor(imagelist, CLR_NONE);

	if (image->GetPixelFormat() == PixelFormat32bppARGB)
	{
		Gdiplus::Rect rcLock(0, 0, w, h);
		Gdiplus::BitmapData bits;
		if (image->LockBits(&rcLock, Gdiplus::ImageLockModeRead, PixelFormat32bppARGB, &bits) == Gdiplus::Ok)
		{
			// TODO: rbg}bṽRs[炷ׂB
			// ݁AGDI+ Bitmap  DIBSection  ImageListobt@ ƕ̃Rs[ĂB
			// DIBSection XLbvł΁ARs[点B̂߂ɂ͈ȉ̂ǂ炩̕@B
			// EGDI+ Bitmap̓f[^𒼐ڎQƂDIBitmap쐬B
			// EImageList̃obt@ɒڏށB
			const int bpp = 4; // sizeof(pixel for PixelFormat32bppARGB)
			BITMAPINFO info = { 0 };
			info.bmiHeader.biSize			= sizeof(info.bmiHeader);
			info.bmiHeader.biWidth			= w; 
			info.bmiHeader.biHeight			= h; 
			info.bmiHeader.biPlanes			= 1; 
			info.bmiHeader.biBitCount		= 32; 
			info.bmiHeader.biCompression	= BI_RGB;
			BYTE* dst = NULL;
			int dststride = ((w * bpp + 3) & ~3);

			HBITMAP bitmap = CreateDIBSection(NULL, &info, DIB_RGB_COLORS, (void**)&dst, NULL, 0);
			CopyBits(dst + dststride * (h-1), -dststride, bits.Scan0, bits.Stride, bpp, w, h);
			image->UnlockBits(&bits);
			ImageList_Add(imagelist, bitmap, NULL);
			DeleteObject(bitmap);
			return imagelist;
		}
	}

	// At@Ȃrbg}bv
	HBITMAP bitmap;
	if (image->GetHBITMAP(SysColor(COLOR_3DFACE), &bitmap) != Gdiplus::Ok)
		return NULL;
	ImageList_Add(imagelist, bitmap, NULL);
	DeleteObject(bitmap);
	return imagelist;
}

class GdiplusGraphics : public IGraphics
{
public:
	void Dispose() throw()
	{
		if (theToken)
		{
			Gdiplus::GdiplusShutdown(theToken);
			theToken = 0;
		}
	}

	HBITMAP BitmapLoad(IStream* stream) throw()
	{
		return NULL;
	}

	HBITMAP BitmapLoad(PCWSTR filename) throw()
	{
		return NULL;
	}

	HIMAGELIST ImageListLoad(IStream* stream) throw()
	{
		HIMAGELIST imagelist = NULL;
		Gdiplus::Bitmap* image = new Gdiplus::Bitmap(stream);
		if (image->GetLastStatus() == Gdiplus::Ok)
			imagelist = LoadImageList(image);
		delete image;
		return imagelist;
	}

	HIMAGELIST ImageListLoad(PCWSTR filename) throw()
	{
		HIMAGELIST imagelist = NULL;
		Gdiplus::Bitmap* image = new Gdiplus::Bitmap(filename);
		if (image->GetLastStatus() == Gdiplus::Ok)
			imagelist = LoadImageList(image);
		delete image;
		return imagelist;
	}
};

GdiplusGraphics theGraphics;

extern "C"
__declspec(dllexport) IGraphics* GraphicsCreate()
{
	if (!theToken)
		Gdiplus::GdiplusStartup(&theToken, &theStartupInput, NULL);
	return &theGraphics;
}

BOOL APIENTRY DllMain(HMODULE module, DWORD what, LPVOID)
{
	switch (what)
	{
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
	case DLL_THREAD_DETACH:
	case DLL_PROCESS_DETACH:
		break;
	}
    return TRUE;
}
