// abstractbar.hpp
#pragma once

#include "commctrl.hpp"
#include "dragdrop.hpp"
#include "icons.hpp"
#include "window.hpp"

namespace brownie
{
	class AbstractBar : public Class<Window>
	{
	public:
		struct Item
		{
			object	node;
			unicode	name;
			int		icon;
			bool	enabled;
			bool	checked;
			bool	executable;
			int		offset;
		};

	private:
		object				m_icons;
		std::vector<Item>	m_items;
		ssize_t				m_pushed;	// NbNꂽ^u
		ssize_t				m_hot;		// zbgԂ̃^u
		DropTarget			m_drop;
		Font				m_font;

	public:
		void	set_Dock(DIRECTION value);
		SIZE	get_DefaultSize() const throw();

		Icons*	GetIcons() const throw();
		void	SetIcons(Icons* value) throw();

		      Item* GetItem(ssize_t index);
		const Item*	GetItem(ssize_t index) const	{ return const_cast<AbstractBar*>(this)->GetItem(index); }
		      Item* GetItems();
		const Item*	GetItems() const				{ return const_cast<AbstractBar*>(this)->GetItems(); }
		ssize_t	GetItemCount() const throw()		{ return m_items.size() - 1; }
		ssize_t	GetPushedItem() const throw()		{ return m_pushed; }
		void	SetPushedItem(ssize_t index) throw();
		ssize_t	GetHotItem() const throw()			{ return m_hot; }
		void	SetHotItem(ssize_t index) throw();
		bool	GetItemRect(ssize_t index, RECT* rc) const throw();
		ssize_t	HitTest(int x, int y) const throw();
		ssize_t	HitTest(POINT pt) const throw()				{ return HitTest(pt.x, pt.y); }
		HRESULT DropDown(ssize_t index, POINT pt) throw();
		HRESULT	DropDown(ssize_t index) throw();

		void	GetStatus(int* mode, STATUS* status, ssize_t index, const int modes[]) const throw();

	protected:
		virtual void	OnDraw(HDC dc, const RECT& rcPaint);

	protected:
		LRESULT	OnMessage(UINT msg, WPARAM wParam, LPARAM lParam);
		void	OnUpdate();
		bool	OnGesture(const MSG& msg);
		void	OnDragDrop(POINT pt, RECT* rc, IDropTarget** pp) throw();

	private:
		void	OnMouseUp(UINT vk, WPARAM wParam, LPARAM lParam);

	public:
		static PCSTR __name__() throw()					{ return "brownie.AbstractBar"; }
		void	__init__(Tuple* args, Dict* kw) throw(...);
		void	__init__(Tuple* args, Dict* kw, DIRECTION dock, DWORD style, DWORD styleEx) throw(...);
//		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();
		void	set_icons(PyObject* value);
		object	get_items();
		void	set_items(PyObject* value);
		void	set_theme(PyObject* value);
	};

	//==================================================================================================================================================================
}
