// brownie.hpp
#pragma once

typedef struct _object PyObject;

#ifdef MS_WIN64
typedef __int64 ssize_t;
#else
typedef _W64 int ssize_t;
#endif

//========================================================================================================================
// STATIC_ASSERT : RpC̃AT[V

template < bool x > struct STATIC_ASSERTION_FAILURE;
template <> struct STATIC_ASSERTION_FAILURE<true>{};
template < int x > struct STATIC_ASSERT_TEST{};
#define STATIC_ASSERT(exp)	typedef STATIC_ASSERT_TEST < sizeof(STATIC_ASSERTION_FAILURE<(bool)(exp)>) > __STATIC_ASSERT_TYPEDEF__

//========================================================================================================================
// DEBUG

#ifdef _DEBUG
#define ASSERT(expr)	if (expr); else { if (::IsDebuggerPresent()) __debugbreak(); else _ASSERTE(expr); }
#define VERIFY(expr)	ASSERT(expr)
void TRACE(PCWSTR format, ...) throw();
#else
#define ASSERT(expr)	__noop
#define VERIFY(expr)	((void)(expr))
#define TRACE			__noop
#endif

//========================================================================================================================
// Error codes

#define	E_NOTSUPPORT		HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED)

//========================================================================================================================
// 

namespace brownie
{
	template < typename T > class ValueT;
	template < typename T > class ref;

	typedef ValueT<bool>	Boolean;
	typedef ValueT<int>		Integer;
	typedef ValueT<__int64>	Numeric;
	typedef ValueT<double>	Real;
	typedef ValueT<CHAR>	String;
	typedef ValueT<WCHAR>	Unicode;
	class Iterator;
	class Sequence;
	class Array;
	class Tuple;
	class List;
	class Mapping;
	class Dict;

	typedef ref<PyObject>	object;
	typedef ref<Boolean>	boolean;
	typedef ref<Integer>	integer;
	typedef ref<Numeric>	numeric;
	typedef ref<Real>		real;
	typedef ref<String>		string;
	typedef ref<Unicode>	unicode;
	typedef ref<Iterator>	iterator;
	typedef ref<Array>		array;
	typedef ref<Tuple>		tuple;
	typedef ref<List>		list;
	typedef ref<Dict>		dict;

	HINSTANCE	DllLoad(PCWSTR name) throw();

	PCSTR _(PCSTR name) throw();
	PCSTR _(PCSTR category, PCSTR name) throw();
	PCWSTR _W(PCSTR name) throw();
}

//========================================================================================================================
// basic string operations

#pragma pack(push, 8)

struct UNICODE_STRING
{
	USHORT	Length;
	USHORT	MaximumLength;
	PWSTR	Buffer;
};

#pragma pack(pop)

inline bool operator == (const UNICODE_STRING& lhs, const UNICODE_STRING& rhs) throw()	{ return wcscmp(lhs.Buffer, rhs.Buffer) == 0; }
inline bool operator == (const UNICODE_STRING& lhs, PCWSTR rhs) throw()					{ return wcscmp(lhs.Buffer, rhs) == 0; }
inline bool operator == (PCWSTR lhs, const UNICODE_STRING& rhs) throw()					{ return wcscmp(lhs, rhs.Buffer) == 0; }
inline bool operator <  (const UNICODE_STRING& lhs, const UNICODE_STRING& rhs) throw()	{ return wcscmp(lhs.Buffer, rhs.Buffer) < 0; }
inline bool operator <  (const UNICODE_STRING& lhs, PCWSTR rhs) throw()					{ return wcscmp(lhs.Buffer, rhs) < 0; }
inline bool operator <  (PCWSTR lhs, const UNICODE_STRING& rhs) throw()					{ return wcscmp(lhs, rhs.Buffer) < 0; }

namespace brownie
{
	inline bool empty(PCSTR s) throw()	{ return !s || s[0] == 0; }
	inline bool empty(PCWSTR s) throw()	{ return !s || s[0] == 0; }
	inline ssize_t len(PCSTR     o) throw()	{ return !o ? 0 : (ssize_t)strlen(o); }
	inline ssize_t len(PCWSTR    o) throw()	{ return !o ? 0 : (ssize_t)wcslen(o); }
	inline ssize_t len(const UNICODE_STRING& o) throw()	{ return o.Length; }
	inline ssize_t len(const UNICODE_STRING* o) throw()	{ return !o ? 0 : o->Length; }

	struct Less
	{
		template < typename L, typename R >
		bool operator () (const L& lhs, const R& rhs) const throw()	{ return lhs < rhs; }
		bool operator () (PCSTR  lhs, PCSTR  rhs) const throw()	{ return strcmp(lhs, rhs) < 0; }
		bool operator () (PCWSTR lhs, PCWSTR rhs) const throw()	{ return wcscmp(lhs, rhs) < 0; }
	};

	struct Greater
	{
		template < typename L, typename R >
		bool operator () (const L& lhs, const R& rhs) const throw()	{ return lhs > rhs; }
		bool operator () (PCSTR  lhs, PCSTR  rhs) const throw()	{ return strcmp(lhs, rhs) > 0; }
		bool operator () (PCWSTR lhs, PCWSTR rhs) const throw()	{ return wcscmp(lhs, rhs) > 0; }
	};
}
