// commctrl.hpp
#pragma once

#include "hwnd.hpp"

#define		__hwnd__	(static_cast<const T*>(this)->Handle)

namespace brownie
{
	template < template < typename > class Wrapper, typename Base = Hwnd >
	class WindowH : public Base, public Wrapper< WindowH<Wrapper, Base>  >
	{
	public:
		WindowH(HWND hwnd = null)	{ m_hwnd = hwnd; }
	};

	//==================================================================================================================================================================
	// Edit

	inline void Edit_GetSelection(HWND hwnd, int* beg, int* end) throw()	{ ::SendMessage(hwnd, EM_GETSEL, (WPARAM)beg, (LPARAM)end); }
	inline void Edit_SetSelection(HWND hwnd, int beg, int end) throw()		{ ::PostMessage(hwnd, EM_SETSEL, beg, end); }

	//==================================================================================================================================================================
	// ListView

	void	ListView_Adjust(HWND hwnd) throw();
	void	ListView_SelectItem_SVSI(HWND hwnd, int i, UINT svsi) throw();
	int		ListView_GetColumnCount(HWND hwnd) throw();
#undef ListView_GetItemText
	int		ListView_GetItemText(HWND hwnd, int row, int col, WCHAR buf[], size_t bufsize);
	HRESULT	ListView_GetSortedColumn(HWND hwnd, int* column, bool* ascending);
	HRESULT	ListView_SetSortedColumn(HWND hwnd, int column, bool ascending);
	bool	ListView_HitTestColumn(HWND hwnd, int column, int x);

	struct LVDRAWDESC
	{
		HIMAGELIST	ImageList;
		int			NameColumnRight;
		int			SortedColumnLeft;
		int			SortedColumnRight;
		int			SizeColumn;
		COLORREF	BackgroundColor[4];
		bool		HasFocus;
		bool		DragDrop;
		bool		DragDropNow;
	};
	void	ListView_BeginDraw(HWND hwnd, LVDRAWDESC& desc) throw();
	void	ListView_DrawItem(const DRAWITEMSTRUCT* draw, const LVDRAWDESC& desc) throw();

	template < typename T >
	class ListViewT
	{
	public:
		void	InsertColumn(const LVCOLUMN& col) throw()							{ ListView_InsertColumn(__hwnd__, col.iSubItem, &col); }

		int		GetItemCount() const throw()										{ return ListView_GetItemCount(__hwnd__); }
		int		GetSelectedCount() const throw()									{ return ListView_GetSelectedCount(__hwnd__); }
		int		GetNextItem(int start, UINT lvni) const throw()						{ return ListView_GetNextItem(__hwnd__, start, lvni); }
		BOOL	GetItemRect(int i, RECT* rc, int lvir) const throw()				{ return ListView_GetItemRect(__hwnd__, i, rc, lvir); }
		UINT	GetItemState(int i, UINT mask) throw()								{ return ListView_GetItemState(__hwnd__, i, mask); }
		void	SetItemState(int i, UINT data, UINT mask) throw()					{ ListView_SetItemState(__hwnd__, i, data, mask); }
		int		GetItemText(int i, int col, WCHAR buf[], size_t cch) const throw()	{ return ListView_GetItemText(__hwnd__, i, col, buf, cch); }
		int		GetHotItem() const throw()											{ return ListView_GetHotItem(__hwnd__); }
		void	DeleteAllItems() throw()											{ ListView_DeleteAllItems(__hwnd__); }
		int		InsertItem(const LVITEM& item) throw()								{ return ListView_InsertItem(__hwnd__, &item); }
		int		HitTest(LVHITTESTINFO* hit) const throw()							{ return ListView_HitTest(__hwnd__, hit); }
		int		HitTest(int x, int y) const throw()									{ LVHITTESTINFO hit = { x, y }; return HitTest(&hit); }
		bool	HitTestColumn(int column, int x) const throw()						{ return ListView_HitTestColumn(__hwnd__, column, x); }
		void	EnsureVisible(int i, bool partialOk = false) throw()				{ ListView_EnsureVisible(__hwnd__, i, partialOk); }
		HWND	GetEdit() const throw()												{ return ListView_GetEditControl(__hwnd__); }
		HWND	GetHeader() const throw()
		{
			if (HWND header = ListView_GetHeader(__hwnd__))
				if (IsVisible(header))
					return header;
			return null;
		}
		void	SelectItem(int i) throw()
		{
			SetItemState(-1, 0, LVIS_SELECTED);
			if (i >= 0)
			{
				SetItemState(i, LVIS_FOCUSED | LVIS_SELECTED, LVIS_FOCUSED | LVIS_SELECTED);
				EnsureVisible(i);
			}
		}
		void	SetExtendedStyle(DWORD mask, DWORD style) throw()					{ ListView_SetExtendedListViewStyleEx(__hwnd__, mask, style); }
		bool	IsGroupViewEnabled() const throw()									{ return ListView_IsGroupViewEnabled(__hwnd__) != 0; }
		void	EnableGroupView(bool value) throw()									{ ListView_EnableGroupView(__hwnd__, value); }
		HRESULT	GetSortedColumn(int* column, bool* ascending) const throw()			{ return ListView_GetSortedColumn(__hwnd__, column, ascending); }
		HRESULT	SetSortedColumn(int column, bool ascending) throw()					{ return ListView_SetSortedColumn(__hwnd__, column, ascending); }
	};

	typedef WindowH<ListViewT> ListViewH;

	//==================================================================================================================================================================
	// TreeView

	void*		TreeView_GetItemData(HWND hwnd, HTREEITEM item) throw();
	HTREEITEM	TreeView_Insert(HWND hwnd, HTREEITEM parent, const TVITEM& item);
	HTREEITEM	TreeView_Insert(HWND hwnd, HTREEITEM parent, void* data, PCWSTR name, int icon = I_IMAGENONE) throw();
	HTREEITEM	TreeView_SelectItemByPos(HWND hwnd, int x, int y);

	template < typename T >
	class TreeViewT
	{
	public:
		void*		GetItemData(HTREEITEM item) const throw()												{ return TreeView_GetItemData(__hwnd__, item); }
		BOOL		GetItemRect(HTREEITEM item, RECT* rc, BOOL textonly = TRUE) const throw()				{ return TreeView_GetItemRect(__hwnd__, item, rc, textonly); }
		HTREEITEM	GetNextItem(HTREEITEM item, UINT flags) const throw()									{ return TreeView_GetNextItem(__hwnd__, item, flags); }
		HTREEITEM	GetChildItem(HTREEITEM item) const throw()												{ return TreeView_GetChild(__hwnd__, item); }
		void		SetItem(const TVITEM& item) throw()														{ VERIFY( TreeView_SetItem(__hwnd__, &item) ); }
		HTREEITEM	GetSelection() const throw()															{ return TreeView_GetSelection(__hwnd__); }
		HTREEITEM	GetDropHilight() const throw()															{ return TreeView_GetDropHilight(__hwnd__); }
		HTREEITEM	HitTest(TVHITTESTINFO* hit) const throw()												{ return TreeView_HitTest(__hwnd__, hit); }
		HTREEITEM	InsertItem(HTREEITEM parent, const TVITEM& item) throw()								{ return TreeView_Insert(__hwnd__, parent, item); }
		HTREEITEM	InsertItem(HTREEITEM parent, void* data, PCWSTR name, int icon = I_IMAGENONE) throw()	{ return TreeView_Insert(__hwnd__, parent, data, name, icon); }
		void		DeleteAllItems() throw()																{ TreeView_DeleteAllItems(__hwnd__); }
		void		Expand(HTREEITEM item, UINT tve) throw()												{ VERIFY( TreeView_Expand(__hwnd__, item, tve) ); }
		void		SelectItem(HTREEITEM item) throw()														{ VERIFY( TreeView_SelectItem(__hwnd__, item) ); }
		HTREEITEM	SelectItem(int x, int y) throw()														{ return TreeView_SelectItemByPos(__hwnd__, x, y); }
		HWND		GetEdit() const throw()																	{ return TreeView_GetEditControl(__hwnd__); }
		HIMAGELIST	SetImageList(HIMAGELIST image, INT what = TVSIL_NORMAL) throw()							{ return TreeView_SetImageList(__hwnd__, image, what); }
		HWND		EditLabel(HTREEITEM item) throw()														{ return TreeView_EditLabel(__hwnd__, item); }
		void		EndEditLabelNow(bool cancel) throw()													{ TreeView_EndEditLabelNow(__hwnd__, cancel); }
	};

	typedef WindowH<TreeViewT> TreeViewH;

	//==================================================================================================================================================================
	// ToolBar

	const int TOOLBAR_INDENT = 2;	// FIXME: magic number for XP
	const int TOOLBAR_MARGIN = 8;	// FIXME: magic number for XP

	int		ToolBar_GetItemCount(HWND hwnd) throw();
	UINT	ToolBar_GetItemState(HWND hwnd, int nID) throw();
	bool	ToolBar_SetItemState(HWND hwnd, int nID, UINT state) throw();
	int		ToolBar_GetHotItem(HWND hwnd) throw();
	int		ToolBar_SetHotItem(HWND hwnd, int index) throw();
	bool	ToolBar_GetItemRect(HWND hwnd, int i, RECT* rc) throw();
	void	ToolBar_InsertItem(HWND hwnd, int image, BYTE style, BYTE state, INT_PTR text = -1) throw();
	void	ToolBar_InsertItems(HWND hwnd, size_t count, const TBBUTTON buttons[]) throw();
	SIZE	ToolBar_GetItemSize(HWND hwnd) throw();
	SIZE	ToolBar_GetDefaultSize(HWND hwnd) throw();
	void	ToolBar_AutoSize(HWND hwnd) throw();
	int		ToolBar_HitTest(HWND hwnd, POINT pt) throw();
	int		ToolBar_HitTest(HWND hwnd, int x, int y) throw();

	template < typename T >
	class ToolBarT
	{
	public:
		int		GetItemCount() const throw()												{ return ToolBar_GetItemCount(__hwnd__); }
		UINT	GetItemState(int nID) const throw()											{ return ToolBar_GetItemState(__hwnd__, nID); }
		bool	SetItemState(int nID, UINT state) throw()									{ return ToolBar_SetItemState(__hwnd__, nID, state); }
		int		GetHotItem() const throw()													{ return ToolBar_GetHotItem(__hwnd__); }
		int		SetHotItem(int index) throw()												{ return ToolBar_SetHotItem(__hwnd__, index); }
		bool	GetItemRect(int i, RECT* rc) const throw()									{ return ToolBar_GetItemRect(__hwnd__, i, rc); }
		void	InsertItem(int image, BYTE style, BYTE state, INT_PTR text = -1) throw()	{ ToolBar_InsertItem(__hwnd__, image, style, state, text); }
		void	InsertItems(size_t count, const TBBUTTON buttons[]) throw()					{ ToolBar_InsertItems(__hwnd__, count, buttons); }
		SIZE	GetItemSize() const throw()													{ return ToolBar_GetItemSize(__hwnd__); }
		SIZE	GetDefaultSize() const throw()												{ return ToolBar_GetDefaultSize(__hwnd__); }
		void	AutoSize() throw()															{ ToolBar_AutoSize(__hwnd__); }
		int		HitTest(POINT pt) const throw()												{ return ToolBar_HitTest(__hwnd__, pt); }
		int		HitTest(int x, int y) const throw()											{ return ToolBar_HitTest(__hwnd__, x, y); }
	};

	typedef WindowH<ToolBarT> ToolBarH;

	//==================================================================================================================================================================
	// ReBar

	struct ReBar_Band : REBARBANDINFO
	{
		ReBar_Band(DWORD rbbim);

		__declspec(property(get=get_Visible, put=set_Visible)) bool Visible;
		bool get_Visible() const		{ return (fStyle & RBBS_HIDDEN) == 0; }
		void set_Visible(bool value)
		{
			if (value)
				fStyle &= ~RBBS_HIDDEN;
			else
				fStyle |= RBBS_HIDDEN;
		}

		__declspec(property(get=get_Break)) bool Break;
		bool get_Break() const		{ return (fStyle & RBBS_BREAK) != 0; }

		__declspec(property(get=get_Locked, put=set_Locked)) bool Locked;
		bool get_Locked() const		{ return (fStyle & RBBS_NOGRIPPER) != 0; }
		void set_Locked(bool value)
		{
			if (value)
			{	// locked, without grip
				fStyle &= ~RBBS_GRIPPERALWAYS;
				fStyle |= RBBS_NOGRIPPER;
			}
			else
			{	// unlocked, with grip
				fStyle |= RBBS_GRIPPERALWAYS;
				fStyle &= ~RBBS_NOGRIPPER;
			}
		}
	};

	BOOL	ReBar_GetBarInfo(HWND hwnd, REBARINFO* bar);
	void	ReBar_SetBarInfo(HWND hwnd, const REBARINFO* bar);
	void	ReBar_ShowBand(HWND hwnd, int i, bool show);
	int		ReBar_GetBandCount(HWND hwnd);
	bool	ReBar_GetBandInfo(HWND hwnd, int i, REBARBANDINFO* band);
	void	ReBar_SetBandInfo(HWND hwnd, int i, const REBARBANDINFO* band);
	int		ReBar_FindBand(HWND hwnd, HWND target);
	void	ReBar_InsertBand(HWND hwnd, const REBARBANDINFO* band, int i = -1);
	void	ReBar_InsertBand(HWND hwnd, HWND child, SIZE size, int i = -1);
	bool	ReBar_DeleteBand(HWND hwnd, int i);
	bool	ReBar_DeleteBand(HWND hwnd, HWND child);
	bool	ReBar_SyncBands(HWND hwnd);
	bool	ReBar_GetLocked(HWND hwnd);
	void	ReBar_SetLocked(HWND hwnd, bool value);

	template < typename T >
	class ReBarT
	{
	public:
		BOOL	GetBarInfo(REBARINFO* bar) const					{ return ReBar_GetBarInfo(__hwnd__, bar); }
		void	SetBarInfo(const REBARINFO* bar)					{ ReBar_SetBarInfo(__hwnd__, bar); }
		void	ShowBand(int i, bool show)							{ ReBar_ShowBand(__hwnd__, i, show); }
		int		GetBandCount() const								{ return ReBar_GetBandCount(__hwnd__); }
		bool	GetBandInfo(int i, REBARBANDINFO* band) const		{ return ReBar_GetBandInfo(__hwnd__, i, band); }
		void	SetBandInfo(int i, const REBARBANDINFO* band)		{ ReBar_SetBandInfo(__hwnd__, i, band); }
		int		FindBand(HWND target) const							{ return ReBar_FindBand(__hwnd__, target); }
		void	InsertBand(const REBARBANDINFO* band, int i = -1)	{ ReBar_InsertBand(__hwnd__, band, i); }
		void	InsertBand(HWND child, SIZE size, int i = -1)		{ ReBar_InsertBand(__hwnd__, child, size, i); }
		bool	DeleteBand(int i)									{ return ReBar_DeleteBand(__hwnd__, i); }
		bool	DeleteBand(HWND child)								{ return ReBar_DeleteBand(__hwnd__, child); }
		bool	SyncBands()											{ return ReBar_SyncBands(__hwnd__); }
		bool	GetLocked() const									{ return ReBar_GetLocked(__hwnd__); }
		void	SetLocked(bool value)								{ ReBar_SetLocked(__hwnd__, value); }
	};

	typedef WindowH<ReBarT> ReBarH;

	//==================================================================================================================================================================
	// StatusBar

	bool	StatusBar_GetItemRect(HWND hwnd, int part, RECT* rc);
	int		StatusBar_HitTest(HWND hwnd, int x) throw();

	template < typename T >
	class StatusBarT
	{
	public:
		int		GetParts() const throw()							{ return (int)::SendMessage(__hwnd__, SB_GETPARTS, 0, 0); }
		int		GetParts(size_t ln, int parts[]) const throw()		{ return (int)::SendMessage(__hwnd__, SB_GETPARTS, ln, (LPARAM)parts); }
		void	SetParts(size_t ln, const int parts[]) throw()		{ VERIFY( ::SendMessage(__hwnd__, SB_SETPARTS, ln, (LPARAM)parts) ); }
		ssize_t GetItemTextLength(int part) const throw()			{ return (ssize_t)::SendMessage(__hwnd__, SB_GETTEXTLENGTH, part, 0); }
		DWORD	GetItemText(int part, WCHAR text[]) const throw()	{ return (DWORD)::SendMessage(__hwnd__, SB_GETTEXT, part, (LPARAM)text); }
		HICON	GetItemIcon(int part) const throw()					{ return (HICON)::SendMessage(__hwnd__, SB_GETICON, part, 0); }
		bool	GetItemRect(int part, RECT* rc) const throw()		{ return StatusBar_GetItemRect(__hwnd__, part, rc); }
		void	GetItemTip(int part, WCHAR text[]) const throw()	{ ::SendMessage(__hwnd__, SB_GETTIPTEXT, part, (LPARAM)text); }
		void	SetItemText(int part, PCWSTR text) throw()			{ VERIFY( ::SendMessage(__hwnd__, SB_SETTEXT, part, (LPARAM)text) ); }
		void	SetItemIcon(int part, HICON icon) throw()			{ VERIFY( ::SendMessage(__hwnd__, SB_SETICON, part, (LPARAM)icon) ); }
		void	SetItemTip(int part, PCWSTR text) throw()			{ ::SendMessage(__hwnd__, SB_SETTIPTEXT, part, (LPARAM)text); }
		int		HitTest(int x) const throw()						{ return StatusBar_HitTest(__hwnd__, x); }
	};

	typedef WindowH<StatusBarT> StatusBarH;

	//==================================================================================================================================================================
	// ToolTip

	int		ToolTip_GetCount(HWND hwnd) throw();
	void	ToolTip_Add(HWND hwnd, const TOOLINFO* info) throw();
	void	ToolTip_Add(HWND hwnd, HWND owner, UINT nID, PCWSTR text, const RECT& rc) throw();
	void	ToolTip_Remove(HWND hwnd, HWND owner, UINT nID) throw();

	template < typename T >
	class ToolTipT
	{
	public:
		int		GetCount() const throw()										{ return ToolTip_GetCount(__hwnd__); }
		void	Add(const TOOLINFO* info) throw()								{ ToolTip_Add(__hwnd__, info); }
		void	Add(HWND owner, UINT nID, PCWSTR text, const RECT& rc) throw()	{ ToolTip_Add(__hwnd__, owner, nID, text, rc); }
		void	Remove(HWND owner, UINT nID) throw()							{ ToolTip_Remove(__hwnd__, owner, nID); }
	};

	typedef WindowH<ToolTipT> ToolTipH;

	//==================================================================================================================================================================
	// Header

	int		Header_HitTest(HWND hwnd, int x, int y) throw();
	HRESULT	Header_GetSortedColumn(HWND hwnd, int* column, bool* ascending) throw();
	HRESULT	Header_SetSortedColumn(HWND hwnd, int column, bool ascending) throw();

	template < typename T >
	class HeaderT
	{
	public:
		int		GetItemCount() const throw()										{ return Header_GetItemCount(__hwnd__); }
		BOOL	GetItemRect(int index, RECT* rc) const throw()						{ return Header_GetItemRect(__hwnd__, index, rc); }
		BOOL	GetItem(int index, HDITEM* item) const throw()						{ return Header_GetItem(__hwnd__, index, item); }
		int		InsertItem(const HDITEM& item, int index = INT_MAX) const throw()	{ return Header_InsertItem(__hwnd__, index, &item); }
		BOOL	DeleteItem(int index) throw()										{ return Header_DeleteItem(__hwnd__, index); }
		void	DeleteAllItems() throw()											{ while (DeleteItem(0)) {} }
		int		HitTest(int x, int y) const throw()									{ return Header_HitTest(__hwnd__, x, y); }
		HRESULT	GetSortedColumn(int* column, bool* ascending) const throw()			{ return Header_GetSortedColumn(__hwnd__, column, ascending); }
		HRESULT	SetSortedColumn(int column, bool ascending) throw()					{ return Header_SetSortedColumn(__hwnd__, column, ascending); }
		HIMAGELIST	GetImageList() const throw()									{ return Header_GetImageList(__hwnd__); }
		HIMAGELIST	SetImageList(HIMAGELIST image) throw()							{ return Header_SetImageList(__hwnd__, image); }
	};

	typedef WindowH<HeaderT> HeaderH;
}

#undef __hwnd__
