// config.hpp
#pragma once

#include "brownie.hpp"
#undef ERROR

//==================================================================================================================================================================
// ENUMERATIONS

namespace brownie
{
	enum DIRECTION
	{
		NORTH,
		SOUTH,
		WEST,
		EAST,
		CENTER,
	};

	enum STATUS
	{
		NORMAL	 = (1 << 0),
		FOCUSED	 = (1 << 1),
		SELECTED = (1 << 2),
		CHECKED	 = (1 << 3),
		HOT		 = (1 << 4),
		DISABLED = (1 << 5),
	};

	enum EXITCODE
	{
		EXIT_ERROR_BUSY	= -3,
		EXIT_ERROR_SEND	= -2,
		EXIT_ERROR_LOAD	= -1,
		EXIT_OK			= 0,
		EXIT_SHUTDOWN	= 1,
		EXIT_RESTART	= 2,
		EXIT_UPDATE		= 3,
	};
}

//==================================================================================================================================================================
// INTERGER CONSTANTS

const DWORD INVALID_TLS = (DWORD)-1;
const DWORD	COPYDATA_MAGIC = 'TEXT';

#define VK_BUTTON_L			VK_LBUTTON	// 0x01 
#define VK_BUTTON_R			VK_RBUTTON	// 0x02
#define VK_BUTTON_M			VK_MBUTTON	// 0x04
#define VK_BUTTON_X			VK_XBUTTON1	// 0x05
#define VK_BUTTON_Y			VK_XBUTTON2	// 0x06
#define VK_DBLCLK(n)		(n + 0x96)
#define VK_BUTTON_LL		VK_DBLCLK(VK_BUTTON_L)
#define VK_BUTTON_RR		VK_DBLCLK(VK_BUTTON_R)
#define VK_BUTTON_MM		VK_DBLCLK(VK_BUTTON_M)
#define VK_BUTTON_XX		VK_DBLCLK(VK_BUTTON_X)
#define VK_BUTTON_YY		VK_DBLCLK(VK_BUTTON_Y)
#define VK_WHEEL_LEFT		0x88
#define VK_WHEEL_UP			0x89
#define VK_WHEEL_RIGHT		0x8A
#define VK_WHEEL_DOWN		0x8B

/*
#define MK_LBUTTON          0x0001
#define MK_RBUTTON          0x0002
#define MK_SHIFT            0x0004
#define MK_CONTROL          0x0008
#define MK_MBUTTON          0x0010
#define MK_XBUTTON1         0x0020
#define MK_XBUTTON2         0x0040
*/
#define MK_ALTKEY		0x0080	// MK_ALT ͎gĂ͂ȂȂI
#define MK_WINKEY		0x0100		
#define MK_KEYS			(MK_SHIFT | MK_CONTROL | MK_ALTKEY | MK_WINKEY)
#define MK_BUTTONS		(MK_LBUTTON | MK_RBUTTON | MK_MBUTTON | MK_XBUTTON1 | MK_XBUTTON2)

// NOTIFICATION LEVEL
const int LOG		= 0x10000000L;
const int INFO		= MB_ICONINFORMATION;
const int WARNING	= MB_ICONWARNING;
const int ERROR		= MB_ICONERROR;

//==================================================================================================================================================================
// SPECIAL FOLDERS

#define FOLDER_DRIVE		L"::{20D04FE0-3AEA-1069-A2D8-08002B30309D}"	// My Computer
#define FOLDER_HOME			L"::{ECF03A32-103D-11d2-854D-006008059367}"	// 
#define FOLDER_DOCUMENT		L"::{450D8FBA-AD25-11D0-98A8-0800361B1103}"	// My Documents
#define FOLDER_NETWORK		L"::{208D2C60-3AEA-1069-A2D7-08002B30309D}"
#define FOLDER_INTERNET		L"::{871C5380-42A0-1069-A2EA-08002B30309D}"
#define FOLDER_TRASH		L"::{645FF040-5081-101B-9F08-00AA002F954E}"
#define FOLDER_CONTROL		L"::{21EC2020-3AEA-1069-A2DD-08002B30309D}"	// Control Panel
#define FOLDER_PRINTER		L"::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\\::{2227A280-3AEA-1069-A2DE-08002B30309D}"
#define FOLDER_WEB			L"::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\\::{BDEADF00-C265-11d0-BCED-00A0C90AB50F}"	// Web Folder
#define FOLDER_DIALUP		L"::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\\::{992CFFA0-F557-101A-88EC-00DD010CCC48}"
#define FOLDER_TASK			L"::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\\::{D6277990-4C6A-11CF-8D87-00AA0060F5BF}"
#define FOLDER_CURRENT		L"."

const int FOLDER_PARENT		= 0;
const int FOLDER_BACK		= -1;
const int FOLDER_FORWARD	= +1;

//==================================================================================================================================================================
// STRING CONSTANTS

#define STR_ERR_PATHCHAR	L"\\/:*?\"<>|"

//==================================================================================================================================================================
// Windows message macros

#ifndef GET_X_LPARAM
#	define GET_X_LPARAM(lParam)		((INT)(INT16)LOWORD(lParam))
#	define GET_Y_LPARAM(lParam)		((INT)(INT16)HIWORD(lParam))
#endif
#ifndef GET_XY_LPARAM
#	define GET_XY_LPARAM(lParam)	GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam)
#endif

#define GET_WHEEL(wParam, up, down)	((GET_WHEEL_DELTA_WPARAM(wParam) > 0) ? (up) : (down))
#define GET_XBUTTON(wParam, x1, x2)	((GET_XBUTTON_WPARAM(wParam) == XBUTTON1) ? (x1) : (x2))

inline INT MSG2VK(UINT msg, WPARAM wParam) throw()
{
	switch (msg)
	{
	case WM_LBUTTONDOWN: case WM_LBUTTONDBLCLK:	case WM_LBUTTONUP:
		return VK_BUTTON_L;
	case WM_RBUTTONDOWN: case WM_RBUTTONDBLCLK:	case WM_RBUTTONUP:
		return VK_BUTTON_R;
	case WM_MBUTTONDOWN: case WM_MBUTTONDBLCLK:	case WM_MBUTTONUP:
		return VK_BUTTON_M;
	case WM_XBUTTONDOWN: case WM_XBUTTONDBLCLK:	case WM_XBUTTONUP:
		return GET_XBUTTON(wParam, VK_BUTTON_X, VK_BUTTON_Y);
	case WM_MOUSEWHEEL:
		return GET_WHEEL(wParam, VK_WHEEL_UP, VK_WHEEL_DOWN);
	case WM_MOUSEHWHEEL:
		return GET_WHEEL(wParam, VK_WHEEL_LEFT, VK_WHEEL_RIGHT);
	default:
		return 0;
	}
}

inline INT MSG2VK(const MSG& msg) throw()
{
	return MSG2VK(msg.message, msg.lParam);
}

inline INT MSG2MK(UINT msg, WPARAM wParam) throw()
{
	switch (msg)
	{
	case WM_LBUTTONDOWN: case WM_LBUTTONDBLCLK: case WM_LBUTTONUP:
		return MK_LBUTTON;
	case WM_RBUTTONDOWN: case WM_RBUTTONDBLCLK: case WM_RBUTTONUP:
		return MK_RBUTTON;
	case WM_MBUTTONDOWN: case WM_MBUTTONDBLCLK: case WM_MBUTTONUP:
		return MK_MBUTTON;
	case WM_XBUTTONDOWN: case WM_XBUTTONDBLCLK: case WM_XBUTTONUP:
		return GET_XBUTTON(wParam, MK_XBUTTON1, MK_XBUTTON2);
	default:
		return 0;
	}
}

inline INT MSG2MK(const MSG& msg) throw()
{
	return MSG2MK(msg.message, msg.wParam);
}

//==================================================================================================================================================================
// config

#define MAKE_VERSION(a, b, c, d)	(((a) << 24) + ((b) << 16) + ((c) << 8) + (d))

namespace brownie
{
	class Module;

	const DWORD	VERSION				= MAKE_VERSION(0, 0, 0, 16);
	const DWORD PROTOCOL_VERSION	= 0x0001;
	const WCHAR	APPNAME[]			= L"brownie";

	extern int			WINDOWS_VERSION;
	extern WCHAR		PATH_ROOT[MAX_PATH];
	extern WCHAR		PATH_BIN[MAX_PATH];
	extern WCHAR		PATH_USER[MAX_PATH];
	extern WCHAR		PATH_FAVORITE[MAX_PATH];
	extern WCHAR		PATH_TOOLS[MAX_PATH];
	extern WCHAR		PATH_FOLDER_STATE[MAX_PATH];
	extern WCHAR		PATH_SOUND_NAVIGATION[MAX_PATH];
	extern int			GESTURE_NAPTIME;
	extern int			GESTURE_TIMEOUT;
	extern int			GESTURE_MAXLEN;
	extern PyObject*	EDITOR;
	extern int			LISTVIEW_AUTOADJUST;
	extern bool			LISTVIEW_AUTOUNIT;
	extern COLORREF		LISTVIEW_BACKGROUND;
	extern bool			LISTVIEW_LOOP;
	extern bool			LISTVIEW_WHEELSELECT;
	extern bool			RENAME_EXTENSION;
	extern int			TAB_MAXLEN;

	void ConfigInit() throw();
	void config_export(Module *m);

	const int	WINDOWS_2000	= 0x05000000;
	const int	WINDOWS_XP		= 0x05010000;
	const int	WINDOWS_2003	= 0x05020000;
	const int	WINDOWS_VISTA	= 0x06000000;

	const int	DIALOG_MARGIN = 8;
}
