// cast.inl
#pragma once

namespace brownie
{
	namespace detail
	{
		Object* ObjectFrom(PyObject* o, PyTypeObject* type) throw(...);

		template < typename T > struct caster
		{
			typedef T* result;
			static result convert(PyObject* o) throw(...)
			{
				STATIC_ASSERT( offsetof(detail::ObjectImpl<T>, __self__) == sizeof(PyObject) );
				return static_cast<T*>(ObjectFrom(o, typeof<T>()));
			}
		};
	}

	//==================================================================================================================================================================

	template < typename T, typename U >
	inline typename detail::caster<T>::result cast(const U& o) throw(...)
	{
		return detail::caster<T>::convert(o);
	}

	template < typename T, typename U >
	inline typename detail::caster<T>::result cast(const ref<U>& o) throw(...)
	{
		return detail::caster<T>::convert(static_cast<U*>(o));
	}

	template < typename T > inline typename detail::caster<T>::result getattr(PyObject* self, PCSTR name) throw(...)	{ return cast<T>(getattr(self, name)); }
	template < typename T > inline typename detail::caster<T>::result getitem(PyObject* self, ssize_t index) throw(...)	{ return cast<T>(getitem(self, index)); }
	template < typename T > inline typename detail::caster<T>::result getitem(Array* self, ssize_t index) throw(...)	{ return cast<T>(getitem(self, index)); }
	template < typename T > inline typename detail::caster<T>::result getitem(Tuple* self, ssize_t index) throw(...)	{ return cast<T>(getitem(self, index)); }
	template < typename T > inline typename detail::caster<T>::result getitem(List* self, ssize_t index) throw(...)		{ return cast<T>(getitem(self, index)); }
	template < typename T > inline typename detail::caster<T>::result getitem(Dict* self, PCSTR name) throw(...)		{ return cast<T>(getitem(self, name)); }

	//==================================================================================================================================================================

	namespace detail
	{
		template < typename T > struct caster< ref<T> >
		{
			typedef ref<T> result;
			template < typename U >
			static result convert(const U& o) throw(...)	{ return T::newobj(o); }
		};

		template <> struct caster<PyObject>
		{
			typedef PyObject* result;
			static result convert(Object* o) throw(...)		{ return o ? o->self : Py_None; }
			static result convert(bool o) throw(...)		{ return o ? Py_True : Py_False; }
		};

		template <> struct caster<bool>
		{
			typedef bool result;
			static result convert(PyObject* o) throw(...)	{ return Boolean::value(o); }
		};

		template < typename T > struct caster_integer
		{
			typedef T result;
			static result convert(Integer* o) throw()		{ return (T)Integer::value(o); }
			static result convert(PyObject* o) throw(...)	{ return (T)Integer::value(o); }
			static result convert(PCSTR o) throw()			{ return (T)atoi(o); }
			static result convert(PCWSTR o) throw()			{ return (T)_wtoi(o); }
		};

		template <> struct caster<INT8>   : caster_integer<INT8>	{};
		template <> struct caster<INT16>  : caster_integer<INT16>	{};
		template <> struct caster<INT32>  : caster_integer<INT32>	{};
		template <> struct caster<UINT8>  : caster_integer<UINT8>	{};
		template <> struct caster<UINT16> : caster_integer<UINT16>	{};
		template <> struct caster<UINT32> : caster_integer<UINT32>	{};
		template <> struct caster<DWORD>  : caster_integer<DWORD>	{};

		template < typename T > struct caster_numeric
		{
			typedef T result;
			static result convert(Integer* o) throw()		{ return (T)Integer::value(o); }
			static result convert(Numeric* o) throw()		{ return (T)Numeric::value(o); }
			static result convert(PyObject* o) throw(...)	{ return (T)Numeric::value(o); }
			static result convert(PCSTR o) throw()			{ return (T)_atoi64(o); }
			static result convert(PCWSTR o) throw()			{ return (T)_wtoi64(o); }
		};

		template <> struct caster<INT64>  : caster_integer<INT64>	{};
		template <> struct caster<UINT64> : caster_integer<UINT64>	{};

		template < typename T, typename U = int > struct caster_two
		{
			typedef T result;
			static result convert(PyObject* o) throw(...)
			{
				T r = { getitem<U>(o, 0), getitem<U>(o, 1) };
				return r;
			}
		};

		template <> struct caster<POINT> : caster_two<POINT>	{};
		template <> struct caster<SIZE>  : caster_two<SIZE>		{};
	}
}
