// class.inl
#pragma once

#include "exception.hpp"

//============================================================================

#define PYTHON_INTERNAL_DEF(name)	name
#define PYTHON_INTERNAL_VAR(name)	m_##name
#define PYTHON_INTERNAL_GET(name)	get_##name
#define PYTHON_INTERNAL_SET(name)	set_##name
#define PYTHON_INTERNAL_DOC(name)	((PSTR)_(Self::__name__(), #name))

//============================================================================

#define PYTHON_END()								\
			{ null }								\
		};											\
		return defs;								\
	}

//============================================================================
// method

#define PYTHON_METHODS(class)						\
	PyMethodDef* class::__method__() throw()		\
	{												\
		typedef class Self;							\
		static PyMethodDef defs[] =					\
		{

#define PYTHON_METHOD_(name, iname)		{ #name, PYTHON_METHOD_PROC(iname), PYTHON_METHOD_FLAG(iname), PYTHON_INTERNAL_DOC(name) },
#define PYTHON_METHOD(name)				PYTHON_METHOD_(name, name)

#define PYTHON_METHOD_THUNK(iname)		(sizeof(detail::MethodOf(&Self::iname))-1)
#define PYTHON_METHOD_FLAG(iname)		(PYTHON_METHOD_THUNK(iname) & ~METH_NORESULT)
#define PYTHON_METHOD_PROC(iname)		(PyCFunction)(detail::Method<PYTHON_METHOD_THUNK(iname)>::Fn<Self, &Self::iname>::Do)

//============================================================================
// member
//
// object^ϐ̏ꍇAPyGetSetDef.closure ɃItZbgBȂ 0B
// GET  : get=get_member()
// SET  : get=m_member     / set=set_member()
// BOTH : get=get_member() / set=set_member()
// VAR  : get=m_member     / set=m_member

#define PYTHON_MEMBERS(class)						\
	PyGetSetDef* class::__member__() throw()		\
	{												\
		typedef class Self;							\
		static PyGetSetDef defs[] =					\
		{

#define PYTHON_MEMBER(name, mode)					\
			{ #name, PYTHON_GETTER_##mode(name), PYTHON_SETTER_##mode(name), PYTHON_INTERNAL_DOC(name), PYTHON_OFFSET_##mode(name) },

#define PYTHON_GETTER_GET(name)		brownie::detail::Get<Self, &Self::PYTHON_INTERNAL_GET(name)>::Do
#define PYTHON_GETTER_SET(name)		PYTHON_GETTER_VAR(name)
#define PYTHON_GETTER_BOTH(name)	PYTHON_GETTER_GET(name)
#define PYTHON_GETTER_VAR(name)		PYTHON_VAR_CLASS(name)::Get

#define PYTHON_SETTER_GET(name)		(0)
#define PYTHON_SETTER_SET(name)		brownie::detail::Set<Self, &Self::PYTHON_INTERNAL_SET(name)>::Do
#define PYTHON_SETTER_BOTH(name)	PYTHON_SETTER_SET(name)
#define PYTHON_SETTER_VAR(name)		PYTHON_VAR_CLASS(name)::Set

#define PYTHON_OFFSET_GET(name)		(0)
#define PYTHON_OFFSET_SET(name)		PYTHON_OFFSET_VAR(name)
#define PYTHON_OFFSET_BOTH(name)	(0)
#define PYTHON_OFFSET_VAR(name)		((void*)(offsetof(Self, PYTHON_INTERNAL_VAR(name)) + sizeof(PyObject)))

#define PYTHON_VAR_CLASS(name)		brownie::detail::Var<Self, &Self::PYTHON_INTERNAL_VAR(name)>

//============================================================================

namespace brownie
{
	namespace detail
	{
		//==================================================================================================================================================================

		template < typename T, typename object (T::*fn)() >
		inline PyObject* __proxy1__(PyObject* self) throw()
		{
			try
			{
				return (cast<T>(self)->*fn)().detach();
			}
			catch (Exception&) { return null; }
		}

		template < typename T, typename object (T::*fn)(Tuple*, Dict*) >
		inline PyObject* __proxy3__(PyObject* self, PyObject* args, PyObject* kw) throw()
		{
			try
			{
				if (PyObject* o = (cast<T>(self)->*fn)(static_cast<Tuple*>(args), static_cast<Dict*>(kw)).detach())
					return o;
				else
					Py_RETURN_NONE;
			}
			catch (Exception&) { return null; }
		}

		//==================================================================================================================================================================

		template < typename T >
		class ObjectImpl
		{
		public:
			PyObject_HEAD
			T	__self__;
		public:
			static PyTypeObject* __type__() throw()
			{
				static PyTypeObject type =
				{
					PyVarObject_HEAD_INIT(0, 0)
					T::__name__(),			// tp_name
					sizeof(ObjectImpl<T>),	// tp_basicsize
					0,						// tp_itemsize
					__dealloc__,			// tp_dealloc
					0,						// tp_print
					__tp_getattr__(),		// tp_getattr
					__tp_setattr__(),		// tp_setattr
					0,						// tp_compare
					0,						// tp_repr
					T::__number__(),		// tp_as_number
					T::__sequence__(),		// tp_as_sequence
					T::__mapping__(),		// tp_as_mapping
					0,						// tp_hash 
					__tp_call__(),			// tp_call
					__tp_str__(),			// tp_str
					0,						// tp_getattro
					0,						// tp_setattro
					0,						// tp_as_buffer
					__tp_flags__(),			// tp_flags
					_(T::__name__()),		// tp_doc 
					py::GenericTraverse,	// tp_traverse
					py::GenericClear,		// tp_clear
					0,						// tp_richcompare
					0,						// tp_weaklistoffset
					__tp_iter__(),			// tp_iter
					__tp_iternext__(),		// tp_iternext
					T::__method__(),		// tp_methods
					0,						// tp_members
					T::__member__(),		// tp_getset
					typeof<T::__base__>(),	// tp_base
					0,						// tp_dict
					0,						// tp_descr_get
					0,						// tp_descr_set
					0,						// tp_dictoffset
					__init__,				// tp_init
					0,						// tp_alloc
					__new__,				// tp_new
					0,						// tp_free
					0,						// tp_is_gc
					0,						// tp_bases
					0,						// tp_mro -- method resolution order 
					0,						// tp_cache
					0,						// tp_subclasses
					0,						// tp_weaklist
					0,						// tp_del
				};
				return &type;
			}

		private:
			static PyObject* __new__(PyTypeObject *type, PyObject* args, PyObject* kw) throw()
			{
				PyObject* self = py::GenericNew(type);
				new(&reinterpret_cast<ObjectImpl*>(self)->__self__) T();
				return self;
			}

			static void __dealloc__(PyObject* self) throw()
			{
				py::GenericDelete(self);
				reinterpret_cast<ObjectImpl*>(self)->__self__.~T();
				self->ob_type->tp_free(self);
			}

			static int __init__(PyObject* self, PyObject* args, PyObject* kw) throw()
			{
				// __init__ z֐ɂ cast<T> ͕svB
				// ARpCɂׂĉł̂ŁALł͂ȂB
				try
				{
					cast<T>(self)->__init__(static_cast<Tuple*>(args), static_cast<Dict*>(kw));
					return 0;
				}
				catch (Exception&)
				{
					return -1;
				}
			}

			static reprfunc __tp_str__() throw()
			{
				return py::GenericStr(T::__method__());
			}

			static ternaryfunc __tp_call__() throw()
			{
				__if_exists(T::__call__)		{ return __proxy3__<T, &T::__call__>; }
				__if_not_exists(T::__call__)	{ return null; }
			}

			static int __tp_flags__() throw()
			{
				return py::GenericTypeFlags(T::__member__(), typeof<T::__base__>());
			}

			static getiterfunc __tp_iter__() throw()
			{
				__if_exists(T::__iter__)		{ return __proxy1__<T, &T::__iter__>; }
				__if_not_exists(T::__iter__)	{ return null; }
			}

			static iternextfunc __tp_iternext__() throw()
			{
				__if_exists(T::__next__)		{ return __proxy1__<T, &T::__next__>; }
				__if_not_exists(T::__next__)	{ return null; }
			}

			static getattrfunc	__tp_getattr__() throw()
			{
				__if_exists(T::__getattr__)		{ return &T::__getattr__; }
				__if_not_exists(T::__getattr__)	{ return null; }
			}

			static setattrfunc	__tp_setattr__() throw()
			{
				__if_exists(T::__setattr__)		{ return &T::__setattr__; }
				__if_not_exists(T::__setattr__)	{ return null; }
			}
		};
	}
}

//==================================================================================================================================================================

#include "cast.inl"

//==================================================================================================================================================================

namespace brownie
{
	namespace detail
	{
#pragma pack(push, 1)
		template < int n > struct SizeOf { char _[n+1]; };
#pragma pack(pop)

		SizeOf<T_BYTE  > extern TypeOf(INT8);
		SizeOf<T_UBYTE > extern TypeOf(UINT8);
		SizeOf<T_SHORT > extern TypeOf(INT16);
		SizeOf<T_USHORT> extern TypeOf(UINT16);
		SizeOf<T_INT   > extern TypeOf(INT32);
		SizeOf<T_UINT  > extern TypeOf(UINT32);
		SizeOf<T_LONG  > extern TypeOf(LONG);
		SizeOf<T_ULONG > extern TypeOf(ULONG);
		SizeOf<T_FLOAT > extern TypeOf(FLOAT);
		SizeOf<T_DOUBLE> extern TypeOf(DOUBLE);
		SizeOf<T_CHAR  > extern TypeOf(CHAR);
		SizeOf<T_STRING> extern TypeOf(PCSTR);
		SizeOf<T_OBJECT> extern TypeOf(PyObject*);
	}

	//============================================================================

	namespace detail
	{
#define METH_NORESULT	0x80

		template < typename T > SizeOf<METH_NOARGS  > extern MethodOf(object (T::*)());
		template < typename T > SizeOf<METH_O       > extern MethodOf(object (T::*)(PyObject*));
		template < typename T > SizeOf<METH_VARARGS > extern MethodOf(object (T::*)(Tuple*));
		template < typename T > SizeOf<METH_KEYWORDS> extern MethodOf(object (T::*)(Tuple*, Dict*));
		template < typename T > SizeOf<METH_NOARGS   | METH_NORESULT> extern MethodOf(void (T::*)());
		template < typename T > SizeOf<METH_O        | METH_NORESULT> extern MethodOf(void (T::*)(PyObject*));
		template < typename T > SizeOf<METH_VARARGS  | METH_NORESULT> extern MethodOf(void (T::*)(Tuple*));
		template < typename T > SizeOf<METH_KEYWORDS | METH_NORESULT> extern MethodOf(void (T::*)(Tuple*, Dict*));

		extern SizeOf<METH_STATIC | METH_NOARGS  > MethodOf(object (*)());
		extern SizeOf<METH_STATIC | METH_O       > MethodOf(object (*)(PyObject*));
		extern SizeOf<METH_STATIC | METH_VARARGS > MethodOf(object (*)(Tuple*));
		extern SizeOf<METH_STATIC | METH_KEYWORDS> MethodOf(object (*)(Tuple*, Dict*));
		extern SizeOf<METH_STATIC | METH_NOARGS   | METH_NORESULT> MethodOf(void (*)());
		extern SizeOf<METH_STATIC | METH_O        | METH_NORESULT> MethodOf(void (*)(PyObject*));
		extern SizeOf<METH_STATIC | METH_VARARGS  | METH_NORESULT> MethodOf(void (*)(Tuple*));
		extern SizeOf<METH_STATIC | METH_KEYWORDS | METH_NORESULT> MethodOf(void (*)(Tuple*, Dict*));

		//============================================================================

		template < int flag > struct Method;

		template <> struct Method<METH_NOARGS>
		{
			template < typename T, object (T::*fn)() > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args) throw()
				{
					try
					{
						if (PyObject* r = (cast<T>(self)->*fn)().detach())
							return r;
						else
							Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_NOARGS | METH_NORESULT>
		{
			template < typename T, void (T::*fn)() > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args) throw()
				{
					try
					{
						(cast<T>(self)->*fn)();
						Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_O>
		{
			template < typename T, object (T::*fn)(PyObject*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* o) throw()
				{
					try
					{
						if (PyObject* r = (cast<T>(self)->*fn)(o).detach())
							return r;
						else
							Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_O | METH_NORESULT>
		{
			template < typename T, void (T::*fn)(PyObject*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* o) throw()
				{
					try
					{
						(cast<T>(self)->*fn)(o);
						Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_VARARGS>
		{
			template < typename T, object (T::*fn)(Tuple*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args) throw()
				{
					try
					{
						if (PyObject* r = (cast<T>(self)->*fn)(static_cast<Tuple*>(args)).detach())
							return r;
						else
							Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_VARARGS | METH_NORESULT>
		{
			template < typename T, void (T::*fn)(Tuple*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args) throw()
				{
					try
					{
						(cast<T>(self)->*fn)(static_cast<Tuple*>(args));
						Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_KEYWORDS>
		{
			template < typename T, object (T::*fn)(Tuple*, Dict*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args, PyObject* kw)
				{
					return __proxy3__<T, fn>(self, args, kw);
				}
			};
		};

		template <> struct Method<METH_KEYWORDS | METH_NORESULT>
		{
			template < typename T, void (T::*fn)(Tuple*, Dict*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args, PyObject* kw)
				{
					try
					{
						(cast<T>(self)->*fn)(static_cast<Tuple*>(args), static_cast<Dict*>(kw));
						Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_STATIC | METH_NOARGS>
		{
			template < typename T, object (*fn)() > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args) throw()
				{
					try
					{
						if (PyObject* r = fn().detach())
							return r;
						else
							Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_STATIC | METH_NOARGS | METH_NORESULT>
		{
			template < typename T, void (*fn)() > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args) throw()
				{
					try
					{
						fn();
						Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_STATIC | METH_O>
		{
			template < typename T, object (*fn)(PyObject*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* o) throw()
				{
					try
					{
						if (PyObject* r = fn(o).detach())
							return r;
						else
							Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_STATIC | METH_O | METH_NORESULT>
		{
			template < typename T, void (*fn)(PyObject*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* o) throw()
				{
					try
					{
						fn(o);
						Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_STATIC | METH_VARARGS>
		{
			template < typename T, object (*fn)(Tuple*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args) throw()
				{
					try
					{
						if (PyObject* r = fn(static_cast<Tuple*>(args)).detach())
							return r;
						else
							Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_STATIC | METH_VARARGS | METH_NORESULT>
		{
			template < typename T, void (*fn)(Tuple*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args) throw()
				{
					try
					{
						fn(static_cast<Tuple*>(args));
						Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		template <> struct Method<METH_STATIC | METH_KEYWORDS>
		{
			template < typename T, object (*fn)(Tuple*, Dict*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args, PyObject* kw)
				{
					return __proxy3__<T, fn>(self, args, kw);
				}
			};
		};

		template <> struct Method<METH_STATIC | METH_KEYWORDS | METH_NORESULT>
		{
			template < typename T, void (*fn)(Tuple*, Dict*) > struct Fn
			{
				static PyObject* Do(PyObject* self, PyObject* args, PyObject* kw)
				{
					try
					{
						fn(static_cast<Tuple*>(args), static_cast<Dict*>(kw));
						Py_RETURN_NONE;
					}
					catch (Exception&) { return null; }
				}
			};
		};

		//============================================================================

		template < typename T, object (T::*fn)() > struct Get
		{
			static PyObject* Do(PyObject* self, void* closure)
			{
				try
				{
					if (PyObject* o = (cast<T>(self)->*fn)().detach())
						return o;
					else
						Py_RETURN_NONE;
				}
				catch (Exception&) { return null; }
			}
		};

		template < typename T, void (T::*fn)(PyObject*) > struct Set
		{
			static int Do(PyObject* self, PyObject* value, void* closure)
			{
				try
				{
					(cast<T>(self)->*fn)(value);
					return 0;
				}
				catch (Exception&) { return -1; }
			}
		};

		template < typename T, object T::*var > struct Var
		{
			static PyObject* Get(PyObject* self, void* closure)
			{
				PyObject* o = cast<T>(self)->*var;
				if (!o)
					o = Py_None;
				Py_INCREF(o);
				return o;
			}

			static int Set(PyObject* self, PyObject* value, void* closure)
			{
				try
				{
					cast<T>(self)->*var = (is<Null>(value) ? null : value);
					return 0;
				}
				catch (Exception&) { return -1; }
			}
		};
	}
}
