// drwaing.hpp
#pragma once

#include "handle.hpp"
#include "struct.hpp"

BOOL __stdcall PenDispose(HPEN handle) throw();
BOOL __stdcall BrushDispose(HBRUSH handle) throw();
BOOL __stdcall FontDispose(HFONT handle) throw();
BOOL __stdcall BitmapDispose(HBITMAP handle) throw();

namespace brownie
{
#ifdef SelectBitmap
#	undef SelectBitmap
#endif
	inline HBITMAP SelectBitmap(HDC dc, HBITMAP obj) throw()
	{
		ASSERT(dc);
		return (HBITMAP)::SelectObject(dc, obj);
	}

#ifdef SelectFont
#	undef SelectFont
#endif
	inline HFONT SelectFont(HDC dc, HFONT obj) throw()
	{
		ASSERT(dc);
		return (HFONT)::SelectObject(dc, obj);
	}

	typedef HandleT<HFONT	, FontDispose>		Font;
	typedef HandleT<HBITMAP	, BitmapDispose>	Bitmap;
	typedef HandleT<HBRUSH	, BrushDispose>		Brush;
	typedef HandleT<HPEN	, PenDispose>		Pen;
	typedef HandleT<HDC		, DeleteDC>			DC;

	void	FillRect(HDC dc, const RECT& rc, COLORREF color) throw();
	void	FillRect(HDC dc, int x, int y, int w, int h, COLORREF color) throw();
	void	FillRect(HDC dc, const RECT& rc, HBRUSH brush) throw();
	void	FillRect(HDC dc, int x, int y, int w, int h, HBRUSH brush) throw();
	void	DrawLine(HDC dc, int x1, int y1, int x2, int y2) throw();
	void	GetNonClientMetrics(NONCLIENTMETRICS* metrics);

	inline HFONT	GetStockFont(int i) throw()		{ return (HFONT)::GetStockObject(i); }

	HRESULT	GraphicsCreate(PCWSTR name) throw();
	void	GraphicsDispose() throw();

	//========================================================================================================================
	// Bitmap

	HBITMAP	BitmapLoad(IStream* stream) throw();
	HBITMAP	BitmapLoad(PCWSTR filename) throw();
	HBITMAP	BitmapLoad(IShellItem* item, SIZE sz) throw();

	inline SIZE BitmapSize(HBITMAP bmp) throw()
	{
		BITMAP info;
		GetObject(bmp, sizeof(info), &info);
		SIZE sz = { info.bmWidth, info.bmHeight };
		return sz;
	}

	//========================================================================================================================
	// Font

	SIZE	FontSize(HFONT font, PCWSTR text, HDC dc) throw();
	int		FontHeight(HFONT font, HDC dc = null) throw();

	//========================================================================================================================
	// ImageList

	HIMAGELIST	ImageListLoad(IStream* stream) throw();
	HIMAGELIST	ImageListLoad(PCWSTR filename) throw();
	HIMAGELIST	SHGetImageList(int size) throw();
	inline bool	IsShellIcon(int icon) throw()		{ return icon >= MAKELONG(0, 1); }

	//========================================================================================================================
	// etc.

	HICON	IconLoad(PCWSTR filename, int index, int w, int h) throw();

	//========================================================================================================================
	// Theme

	class Theme
	{
	private:
		HTHEME	m_theme;

	public:
		Theme();
		~Theme();
		bool Create(HWND hwnd, PCWSTR scheme);
		void Dispose() throw();
		void Background(HDC hDC, int part, int state, const RECT* rc, const RECT* clip = null);
		void Edge(HDC hDC, int part, int state, const RECT* rc, UINT uEdge, UINT uFlags, RECT* rcAdjust = null);
		void Text(HDC hDC, int part, int state, const RECT* rc, PCWSTR text, DWORD dwFlagsDT, STATUS status = NORMAL);
	};
}

//========================================================================================================================
// 

inline void	brownie::FillRect(HDC dc, int x, int y, int w, int h, COLORREF color) throw()
{
	FillRect(dc, RECT_FROM_XYWH(x, y, w, h), color);
}

inline void	brownie::FillRect(HDC dc, const RECT& rc, HBRUSH brush) throw()
{
	::FillRect(dc, &rc, brush);
}

inline void	brownie::FillRect(HDC dc, int x, int y, int w, int h, HBRUSH brush) throw()
{
	FillRect(dc, RECT_FROM_XYWH(x, y, w, h), brush);
}
