// exception.hpp
#pragma once

#include "brownie.hpp"

#if PY_VERSION_HEX < 0x03000000
#	define PyExc_Exception			PyExc_StandardError
#endif

namespace brownie
{
	class Exception
	{
	protected:
		Exception();
		Exception(PyObject* type, PyObject* msg);
		Exception(PyObject* type, PCSTR     msg);
		Exception(PyObject* type, PCWSTR    msg);
	public:
		static void __declspec(noreturn) raise() throw(...);
		static void notify() throw();
		static void clear() throw();
	};

#define EXCEPTION(name, base)										\
	class name : public base										\
	{																\
	protected:														\
		name(PyObject* type, PyObject* msg) : base(type, msg) {}	\
		name(PyObject* type, PCSTR     msg) : base(type, msg) {}	\
		name(PyObject* type, PCWSTR    msg) : base(type, msg) {}	\
	public:															\
		name()	{}													\
		name(PyObject* msg) : base(PyExc_##name, msg) {}			\
		name(PCSTR     msg) : base(PyExc_##name, msg) {}			\
		name(PCWSTR    msg) : base(PyExc_##name, msg) {}			\
	}

//	EXCEPTION( SystemExit					, BaseException		);
//	EXCEPTION( StopIteration				, BaseException		);

	EXCEPTION( KeyboardInterrupt		, Exception			);
	EXCEPTION( ImportError				, Exception			);
	EXCEPTION( EnvironmentError			, Exception			);
	EXCEPTION(   IOError				, EnvironmentError	);
	EXCEPTION(   OSError				, EnvironmentError	);
	EXCEPTION(	 WindowsError			, OSError			);
	EXCEPTION( EOFError					, Exception			);
	EXCEPTION( RuntimeError				, Exception			);
	EXCEPTION(   NotImplementedError	, RuntimeError		);
	EXCEPTION( NameError				, Exception			);
	EXCEPTION(   UnboundLocalError		, NameError			);
	EXCEPTION( AttributeError			, Exception			);
	EXCEPTION( SyntaxError				, Exception			);
	EXCEPTION(   IndentationError		, SyntaxError		);
	EXCEPTION(	 TabError				, IndentationError	);
	EXCEPTION( TypeError				, Exception			);
	EXCEPTION( AssertionError			, Exception			);
	EXCEPTION( LookupError				, Exception			);
	EXCEPTION(   IndexError				, LookupError		);
	EXCEPTION(   KeyError				, LookupError		);
	EXCEPTION( ArithmeticError			, Exception			);
	EXCEPTION(   OverflowError			, ArithmeticError	);
	EXCEPTION(   ZeroDivisionError		, ArithmeticError	);
	EXCEPTION(   FloatingPointError		, ArithmeticError	);
	EXCEPTION( ValueError				, Exception			);
	EXCEPTION(   UnicodeError			, ValueError		);
	EXCEPTION(	 UnicodeEncodeError		, UnicodeError		);
	EXCEPTION(	 UnicodeDecodeError		, UnicodeError		);
	EXCEPTION(	 UnicodeTranslateError	, UnicodeError		);
	EXCEPTION( ReferenceError			, Exception			);
	EXCEPTION( SystemError				, Exception			);
	EXCEPTION( MemoryError				, Exception			);

//	EXCEPTION( Warning						, BaseException		);
//	EXCEPTION(	 UserWarning				, Warning			);
//	EXCEPTION(	 DeprecationWarning			, Warning			);
//	EXCEPTION(	 PendingDeprecationWarning	, Warning			);
//	EXCEPTION(	 SyntaxWarning				, Warning			);
//	EXCEPTION(	 OverflowWarning			, Warning			);
//	EXCEPTION(	 RuntimeWarning				, Warning			);
//	EXCEPTION(	 FutureWarning				, Warning			);

#undef EXCEPTION
}
