// folderview.hpp
#pragma once

#include "listview.hpp"
#include "shell.hpp"
#include "unknown.hpp"

namespace brownie
{
	class FolderView :
		public Class<AbstractListView>,
		public Unknown< implements<IServiceProvider, IExplorerBrowserEvents, IOleWindow, IShellBrowser, ICommDlgBrowser, ICommDlgBrowser2/*, ICommDlgBrowser3*/>, mixin<StaticLife> >
	{
	private:
		LRESULT	DefView_OnMessage(Hwnd& hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
		LRESULT	ListView_OnMessage(ListViewH& hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

	private:
		ref<IExplorerBrowser>	m_browser;
		DWORD					m_cookie;	// for m_browser->Advise()
		ILPtr					m_item;
		object					m_path;		// Path(m_item) ̃LbV
		std::vector<WCHAR>		m_notification;

		HwndT<FolderView, Hwnd, &FolderView::DefView_OnMessage>			m_defview;
		HwndT<FolderView, ListViewH, &FolderView::ListView_OnMessage>	m_listview;

	public:
		RECT	get_CenterArea() const throw();
		HRESULT	GetCurrentView(REFINTF pp) const throw();
		HRESULT	GetCurrentFolder(REFINTF pp) const throw();
		HRESULT	PopupContextMenu(int x, int y) throw();
		CIDA*	ExecuteSelection(IShellView* view) throw();
		HRESULT	Rename() throw();
		HRESULT	Refresh() throw();
		UINT	GetMode();
		HWND	ListView_Handle() const throw()		{ return m_listview; }

		bool	get_AutoArrange() const;
		void	set_AutoArrange(bool value);
		__declspec(property(get=get_AutoArrange, put=set_AutoArrange)) bool AutoArrange;

		void	set_Grouping(bool value);
		__declspec(property(get=get_Grouping, put=set_Grouping)) bool Grouping;

	protected:
		LRESULT	OnMessage(UINT msg, WPARAM wParam, LPARAM lParam);
		bool	OnPress(const MSG& msg);
		bool	OnGesture(const MSG& msg);
		void	QueryLayout(ILayout& layout) const;
		void	OnDragDrop(POINT pt, RECT* rc, IDropTarget** pp) throw();

	private:
		void	OnNotify(PCWSTR message);
		void	OnAdjust();
		bool	OnTimer(UINT wID);
		void	DefView_OnBeginLabelEdit(NMLVDISPINFO* info);
		bool	DefView_OnDrawItem(DRAWITEMSTRUCT* draw);
		void	ListView_UpdateOwnerDraw();
		LRESULT	ListView_DefProcNoOwnerDraw(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
		void	SetCurrentFolder(const ITEMIDLIST* folder);

	public:	// IServiceProvider
		STDMETHODIMP QueryService(REFGUID guidService, REFIID riid, void** ppv);

	public: // IExplorerBrowserEvents
		STDMETHODIMP OnNavigationPending(const ITEMIDLIST* folder);
		STDMETHODIMP OnViewCreated(IShellView* view);
		STDMETHODIMP OnNavigationComplete(const ITEMIDLIST* folder);
		STDMETHODIMP OnNavigationFailed(const ITEMIDLIST* folder);

	public: // IOleWindow
		STDMETHODIMP GetWindow(HWND* phwnd);
		STDMETHODIMP ContextSensitiveHelp(BOOL fEnterMode)	{ return E_NOTIMPL; }

	public: // ICommDlgBrowser
		STDMETHODIMP OnDefaultCommand(IShellView* view);
		STDMETHODIMP OnStateChange(IShellView* view, ULONG uChange);
		STDMETHODIMP IncludeObject(IShellView* view, const ITEMIDLIST* pidl);

	public: // ICommDlgBrowser2
		STDMETHODIMP GetDefaultMenuText(IShellView* view, WCHAR* pszText, int cchMax);
		STDMETHODIMP Notify(IShellView* view, DWORD dwNotifyType)						{ return S_OK; }
		STDMETHODIMP GetViewFlags(DWORD* pdwFlags);

	//public:	// ICommDlgBrowser3
	//	STDMETHODIMP OnColumnClicked(IShellView* view, int iColumn)				{ return E_NOTIMPL; }
	//	STDMETHODIMP GetCurrentFilter(PWSTR pszFileSpec, int cchFileSpec)		{ return E_NOTIMPL; }
	//	STDMETHODIMP OnPreViewCreated(IShellView* view)							{ return E_NOTIMPL; }

	public: // IShellBrowser
		STDMETHODIMP InsertMenusSB(HMENU hmenuShared, LPOLEMENUGROUPWIDTHS lpMenuWidths)				{ return E_NOTIMPL; }
		STDMETHODIMP SetMenuSB(HMENU hmenuShared, HOLEMENU holemenuReserved,HWND hwndActiveObject)		{ return E_NOTIMPL; }
		STDMETHODIMP RemoveMenusSB(HMENU hmenuShared)													{ return E_NOTIMPL; }
		STDMETHODIMP SetStatusTextSB(LPCOLESTR text)													{ return E_NOTIMPL; }
		STDMETHODIMP EnableModelessSB(BOOL fEnable)														{ return E_NOTIMPL; }
		STDMETHODIMP SetToolbarItems(LPTBBUTTON lpButtons, UINT nButtons,UINT uFlags)					{ return E_NOTIMPL; }
		STDMETHODIMP TranslateAcceleratorSB(LPMSG lpmsg, WORD wID)										{ return E_NOTIMPL; }
		STDMETHODIMP OnViewWindowActive(IShellView* pShellView)											{ return E_NOTIMPL; }
		STDMETHODIMP GetControlWindow(UINT id, HWND* pHwnd);
		STDMETHODIMP SendControlMsg(UINT id, UINT msg, WPARAM wParam, LPARAM lParam, LRESULT* lResult);
		STDMETHODIMP GetViewStateStream(DWORD stgm, IStream** pp)										{ return E_NOTIMPL; }
		STDMETHODIMP QueryActiveShellView(IShellView** pp);
		STDMETHODIMP BrowseObject(const ITEMIDLIST* pidl, UINT wFlags);

	public: // IOleCommandTarget
	//	STDMETHODIMP Exec(const GUID* group, DWORD nCmdId, DWORD option, VARIANTARG* inargs, VARIANTARG* outargs);
	//	STDMETHODIMP QueryStatus(const GUID* group, ULONG cCmds, OLECMD* prgCmds, OLECMDTEXT *pCmdText);

	public: // python 
		static PCSTR __name__() throw()					{ return "brownie.FolderView"; }
		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();

		object	get_path();
		object	get_items(STATUS status);

		void	cut();
		void	copy();
		object	history(PyObject* value);
		object	parse(PyObject* value);
		void	paste();
		void	delete_();
		void	bury();
		void	rename();
		void	refresh();
		int		select(PyObject* o, UINT svsi);
		void	undo();
		void	go(PyObject* value);
		void	new_(Tuple* args);
		object	mode(Tuple* args, Dict* kw);
	};
}
