// handle.hpp
#pragma once

#include "brownie.hpp"

namespace brownie
{
	template < typename T, BOOL (STDMETHODCALLTYPE *Free)(T)  >
	class HandleT
	{
	protected:
		T	m_handle;

	public:
		HandleT(T handle = null) : m_handle(handle)	{}
		~HandleT()	{ Dispose(); }

		HandleT& operator = (T handle) throw()	{ Attach(handle); return *this; }

		operator T () const throw()		{ return m_handle; }
		T	get_Handle() const throw()	{ return m_handle; }
		__declspec(property(get=get_Handle)) T Handle;

		void Dispose() throw()
		{
			if (m_handle)
			{
				Free(m_handle);
				m_handle = null;
			}
		}

		void Attach(T handle) throw()
		{
			if (m_handle != handle)
			{
				Dispose();
				m_handle = handle;
			}
		}

		T Detach() throw()
		{
			T tmp = m_handle;
			m_handle = null;
			return tmp;
		}

	private:
		HandleT(const HandleT&);
		HandleT& operator = (const HandleT&);
	};
}
