// hwnd.hpp
#pragma once

#include "config.hpp"
#include "struct.hpp"

inline POINT	GetCursorPos() throw()						{ POINT pt; ::GetCursorPos(&pt); return pt; }
inline DWORD	GetStyle(HWND hwnd) throw()					{ return ::GetWindowLong(hwnd, GWL_STYLE); }
inline DWORD	SetStyle(HWND hwnd, DWORD style) throw()	{ return ::SetWindowLong(hwnd, GWL_STYLE, style); }
inline bool		HasStyle(HWND hwnd, UINT style) throw()		{ return (GetStyle(hwnd) & style) != 0; }
inline DWORD	GetStyleEx(HWND hwnd) throw()				{ return ::GetWindowLong(hwnd, GWL_EXSTYLE); }
inline DWORD	SetStyleEx(HWND hwnd, DWORD style) throw()	{ return ::SetWindowLong(hwnd, GWL_EXSTYLE, style); }
inline bool		HasStyleEx(HWND hwnd, UINT style) throw()	{ return (GetStyleEx(hwnd) & style) != 0; }
inline bool		IsVisible(HWND hwnd) throw()				{ return HasStyle(hwnd, WS_VISIBLE); }
bool	HasFocus(HWND hwnd) throw();
SIZE	GetClientSize(HWND hwnd) throw();
void	SetClientSize(HWND hwnd, int w, int h) throw();
int		MapWindowRect(HWND hwndFrom, HWND hwndTo, RECT* rc);
void	CenterWindow(HWND hwnd, HWND outer = null) throw();
void	SetRedraw(HWND hwnd, bool redraw) throw();
WNDPROC	SetWindowProc(HWND hwnd, WNDPROC wndproc);

inline void ClientToScreen(HWND hwnd, RECT* rc)
{
	::ClientToScreen(hwnd, (POINT*)rc + 0);
	::ClientToScreen(hwnd, (POINT*)rc + 1);
}

inline void ScreenToClient(HWND hwnd, RECT* rc)
{
	::ScreenToClient(hwnd, (POINT*)rc + 0);
	::ScreenToClient(hwnd, (POINT*)rc + 1);
}

void	PumpMessage() throw();
void	BoardcastMessage(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0) throw();

void	InitCommonControls(UINT flags) throw();

void	DoubleClick(POINT ptScreen);
UINT	GetButtonState() throw();
UINT	GetModifiers() throw();
void	SetKeyState(UINT vkey, UINT mods);
void	ResetKeyState(UINT vkey);
inline BOOL IsKeyPressed(int vkey) throw()	{ return ::GetKeyState(vkey) & 0x8000; }

class LockRedraw
{
private:
	HWND	m_hwnd;
public:
	LockRedraw(HWND hwnd) : m_hwnd(hwnd)	{ SetRedraw(m_hwnd, false); }
	~LockRedraw()							{ SetRedraw(m_hwnd, true); }
};

namespace brownie
{
	class Hwnd
	{
	protected:
		HWND	m_hwnd;

	public:
		Hwnd(HWND hwnd = null) : m_hwnd(hwnd)	{}

		operator HWND () const throw()		{ return m_hwnd; }
		HWND	get_Handle() const throw()	{ return m_hwnd; }
		__declspec(property(get=get_Handle)) HWND Handle;

		Hwnd& operator = (HWND hwnd)		{ m_hwnd = hwnd; return *this; }

		HWND	get_Parent() const throw()			{ return ::GetParent(m_hwnd); }
		__declspec(property(get=get_Parent)) HWND Parent;

		bool	get_Visible() const throw()			{ return IsVisible(m_hwnd); }
		void	set_Visible(bool value) throw()		{ ::ShowWindow(m_hwnd, value ? SW_SHOW : SW_HIDE); }
		__declspec(property(get=get_Visible, put=set_Visible)) bool Visible;

		POINT	get_Location() const throw()		{ RECT r; GetPos(&r); return RECT_LOCATION(r); }
		void	set_Location(POINT value) throw()	{ SetPos(POINT_XY(value), 0, 0, SWP_NOSIZE); }
		__declspec(property(get=get_Location, put=set_Location)) POINT Location;

		SIZE	get_Size() const throw()			{ RECT r; ::GetWindowRect(m_hwnd, &r); SIZE s = { RECT_WH(r) }; return s; }
		void	set_Size(SIZE value) throw()		{ SetPos(0, 0, SIZE_WH(value), SWP_NOMOVE); }
		__declspec(property(get=get_Size, put=set_Size)) SIZE Size;

		SIZE	get_ClientSize() const throw()		{ return ::GetClientSize(m_hwnd); }
		void	set_ClientSize(SIZE value) throw()	{ ::SetClientSize(m_hwnd, SIZE_WH(value)); }
		__declspec(property(get=get_ClientSize, put=set_ClientSize)) SIZE ClientSize;

		BOOL	IsWindow() const throw()			{ return ::IsWindow(m_hwnd); }

		LRESULT	SendMessage(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0) const throw()
		{
			if (!m_hwnd)
				return 0;
			return ::SendMessage(m_hwnd, msg, wParam, lParam);
		}

		BOOL PostMessage(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0) const throw()
		{
			if (!m_hwnd)
				return false;
			return ::PostMessage(m_hwnd, msg, wParam, lParam);
		}

		void	SendCommand(UINT wID) throw()					{ SendMessage(WM_COMMAND, MAKEWPARAM(wID, 0), 0); }
		void	PostCommand(UINT wID) throw()					{ PostMessage(WM_COMMAND, MAKEWPARAM(wID, 0), 0); }
		void	SetTimer(UINT wID, UINT delay) throw()			{ ::SetTimer(m_hwnd, wID, delay, null); }
		void	KillTimer(UINT wID) throw()						{ ::KillTimer(m_hwnd, wID); }
		int		GetTextLength() const throw()					{ return ::GetWindowTextLength(m_hwnd); }
		int		GetText(WCHAR value[], int ln) const throw()	{ return ::GetWindowText(m_hwnd, value, ln); }
		void	SetText(PCWSTR value) throw()					{ ::SetWindowText(m_hwnd, value); }
		void	GetPos(RECT* bounds) const throw();
		void	SetPos(INT x, INT y, INT w, INT h, UINT swp = 0) throw()	{ ::SetWindowPos(m_hwnd, null, x, y, w, h, SWP_NOZORDER | SWP_NOACTIVATE | swp); }
		HFONT	GetFont() const throw()					{ return (HFONT)SendMessage(WM_GETFONT, 0, 0); }
		void	SetFont(HFONT value) throw()			{ SendMessage(WM_SETFONT, (WPARAM)value, TRUE); }
		HWND	GetAncestor(UINT type) const throw()	{ return ::GetAncestor(m_hwnd, type); }

		void	ModifyStyle(DWORD dwRemove, DWORD dwAdd, DWORD dwRemoveEx = 0, DWORD dwAddEx = 0) throw();
		void	DisableIME() throw()									{ ::ImmAssociateContext(m_hwnd, null); }

		void	Close() throw()
		{
			PostMessage(WM_CLOSE, 0, 0);
		}

		void	PopupMenu(HWND hwnd = null, POINT ptScreen = GetCursorPos()) throw()
		{
			PostMessage(WM_CONTEXTMENU, (WPARAM)hwnd, MAKELPARAM(ptScreen.x, ptScreen.y));
		}

		static HWND From(IOleWindow* w) throw();
		static HWND From(IUnknown* p) throw();

		BOOL	IsParentOf(HWND hwnd) const			{ return ::IsChild(m_hwnd, hwnd); }
		BOOL	IsSelfOrParentOf(HWND hwnd) const	{ return m_hwnd == hwnd || ::IsChild(m_hwnd, hwnd); }

		POINT	ScreenToClient(POINT pt) const	{ ::ScreenToClient(m_hwnd, &pt); return pt; }
		POINT	ClientToScreen(POINT pt) const	{ ::ClientToScreen(m_hwnd, &pt); return pt; }
		void	ScreenToClient(POINT* pt) const	{ VERIFY( ::ScreenToClient(m_hwnd, pt) ); }
		void	ClientToScreen(POINT* pt) const	{ VERIFY( ::ClientToScreen(m_hwnd, pt) ); }
		void	ScreenToClient(RECT* rc) const	{ ::ScreenToClient(m_hwnd, rc); }
		void	ClientToScreen(RECT* rc) const	{ ::ClientToScreen(m_hwnd, rc); }

		HWND	Create(PCWSTR className,
					   HWND parent,
					   DWORD style,
					   DWORD styleEx,
					   int x = CW_USEDEFAULT,
					   int y = CW_USEDEFAULT,
					   int w = CW_USEDEFAULT,
					   int h = CW_USEDEFAULT) throw();
		void	Dispose() throw()				{ if (m_hwnd) { ::DestroyWindow(m_hwnd); m_hwnd = null; } }
		void	Activate() throw();
		HWND	SetFocus() throw()				{ return ::GetFocus() != m_hwnd ? ::SetFocus(m_hwnd) : null; }
		void	SetRedraw(bool redraw) throw()	{ ::SetRedraw(m_hwnd, redraw); }

		HWND	SetCapture()					{ return ::SetCapture(m_hwnd); }
		void	ReleaseCapture()				{ if (::GetCapture() == m_hwnd) { ::ReleaseCapture(); } }

		HDC		GetDC() throw()					{ return ::GetDC(m_hwnd); }
		void	ReleaseDC(HDC dc) throw()		{ ::ReleaseDC(m_hwnd, dc); }
		void	Invalidate() throw()				{ ::InvalidateRect(m_hwnd, null, TRUE); }
		void	Invalidate(const RECT& rc) throw()	{ ::InvalidateRect(m_hwnd, &rc, TRUE); }
		void	TrackMouseEvent(DWORD tme, DWORD time) throw();
	};

	//==================================================================================================================================================================

	template < typename T, typename U, LRESULT (T::*fn)(U&, UINT, WPARAM, LPARAM) >
	class HwndT : public U
	{
	public:
		~HwndT()	{ Detach(); }

		bool Attach(HWND hwnd, T* owner) throw()
		{
			if (m_hwnd == hwnd)
				return false;
			Detach();
			if (!hwnd)
				return false;
			m_hwnd = hwnd;
			if (::SetWindowSubclass(hwnd, OnMessage, (UINT_PTR)this, (DWORD_PTR)owner))
				return true; // ok
			m_hwnd = null;
			return false;
		}

		bool Attach(IOleWindow* w, T* owner) throw()
		{
			HWND hwnd;
			if FAILED(w->GetWindow(&hwnd))
				return false;
			return Attach(hwnd, owner);
		}

		bool Detach() throw()
		{
			if (!m_hwnd)
				return false;
			if (!::RemoveWindowSubclass(m_hwnd, OnMessage, (UINT_PTR)this))
				return false;
			m_hwnd = null;
			return true;
		}

	private:
		static LRESULT CALLBACK OnMessage(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam, UINT_PTR id, DWORD_PTR data)
		{
			HwndT* self = (HwndT*)id;
			ASSERT(self->m_hwnd == hwnd);
			LRESULT lResult = (((T*)data)->*fn)(*self, msg, wParam, lParam);
			if (msg == WM_NCDESTROY)
				self->Detach();
			return lResult;
		}
	};
}
