// listview.hpp
#pragma once

#include "commctrl.hpp"
#include "icons.hpp"
#include "window.hpp"

namespace brownie
{
	//==================================================================================================================================================================

	class AbstractListView : public Class<Window>
	{
	private:
		bool	m_adjustRequested;

	public:
		void	Adjust() throw();

		bool	get_FullRowSelect() const;
		void	set_FullRowSelect(bool value);
		__declspec(property(get=get_FullRowSelect, put=set_FullRowSelect)) bool FullRowSelect;

		bool	get_GridLine() const;
		void	set_GridLine(bool value);
		__declspec(property(get=get_GridLine, put=set_GridLine)) bool GridLine;

		bool	get_Grouping() const;
		virtual void	set_Grouping(bool value);
		__declspec(property(get=get_Grouping, put=set_Grouping)) bool Grouping;

	public:
		static PCSTR __name__() throw()					{ return "brownie.AbstractListView"; }
		void __init__(Tuple* args, Dict* kw) throw(...);
		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();
		void	adjust();
		object	get_selection();
		void	set_selection(PyObject* o);
		object	get_grouping();
		void	set_grouping(PyObject* value);
		object	select(Tuple* args);
		int		select_by_callable(PyObject* callable, UINT svsi);

	protected:
		void	ListView_OnMessage(UINT msg, WPARAM wParam, LPARAM lParam);
	public:
		virtual HWND	ListView_Handle() const throw()	{ return m_hwnd; }
		virtual object	get_items(STATUS status)		{ return null; }
		virtual int		select(PyObject* o, UINT svsi);
	};

	//==================================================================================================================================================================

	class ListView : public Class<AbstractListView>, ListViewT<ListView>
	{
	private:
		object		m_columns;
		object		m_icons;
		object		m_items;

	public:
		Icons*	GetIcons() const throw();
		void	SetIcons(Icons* value) throw();

	protected:
		LRESULT	OnMessage(UINT msg, WPARAM wParam, LPARAM lParam);
		LRESULT	OnMessage(Window* from, NMHDR& nm);
		bool	OnGesture(const MSG& msg);

	private:
		bool	OnLButtonDblClk(WPARAM wParam, LPARAM lParam);
		bool	OnMButtonUp(WPARAM wParam, LPARAM lParam);
		bool	OnContextMenu(LPARAM lParam);

	public:
		PyObject*	GetItem(int index) const throw();
		void		Clear() throw();

	public:
		static PCSTR __name__() throw()					{ return "brownie.ListView"; }
		void __init__(Tuple* args, Dict* kw) throw(...);
//		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();

		void	set_columns(PyObject* value);
		void	set_icons(PyObject* value);
		object	get_items(STATUS status);
		void	set_items(PyObject* value);
		void	clear();
	};
}
