// ntddk.hpp
#pragma once

#include "brownie.hpp"

#pragma pack(push, 8)

typedef DWORD	NTSTATUS;

#ifndef STATUS_SUCCESS
#define STATUS_SUCCESS					((NTSTATUS)0x00000000L)
#define STATUS_INFO_LENGTH_MISMATCH		((NTSTATUS)0xC0000004L)
#define STATUS_BUFFER_OVERFLOW			((NTSTATUS)0xC0000005L)
#define STATUS_OBJECT_NAME_NOT_FOUND	((NTSTATUS)0xC0000034L)
#define STATUS_INVALID_SID				((NTSTATUS)0xC0000078L)
#endif

enum OBJECT_TYPE
{
	Adapter,
	Callback,
	Controller,
	DebugObject,
	Desktop,
	Device,
	Directory,
	Driver,
	Event,
	EventPair,
	File,
	FilterCommunicationPort,
	FilterConnectionPort,
	IoCompletion,
	Job,
	Key,
	KeyedEvent,
	Mutant,
	Port,
	Process,
	Profile,
	Section,
	Semaphore,
	SymbolicLink,
	Thread,
	Timer,
	Token,
	Type,
	WaitablePort,
	WindowStation,
	WmiGuid,
	ObjectTypeMax
};

enum OBJECT_INFORMATION_CLASS
{
	ObjectBasicInformation,
	ObjectNameInformation,
	ObjectTypeInformation,
	ObjectAllTypesInformation,
	ObjectHandleInformation
};

enum SYSTEM_INFORMATION_CLASS
{
	SystemBasicInformation = 0,
	SystemCpuInformation = 1,
	SystemPerformanceInformation = 2,
	SystemTimeOfDayInformation = 3, /* was SystemTimeInformation */
	Unknown4,
	SystemProcessInformation = 5,
	Unknown6,
	Unknown7,
	SystemProcessorPerformanceInformation = 8,
	Unknown9,
	Unknown10,
	SystemModuleInformation = 11,
	Unknown12,
	Unknown13,
	Unknown14,
	Unknown15,
	SystemHandleInformation = 16,
	Unknown17,
	SystemPageFileInformation = 18,
	Unknown19,
	Unknown20,
	SystemCacheInformation = 21,
	Unknown22,
	SystemInterruptInformation = 23,
	SystemDpcBehaviourInformation = 24,
	SystemFullMemoryInformation = 25,
	SystemNotImplemented6 = 25,
	SystemLoadImage = 26,
	SystemUnloadImage = 27,
	SystemTimeAdjustmentInformation = 28,
	SystemTimeAdjustment = 28,
	SystemSummaryMemoryInformation = 29,
	SystemNotImplemented7 = 29,
	SystemNextEventIdInformation = 30,
	SystemNotImplemented8 = 30,
	SystemEventIdsInformation = 31,
	SystemCrashDumpInformation = 32,
	SystemExceptionInformation = 33,
	SystemCrashDumpStateInformation = 34,
	SystemKernelDebuggerInformation = 35,
	SystemContextSwitchInformation = 36,
	SystemRegistryQuotaInformation = 37,
	SystemCurrentTimeZoneInformation = 44,
	SystemTimeZoneInformation = 44,
	SystemLookasideInformation = 45,
	SystemSetTimeSlipEvent = 46,
	SystemCreateSession = 47,
	SystemDeleteSession = 48,
	SystemInvalidInfoClass4 = 49,
	SystemRangeStartInformation = 50,
	SystemVerifierInformation = 51,
	SystemAddVerifier = 52,
	SystemSessionProcessesInformation   = 53,
	SystemInformationClassMax
};

struct OBJECT_TYPE_INFORMATION
{
    UNICODE_STRING	TypeName;
    ULONG			TotalNumberOfObjects;
    ULONG			TotalNumberOfHandles;
    ULONG			TotalPagedPoolUsage;
    ULONG			TotalNonPagedPoolUsage;
    ULONG			TotalNamePoolUsage;
    ULONG			TotalHandleTableUsage;
    ULONG			HighWaterNumberOfObjects;
    ULONG			HighWaterNumberOfHandles;
    ULONG			HighWaterPagedPoolUsage;
    ULONG			HighWaterNonPagedPoolUsage;
    ULONG			HighWaterNamePoolUsage;
    ULONG			HighWaterHandleTableUsage;
    ULONG			InvalidAttributes;
    GENERIC_MAPPING	GenericMapping;
    ULONG			ValidAccessMask;
    BOOLEAN			SecurityRequired;
    BOOLEAN			MaintainHandleCount;
    ULONG			PoolType;
    ULONG			DefaultPagedPoolCharge;
    ULONG			DefaultNonPagedPoolCharge;
};

struct OBJECT_ALL_TYPES_INFORMATION
{
    ULONG					NumberOfTypes;
    OBJECT_TYPE_INFORMATION	TypeInformation[1];
};

struct SYSTEM_HANDLE_TABLE_ENTRY_INFO
{
	USHORT	UniqueProcessId;
	USHORT	CreatorBackTraceIndex;
	UCHAR	ObjectTypeIndex;
	UCHAR	HandleAttributes;
	USHORT	HandleValue;
	PVOID	Object;
	ULONG	GrantedAccess;
};

struct SYSTEM_HANDLE_INFORMATION
{
	ULONG							NumberOfHandles;
	SYSTEM_HANDLE_TABLE_ENTRY_INFO	Handles[1];
};

NTSTATUS NTAPI NtQueryObject(
	IN  HANDLE						ObjectHandle OPTIONAL,
	IN  OBJECT_INFORMATION_CLASS	ObjectInformationClass,
	OUT PVOID						ObjectInformation,
	IN  ULONG						Length,
	OUT PULONG						ResultLength);

NTSTATUS NTAPI NtQuerySystemInformation(
	IN  SYSTEM_INFORMATION_CLASS	SystemInformationClass,
    OUT PVOID						SystemInformation,
    IN  ULONG						SystemInformationLength,
    OUT PULONG						ReturnLength OPTIONAL);

#pragma pack(pop)

namespace NT
{
	extern DWORD OBJECT_TYPE_ID[ObjectTypeMax];

	HRESULT	Initialize() throw();
	SYSTEM_HANDLE_INFORMATION* QuerySystemHandleInformation() throw();
	UNICODE_STRING* GetObjectName(HANDLE handle);
	UNICODE_STRING* GetObjectName(HANDLE process, const SYSTEM_HANDLE_TABLE_ENTRY_INFO& handle);
}
