// path.hpp
#pragma once

#include "class.hpp"

//========================================================================================================================

namespace brownie
{
	//========================================================================================================================

	enum BROWSE
	{
		BROWSE_NG = 0,
		BROWSE_OK = 1,
		BROWSE_PREFERRED = 2,
	};

	class Path : public Object
	{
	private:
		ITEMIDLIST*		m_id;
		ref<IShellItem>	m_item;

	public:
		Path() : m_id(null)	{}
		~Path();
		static ref<Path>	newobj(PyObject* o) throw();
		static ref<Path>	newobj(PCWSTR path) throw();
		static ref<Path>	newobj(const ITEMIDLIST* id) throw();
		static ref<Path>	newobj(const ITEMIDLIST* parent, const ITEMIDLIST* leaf) throw();
		static ref<Path>	newobj(IShellFolder* folder, const ITEMIDLIST* parent, const ITEMIDLIST* leaf) throw();
		static ref<Path>	newobj(IShellItem* item) throw();
		static ref<Path>	newobj_nocopy(ITEMIDLIST* id) throw();

		const ITEMIDLIST*	get_ID() const throw()	{ return m_id; }
		__declspec(property(get=get_ID)) const ITEMIDLIST* ID;

		PWSTR			GetDisplayName(SIGDN sigdn) throw();
		PWSTR			GetName() throw()		{ return GetDisplayName(SIGDN_PARENTRELATIVE); }
		PWSTR			GetPath() throw()		{ return GetDisplayName(SIGDN_DESKTOPABSOLUTEPARSING); }
		bool			HasAttribute(SFGAOF sfgaof) throw();
		BROWSE			GetBrowsable() throw();
		HRESULT			GetObject(REFGUID rbhid, REFINTF pp) throw();
		ref<IStream>	Open(DWORD stgm = STGM_READ) throw();

	public: // python
		static PCSTR __name__() throw()					{ return "brownie.Path"; }
		void		__init__(Tuple* args, Dict* kw) throw(...);
		object		__iter__() throw();
		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();

	public:
		unicode	get_displayname(SIGDN sigdn) throw(...);
		HRESULT	get_menu(IContextMenu** pp) throw();
		ref<Path>	parent();

	public:
		void	execute(Tuple* args, Dict* kw);
		object	open(Tuple* args);
		void	symlink(PyObject* o);
		object	get_name();
		object	get_path();
		object	get_browsable();
		object	get_parent();
		object	get_children(DWORD flags);
		object	get_children();
		object	get_has_parent();
		object	get_has_subfolder();
		object	get_icon();
		object	get_linked();
		object	get_subfolders();

	private:
		HRESULT	Verify() throw();
	};

	//========================================================================================================================

	class Paths : public Object, public IShellItemArray
	{
	private:
		CIDA*			m_items;
		ref<IUnknown>	m_object;	// m_items is null ? IShellItemArray : IShellFolder
		bool			m_iscopy;

		IShellItemArray* get_array() const throw();
		__declspec(property(get=get_array)) IShellItemArray* m_array;

		IShellFolder* get_folder() const throw();
		__declspec(property(get=get_folder)) IShellFolder* m_folder;

	public:
		Paths() : m_items(null), m_iscopy(false)	{}
		static ref<Paths>	newobj(const CIDA* items, IShellFolder* folder, bool iscopy = false) throw();
		static ref<Paths>	newobj(IShellItemArray* items, bool iscopy = false) throw();

		const CIDA*		GetCIDA() const throw()	{ return m_items; }

	public: // IUnknown
		STDMETHODIMP QueryInterface(REFIID iid, void** pp);
		STDMETHODIMP_(ULONG) AddRef();
		STDMETHODIMP_(ULONG) Release();

	public: // IShellItemArray
        STDMETHODIMP BindToHandler(IBindCtx *pbc, REFGUID rbhid, REFIID riid, void **pp);
        STDMETHODIMP GetPropertyStore(GETPROPERTYSTOREFLAGS flags, REFIID riid, void **ppv);
        STDMETHODIMP GetPropertyDescriptionList(REFPROPERTYKEY keyType, REFIID riid, void **ppv);
        STDMETHODIMP GetAttributes(SIATTRIBFLAGS dwAttribFlags, SFGAOF sfgaoMask, SFGAOF *psfgaoAttribs);
        STDMETHODIMP GetCount(DWORD* pdwNumItems);
        STDMETHODIMP GetItemAt(DWORD dwIndex, IShellItem **ppsi);
        STDMETHODIMP EnumItems(IEnumShellItems** pp);

	public: // python
		static PCSTR __name__() throw()					{ return "brownie.Paths"; }
		static PyMethodDef* __method__() throw();
		ssize_t	__len__() throw();
		object	__getitem__(ssize_t index) throw(...);
		object	__iter__() throw();
		void	execute(Tuple* args, Dict* kw);

		static PySequenceMethods* __sequence__() throw();
		static ssize_t		__len__(PyObject* self) throw();
		static PyObject*	__getitem__(PyObject* self, ssize_t index) throw();

	private:
		void InitFolder(const ITEMIDLIST* parent);
	};

	//========================================================================================================================
}
