// shell.hpp
#pragma once

#include <shlobj.h>
#include <shellapi.h>
#include "autoptr.hpp"

__interface IShellFolder;
__interface IShellView;
__interface IImageList;

struct _IDA;
typedef _IDA CIDA;

struct __declspec(uuid("000214F1-0000-0000-C000-000000000046")) ICommDlgBrowser;
struct __declspec(uuid("10339516-2894-11d2-9039-00C04F8EEB3E")) ICommDlgBrowser2;
struct __declspec(uuid("000214E4-0000-0000-C000-000000000046")) IContextMenu;
struct __declspec(uuid("000214F4-0000-0000-C000-000000000046")) IContextMenu2;
struct __declspec(uuid("BCFCE0A0-EC17-11d0-8D10-00A0C90F2719")) IContextMenu3;

#ifndef SHCNRF_InterruptLevel
#	define SHCNRF_InterruptLevel		0x0001
#	define SHCNRF_ShellLevel			0x0002
#	define SHCNRF_RecursiveInterrupt	0x1000	/* Must be combined with SHCNRF_InterruptLevel */
#	define SHCNRF_NewDelivery			0x8000	/* Messages use shared memory */
#endif

namespace brownie
{
	//==================================================================================================================================================================
	// ITEMIDLIST

	// UINT ILGetSize(const ITEMIDLIST* pidl);
	// BOOL ILIsEqual(const ITEMIDLIST* pidl1, const ITEMIDLIST* pidl2);
	// BOOL ILIsParent(const ITEMIDLIST* pidlParent, const ITEMIDLIST* pidlBelow, BOOL fImmediate);
	// ITEMIDLIST* ILGetNext(const ITEMIDLIST* pidl);
	// ITEMIDLIST* ILFindChild(const ITEMIDLIST* pidlParent, const ITEMIDLIST* pidlChild);
	// ITEMIDLIST* ILFindLastID(const ITEMIDLIST* pidl);

	// BOOL ILRemoveLastID(ITEMIDLIST* pidl);
	// ITEMIDLIST* ILAppendID(ITEMIDLIST* pidl, LPSHITEMID pmkid, BOOL fAppend_or_Prepend);
	// ITEMIDLIST* ILClone(const ITEMIDLIST* pidl);
	// ITEMIDLIST* ILCloneFirst(const ITEMIDLIST* pidl);
	// ITEMIDLIST* ILCombine(const ITEMIDLIST* pidl1, const ITEMIDLIST* pidl2);

	inline bool		ILIsRoot(const ITEMIDLIST* item) throw()	{ return item->mkid.cb == 0; }
	inline void		ILDelete(const ITEMIDLIST* item) throw()	{ ::CoTaskMemFree((void*)item); }
	ITEMIDLIST*		ILCreate(PCWSTR path, DWORD* pdwAttr = null) throw();
	ITEMIDLIST*		ILCreate(IShellFolder* folder, PCWSTR relname) throw();
	ITEMIDLIST*		ILCreate(int csidl) throw();
	ITEMIDLIST*		ILCreate(const CIDA* items, size_t index) throw();
	ITEMIDLIST*		ILCreate(IShellItem* item) throw();
	ITEMIDLIST*		ILCreate(HANDLE mem, DWORD pid) throw();
	ITEMIDLIST*		ILCreate(HWND explorer);
	ITEMIDLIST*		ILCloneParent(const ITEMIDLIST* item) throw();
	inline HRESULT	ILPath(const ITEMIDLIST* pidl, CHAR  path[MAX_PATH]) throw()	{ return SHGetPathFromIDListA(pidl, path) ? S_OK : E_FAIL; }
	inline HRESULT	ILPath(const ITEMIDLIST* pidl, WCHAR path[MAX_PATH]) throw()	{ return SHGetPathFromIDListW(pidl, path) ? S_OK : E_FAIL; }
	bool			ILIsFolder(const ITEMIDLIST* item) throw();
	HRESULT			ILInfo(const ITEMIDLIST* item, WIN32_FIND_DATA* info) throw();
	HRESULT			ILParentFolder(IShellFolder** pp, const ITEMIDLIST* item, const ITEMIDLIST** leaf = null) throw();
	HRESULT			ILFolder(IShellFolder** pp, const ITEMIDLIST* item) throw();
	HRESULT			ILContextMenu(IContextMenu** pp, const ITEMIDLIST* item) throw();
	HRESULT			ILDisplayName(IShellFolder* folder, const ITEMIDLIST* leaf, DWORD shgdn, WCHAR name[MAX_PATH]);
	DWORD_PTR		ILFileInfo(const ITEMIDLIST* pidl, SHFILEINFO* info, DWORD flags, DWORD dwFileAttr = 0);
	int				ILIconIndex(const ITEMIDLIST* pidl);
	HRESULT			ILExecute(const ITEMIDLIST* pidl, PCWSTR verb = null, PCWSTR args = null, PCWSTR dir = null);
	int				ILCompare(const ITEMIDLIST* lhs, const ITEMIDLIST* rhs) throw();

	typedef AutoPtr<ITEMIDLIST, CoTaskMem> ILPtr;

	//==================================================================================================================================================================
	// CIDA

	CIDA*				CIDACreate(IDataObject* data) throw();
	CIDA*				CIDACreate(IShellView* view, int svgio) throw();
	CIDA*				CIDACreate(IFolderView* view, int svgio) throw();
	HGLOBAL				CIDACreate(const ITEMIDLIST* pidl);
	CIDA*				CIDAClone(const CIDA* items) throw();
	const ITEMIDLIST*	CIDAParent(const CIDA* items) throw();
	const ITEMIDLIST*	CIDAChild(const CIDA* items, size_t index) throw();
	inline size_t		CIDACount(const CIDA* items) throw()	{ return items ? items->cidl : 0; }
	size_t				CIDASize(const CIDA* items) throw();
	inline void			CIDAFree(const CIDA* items) throw()		{ CoTaskMemFree((void*)items); }

//	HGLOBAL CIDACreate(const ITEMIDLIST* pidl);
//	HGLOBAL CIDAToHDROP(const CIDA* pCIDA);	// CIDAHDROPɕϊ.
//	HGLOBAL CIDAToString(const CIDA* pCIDA);	// CIDAWCSɕϊ.

	//==================================================================================================================================================================
	// IShellItem

	HRESULT ItemCreate(IShellItem** pp, PCWSTR path);
	HRESULT	ItemCreate(IShellItem** pp, IShellFolder* folder, const ITEMIDLIST* parent, const ITEMIDLIST* leaf);
	HRESULT	ItemCreate(IShellItem** pp, const ITEMIDLIST* item);
	HRESULT	ItemCreate(IShellItem** pp, IShellItem* parent, PCWSTR name);

	//==================================================================================================================================================================
	// HDROP

	HRESULT HDropToString(HDROP hDrop, IStream* stream, PCWSTR sep = L"\r\n");

	//==================================================================================================================================================================
	// Registry

	HRESULT	RegGetDWORD(HKEY hKey, PCWSTR subkey, PCWSTR value, DWORD* outDWORD);
	HRESULT	RegSetDWORD(HKEY hKey, PCWSTR subkey, PCWSTR value, DWORD  inDWORD);
	HRESULT	RegGetString(HKEY hKey, PCWSTR subkey, PCWSTR value, WCHAR outString[], size_t cch);
	template < size_t sz >
	inline HRESULT RegGetString(HKEY hKey, PCWSTR subkey, PCWSTR value, WCHAR (&outString)[sz])
	{
		return RegGetString(hKey, subkey, value, outString, sz);
	}
	HRESULT	RegGetAssocExe(PCWSTR extension, WCHAR exe[MAX_PATH]);

	//==================================================================================================================================================================
	// FolderState

	struct FolderState : FOLDERSETTINGS
	{
		FILETIME	mtime;
	};

	void	FolderStateDispose() throw();
	HRESULT	FolderStateRestore(PCWSTR path) throw();
	HRESULT	FolderStateDump(PCWSTR path) throw();
	HRESULT	FolderStateLoad(const ITEMIDLIST* item, FolderState* state, IStream** stream) throw();
	HRESULT	FolderStateSave(const ITEMIDLIST* item, IShellView* view, IStream** stream) throw();
	void	FolderStateReset(const ITEMIDLIST* item);

	//==================================================================================================================================================================
	// Link (Shortcut)

	HRESULT	LinkResolve(PCWSTR src, ITEMIDLIST** dst, IShellLink** pp = null) throw();
	HRESULT	LinkResolve(PCWSTR src, WCHAR dst[MAX_PATH], IShellLink** pp = null) throw();
	HRESULT	LinkCreate(const ITEMIDLIST* item, PCWSTR dst, IShellLink** pp = null) throw();

	//==================================================================================================================================================================
	// Filesystem observer

	void	SHObserveBegin(HWND hwnd, UINT msg);
	void	SHObserveEnd();
	HRESULT	SHObserveNotify(WPARAM wParam, LPARAM lParam);

	//==================================================================================================================================================================
	// File operations

	HRESULT FileRename(IFileOperation* op, IShellItem* item, PCWSTR name);
	HRESULT FileRename(IFileOperation* op, IUnknown* items, PCWSTR name);
	HRESULT FileRename(IFileOperation* op, PCWSTR item, PCWSTR name);
	HRESULT FileRename(IFileOperation* op, size_t num, PCWSTR items[], PCWSTR name);

	HRESULT FileMove(IFileOperation* op, IShellItem* item, IShellItem* folder, PCWSTR name);
	HRESULT FileMove(IFileOperation* op, IUnknown* items, IShellItem* folder);
	HRESULT FileMove(IFileOperation* op, PCWSTR item, PCWSTR path);
	HRESULT FileMove(IFileOperation* op, size_t num, PCWSTR items[], PCWSTR folder);

	HRESULT FileCopy(IFileOperation* op, IShellItem* item, IShellItem* folder, PCWSTR name);
	HRESULT FileCopy(IFileOperation* op, IUnknown* items, IShellItem* folder);
	HRESULT FileCopy(IFileOperation* op, PCWSTR item, PCWSTR path);
	HRESULT FileCopy(IFileOperation* op, size_t num, PCWSTR items[], PCWSTR folder);

	HRESULT FileDelete(IFileOperation* op, IShellItem* item);
	HRESULT FileDelete(IFileOperation* op, IUnknown* items);
	HRESULT FileDelete(IFileOperation* op, PCWSTR item);
	HRESULT FileDelete(IFileOperation* op, size_t num, PCWSTR items[]);

	HRESULT	FileNew(IFileOperation* op, IShellItem* folder, PCWSTR name) throw();
	HRESULT	FileNew(IFileOperation* op, PCWSTR path) throw();

	HRESULT	FileCut(PCWSTR src);
	HRESULT	FileCopy(PCWSTR src);
	HRESULT	FilePaste(PCWSTR dst);
	HRESULT	FileDialogSetPath(PCWSTR path);

	//==================================================================================================================================================================
	// etc

	BOOL		AwareDPI() throw();

/*
	HRESULT ExpEnumExplorers(HRESULT (*fnEnum)(HWND hExplorer, const ITEMIDLIST* pidl, LPARAM lParam), LPARAM lParam);
	DWORD   ExpGetThumbnailSize();
	HRESULT ExpSetThumbnailSize(DWORD dwSize);
	HRESULT ExpGetNavigateSound(WCHAR path[MAX_PATH]);
	HRESULT NewDocument(PCWSTR path);

	bool	PathIsFolder(PCWSTR path);
	bool	PathIsRegistry(PCWSTR name);
*/
}

IStream& operator >> (IStream& stream, const ITEMIDLIST*& v) throw(HRESULT);
IStream& operator << (IStream& stream, const ITEMIDLIST* v) throw(HRESULT);

inline bool operator == (const FILETIME& lhs, const FILETIME& rhs) throw()
{
	return memcmp(&lhs, &rhs, sizeof(FILETIME)) == 0;
}
