// window.hpp
#pragma once

#include "hwnd.hpp"
#include "class.hpp"
#include "layout.hpp"
#include "ref.hpp"

namespace brownie
{
	class Thread;

	//==================================================================================================================================================================

	class Window : public Object, public Hwnd
	{
		friend class Thread;

	private:
		DIRECTION	m_dock;			// eEBhEł̔zu
		Hwnd		m_focus;		// ŌɃtH[JXĂq
		SIZE		m_size;			// NCAgTCỸLbV
		object		m_keymap;		// L[}bv
		object		m_gesture;		// WFX`
		bool		m_rearrange;	// ĔzuKvH
		bool		m_pressed;		// L[H
		bool		m_activated;	// tH[JX𓾂

	public:
		Window() throw();
		~Window() throw();
		static Window*	From(HWND hwnd) throw();
		Window* FromPoint(int x, int y) const throw();

	public:
		DIRECTION		get_Dock() const throw()	{ return m_dock; }
		virtual void	set_Dock(DIRECTION value) throw();
		__declspec(property(get=get_Dock, put=set_Dock)) DIRECTION Dock;

		SIZE	get_ClientSize() const throw()		{ return m_size; }
		__declspec(property(get=get_ClientSize, put=set_ClientSize)) SIZE ClientSize;

		RECT	get_ClientArea() const throw()		{ RECT r = { 0, 0, SIZE_WH(m_size) }; return r; }
		__declspec(property(get=get_ClientArea)) RECT ClientArea;

		RECT	get_CenterArea() const throw();
		__declspec(property(get=get_CenterArea)) RECT CenterArea;

		virtual SIZE get_DefaultSize() const throw();
		__declspec(property(get=get_DefaultSize)) SIZE DefaultSize;

		Window*	get_Parent() const throw();
		__declspec(property(get=get_Parent)) Window* Parent;

		Window*	get_Root() const throw();
		__declspec(property(get=get_Root)) Window* Root;

		HWND	get_Focus() const throw();
		void	set_Focus(HWND hwnd) throw();
		__declspec(property(get=get_Focus, put=set_Focus)) HWND Focus;

		bool	Press(UINT vkey, UINT mods) throw();
		bool	Gesture(UINT vkey, const DIRECTION dirs[], size_t ndirs) throw();
		void	Activate(DIRECTION dir);
		void	Scroll(DIRECTION dir);
		void	Update(bool needsLayout) throw();
		UINT	SendMessageToChildren(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0) const throw();
		UINT	PostMessageToChildren(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0) const throw();

	public: // override
		virtual void	QueryLayout(ILayout& layout) const;
		virtual void	OnDragDrop(POINT pt, RECT* rc, IDropTarget** pp) throw();	// override: 

	protected: // override
		virtual LRESULT	OnMessage(UINT msg, WPARAM wParam, LPARAM lParam);
		virtual LRESULT	OnMessage(Window* from, NMHDR& nm);
		virtual	void	OnUpdate();									// UPDATEUISTATE
		virtual void	OnLayout();									// zu@ύXł.
		virtual bool	OnPress(const MSG& msg);					// L[͂ł.
		virtual bool	OnGesture(const MSG& msg);	// WFX`̎󂯓ۂύXł.
		virtual void	OnChildFocus(HWND child);					// qtH[JXꂽ.
		virtual void	OnChildResize(Window* child);				// q̃TCYύXꂽ.

	protected: // internal message
		void	NotifyResize();

	private:
		void	UpdateLayout() throw();
		static LRESULT CALLBACK OnMessage_Bootstrap(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
		static LRESULT CALLBACK OnMessage_Main(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam, UINT_PTR id, DWORD_PTR data);

	protected: // init
		void	__init__(Tuple* args, Dict* kw, DIRECTION dock, DWORD style = 0, DWORD styleEx = 0) throw(...);
		HWND	__init__parse(Tuple* args, Dict* kw, DIRECTION dock) throw();
		void	create(WNDCLASSEX wc, DWORD style, DWORD styleEx, HWND parent = null);
		void	create(PCWSTR cls, DWORD style, DWORD styleEx, HWND parent = null);
		void	create(DWORD style, DWORD styleEx, DWORD ctyleCls, HWND parent = null);

	public: // python
		static PCSTR __name__() throw()					{ return "brownie.Window"; }
		void __init__(Tuple* args, Dict* kw) throw(...);
		static PyMethodDef* __method__() throw();
		static PyGetSetDef* __member__() throw();
		void	activate(Tuple* args);
		void	close() throw();
		void	move(Tuple* args);
		void	resize(Tuple* args);
		object	press(UINT vkey, UINT mods);
		object	press(Tuple* args);
		void	scroll(PyObject* o);
		object	get_focus();
		object	get_parent();
		object	get_root();
		object	get_name();
		void	set_name(PyObject* value);
		object	get_visible();
		void	set_visible(PyObject* value);
		object	get_location();
		void	set_location(PyObject* value);
		object	get_size();
		void	set_size(PyObject* value);
		object	get_dock();
		void	set_dock(PyObject* value);
		object	get_value();
		void	set_value(PyObject* value);
		object	get_children();
	};

	//==================================================================================================================================================================

	inline DWORD DockToCCS(DIRECTION dock) throw()
	{
		switch (dock)
		{
		case NORTH: return CCS_TOP;
		case SOUTH: return CCS_BOTTOM;
		case WEST:  return CCS_LEFT;
		case EAST:  return CCS_RIGHT;
		default:    return 0;
		}
	}

	void SubclassSingleLineEdit(HWND hwnd);
}
