// main.cpp

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <psapi.h>
#include "../include/config.hpp"

#pragma comment(linker, "/nodefaultlib")

using namespace brownie;

namespace
{
#ifdef _DEBUG
#	define BROWNIE_DLL		L"brownie_d.pyd"
#else
#	define BROWNIE_DLL		L"brownie.pyd"
#endif

	const WCHAR	LIB[] = L"bin\\" BROWNIE_DLL;

	//======================================================================

	inline static PWSTR GetArguments()
	{
		PWSTR args = GetCommandLine();
		if (*args == L'"')
		{
			++args;
			while (*args != L'"') { ++args; }
			++args;
		}
		else
		{
			while (*args != L' ') { ++args; }
		}
		while (*args == L' ') { ++args; }
		return args;
	}

	inline static int GetShowCommand()
	{
		STARTUPINFO info = { 0 };
		GetStartupInfo(&info);
		return info.dwFlags & STARTF_USESHOWWINDOW ? info.wShowWindow : SW_SHOWDEFAULT;
	}

	//======================================================================

	inline HWND Find(PCWSTR exe)
	{
		HWND hwndChildAfter = NULL;
		while (HWND hwnd = FindWindowEx(HWND_MESSAGE, hwndChildAfter, APPNAME, NULL))
		{
			DWORD dwProcessId;
			GetWindowThreadProcessId(hwnd , &dwProcessId);
			if (HANDLE hProcess = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, dwProcessId))
			{
				const DWORD MAX_MODULE = 16;
				HMODULE hModule[MAX_MODULE];
				DWORD dwNeed;
				if (EnumProcessModules(hProcess, hModule, sizeof(hModule), &dwNeed))
				{	// XXX: ǂꍇɁA̃vZX̃W[\̂H
					for (DWORD i = 0; i < dwNeed; ++i)
					{
						WCHAR path[MAX_PATH];
						if (GetModuleFileNameEx(hProcess, hModule[i], path, MAX_PATH) && lstrcmpi(exe, path) == 0)
						{	// ok, found
							CloseHandle(hProcess);
							return hwnd;
						}
					}
				}
				// not my avesta
				CloseHandle(hProcess);
			}
			hwndChildAfter = hwnd;
		}
		return NULL;
	}

	inline int Send(HWND hwnd, PCWSTR args)
	{
		COPYDATASTRUCT data;
		data.dwData = COPYDATA_MAGIC;
		data.cbData = (1 + lstrlen(args)) * sizeof(WCHAR);
		data.lpData = const_cast<PWSTR>(args);
		if ((BOOL)::SendMessage(hwnd, WM_COPYDATA, NULL, (LPARAM)&data))
			return EXIT_OK;
		else
			return EXIT_ERROR_SEND;
	}

	inline int Run(WCHAR path[])
	{
		typedef int (*Main)(DWORD version);
		int ret;

		// Create DLL path. $ROOT/*.exe -> $ROOT/LIB
		size_t sep = 0;
		for (size_t i = 0; path[i]; ++i)
		{
			if (path[i] == L'\\')
				sep = i;
		}
		memcpy(path + sep + 1, LIB, sizeof(LIB));

retry:	// Main
		ret = EXIT_ERROR_LOAD;
		HINSTANCE lib = ::LoadLibrary(path);
		if (lib)
		{
			if (Main main = (Main)::GetProcAddress(lib, "main"))
				ret = main(PROTOCOL_VERSION);
		}

		switch (ret)
		{
		case EXIT_RESTART:
			// SɃA[h܂ŉxsB
			// Pythonǂݍ񂾃W[A[hȂ߁ǍnB
			while (::FreeLibrary(lib)) {}
			goto retry;
		}

		return ret;
	}
}

INT APIENTRY WinMainCRTStartup() throw()
{
	int		ret;

	WCHAR path[MAX_PATH];
	::GetModuleFileName(NULL, path, MAX_PATH);

	PWSTR	args = GetArguments();
	HWND	hwnd = Find(path);

	if (hwnd)
		ret = Send(hwnd, args);
	else
		ret = Run(path);

	if (ret < EXIT_OK)
	{
		WCHAR message[1024];

		if (ret == EXIT_ERROR_LOAD)
		{
			DWORD errcode = ::GetLastError();
			PWSTR errmsg = NULL;
			::FormatMessageW(
				FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
				NULL, errcode, LANG_USER_DEFAULT, (PWSTR)&errmsg, 0, NULL);
			for (int ln = lstrlen(errmsg)-1; ln >= 0; ln--)
			{
				if (errmsg[ln] == L'\n' || errmsg[ln] == L'\r')
					errmsg[ln] = L'\0';
				else
					break;
			}
			wsprintf(message, BROWNIE_DLL L" ǂݍ߂܂B\n\n%s (0x%08X)", errmsg, errcode);
			::LocalFree(errmsg);
		}
		else if (ret == EXIT_ERROR_SEND)
			lstrcpy(message, L"̎󂯓nɎs܂B");
		else
			wsprintf(message, L"G[܂B (%d)", ret);

		::MessageBox(NULL, message, APPNAME, MB_OK | MB_ICONERROR);
	}

	::ExitProcess(ret);
}
